/******************************************************************************
 **
 ** FILE NAME    : ifxmips_ppa_datapath_ar10_a5.c
 ** PROJECT      : UEIP
 ** MODULES      : ATM + MII0/1 Acceleration Package (AR10 PPA A5)
 **
 ** DATE         : 18 FEB 2010
 ** AUTHOR       : Xu Liang
 ** DESCRIPTION  : ATM + MII0/1 Driver with Acceleration Firmware (A5)
 ** COPYRIGHT    :   Copyright (c) 2006
 **          Infineon Technologies AG
 **          Am Campeon 1-12, 85579 Neubiberg, Germany
 **
 **    This program is free software; you can redistribute it and/or modify
 **    it under the terms of the GNU General Public License as published by
 **    the Free Software Foundation; either version 2 of the License, or
 **    (at your option) any later version.
 **
 ** HISTORY
 ** $Date        $Author         $Comment
 ** 18 FEB 2010  Xu Liang        Initiate Version
 *******************************************************************************/

/*
 * ####################################
 *              Version No.
 * ####################################
 */
#define VER_FAMILY      0x40        //  bit 0: res
//      1: Danube
//      2: Twinpass
//      3: Amazon-SE
//      4: res
//      5: AR9
//      6: VR9
//      7: AR10
#define VER_DRTYPE      0x03        //  bit 0: Normal Data Path driver
//      1: Indirect-Fast Path driver
//      2: HAL driver
//      3: Hook driver
//      4: Stack/System Adaption Layer driver
//      5: PPA API driver
#define VER_INTERFACE   0x07        //  bit 0: MII 0
//      1: MII 1
//      2: ATM WAN
//      3: PTM WAN
#define VER_ACCMODE     0x01        //  bit 0: Routing
//      1: Bridging
#define VER_MAJOR       0
#define VER_MID         1
#define VER_MINOR       4

#define DMA_CHAN_NR_DIRECTPATH_CPU_TO_PPE  3
#define DMA_CHAN_NR_DEFAULT_CPU_TX         2

/*
 * ####################################
 *              Head File
 * ####################################
 */

/*
 *  Common Head File
 */
#include <linux/autoconf.h>
#include <linux/vmalloc.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/miscdevice.h>
#include <linux/atmdev.h>
#include <linux/init.h>
#include <linux/etherdevice.h>  /*  eth_type_trans  */
#include <linux/ethtool.h>      /*  ethtool_cmd     */
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <asm/uaccess.h>
#include <asm/addrspace.h>
#include <asm/unistd.h>
#include <asm/irq.h>
#include <asm/delay.h>
#include <asm/io.h>
#include <asm/prom.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/avm_pa_ifx_multiplexer.h>
#include <linux/avmnet.h>
#include <linux/pkt_sched.h>

/*
 *  Chip Specific Head File
 */
#include <ifx_types.h>
#include <ifx_regs.h>
#include <common_routines.h>
#include <ifx_pmu.h>
#include <ifx_rcu.h>
#include <ifx_gpio.h>
#include <ifx_led.h>
#include <ifx_dma_core.h>
#include <ifx_clk.h>
#include <switch_api/ifx_ethsw_kernel_api.h>
#include <switch_api/ifx_ethsw_api.h>

#ifdef CONFIG_IFX_ETH_FRAMEWORK
  #include <ifx_eth_framework.h>
#endif
#include <net/ifx_ppa_api.h>
#include <net/ifx_ppa_stack_al.h>
#include <net/ifx_ppa_api_directpath.h>
#include <net/ifx_ppa_ppe_hal.h>

#include <avmnet_config.h>

#include "../common/swi_ifx_common.h"
#include "../common/ifx_ppa_datapath.h"
#include "../7port/swi_7port.h"
#include "../7port/swi_7port_reg.h"
#include "ifxmips_ppa_datapath_fw_ar10_a5.h"

#include "ifxmips_ppa_hal_ar10_a5.h"
#include "../7port/ifxmips_ppa_datapath_7port_common.h"



/*
 * ####################################
 *   Parameters to Configure PPE
 * ####################################
 */
static int port_cell_rate_up[2] = { 3200, 3200 }; /*  Maximum TX cell rate for ports                  */

static int qsb_tau = 1; /*  QSB cell delay variation due to concurrency     */
static int qsb_srvm = 0x0F; /*  QSB scheduler burst length                      */
static int qsb_tstep = 4; /*  QSB time step, all legal values are 1, 2, 4     */

static int write_descriptor_delay = 0x20; /*  Write descriptor delay                          */

static int aal5r_max_packet_size = 0x0630; /*  Max frame size for RX                           */
static int aal5r_min_packet_size = 0x0000; /*  Min frame size for RX                           */
static int aal5s_max_packet_size = 0x0630; /*  Max frame size for TX                           */

static int ethwan = 0;
static int wanitf = ~0;

static int ipv6_acc_en = 1;
static unsigned int oam_cell_drop_count = 0;

static int wanqos_en = 8; //bis UGW 5.1: 0
static char *avm_ata_dev = NULL;

#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_SIZE)
static int directpath_tx_queue_size = 1524 * 50;
#endif
#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_PKTS)
static int directpath_tx_queue_pkts = 50;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
#define MODULE_PARM_ARRAY(a, b)   module_param_array(a, int, NULL, 0)
#define MODULE_PARM(a, b)         module_param(a, int, 0)
#else
#define MODULE_PARM_ARRAY(a, b)   MODULE_PARM(a, b)
#endif

MODULE_PARM_ARRAY(port_cell_rate_up, "2-2i");
MODULE_PARM_DESC(port_cell_rate_up, "ATM port upstream rate in cells/s");

MODULE_PARM(qsb_tau, "i");
MODULE_PARM_DESC(qsb_tau, "Cell delay variation. Value must be > 0");
MODULE_PARM(qsb_srvm, "i");
MODULE_PARM_DESC(qsb_srvm, "Maximum burst size");
MODULE_PARM(qsb_tstep, "i");
MODULE_PARM_DESC(qsb_tstep, "n*32 cycles per sbs cycles n=1,2,4");

MODULE_PARM(write_descriptor_delay, "i");
MODULE_PARM_DESC(
		write_descriptor_delay,
		"PPE core clock cycles between descriptor write and effectiveness in external RAM");

MODULE_PARM(aal5r_max_packet_size, "i");
MODULE_PARM_DESC(aal5r_max_packet_size,
		"Max packet size in byte for downstream AAL5 frames");
MODULE_PARM(aal5r_min_packet_size, "i");
MODULE_PARM_DESC(aal5r_min_packet_size,
		"Min packet size in byte for downstream AAL5 frames");
MODULE_PARM(aal5s_max_packet_size, "i");
MODULE_PARM_DESC(aal5s_max_packet_size,
		"Max packet size in byte for upstream AAL5 frames");

MODULE_PARM(ethwan, "i");
MODULE_PARM_DESC(ethwan, "WAN mode, 2 - ETH WAN, 1 - ETH mixed, 0 - DSL WAN.");

module_param(avm_ata_dev, charp, 0);
MODULE_PARM_DESC(avm_ata_dev, "eth0: ethernet_wan with NAT, ath0: WLAN_WAN with NAT");

MODULE_PARM(wanitf, "i");
MODULE_PARM_DESC(
		wanitf,
		"WAN interfaces, bit 0 - ETH0, 1 - ETH1, 2 - reserved for CPU0, 3/4/5 - DirectPath, 6/7 - DSL");

MODULE_PARM(ipv6_acc_en, "i");
MODULE_PARM_DESC(ipv6_acc_en, "IPv6 support, 1 - enabled, 0 - disabled.");

MODULE_PARM(wanqos_en, "i");
MODULE_PARM_DESC(wanqos_en, "WAN QoS support, 1 - enabled, 0 - disabled.");

#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_SIZE)
MODULE_PARM(directpath_tx_queue_size, "i");
MODULE_PARM_DESC(directpath_tx_queue_size, "directpath TX queue size in total buffer size");
#endif
#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_PKTS)
MODULE_PARM(directpath_tx_queue_pkts, "i");
MODULE_PARM_DESC(directpath_tx_queue_pkts, "directpath TX queue size in total packet number");
#endif

int mpoa_hack = 0; // falls mpoa_type == 2 && mpoa_mode == 0 -> setup(3,0)
module_param(mpoa_hack, int, S_IRUSR | S_IWUSR);

/*
 * ####################################
 *              Board Type
 * ####################################
 */

#define BOARD_AR10_REFERENCE                     0x01

/*
 * ####################################
 *              Definition
 * ####################################
 */

#define BOARD_CONFIG                            BOARD_AR10_REFERENCE

#define DEBUG_SKB_SWAP                          0

#define AVM_SETUP_MAC 							1

#define ENABLE_P2P_COPY_WITHOUT_DESC            1

#define ENABLE_LED_FRAMEWORK                    0

#define ENABLE_CONFIGURABLE_DSL_VLAN            1

#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_SIZE)
#define ENABLE_DIRECTPATH_TX_QUEUE            1
#define ENABLE_DIRECTPATH_TX_QUEUE_SIZE       directpath_tx_queue_size
#elif defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_PKTS)
#define ENABLE_DIRECTPATH_TX_QUEUE            1
#define ENABLE_DIRECTPATH_TX_QUEUE_PKTS       directpath_tx_queue_pkts
#endif

#ifdef CONFIG_IFX_PPA_NAPI_ENABLE
#define ENABLE_NAPI                           1
#else
#error driver just working with napi support so far
#endif

//TODO UGW 5.1.1: ENABLE_SWITCH_FLOW_CONTROL 1
#define ENABLE_SWITCH_FLOW_CONTROL              0

#define ENABLE_STATS_ON_VCC_BASIS               1

#define ENABLE_DSL_LOOPBACK_TEST                0   //  driver level loopback

#define ENABLE_DFE_LOOPBACK_TEST                0   //  DSL firmware level loopback

#define DISABLE_VBR                             0

#define ENABLE_MY_MEMCPY                        0

#define ENABLE_FW_MODULE_TEST                   0

#if defined(CONFIG_AVMNET_DEBUG) 
#define ENABLE_ASSERT                           1

#define DEBUG_QOS                               1

#define DEBUG_FIRMWARE_TABLES_PROC              1

#define PROC_WRITE_ROUTE						1

#define DEBUG_HTU_PROC                          1

#define DEBUG_MEM_PROC                          1

#define DEBUG_FW_PROC                           1

#define DEBUG_SEND_PROC                         0

#define DEBUG_MIRROR_PROC                       0

#else // defined(CONFIG_AVMNET_DEBUG) 

#define ENABLE_ASSERT                           0

#define DEBUG_QOS                               0

#define DEBUG_FIRMWARE_TABLES_PROC              0

#define PROC_WRITE_ROUTE                        0

#define DEBUG_HTU_PROC                          0

#define DEBUG_MEM_PROC                          0

#define DEBUG_FW_PROC                           0

#define DEBUG_SEND_PROC                         0

#define DEBUG_MIRROR_PROC                       0
#endif // defined(CONFIG_AVMNET_DEBUG) 

#if 0
#define PPE_MAILBOX_IGU0_INT                    INT_NUM_IM2_IRL23
#define PPE_MAILBOX_IGU1_INT                    INT_NUM_IM1_IRL29
#endif

#if defined(CONFIG_DSL_MEI_CPE_DRV) && !defined(CONFIG_IFXMIPS_DSL_CPE_MEI)
#define CONFIG_IFXMIPS_DSL_CPE_MEI            1
#endif

//  specific board related configuration
#if defined (BOARD_CONFIG)
#if BOARD_CONFIG == BOARD_AR10_REFERENCE
#endif
#endif

#if defined(ENABLE_STATS_ON_VCC_BASIS) && ENABLE_STATS_ON_VCC_BASIS
#define UPDATE_VCC_STAT(conn, field, num)     do { g_atm_priv_data.connection[conn].field += (num); } while (0)
#else
#define UPDATE_VCC_STAT(conn, field, num)
#endif


#if !defined(DISABLE_INLINE) || !DISABLE_INLINE
#define INLINE                                inline
#else
#define INLINE
#endif

/*
 *  Mailbox Signal Bit
 */
#define DMA_TX_CH1_SIG                          (1 << 31)   //  IGU0
#define DMA_TX_CH0_SIG                          (1 << 30)   //  IGU0
#define CPU_TO_ETH_WAN_TX_SIG                   (1 << 17)   //  IGU1
#define CPU_TO_DSL_WAN_TX_SIG                   (1 << 16)   //  IGU1
#define CPU_TO_WAN_SWAP_SIG                     (1 << 16)   //  IGU1, ETH WAN QoS Mode only
#define WAN_RX_SIG(i)                           (1 << (i))  //  IGU1, 0: AAL5, 1: OAM
/*
 *  Constant Definition
 */
#define ETH_WATCHDOG_TIMEOUT                    (10 * HZ)
#define ETOP_MDIO_DELAY                         1
#define IDLE_CYCLE_NUMBER                       30000

/*change DMA packet size for IP tunnelling feature */
// #define DMA_PACKET_SIZE                         1600    //  ((1518 + 62 + 8) + 31) & ~31
#define DMA_PACKET_SIZE                         1632

#define DMA_RX_ALIGNMENT                        128
#define DMA_TX_ALIGNMENT                        32




/*
 *  ATM Cell
 */
#define CELL_SIZE                               ATM_AAL0_SDU

/*
 *  Ethernet Frame Definitions
 */
#define ETH_CRC_LENGTH                          4
#define ETH_MAX_DATA_LENGTH                     ETH_DATA_LEN
#define ETH_MIN_TX_PACKET_LENGTH                ETH_ZLEN

/*
 *  ATM Port, QSB Queue, DMA RX/TX Channel Parameters
 */
#define ATM_PORT_NUMBER                         2
#define ATM_QUEUE_NUMBER                        15
#define ATM_SW_TX_QUEUE_NUMBER                  16
#define QSB_QUEUE_NUMBER_BASE                   1

#define DEFAULT_RX_HUNT_BITTH                   4

/*
 *  QSB Queue Scheduling and Shaping Definitions
 */
#define QSB_WFQ_NONUBR_MAX                      0x3f00
#define QSB_WFQ_UBR_BYPASS                      0x3fff
#define QSB_TP_TS_MAX                           65472
#define QSB_TAUS_MAX                            64512
#define QSB_GCR_MIN                             18

/*
 *  QSB Command Set
 */
#define QSB_RAMAC_RW_READ                       0
#define QSB_RAMAC_RW_WRITE                      1

#define QSB_RAMAC_TSEL_QPT                      0x01
#define QSB_RAMAC_TSEL_SCT                      0x02
#define QSB_RAMAC_TSEL_SPT                      0x03
#define QSB_RAMAC_TSEL_VBR                      0x08

#define QSB_RAMAC_LH_LOW                        0
#define QSB_RAMAC_LH_HIGH                       1

#define QSB_QPT_SET_MASK                        0x0
#define QSB_QVPT_SET_MASK                       0x0
#define QSB_SET_SCT_MASK                        0x0
#define QSB_SET_SPT_MASK                        0x0
#define QSB_SET_SPT_SBVALID_MASK                0x7FFFFFFF

#define QSB_SPT_SBV_VALID                       (1 << 31)
#define QSB_SPT_PN_SET(value)                   (((value) & 0x01) ? (1 << 16) : 0)
#define QSB_SPT_INTRATE_SET(value)              SET_BITS(0, 13, 0, value)

/*
 *  OAM Definitions
 */
#define OAM_HTU_ENTRY_NUMBER                    3
#define OAM_F4_SEG_HTU_ENTRY                    0
#define OAM_F4_TOT_HTU_ENTRY                    1
#define OAM_F5_HTU_ENTRY                        2
#define OAM_F4_CELL_ID                          0
#define OAM_F5_CELL_ID                          15

/*
 *  EMA Settings
 */
#define EMA_CMD_BUF_LEN                         0x0010
#define EMA_CMD_BASE_ADDR                       (0x1710 << 2)
#define EMA_DATA_BUF_LEN                        0x0040
#define EMA_DATA_BASE_ADDR                      (0x16d0 << 2)
#define EMA_WRITE_BURST                         0x02
#define EMA_READ_BURST                          0x02

/*
 *  Firmware Settings
 */
#define CPU_TO_WAN_TX_DESC_NUM                  32  //  WAN CPU TX
#define DSL_WAN_TX_DESC_NUM(i)                  32  //  WAN QoS TX, i < 15
#define ETH_WAN_TX_DESC_NUM(i)                  __ETH_WAN_TX_QUEUE_LEN  //  ETH WAN QoS TX, i < __ETH_WAN_TX_QUEUE_NUM
#define CPU_TO_WAN_SWAP_DESC_NUM                32
#define WAN_TX_DESC_NUM_TOTAL                   (DSL_WAN_TX_DESC_NUM(0) * 15)   //  DSL WAN TX desc overlap ETH WAN TX desc + ETH WAN Swap desc
#define WAN_RX_DESC_NUM(i)                      32  //  DSL RX PKT/OAM, i < 2, 0 PKT, 1 OAM
#define DMA_TX_CH1_DESC_NUM                     (g_eth_wan_mode ? (g_eth_wan_mode == 3 /* ETH WAN */ && g_wanqos_en ? 32 : DMA_RX_CH1_DESC_NUM) : WAN_RX_DESC_NUM(0))
#define DMA_RX_CH1_DESC_NUM                     32  //  DSL/MII1 WAN mode fast path
#define DMA_RX_CH2_DESC_NUM                     32  //  Switch/MII0/MII1 LAN mode fast path
#define DMA_TX_CH0_DESC_NUM                     DMA_RX_CH2_DESC_NUM //  DMA_TX_CH0_DESC_NUM is used to replace DMA_TX_CH2_DESC_NUM

#define IFX_PPA_PORT_NUM                        8

/*
 *  Bits Operation
 */
//#define GET_BITS(x, msb, lsb)                   (((x) & ((1 << ((msb) + 1)) - 1)) >> (lsb))
//#define SET_BITS(x, msb, lsb, value)            (((x) & ~(((1 << ((msb) + 1)) - 1) ^ ((1 << (lsb)) - 1))) | (((value) & ((1 << (1 + (msb) - (lsb))) - 1)) << (lsb)))

#define SWITCH_CLASS_VALUE                      4



/*
 *  DMA/EMA Descriptor Base Address
 */
#define CPU_TO_WAN_TX_DESC_BASE                 ((volatile struct tx_descriptor *)SB_BUFFER(0x3C00))                /*         32 each queue    */
#define DSL_WAN_TX_DESC_BASE(i)                 ((volatile struct tx_descriptor *)SB_BUFFER(0x3C40 + (i) * 2 * 32)) /*  i < 15,32 each queue    */
#define __ETH_WAN_TX_QUEUE_NUM                  g_wanqos_en
#define __ETH_WAN_TX_QUEUE_LEN                  (448 / __ETH_WAN_TX_QUEUE_NUM < 256 ? (448 / __ETH_WAN_TX_QUEUE_NUM) : 255)
#define __ETH_WAN_TX_DESC_BASE(i)               (0x3C80 + (i) * 2 * __ETH_WAN_TX_QUEUE_LEN)
#define ETH_WAN_TX_DESC_BASE(i)                 ((volatile struct tx_descriptor *)SB_BUFFER(__ETH_WAN_TX_DESC_BASE(i))) /*  i < __ETH_WAN_TX_QUEUE_NUM, __ETH_WAN_TX_QUEUE_LEN each queue    */
#define CPU_TO_WAN_SWAP_DESC_BASE               ((volatile struct tx_descriptor *)SB_BUFFER(0x3C40))                /*         32 each queue    */
#define WAN_RX_DESC_BASE(i)                     ((volatile struct rx_descriptor *)SB_BUFFER(0x3B80 + (i) * 2 * 32)) /*  i < 2, 32 each queue    */
#define DMA_TX_CH1_DESC_BASE                    (g_eth_wan_mode ? (g_eth_wan_mode == 3 /* ETH WAN */ && g_wanqos_en ? ((volatile struct rx_descriptor *)SB_BUFFER(0x3AC0)): DMA_RX_CH1_DESC_BASE) : WAN_RX_DESC_BASE(0))
#define DMA_RX_CH1_DESC_BASE                    ((volatile struct rx_descriptor *)SB_BUFFER(0x3B40))                /*         32 each queue    */
#define DMA_RX_CH2_DESC_BASE                    ((volatile struct rx_descriptor *)SB_BUFFER(0x3B00))                /*         32 each queue    */
#define DMA_TX_CH0_DESC_BASE                    DMA_RX_CH2_DESC_BASE


#if defined(DEBUG_FW_PROC) && DEBUG_FW_PROC
/*
 *  Firmware Proc
 */
//  need check with hejun
#define TIR(i)                                ((volatile u32 *)(0xBE198800) + (i))

#define __CPU_TX_SWAPPER_DROP_PKTS            SB_BUFFER(0x32C0)
#define __CPU_TX_SWAPPER_DROP_BYTES           SB_BUFFER(0x32C1)
#define __DSLWAN_FP_SWAPPER_DROP_PKTS         SB_BUFFER(0x32C2)
#define __DSLWAN_FP_SWAPPER_DROP_BYTES        SB_BUFFER(0x32C3)

#define __CPU_TXDES_SWAP_RDPTR                SB_BUFFER(0x32C4)
#define __DSLWAN_FP_RXDES_SWAP_RDPTR          SB_BUFFER(0x32C5)
#define __DSLWAN_FP_RXDES_DPLUS_WRPTR         SB_BUFFER(0x32C6)

#define __DSLWAN_TX_PKT_CNT(i)                SB_BUFFER(0x32D0+(i))   //  i < 8

#define __DSLWAN_TXDES_SWAP_PTR(i)            SB_BUFFER(0x32E0+(i))   //  i < 8

#if 0
// SB_BUFFER(0x29C0) - SB_BUFFER(0x29C7)
#define __VLAN_PRI_TO_QID_MAPPING             SB_BUFFER(0x32E0)
#endif

//
// etx MIB: SB_BUFFER(0x32F0) - SB_BUFFER(0x32FF)
#define __DSLWAN_TX_THRES_DROP_PKT_CNT(i)     SB_BUFFER(0x2D80+(i))   //  i < 8

#define __CPU_TO_DSLWAN_TX_PKTS               SB_BUFFER(0x32C8)
#define __CPU_TO_DSLWAN_TX_BYTES              SB_BUFFER(0x32C9)

#define ACC_ERX_PID                           SB_BUFFER(0x2B00)
#define ACC_ERX_PORT_TIMES                    SB_BUFFER(0x2B01)
#define SLL_ISSUED                            SB_BUFFER(0x2B02)
#define BMC_ISSUED                            SB_BUFFER(0x2B03)
#define DPLUS_RX_ON                           SB_BUFFER(0x3283)
#define ISR_IS                                SB_BUFFER(0x3284)

#define __ENETS_PGNUM                         20
#define __ERX_CBUF_BASE                       SB_BUFFER(0x343C)
#define __ERX_DBUF_BASE                       SB_BUFFER(0x3450)

#define PRESEARCH_RDPTR                       SB_BUFFER(0x2B06)

#define SLL_ERX_PID                           SB_BUFFER(0x2B04)
#define SLL_PKT_CNT                           SB_BUFFER(0x2B08)
#define SLL_RDPTR                             SB_BUFFER(0x2B0A)

#define EDIT_PKT_CNT                          SB_BUFFER(0x2B0C)
#define EDIT_RDPTR                            SB_BUFFER(0x2B0E)

#define DPLUSRX_PKT_CNT                       SB_BUFFER(0x2B10)
#define DPLUS_RDPTR                           SB_BUFFER(0x2B12)

#define SLL_STATE_NULL                        0
#define SLL_STATE_DA                          1
#define SLL_STATE_SA                          2
#define SLL_STATE_DA_BC                       3
#define SLL_STATE_ROUTER                      4

#define PRE_DPLUS_PTR                         SB_BUFFER(0x3281)
#define DPLUS_PTR                             SB_BUFFER(0x3282)
#define DPLUS_CNT                             SB_BUFFER(0x3280)
#endif



#define PPA_VDEV_ID(hw_dev_id) (hw_dev_id +   IFX_VIRT_DEV_OFFSET)
#define HW_DEV_ID(ppa_vdev_id) (ppa_vdev_id - IFX_VIRT_DEV_OFFSET)

/*
 *  Clock Generation Unit Registers
 */
#define AR10_CGU                                 (KSEG1 | 0x1F103000)
#define AR10_CGU_CLKFSR                          ((volatile u32*)(AR10_CGU + 0x0010))

/*
 * ####################################
 *              Data Type
 * ####################################
 */

/*
 *  64-bit Data Type
 */
typedef struct {
	unsigned int h :32;
	unsigned int l :32;
} ppe_u64_t;

/*
 *  PPE ATM Cell Header
 */
#if defined(__BIG_ENDIAN)
struct uni_cell_header {
	unsigned int gfc :4;
	unsigned int vpi :8;
	unsigned int vci :16;
	unsigned int pti :3;
	unsigned int clp :1;
};
#else
struct uni_cell_header {
	unsigned int clp :1;
	unsigned int pti :3;
	unsigned int vci :16;
	unsigned int vpi :8;
	unsigned int gfc :4;
};
#endif

struct rx_descriptor {
	//  0 - 3h
	unsigned int own :1; //  0: PPE, 1: MIPS, this value set is for DSL PKT RX and DSL OAM RX, it's special case.
	unsigned int c :1;
	unsigned int sop :1;
	unsigned int eop :1;
	unsigned int res1 :3;
	unsigned int byteoff :2;
	unsigned int res2 :2;
	unsigned int qid :4;
	unsigned int err :1; //  0: no error, 1: error (RA, IL, CRC, USZ, OVZ, MFL, CPI, CPCS-UU)
	unsigned int datalen :16;
	//  4 - 7h
	unsigned int res3 :3;
	unsigned int dataptr :29;
};

struct tx_descriptor {
	//  0 - 3h
	unsigned int own :1; //  0: MIPS, 1: PPE, for CPU to ATM WAN TX, it's inverse, 0: PPE, 1: MIPS
	unsigned int c :1;
	unsigned int sop :1;
	unsigned int eop :1;
	unsigned int byteoff :5;
	unsigned int mpoa_pt :1; //  not for WLAN TX, 0: MPoA is dterminated in FW, 1: MPoA is transparent to FW.
	unsigned int mpoa_type :2; //  not for WLAN TX, 0: EoA without FCS, 1: reserved, 2: PPPoA, 3: IPoA
	unsigned int pdu_type :1; //  not for WLAN TX, 0: AAL5, 1: Non-AAL5 cell
	unsigned int qid :4; //  not for WLAN TX, careful, in DSL it's 4 bits, in ETH it's 3 bits
	unsigned int datalen :15; //  careful, in DSL it's 15 bits, in ETH it's 16 bits
	//  4 - 7h
	unsigned int small :1;
	unsigned int res1 :2;
	unsigned int dataptr :29;
};

//    ETH TX is different compare to DSL TX
struct eth_tx_descriptor {
    //  0 - 3h
    unsigned int    own                 :1; //  0: MIPS, 1: PPE
    unsigned int    c                   :1;
    unsigned int    sop                 :1;
    unsigned int    eop                 :1;
    unsigned int    byteoff             :5;
    unsigned int    qid                 :4;
    unsigned int    res1                :3;
    unsigned int    datalen             :16;
    //  4 - 7h
    unsigned int    small               :1;
    unsigned int    res2                :2;
    unsigned int    dataptr             :29;
  };

/*
 *  QSB Queue Parameter Table Entry and Queue VBR Parameter Table Entry
 */
union qsb_queue_parameter_table {
	struct {
		unsigned int res1 :1;
		unsigned int vbr :1;
		unsigned int wfqf :14;
		unsigned int tp :16;
	} bit;
	u32 dword;
};

union qsb_queue_vbr_parameter_table {
	struct {
		unsigned int taus :16;
		unsigned int ts :16;
	} bit;
	u32 dword;
};

/*
 *  Internal Structure of Devices (ETH/ATM)
 */
struct atm_port {
	struct atm_dev *dev;

	u32 tx_max_cell_rate; /*  maximum cell rate                       */
	u32 tx_used_cell_rate; /*  currently used cell rate                */
};

struct atm_connection {
	struct atm_vcc *vcc; /*  opened VCC                              */
	struct timespec access_time; /*  time when last F4/F5 user cell arrives  */

	int prio_queue_map[8];
	u32 prio_tx_packets[8];

	u32 rx_packets;
	u32 rx_bytes;
	u32 rx_errors;
	u32 rx_dropped;
	u32 tx_packets;
	u32 tx_bytes;
	u32 tx_errors;
	u32 tx_dropped;

	u32 port; /*  to which port the connection belongs    */
	u32 sw_tx_queue_table; /*  software TX queues used for this        */
/*  connection                              */
};

struct atm_priv_data {
	struct atm_port port[ATM_PORT_NUMBER];
	volatile struct atm_connection connection[ATM_QUEUE_NUMBER];
	u32 max_connections; /*  total connections available             */
	u32 connection_table; /*  connection opened status, every bit     */
	/*  stands for one connection as well as    */
	/*  the QSB queue used by this connection   */
	u32 max_sw_tx_queues; /*  total software TX queues available      */
	u32 sw_tx_queue_table; /*  software TX queue occupations status    */

	ppe_u64_t wrx_total_byte; /*  bit-64 extention of MIB table member    */
	ppe_u64_t wtx_total_byte; /*  bit-64 extention of MIB talbe member    */

	u32 wrx_pdu; /*  successfully received AAL5 packet       */
	u32 wrx_drop_pdu; /*  AAL5 packet dropped by driver on RX     */
	u32 wtx_pdu;
	u32 wtx_err_pdu; /*  error AAL5 packet                       */
	u32 wtx_drop_pdu; /*  AAL5 packet dropped by driver on TX     */

	struct dsl_wan_mib_table prev_mib;
};

/*
 * ####################################
 *             Declaration
 * ####################################
 */

#if 0 && defined(ENABLE_NAPI) && ENABLE_NAPI // wird in AVMNET implementiert
static int eth_poll(struct napi_struct* napi, int);
#endif
/*
 *  ATM Operations
 */
static int ppe_ioctl(struct atm_dev *, unsigned int, void *);
static int ppe_open(struct atm_vcc *);
static void ppe_close(struct atm_vcc *);
static int ppe_send(struct atm_vcc *, struct sk_buff *);
static int ppe_send_oam(struct atm_vcc *, void *, int);
static int ppe_change_qos(struct atm_vcc *, struct atm_qos *, int);

void dump_atm_tx_desc( struct tx_descriptor *d );
/*
 *  ATM Upper Layer Hook Function
 */
static void mpoa_setup(struct atm_vcc *, int, int);

/*
 *  Network Operations
 */

int ifx_ppa_eth_hard_start_xmit_a5(struct sk_buff *, struct net_device *);
#if 0
int ifx_ppa_eth_qos_hard_start_xmit_a5(struct sk_buff *, struct net_device *);
#endif
int ifx_ppa_eth_ioctl_a5(struct net_device *, struct ifreq *, int);
void ifx_ppa_eth_tx_timeout_a5(struct net_device *);

/*
 *  Network operations help functions
 */
static INLINE int eth_xmit(struct sk_buff *, unsigned int, int, int, int);

/*
 *  ioctl help functions
 */
static INLINE int ethtool_ioctl(struct net_device *, struct ifreq *);

/*
 *  64-bit operation used by MIB calculation
 */
static INLINE void u64_add_u32(ppe_u64_t opt1, u32 opt2, ppe_u64_t *ret);

/*
 *  DSL led flash function
 */
static INLINE void adsl_led_flash(void);

/*
 *  DSL PVC Multiple Queue/Priority Operation
 */
static int sw_tx_queue_add(int);
static int sw_tx_queue_del(int);
static int sw_tx_prio_to_queue(int, int, int, int *);

/*
 *  QSB & HTU setting functions
 */
static void set_qsb(struct atm_vcc *, struct atm_qos *, unsigned int);
static INLINE void set_htu_entry(unsigned int, unsigned int, unsigned int, int);
static INLINE void clear_htu_entry(unsigned int);
static INLINE uint32_t get_qsb_clk(void);

/*
 *  QSB & HTU init functions
 */
static INLINE void qsb_global_set(void);
static INLINE void setup_oam_htu_entry(void);
static INLINE void validate_oam_htu_entry(void);
static INLINE void invalidate_oam_htu_entry(void);

/*
 *  look up for connection ID
 */
static INLINE int find_vpi(int);
static INLINE int find_vpivci(int, int);
static INLINE int find_vcc(struct atm_vcc *);

/*
 *  Buffer manage functions
 */
static INLINE struct sk_buff *alloc_skb_rx(void);
static INLINE struct sk_buff *alloc_skb_tx(int);
static struct sk_buff* skb_break_away_from_protocol(struct sk_buff *);
static INLINE void __get_skb_from_dbg_pool(struct sk_buff *, const char *, unsigned int);
#define get_skb_from_dbg_pool(skb)  __get_skb_from_dbg_pool(skb, __FUNCTION__, __LINE__)
static INLINE struct sk_buff *__get_skb_pointer(unsigned int, const char *, unsigned int);
#define get_skb_pointer(dataptr)    __get_skb_pointer(dataptr, __FUNCTION__, __LINE__)
static INLINE void __put_skb_to_dbg_pool(struct sk_buff *, const char *, unsigned int);
#define put_skb_to_dbg_pool(skb)    __put_skb_to_dbg_pool(skb, __FUNCTION__, __LINE__)
static struct sk_buff* atm_alloc_tx(struct atm_vcc *, unsigned int);
static INLINE void atm_free_tx_skb_vcc(struct sk_buff *skb);

/*
 *  Mailbox handler
 */
static irqreturn_t mailbox0_irq_handler(int, void *);
static irqreturn_t mailbox1_irq_handler(int, void *);

/*
 *  Turn On/Off Dma
 */
static INLINE void turn_on_dma_rx(int);
static INLINE void turn_off_dma_rx(int);

/*
 *  DMA interface functions
 */
static u8 *dma_buffer_alloc(int, int *, void **);
static int dma_buffer_free(u8 *, void *);
static int dma_int_handler(struct dma_device_info *, int, int);
static INLINE int dma_rx_int_handler(struct dma_device_info *, int);
static INLINE int handle_skb_a5(struct sk_buff *skb);
static int process_backlog_a5(struct napi_struct *napi, int quota);
/*
 *  Hardware init & clean-up functions
 */
static INLINE void reset_ppe(void);
static INLINE void init_pmu(void);
static INLINE void start_cpu_port(void);
//static INLINE void stop_datapath(void);
static INLINE void init_internal_tantos_qos_setting(void);
static INLINE void init_internal_tantos(void);
static INLINE void init_dplus(void);
static INLINE void init_ema(void);
static INLINE void init_mailbox(void);
static INLINE void clear_share_buffer(void);
static INLINE void clear_cdm(void);
static INLINE void board_init(void);
static INLINE void init_tc_hw(void);
static INLINE void setup_ppe_conf(void);
static INLINE void init_hw(void);

/*
 *  PP32 specific functions
 */
static INLINE int pp32_download_code(const u32 *, unsigned int, const u32 *, unsigned int);
static INLINE int pp32_specific_init(int, void *);
static INLINE int pp32_start(int);
static INLINE void pp32_stop(void);

/*
 *  Init & clean-up functions
 */
static INLINE int init_local_variables(void);
static INLINE void clear_local_variables(void);
static INLINE void init_communication_data_structures(int);
static INLINE int alloc_dma(void);
static INLINE void free_dma(void);

/*
 *  DSL Data Led help function
 */
static void dsl_led_polling(unsigned long);

/*
 *  local implement memcpy
 */
#if defined(ENABLE_MY_MEMCPY) && ENABLE_MY_MEMCPY
static INLINE void *my_memcpy(unsigned char *, const unsigned char *, unsigned int);
#else
#define my_memcpy             memcpy
#endif

/*
 *  Print Firmware/Driver Version ID
 */
static int print_fw_ver(char *buf, int buf_len);
static int print_driver_ver(char *, int, char *, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int);

/*
 *  Proc File
 */
static INLINE void proc_file_create(void);
static INLINE void proc_file_delete(void);
static int proc_read_ver(char *, char **, off_t, int, int *, void *);
static int proc_read_mib(char *, char **, off_t, int, int *, void *);
static int proc_write_mib(struct file *, const char *, unsigned long, void *);
#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC
#if defined(PROC_WRITE_ROUTE) && PROC_WRITE_ROUTE
  static INLINE int print_route(char *, int, int, struct rout_forward_compare_tbl *, struct rout_forward_action_tbl *);
  static int proc_read_route(char *, char **, off_t, int, int *, void *);
  static int proc_write_route(struct file *, const char *, unsigned long, void *);
#endif
  static int proc_read_mc(char *, char **, off_t, int, int *, void *);
  static int proc_read_genconf(char *, char **, off_t, int, int *, void *);
  static int proc_write_genconf(struct file *, const char *, unsigned long, void *);
  static int proc_read_queue(char *, char **, off_t, int, int *, void *);
  static int proc_write_queue(struct file *, const char *, unsigned long, void *);
  static int proc_read_pppoe(char *, char **, off_t, int, int *, void *);
  static int proc_read_mtu(char *, char **, off_t, int, int *, void *);
  static int proc_read_hit(char *, char **, off_t, int, int *, void *);
  static int proc_write_hit(struct file *, const char *, unsigned long, void *);
  static int proc_read_mac(char *, char **, off_t, int, int *, void *);
  static int proc_read_out_vlan(char *, char **, off_t, int, int *, void *);
  static int proc_read_ipv6_ip(char *, char **, off_t, int, int *, void *);
  static int proc_read_mpoa(char *page, char **start, off_t off, int count, int *eof, void *data);
  static int proc_write_mpoa(struct file *file, const char *buf, unsigned long count, void *data);
#endif
#if defined(DEBUG_HTU_PROC) && DEBUG_HTU_PROC
  static int proc_read_htu(char *, char **, off_t, int, int *, void *);
#endif
#if defined(DEBUG_MEM_PROC) && DEBUG_MEM_PROC
  static int proc_write_mem(struct file *, const char *, unsigned long, void *);
#endif
static int proc_read_burstlen(char *, char **, off_t, int, int *, void *);
static int proc_write_burstlen(struct file *, const char *, unsigned long, void *);

#if defined(DEBUG_FW_PROC) && DEBUG_FW_PROC
  static int proc_read_fw(char *, char **, off_t, int, int *, void *);
  static int proc_write_fw(struct file *, const char *, unsigned long, void *);
  static int proc_write_fwdbg(struct file *, const char *, unsigned long, void *);
  static void fw_dbg_start(char *cmdbuf);
  static void fwdbg_help(char **tokens, int token_num);
  static void dump_ipv4_hdr(struct iphdr * iph);
  static void dump_ipv6_hdr(struct ipv6hdr *iph6);
#endif
#if defined(DEBUG_SEND_PROC) && DEBUG_SEND_PROC
  static int proc_read_send(char *, char **, off_t, int, int *, void *);
  static int proc_write_send(struct file *, const char *, unsigned long, void *);
#endif
static INLINE unsigned int sw_get_rmon_counter(int, int);
static INLINE void sw_clr_rmon_counter(int);
static INLINE int read_port_counters(int, char *);
static int port0_counters_proc_read(char *, char **, off_t, int, int *, void *);
static int port1_counters_proc_read(char *, char **, off_t, int, int *, void *);
static int port2_counters_proc_read(char *, char **, off_t, int, int *, void *);
static int port3_counters_proc_read(char *, char **, off_t, int, int *, void *);
static int port4_counters_proc_read(char *, char **, off_t, int, int *, void *);
static int port5_counters_proc_read(char *, char **, off_t, int, int *, void *);
static int port6_counters_proc_read(char *, char **, off_t, int, int *, void *);
static int port_counters_proc_write(struct file *, const char *, unsigned long, void *);
static int proc_read_directforwarding(char *, char **, off_t, int, int *, void *);
static int proc_write_directforwarding(struct file *, const char *, unsigned long, void *);
static int proc_read_clk(char *, char **, off_t, int, int *, void *);
static int proc_write_clk(struct file *, const char *, unsigned long, void *);
static int proc_read_prio(char *, char **, off_t, int, int *, void *);
static int proc_write_prio(struct file *, const char *, unsigned long, void *);
static int proc_read_ewan(char *, char **, off_t, int, int *, void *);
static int proc_write_ewan(struct file *, const char *, unsigned long, void *);
#ifdef CONFIG_IFX_PPA_QOS
  static int proc_read_qos(char *, char **, off_t, int, int *, void *);
#endif
#if defined(ENABLE_CONFIGURABLE_DSL_VLAN) && ENABLE_CONFIGURABLE_DSL_VLAN
  static int proc_read_dsl_vlan(char *, char **, off_t, int, int *, void *);
  static int proc_write_dsl_vlan(struct file *, const char *, unsigned long, void *);
  static INLINE int ppe_set_vlan(int, unsigned short);
  static INLINE int sw_set_vlan_bypass(int, unsigned int, int);
#endif
#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC
  static int proc_read_mirror(char *, char **, off_t, int, int *, void *);
  static int proc_write_mirror(struct file *, const char *, unsigned long, void *);
#endif
static int proc_read_class(char *, char **, off_t, int, int *, void *);
static int proc_write_class(struct file *, const char *, unsigned long, void *);

static int dma_alignment_atm_good_count = 0;
static int dma_alignment_atm_bad_count = 0;

static int dma_alignment_eth_good_count = 0;
static int dma_alignment_eth_bad_count = 0;
static int padded_frame_count = 0;

/*
 *  Proc File help functions
 */
static INLINE int proc_buf_copy(char **, int, off_t, int *, const char *, int);
#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC
  static INLINE void get_ip_port(char **, int *, unsigned int *);
  static INLINE void get_mac(char **, int *, unsigned int *);
#endif
#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC
  static INLINE char *get_wanitf(int);
  static INLINE int print_mc(char *, int, struct wan_rout_multicast_cmp_tbl *, struct wan_rout_multicast_act_tbl *);
#endif

/*
 *  Debug functions
 */

#if defined(DEBUG_DUMP_INIT) && DEBUG_DUMP_INIT
static INLINE void dump_init(void);
#else
#define dump_init()
#endif

/*
 *  Directpath Help Functions
 */
#if defined(ENABLE_DIRECTPATH_TX_QUEUE) && ENABLE_DIRECTPATH_TX_QUEUE
static int send_directpath_tx_queue(void);
#endif

/*
 *  Local MAC Address Tracking Functions
 */
static INLINE void register_netdev_event_handler(void);
static INLINE void unregister_netdev_event_handler(void);
static int netdev_event_handler(struct notifier_block *, unsigned long, void *);

/*
 *  External Functions
 */
extern void (*ppa_hook_mpoa_setup)(struct atm_vcc *, int, int);
#if defined(CONFIG_IFX_OAM) || defined(CONFIG_IFX_OAM_MODULE)
extern void ifx_push_oam(unsigned char *);
#else
static inline void ifx_push_oam(unsigned char *dummy __attribute__((unused))) {
}
#endif
#if defined(CONFIG_IFXMIPS_DSL_CPE_MEI) || defined(CONFIG_IFXMIPS_DSL_CPE_MEI_MODULE)
#if !defined(ENABLE_LED_FRAMEWORK) || !ENABLE_LED_FRAMEWORK
extern int ifx_mei_atm_led_blink(void) __attribute__ ((weak));
#endif
extern int ifx_mei_atm_showtime_check(int *is_showtime,
		struct port_cell_info *port_cell, void **xdata_addr) __attribute__ ((weak));
#else
#if !defined(ENABLE_LED_FRAMEWORK) || !ENABLE_LED_FRAMEWORK
static inline int ifx_mei_atm_led_blink(void) {return IFX_SUCCESS;}
#endif
static inline int ifx_mei_atm_showtime_check(int *is_showtime, struct port_cell_info *port_cell, void **xdata_addr)
{
	if ( is_showtime != NULL )
	*is_showtime = 0;
	return IFX_SUCCESS;
}
#endif

/*
 *  External Variables
 */
extern struct sk_buff* (*ifx_atm_alloc_tx)(struct atm_vcc *, unsigned int);
#if defined(CONFIG_IFXMIPS_DSL_CPE_MEI) || defined(CONFIG_IFXMIPS_DSL_CPE_MEI_MODULE)
extern int (*ifx_mei_atm_showtime_enter)(struct port_cell_info *, void *) __attribute__ ((weak));
extern int (*ifx_mei_atm_showtime_exit)(void) __attribute__ ((weak));
#else
int (*ifx_mei_atm_showtime_enter)(struct port_cell_info *, void *) = NULL;
EXPORT_SYMBOL(ifx_mei_atm_showtime_enter);
int (*ifx_mei_atm_showtime_exit)(void) = NULL;
EXPORT_SYMBOL(ifx_mei_atm_showtime_exit);
#error
#endif



/*
 * ####################################
 *            Local Variable
 * ####################################
 */

static int g_fwcode = FWCODE_ROUTING_ACC_A5;

static struct semaphore g_sem; //  lock used by open/close function

static int g_eth_wan_mode = 0;
static unsigned int g_wan_itf = 3 << 6;

static int g_ipv6_acc_en = 1;

static unsigned int g_ipv4_lan_collision_rout_fwdc_table_base;
static unsigned int g_ipv4_wan_collision_rout_fwda_table_base;

static int g_wanqos_en = 0;

static int g_cpu_to_wan_tx_desc_pos = 0;
static int g_cpu_to_wan_swap_desc_pos = 0;
static int g_wan_rx1_desc_pos = 0;
struct sk_buff *g_wan_rx1_desc_skb[WAN_RX_DESC_NUM(1)] = { 0 };

static u32 g_mailbox_signal_mask;

static int g_f_dma_opened = 0;

static struct atm_priv_data g_atm_priv_data;

static struct atmdev_ops g_ppe_atm_ops = { owner: THIS_MODULE,
  open: ppe_open,
  close: ppe_close,
  ioctl: ppe_ioctl,
  send: ppe_send,
  send_oam: ppe_send_oam,
  change_qos: ppe_change_qos,
};


#ifdef CONFIG_IFXMIPS_MEI_FW_LOOPBACK
  static int g_showtime = 1;
#else
  static int g_showtime = 0;
#endif
static void *g_xdata_addr = NULL;

static u32 g_dsl_wrx_correct_pdu = 0;
static u32 g_dsl_wtx_total_pdu = 0;
static struct timer_list g_dsl_led_polling_timer;
#if defined(ENABLE_LED_FRAMEWORK) && ENABLE_LED_FRAMEWORK && defined(CONFIG_IFX_LED)
static void *g_data_led_trigger = NULL;
#endif

//static struct eth_priv_data g_eth_priv_data[2];

static int g_eth_prio_queue_map[MAX_ETH_INF][8];
static int g_wan_queue_class_map[8];

static unsigned int g_mib_itf = 0xC3;

//  1:switch forward packets between port0/1 and cpu port directly w/o processing
//  0:pkts go through normal path and are processed by switch central function
static int g_ingress_direct_forward_enable = 0;
static int g_egress_direct_forward_enable = 1;

static int g_pmac_ewan = 0;

/*
 *  variable for directpath
 */
static int g_directpath_dma_full = 0;

#if defined(ENABLE_CONFIGURABLE_DSL_VLAN) && ENABLE_CONFIGURABLE_DSL_VLAN
static unsigned short g_dsl_vlan_qid_map[ATM_QUEUE_NUMBER * 2] = { 0 };
#endif

static struct notifier_block g_netdev_event_handler_nb = { 0 };

#if defined(DEBUG_SKB_SWAP) && DEBUG_SKB_SWAP
  static struct sk_buff *g_dbg_skb_swap_pool[1024] = {0};
#endif

#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC
static struct net_device *g_mirror_netdev = NULL;
#endif

/*
 * ####################################
 *           Global Variable
 * ####################################
 */

/*
 *  variable for directpath
 */
#if defined(ENABLE_DIRECTPATH_TX_QUEUE)
// replaced by common ppa_push_through_data:
// struct ppe_directpath_data g_ppe_directpath_data[3];    //  port 0 - eth0, 1 - eth1, 2 - cpu, 6 - dsl (PPPoA/IPoA), 7 - dsl (EoA)
#error AVM_CONCEPT doesnet support DIRECTPATH_TX_QUEUE so far
#endif

/*
 * ####################################
 *            Local Function
 * ####################################
 */

/*
 *  Description:
 *    Handle all ioctl command. This is the only way, which user level could
 *    use to access PPE driver.
 *  Input:
 *    inode --- struct inode *, file descriptor on drive
 *    file  --- struct file *, file descriptor of virtual file system
 *    cmd   --- unsigned int, device specific commands.
 *    arg   --- unsigned long, pointer to a structure to pass in arguments or
 *              pass out return value.
 *  Output:
 *    int   --- 0:    Success
 *              else: Error Code
 */
static int ppe_ioctl(struct atm_dev *dev __attribute__((unused)),
		unsigned int cmd, void *arg) {
	int ret = 0;
	atm_cell_ifEntry_t mib_cell;
	atm_aal5_ifEntry_t mib_aal5;
	atm_aal5_vcc_x_t mib_vcc;
	u32 value;
	int conn;

	dbg("cmd=%#x", cmd);
	if (_IOC_TYPE(cmd) != PPE_ATM_IOC_MAGIC || _IOC_NR(cmd) >= PPE_ATM_IOC_MAXNR)
		return -ENOTTY;

	if (_IOC_DIR(cmd) & _IOC_READ)
		ret = !access_ok(VERIFY_WRITE, arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		ret = !access_ok(VERIFY_READ, arg, _IOC_SIZE(cmd));
	if (ret)
		return -EFAULT;

	switch (cmd) {
	case PPE_ATM_MIB_CELL: /*  cell level  MIB */
		/*  These MIB should be read at ARC side, now put zero only.    */
		mib_cell.ifHCInOctets_h = 0;
		mib_cell.ifHCInOctets_l = 0;
		mib_cell.ifHCOutOctets_h = 0;
		mib_cell.ifHCOutOctets_l = 0;
		mib_cell.ifInErrors = 0;
		mib_cell.ifInUnknownProtos = DSL_WAN_MIB_TABLE->wrx_drophtu_cell;
		mib_cell.ifOutErrors = 0;

		ret = sizeof(mib_cell) - copy_to_user(arg, &mib_cell, sizeof(mib_cell));
		break;
	case PPE_ATM_MIB_AAL5: /*  AAL5 MIB    */
		value = DSL_WAN_MIB_TABLE->wrx_total_byte;
		u64_add_u32(g_atm_priv_data.wrx_total_byte,
				value - g_atm_priv_data.prev_mib.wrx_total_byte,
				&g_atm_priv_data.wrx_total_byte);
		g_atm_priv_data.prev_mib.wrx_total_byte = value;
		mib_aal5.ifHCInOctets_h = g_atm_priv_data.wrx_total_byte.h;
		mib_aal5.ifHCInOctets_l = g_atm_priv_data.wrx_total_byte.l;

		value = DSL_WAN_MIB_TABLE->wtx_total_byte;
		u64_add_u32(g_atm_priv_data.wtx_total_byte,
				value - g_atm_priv_data.prev_mib.wtx_total_byte,
				&g_atm_priv_data.wtx_total_byte);
		g_atm_priv_data.prev_mib.wtx_total_byte = value;
		mib_aal5.ifHCOutOctets_h = g_atm_priv_data.wtx_total_byte.h;
		mib_aal5.ifHCOutOctets_l = g_atm_priv_data.wtx_total_byte.l;

		mib_aal5.ifInUcastPkts = g_atm_priv_data.wrx_pdu;
		mib_aal5.ifOutUcastPkts = DSL_WAN_MIB_TABLE->wtx_total_pdu;
		mib_aal5.ifInErrors = DSL_WAN_MIB_TABLE->wrx_err_pdu;
		mib_aal5.ifInDiscards = DSL_WAN_MIB_TABLE->wrx_dropdes_pdu
				+ g_atm_priv_data.wrx_drop_pdu;
		mib_aal5.ifOutErros = g_atm_priv_data.wtx_err_pdu;
		mib_aal5.ifOutDiscards = g_atm_priv_data.wtx_drop_pdu;

		ret = sizeof(mib_aal5) - copy_to_user(arg, &mib_aal5, sizeof(mib_aal5));
		break;
	case PPE_ATM_MIB_VCC: /*  VCC related MIB */
		copy_from_user(&mib_vcc, arg, sizeof(mib_vcc));
		conn = find_vpivci(mib_vcc.vpi, mib_vcc.vci);
		if (conn >= 0) {
			mib_vcc.mib_vcc.aal5VccCrcErrors = 0; //  no support any more, g_atm_priv_data.connection[conn].aal5_vcc_crc_err;
			mib_vcc.mib_vcc.aal5VccOverSizedSDUs = 0; //  no support any more, g_atm_priv_data.connection[conn].aal5_vcc_oversize_sdu;
			mib_vcc.mib_vcc.aal5VccSarTimeOuts = 0; /*  no timer support    */
			ret = sizeof(mib_vcc) - copy_to_user(arg, &mib_vcc, sizeof(mib_vcc));
		} else
			ret = -EINVAL;
		break;
	case PPE_ATM_MAP_PKT_PRIO_TO_Q: {
		struct ppe_prio_q_map cmd;

		if (copy_from_user(&cmd, arg, sizeof(cmd)))
			return -EFAULT;

		if (cmd.qid < 0 || cmd.pkt_prio < 0 || cmd.pkt_prio >= 8)
			return -EINVAL;

		conn = find_vpivci(cmd.vpi, cmd.vci);
		if (conn < 0)
			return -EINVAL;

		return sw_tx_prio_to_queue(conn, cmd.pkt_prio, cmd.qid, NULL);
	}
		break;
	case PPE_ATM_TX_Q_OP: {
		struct tx_q_op cmd;

		if (copy_from_user(&cmd, arg, sizeof(cmd)))
			return -EFAULT;

		if (!(cmd.flags & PPE_ATM_TX_Q_OP_CHG_MASK))
			return -EFAULT;

		conn = find_vpivci(cmd.vpi, cmd.vci);
		if (conn < 0)
			return -EINVAL;

		if ((cmd.flags & PPE_ATM_TX_Q_OP_ADD))
			return sw_tx_queue_add(conn);
		else
			return sw_tx_queue_del(conn);
	}
		break;
	case PPE_ATM_GET_MAP_PKT_PRIO_TO_Q: {
		struct ppe_prio_q_map_all cmd;
		u32 sw_tx_queue_table;
		int sw_tx_queue_to_virt_queue_map[ATM_SW_TX_QUEUE_NUMBER] = { 0 };
		int sw_tx_queue;
		int virt_queue;
		int i;

		if (copy_from_user(&cmd, arg, sizeof(cmd)))
			return -EFAULT;

		conn = find_vpivci(cmd.vpi, cmd.vci);
		if (conn >= 0) {
			for (sw_tx_queue_table =
					g_atm_priv_data.connection[conn].sw_tx_queue_table, sw_tx_queue =
					virt_queue = 0; sw_tx_queue_table != 0;
					sw_tx_queue_table >>= 1, sw_tx_queue++)
				if ((sw_tx_queue_table & 0x01)) {
					sw_tx_queue_to_virt_queue_map[sw_tx_queue] = virt_queue++;
				}

			cmd.total_queue_num = virt_queue;
			for (i = 0; i < 8; i++) {
				cmd.pkt_prio[i] = i;
				cmd.qid[i] =
						sw_tx_queue_to_virt_queue_map[g_atm_priv_data.connection[conn].prio_queue_map[i]];
			}

			ret = sizeof(cmd) - copy_to_user(arg, &cmd, sizeof(cmd));
		} else
			ret = -EINVAL;
	}
		break;
	default:
		ret = -ENOIOCTLCMD;
	}

	return ret;
}

static int ppe_open(struct atm_vcc *vcc) {
	int ret;
	short vpi = vcc->vpi;
	int vci = vcc->vci;
	struct atm_port *port = &g_atm_priv_data.port[(int) vcc->dev->dev_data];
	unsigned int sw_tx_queue;
    unsigned int conn;
    int conn_vpi_vci;
	int f_enable_irq = 0;
	struct wrx_queue_config wrx_queue_config = { 0 };
	struct wtx_queue_config wtx_queue_config = { 0 };
	struct uni_cell_header *cell_header;
	int i;

	dbg("vpi %d, vci %d",vpi, vci);

	if (g_eth_wan_mode) {
		err("ETH WAN mode: g_eth_wan_mode = %d", g_eth_wan_mode);
		return -EIO;
	}

	if (vcc->qos.aal != ATM_AAL5 && vcc->qos.aal != ATM_AAL0)
		return -EPROTONOSUPPORT;

	down(&g_sem);

	/*  check bandwidth */
	if ((vcc->qos.txtp.traffic_class == ATM_CBR
			&& vcc->qos.txtp.max_pcr
					> (port->tx_max_cell_rate - port->tx_used_cell_rate))
			|| (vcc->qos.txtp.traffic_class == ATM_VBR_RT
					&& vcc->qos.txtp.max_pcr
							> (port->tx_max_cell_rate - port->tx_used_cell_rate))
#if defined(DISABLE_VBR) && DISABLE_VBR
			|| (vcc->qos.txtp.traffic_class == ATM_VBR_NRT && vcc->qos.txtp.pcr > (port->tx_max_cell_rate - port->tx_used_cell_rate))
#else
			|| (vcc->qos.txtp.traffic_class == ATM_VBR_NRT
					&& vcc->qos.txtp.scr
							> (port->tx_max_cell_rate - port->tx_used_cell_rate))
#endif
			|| (vcc->qos.txtp.traffic_class == ATM_UBR_PLUS
					&& vcc->qos.txtp.min_pcr
							> (port->tx_max_cell_rate - port->tx_used_cell_rate))) {
		err("exceed TX line rate");
		ret = -EINVAL;
		goto PPE_OPEN_EXIT;
	}

	/*  check existing vpi,vci  */
	conn_vpi_vci = find_vpivci(vpi, vci);
	if (conn_vpi_vci >= 0) {
		dbg("existing PVC: %d (vpi=%d, vci=%d)", conn_vpi_vci, (int)vpi, vci);
		ret = -EADDRINUSE;
		goto PPE_OPEN_EXIT;
	} else {
		dbg("no existing PVC (vpi=%d, vci=%d)", (int)vpi, vci);
	}




	/*  check whether it need to enable irq */
	if (g_atm_priv_data.connection_table == 0)
		f_enable_irq = 1;

	/*  allocate software TX queue  */
	for (sw_tx_queue = 0; sw_tx_queue < g_atm_priv_data.max_sw_tx_queues;
			sw_tx_queue++)
		if (!(g_atm_priv_data.sw_tx_queue_table & (1 << sw_tx_queue)))
			break;
	if (sw_tx_queue == g_atm_priv_data.max_sw_tx_queues) {
		err("no free TX queue");
		ret = -EINVAL;
		goto PPE_OPEN_EXIT;
	}

	/*  allocate connection */
	for (conn = 0; conn < g_atm_priv_data.max_connections; conn++)
		if (!(g_atm_priv_data.connection_table & (1 << conn))) {
			g_atm_priv_data.connection_table |= 1 << conn;
			g_atm_priv_data.sw_tx_queue_table |= 1 << sw_tx_queue;
			g_atm_priv_data.connection[conn].vcc = vcc;
			g_atm_priv_data.connection[conn].port = (u32) vcc->dev->dev_data;
			g_atm_priv_data.connection[conn].sw_tx_queue_table = 1
					<< sw_tx_queue;
			for (i = 0; i < 8; i++)
				g_atm_priv_data.connection[conn].prio_queue_map[i] =
						sw_tx_queue;
			break;
		}
	if (conn == g_atm_priv_data.max_connections) {
		err("exceed PVC limit");
		ret = -EINVAL;
		goto PPE_OPEN_EXIT;
	}

	/*  reserve bandwidth   */
	switch (vcc->qos.txtp.traffic_class) {
	case ATM_CBR:
	case ATM_VBR_RT:
		port->tx_used_cell_rate += vcc->qos.txtp.max_pcr;
		break;
	case ATM_VBR_NRT:
#if defined(DISABLE_VBR) && DISABLE_VBR
		port->tx_used_cell_rate += vcc->qos.txtp.pcr;
#else
		port->tx_used_cell_rate += vcc->qos.txtp.scr;
#endif
		break;
	case ATM_UBR_PLUS:
		port->tx_used_cell_rate += vcc->qos.txtp.min_pcr;
		break;
	}

	/*  setup RX queue cfg and TX queue cfg */
	wrx_queue_config.new_vlan = 0x0FF0 | conn; //  use vlan to differentiate PVCs
	wrx_queue_config.vlan_ins = 1; // use vlan to differentiate PVCs, please use outer VLAN in routing table, and do remember to configure switch to handle bridging case
	wrx_queue_config.mpoa_type = 3; //TODO EoA wo FCS wurde von uns geandert ( davor 3 = IPoA)
	wrx_queue_config.ip_ver = 0; //  set IPv4 as default
	wrx_queue_config.mpoa_mode = 0; //TODO 0: VC-mux war das orginal
	wrx_queue_config.oversize = aal5r_max_packet_size;
	wrx_queue_config.undersize = aal5r_min_packet_size;
	wrx_queue_config.mfs = aal5r_max_packet_size;

	wtx_queue_config.same_vc_qmap = 0x00; //  only one TX queue is assigned now, use ioctl to add other TX queues
	wtx_queue_config.sbid = g_atm_priv_data.connection[conn].port;
	wtx_queue_config.qsb_vcid = conn + QSB_QUEUE_NUMBER_BASE; //  qsb qid = firmware qid + 1
	wtx_queue_config.mpoa_mode = 0; //TODO  0: VC-mux war das orginal
	wtx_queue_config.qsben = 1; //  reserved in A4, however put 1 for backward compatible

	cell_header = (struct uni_cell_header *) ((u32 *) &wtx_queue_config + 2);
	cell_header->clp = (vcc->atm_options & ATM_ATMOPT_CLP) ? 1 : 0;
	cell_header->pti = ATM_PTI_US0;
	cell_header->vci = vci;
	cell_header->vpi = vpi;
	cell_header->gfc = 0;

	*WRX_QUEUE_CONFIG(conn) = wrx_queue_config;
	*WTX_QUEUE_CONFIG(sw_tx_queue) = wtx_queue_config;

	/*  set qsb */
	set_qsb(vcc, &vcc->qos, conn);

	/*  update atm_vcc structure    */
	vcc->itf = (int) vcc->dev->dev_data;
	set_bit(ATM_VF_READY, &vcc->flags);

	if (f_enable_irq) {
		ifx_atm_alloc_tx = atm_alloc_tx;
		// turn_on_dma_rx(31); dma laueft immer

		g_dsl_wrx_correct_pdu = DSL_WAN_MIB_TABLE->wrx_correct_pdu;
		g_dsl_wtx_total_pdu = DSL_WAN_MIB_TABLE->wtx_total_pdu;
		g_dsl_led_polling_timer.expires = jiffies + HZ / 10; //  100ms
		add_timer(&g_dsl_led_polling_timer);
	}

	/*  set htu entry   */
	set_htu_entry(vpi, vci, conn, vcc->qos.aal == ATM_AAL5 ? 1 : 0);

	dbg("ppe_open(%d.%d): conn = %d", vcc->vpi, vcc->vci, conn);
	up(&g_sem);
	dbg("get dslwan_qid_with_vcc qid=%d", ifx_ppa_drv_get_dslwan_qid_with_vcc_hook(vcc, 0));
	return 0;

	PPE_OPEN_EXIT: up(&g_sem);
	return ret;
}

static void ppe_close(struct atm_vcc *vcc) {
	int conn;
	struct atm_port *port;
	volatile struct atm_connection *connection;
	unsigned int i;

	if (vcc == NULL)
		return;

	down(&g_sem);

	/*  get connection id   */
	conn = find_vcc(vcc);
	if (conn < 0) {
		err("can't find vcc");
		goto PPE_CLOSE_EXIT;
	}
	dbg("ppe_close(%d.%d): conn = %d", vcc->vpi, vcc->vci, conn);
	connection = &g_atm_priv_data.connection[conn];
	port = &g_atm_priv_data.port[connection->port];

	/*  clear htu   */
	clear_htu_entry(conn);

	/*  release connection  */
	g_atm_priv_data.connection_table &= ~(1 << conn);
	g_atm_priv_data.sw_tx_queue_table &= ~(connection->sw_tx_queue_table);
	memset(connection, 0, sizeof(*connection));

	/*  disable irq */
	if (g_atm_priv_data.connection_table == 0) {
		del_timer(&g_dsl_led_polling_timer);

		// turn_off_dma_rx(31); dma laueft immer
		ifx_atm_alloc_tx = NULL;
	}

	/*  release bandwidth   */
	switch (vcc->qos.txtp.traffic_class) {
	case ATM_CBR:
	case ATM_VBR_RT:
		port->tx_used_cell_rate -= vcc->qos.txtp.max_pcr;
		break;
	case ATM_VBR_NRT:
#if defined(DISABLE_VBR) && DISABLE_VBR
		port->tx_used_cell_rate -= vcc->qos.txtp.pcr;
#else
		port->tx_used_cell_rate -= vcc->qos.txtp.scr;
#endif
		break;
	case ATM_UBR_PLUS:
		port->tx_used_cell_rate -= vcc->qos.txtp.min_pcr;
		break;
	}

	/*  idle for a while to let parallel operation finish   */
	for (i = 0; i < IDLE_CYCLE_NUMBER; i++)
		;

#if defined(ENABLE_CONFIGURABLE_DSL_VLAN) && ENABLE_CONFIGURABLE_DSL_VLAN
	for (i = 0; i < NUM_ENTITY(g_dsl_vlan_qid_map); i += 2)
		if (g_dsl_vlan_qid_map[i] == WRX_QUEUE_CONFIG(conn)->new_vlan) {
			g_dsl_vlan_qid_map[i] = 0;
			break;
		}
#endif

	PPE_CLOSE_EXIT: up(&g_sem);
}

static int ppe_send(struct atm_vcc *vcc, struct sk_buff *skb) {
	int ret;
	int conn;
	unsigned long sys_flag;
	volatile struct tx_descriptor *desc;
	struct tx_descriptor reg_desc;
	struct sk_buff *skb_to_free;
	int byteoff;
	u32 priority;

	if (vcc == NULL || skb == NULL)
		return -EINVAL;

	if (skb->len > aal5s_max_packet_size)
		return -EOVERFLOW;

	ATM_SKB(skb)->vcc = vcc;

    if ( g_stop_datapath != 0 )
    {
        atm_free_tx_skb_vcc(skb);
        return -EBUSY;
    }

	skb_get(skb);
	atm_free_tx_skb_vcc(skb);
	ATM_SKB(skb)->vcc = NULL;

	conn = find_vcc(vcc);
	if (conn < 0) {
		err("not find vcc");
		ret = -EINVAL;
		goto FIND_VCC_FAIL;
	}

	if (!g_showtime) {
		err("not in showtime");
		ret = -EIO;
		goto CHECK_SHOWTIME_FAIL;
	}

	priority = (skb->priority & TC_H_MIN_MASK);
	if (priority >= 8)
		priority = 7;

	/*  reserve space to put pointer in skb */
	byteoff = (u32) skb->data & (DMA_TX_ALIGNMENT - 1);
	if (skb_headroom(skb) < sizeof(struct sk_buff *) + byteoff
			|| skb->end - skb->data < 1564 /* 1518 (max ETH frame) + 22 (reserved for firmware) + 10 (AAL5) + 6 (PPPoE) + 4 x 2 (2 VLAN) */
			|| skb_shared(skb) || skb_cloned(skb)
			|| ( byteoff > 31 ) ) {
		//  this should be few case

		struct sk_buff *new_skb;

		dma_alignment_atm_bad_count++;

		ASSERT(byteoff < 32, "dma-byteoff too big: %d ", byteoff);
		ASSERT(!skb_shared(skb),"skb_shared");
		ASSERT(!skb_cloned(skb),"skb_cloned");

		ASSERT(skb_headroom(skb) >= sizeof(struct sk_buff *) + byteoff, "skb_headroom(skb){%d} < sizeof(struct sk_buff *){%d} + byteoff{%d}"
				, skb_headroom(skb), sizeof(struct sk_buff*), byteoff);

		ASSERT(skb->end - skb->data >= 1564, "skb->end{%#x} - skb->data{%#x} < 1518 + 22 + 10 + 6 + 4 x 2; diff=%d",
				(unsigned int)skb->end, (unsigned int)skb->data, (unsigned int)skb->end - (unsigned int)skb->data);

		ASSERT (0, "skb_headroom(skb){%d}, sizeof(struct sk_buff *){%d}, byteoff{%d}" , skb_headroom(skb), sizeof(struct sk_buff*), byteoff);

#if defined(CONFIG_AVMNET_DEBUG) 
		if ( g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_ALIGNMENT )
			printk( KERN_ERR
				"\n"
				"   skb_headroom :   %d \n"
				"   skb_tailroom :   %d \n"
				"   skb_len      :   %d \n"
				"   skb_total    :   %d \n"
				"   sizeof(struct sk_buff *): %d \n"
				"   byteoff:         %d \n"
				"   skb->head:       %p\n"
				"   skb->data:       %p\n"
				"   in_dev:          %s\n",
					skb_headroom(skb),
					skb_tailroom(skb),
					skb->len,
					skb->end - skb->head,
					sizeof(struct sk_buff *),
					byteoff,
					skb->head,
					skb->data,
					skb->input_dev?skb->input_dev->name:"no_input_device");
#endif

		new_skb = alloc_skb_tx(
				skb->len < DMA_PACKET_SIZE ? DMA_PACKET_SIZE : skb->len); //  may be used by RX fastpath later
		if (!new_skb) {
			err("no memory");
			ret = -ENOMEM;
			goto ALLOC_SKB_TX_FAIL;
		}
		skb_put(new_skb, skb->len);
		my_memcpy(new_skb->data, skb->data, skb->len);
		atm_free_tx_skb_vcc(skb);
		/* AVMGRM: priority wegen QoS auch kopieren */
		new_skb->priority = skb->priority;
		skb = new_skb;
		byteoff = (u32) skb->data & (DMA_TX_ALIGNMENT - 1);
#ifndef CONFIG_MIPS_UNCACHED
		/*  write back to physical memory   */
		dma_cache_wback((u32)skb->data, skb->len);
#endif
	} else {
	    skb_orphan(skb);
		*(struct sk_buff **) ((u32) skb->data - byteoff
				- sizeof(struct sk_buff *)) = skb;
#ifndef CONFIG_MIPS_UNCACHED
		/*  write back to physical memory   */
		dma_cache_wback((u32)skb->data - byteoff - sizeof(struct sk_buff *),
				skb->len + byteoff + sizeof(struct sk_buff *));
#endif
		dma_alignment_atm_good_count++;
#if defined(CONFIG_AVMNET_DEBUG) 
		if ( g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_ALIGNMENT )
			printk( KERN_ERR "len ok: total_skb=%d\n", skb->end - skb->head );
#endif
	}

	/*  allocate descriptor */
	local_irq_save(sys_flag);
	desc = CPU_TO_WAN_TX_DESC_BASE + g_cpu_to_wan_tx_desc_pos;
	if (!desc->own) //  PPE hold
	{
		local_irq_restore(sys_flag);
		ret = -EIO;
		err("PPE hold");
		goto NO_FREE_DESC;
	}
	if (++g_cpu_to_wan_tx_desc_pos == CPU_TO_WAN_TX_DESC_NUM)
		g_cpu_to_wan_tx_desc_pos = 0;
	local_irq_restore(sys_flag);

	/*  load descriptor from memory */
	reg_desc = *desc;

	/*  free previous skb   */
	ASSERT((reg_desc.dataptr & (DMA_TX_ALIGNMENT - 1)) == 0,
			"reg_desc.dataptr (0x%#x) must be 8 DWORDS aligned",
			reg_desc.dataptr);
	skb_to_free = get_skb_pointer(reg_desc.dataptr);
	dev_kfree_skb_any(skb_to_free);
	put_skb_to_dbg_pool(skb);

    /*  detach from protocol    */
    skb_to_free = skb;
    skb = skb_break_away_from_protocol(skb);
    dev_kfree_skb_any(skb_to_free);

    put_skb_to_dbg_pool(skb);
#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC
	if (g_mirror_netdev != NULL) {
		if ((WRX_QUEUE_CONFIG(conn)->mpoa_type & 0x02) == 0) //  handle EoA only, TODO: IPoA/PPPoA
				{
			struct sk_buff *new_skb = skb_copy(skb, GFP_ATOMIC);

			if (new_skb != NULL) {
				if (WRX_QUEUE_CONFIG(conn)->mpoa_mode) //  LLC
					skb_pull(new_skb, 10); //  remove EoA header
				else
					//  VC mux
					skb_pull(new_skb, 2); //  remove EoA header
				if (WRX_QUEUE_CONFIG(conn)->mpoa_type == 1) //  remove FCS
					skb_trim(new_skb, new_skb->len - 4);
				new_skb->dev = g_mirror_netdev;
				dev_queue_xmit(new_skb);
			}
		}
	}
#endif

	ASSERT(byteoff < 32, "dma-byteoff still too big: %d ", byteoff);

	/*  update descriptor   */
	reg_desc.dataptr = (u32) skb->data & (0x1FFFFFFF ^ (DMA_TX_ALIGNMENT - 1)); //  byte address
	reg_desc.byteoff = byteoff;
	reg_desc.datalen = skb->len;
	reg_desc.mpoa_pt = 1;
	reg_desc.mpoa_type = 0; //  this flag should be ignored by firmware
	reg_desc.pdu_type = vcc->qos.aal == ATM_AAL5 ? 0 : 1;
	reg_desc.qid = g_atm_priv_data.connection[conn].prio_queue_map[priority];
	reg_desc.own = 0;
	reg_desc.c = 0;
	reg_desc.sop = reg_desc.eop = 1;

	/*  update MIB  */
	UPDATE_VCC_STAT(conn, tx_packets, 1);
	UPDATE_VCC_STAT(conn, tx_bytes, skb->len);

	if (vcc->stats)
		atomic_inc(&vcc->stats->tx);
	if (vcc->qos.aal == ATM_AAL5)
		g_atm_priv_data.wtx_pdu++;

	g_atm_priv_data.connection[conn].prio_tx_packets[priority]++;

	dump_atm_tx_desc( &reg_desc );
#if defined(CONFIG_AVMNET_DEBUG) 
	if (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_TX)
		dump_skb(skb, DUMP_SKB_LEN, "ppe_send", 7, reg_desc.qid, 1, 1);
#endif

#if defined(ENABLE_DBG_PROC) && ENABLE_DBG_PROC
	if ((g_dbg_datapath & DBG_ENABLE_MASK_MAC_SWAP)) {
		unsigned char tmp[8];
		unsigned char *p = skb->data;

		p += 10; //  ignore AAL5 header

		if (!(p[0] & 0x01)) { //  bypass broadcast/multicast
			//  swap MAC
			memcpy(tmp, p, 6);
			memcpy(p, p + 6, 6);
			memcpy(p + 6, tmp, 6);
			p += 12;

			//  bypass VLAN
			while (p[0] == 0x81 && p[1] == 0x00)
				p += 4;

			//  IP
			if (p[0] == 0x08 && p[1] == 0x00) {
				p += 14;
				memcpy(tmp, p, 4);
				memcpy(p, p + 4, 4);
				memcpy(p + 4, tmp, 4);
				p += 8;
			}

			dma_cache_wback((unsigned long)skb->data,
					(unsigned long)p - (unsigned long)skb->data);
		}
	}
#endif

	/*  write discriptor to memory and write back cache */
	*((volatile u32 *) desc + 1) = *((u32 *) &reg_desc + 1);
	*(volatile u32 *) desc = *(u32 *) &reg_desc;

	adsl_led_flash();

	return 0;

	FIND_VCC_FAIL: if (vcc->stats) {
		atomic_inc(&vcc->stats->tx_err);
	}
	if (vcc->qos.aal == ATM_AAL5) {
		g_atm_priv_data.wtx_err_pdu++;
		g_atm_priv_data.wtx_drop_pdu++;
	}
	atm_free_tx_skb_vcc(skb);
	return ret;

	CHECK_SHOWTIME_FAIL: ALLOC_SKB_TX_FAIL: if (vcc->stats)
		atomic_inc(&vcc->stats->tx_err);
	if (vcc->qos.aal == ATM_AAL5)
		g_atm_priv_data.wtx_drop_pdu++;
	UPDATE_VCC_STAT(conn, tx_dropped, 1);
	atm_free_tx_skb_vcc(skb);
	return ret;

	NO_FREE_DESC: if (vcc->stats)
		atomic_inc(&vcc->stats->tx_err);
	if (vcc->qos.aal == ATM_AAL5)
		g_atm_priv_data.wtx_drop_pdu++;
	UPDATE_VCC_STAT(conn, tx_dropped, 1);
	atm_free_tx_skb_vcc(skb);
	return ret;
}

static int ppe_send_oam(struct atm_vcc *vcc, void *cell, int flags __attribute__((unused))) {
	int conn;
	struct uni_cell_header *uni_cell_header = (struct uni_cell_header *) cell;
	struct sk_buff *skb;
	unsigned long sys_flag;
	volatile struct tx_descriptor *desc;
	struct tx_descriptor reg_desc;
	struct sk_buff *skb_to_free;

	if (((uni_cell_header->pti == ATM_PTI_SEGF5
			|| uni_cell_header->pti == ATM_PTI_E2EF5)
			&& find_vpivci(uni_cell_header->vpi, uni_cell_header->vci) < 0)
			|| ((uni_cell_header->vci == 0x03 || uni_cell_header->vci == 0x04)
					&& find_vpi(uni_cell_header->vpi) < 0))
		return -EINVAL;

	if (!g_showtime) {
		err("not in showtime");
		return -EIO;
	}

	/*  find queue ID   */
	conn = find_vcc(vcc);
	if (conn < 0) {
		err("OAM not find queue");
		return -EINVAL;
	}

	/*  allocate sk_buff    */
	skb = alloc_skb_tx(DMA_PACKET_SIZE); //  may be used by RX fastpath later
	if (skb == NULL)
		return -ENOMEM;

	/*  copy data   */
	skb_put(skb, CELL_SIZE);
	my_memcpy(skb->data, cell, CELL_SIZE);

#ifndef CONFIG_MIPS_UNCACHED
	/*  write back to physical memory   */
	dma_cache_wback((u32)skb->data, skb->len);
#endif

	/*  allocate descriptor */
	local_irq_save(sys_flag);
	desc = CPU_TO_WAN_TX_DESC_BASE + g_cpu_to_wan_tx_desc_pos;
	if (!desc->own) //  PPE hold
	{
		local_irq_restore(sys_flag);
		err("OAM not alloc tx connection");
		return -EIO;
	}
	if (++g_cpu_to_wan_tx_desc_pos == CPU_TO_WAN_TX_DESC_NUM)
		g_cpu_to_wan_tx_desc_pos = 0;
	local_irq_restore(sys_flag);

	/*  load descriptor from memory */
	reg_desc = *desc;

	/*  free previous skb   */
	ASSERT((reg_desc.dataptr & (DMA_TX_ALIGNMENT - 1)) == 0,
			"reg_desc.dataptr (0x%#x) must be 8 DWORDS aligned",
			reg_desc.dataptr);
	skb_to_free = get_skb_pointer(reg_desc.dataptr);
	dev_kfree_skb_any(skb_to_free);
	put_skb_to_dbg_pool(skb);

	/*  setup descriptor    */
	reg_desc.dataptr = (u32) skb->data & (0x1FFFFFFF ^ (DMA_TX_ALIGNMENT - 1)); //  byte address
	reg_desc.byteoff = (u32) skb->data & (DMA_TX_ALIGNMENT - 1);
	reg_desc.datalen = skb->len;
	reg_desc.mpoa_pt = 1;
	reg_desc.mpoa_type = 0; //  this flag should be ignored by firmware
	reg_desc.pdu_type = 1; //  cell
	reg_desc.qid = g_atm_priv_data.connection[conn].prio_queue_map[7]; //  expect priority '7' is highest priority
	reg_desc.own = 0;
	reg_desc.c = 0;
	reg_desc.sop = reg_desc.eop = 1;

	dump_atm_tx_desc(&reg_desc);
#if defined(CONFIG_AVMNET_DEBUG) 
	if (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_TX)
		dump_skb(skb, DUMP_SKB_LEN, "ppe_send_oam", 7, reg_desc.qid, 1, 1);
#endif

	/*  write discriptor to memory and write back cache */
	*((volatile u32 *) desc + 1) = *((u32 *) &reg_desc + 1);
	*(volatile u32 *) desc = *(u32 *) &reg_desc;

	adsl_led_flash();

	return 0;
}

static int ppe_change_qos(struct atm_vcc *vcc, struct atm_qos *qos, int flags __attribute__((unused))) {
	int conn;

	if (vcc == NULL || qos == NULL)
		return -EINVAL;

	conn = find_vcc(vcc);
	if (conn < 0)
		return -EINVAL;

	set_qsb(vcc, qos, conn);

	return 0;
}

#ifdef AVMNET_VCC_QOS

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

/*
* returns res:
*   res == 0: queue setup successful
*   res <  0: setup qos queue failure
*/
static int vcc_set_nr_qos_queues_datapath(struct atm_vcc *vcc, unsigned int nr_queues){
	int conn;
	unsigned int new_queue_count = 1;
	int res;

	if (vcc == NULL )
		return -EINVAL;

	conn = find_vcc(vcc);
	if (conn < 0)
		return -EINVAL;
    while( new_queue_count < nr_queues ){

	    // add lower priority queue
        res = sw_tx_queue_add(conn);
	    if (res != 0) {
	        // something went wrong: delete all the queues, we already added
            while( --new_queue_count ){
                // del lowest priority queue
                sw_tx_queue_del(conn);

            }
            return -EINVAL;
	    }
	    ++new_queue_count;
    }

    return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
/*
* returns ret:
*   ret == 0: success
*   ret != 0: failure
*/
int vcc_map_skb_prio_qos_queue_datapath(struct atm_vcc *vcc, unsigned int skb_prio, unsigned int qid){
	int conn;

	if (vcc == NULL )
		return -EINVAL;
	conn = find_vcc(vcc);
	if (conn < 0)
		return -EINVAL;

    return sw_tx_prio_to_queue(conn, skb_prio, qid, NULL);
}

#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/


static void mpoa_setup_conn(int conn, int mpoa_type, int f_llc) {

	u32 sw_tx_queue_table;
	int sw_tx_queue;

	printk(KERN_ERR "%s conn=%d, mpoa_type=%d, mpoa_mode=%d\n ", __func__, conn, mpoa_type, f_llc);
	if (mpoa_hack && mpoa_type == 2 && f_llc == 0){
		printk(KERN_ERR "%s hack!\n", __func__);
		mpoa_type = 3;
		f_llc = 0;
	}

	WRX_QUEUE_CONFIG(conn)->mpoa_type = mpoa_type; //mpoa_type = 0: funktionierts im AVM Netz
	WRX_QUEUE_CONFIG(conn)->mpoa_mode = f_llc;

	for (sw_tx_queue_table = g_atm_priv_data.connection[conn].sw_tx_queue_table, sw_tx_queue =
			0; sw_tx_queue_table != 0; sw_tx_queue_table >>= 1, sw_tx_queue++)
		if ((sw_tx_queue_table & 0x01))
			WTX_QUEUE_CONFIG(sw_tx_queue)->mpoa_mode = f_llc;
}

static void mpoa_setup(struct atm_vcc *vcc, int mpoa_type, int f_llc) { //todo geeignete Aufrufparameter im kdsld waeheln (bei uns funktioniert (vcc, 0, 1)
	int conn;

	printk(KERN_ERR "%s vcc=%p, mpoa_type=%d, mpoa_mode=%d\n ", __func__, vcc, mpoa_type, f_llc);
	conn = find_vcc(vcc);
	if (conn < 0)
		return;

	mpoa_setup_conn(conn, mpoa_type, f_llc);
}

#if 0
int ifx_ppa_eth_init(struct net_device *dev) {
	dbg(KERN_ERR "[%s] %d\n", __FUNCTION__, __LINE__);

	ether_setup(dev); /*  assign some members */

	ifx_ppa_eth_set_mac_address(dev, NULL);
	dbg(KERN_ERR "[%s] %d\n", __FUNCTION__, __LINE__);

	// printk("Activating DMA rx channels\n");
	 // turn_on_dma_rx(1); dma laueft immer

	dbg(KERN_ERR "[%s] %d done\n", __FUNCTION__, __LINE__);
	return 0;
}

static struct net_device_stats *eth_get_stats(struct net_device *dev)
{
	int port;
	struct eth_priv_data *priv = netdev_priv(dev);

	port = get_eth_port(dev);
	if ( port < 0 )
	return NULL;

	priv->stats.rx_packets = priv->rx_packets
	+ ITF_MIB_TBL(port)->ig_fast_brg_pkts
	+ ITF_MIB_TBL(port)->ig_fast_rt_ipv4_udp_pkts
	+ ITF_MIB_TBL(port)->ig_fast_rt_ipv4_tcp_pkts
	+ ITF_MIB_TBL(port)->ig_fast_rt_ipv4_mc_pkts
	+ ITF_MIB_TBL(port)->ig_fast_rt_ipv6_udp_pkts
	+ ITF_MIB_TBL(port)->ig_fast_rt_ipv6_tcp_pkts;

	priv->stats.rx_bytes = priv->rx_bytes
	+ ITF_MIB_TBL(port)->ig_fast_brg_bytes
	+ ITF_MIB_TBL(port)->ig_fast_rt_ipv4_bytes
	+ ITF_MIB_TBL(port)->ig_fast_rt_ipv6_bytes;
	priv->stats.rx_errors = 0;
	priv->stats.rx_dropped = priv->rx_dropped
	+ ITF_MIB_TBL(port)->ig_drop_pkts;

	priv->stats.tx_packets = priv->tx_packets + ITF_MIB_TBL(port)->eg_fast_pkts;
	priv->stats.tx_bytes = priv->tx_bytes;
	priv->stats.tx_errors = priv->tx_errors;
	priv->stats.tx_dropped = priv->tx_dropped;

	return &priv->stats;
}
#endif

#if 0 && 0 && defined(ENABLE_NAPI) && ENABLE_NAPI
static int eth_poll(struct napi_struct *napi __attribute__ ((unused)),
		int work_to_do) {
	int work_done = 0;
#if defined(TRACE_RX_PATH)
	static int once = 5;
	if (once) {
		dump_stack();
		once--;
	}
#endif
	dma_device_poll(g_dma_device_ppe, work_to_do, &work_done);
	return work_done;
}
#endif

int ifx_ppa_eth_hard_start_xmit_a5(struct sk_buff *skb, struct net_device *dev) {
	int port;
	int qid;

	port = get_eth_port(dev);
	if (port < 0)
		return -ENODEV;

	if (skb->priority >= NUM_ENTITY(g_eth_prio_queue_map[port]))
		qid = g_eth_prio_queue_map[port][NUM_ENTITY(g_eth_prio_queue_map[port])
				- 1];
	else
		qid = g_eth_prio_queue_map[port][skb->priority];

	eth_xmit(skb, port, DMA_CHAN_NR_DEFAULT_CPU_TX, 2, qid); //  spid - CPU, qid - taken from prio_queue_map

	return 0;
}

#if 0
int ifx_ppa_eth_qos_hard_start_xmit_a5(struct sk_buff *skb, struct net_device *dev) {
	int if_id;
	int qid;
	unsigned long sys_flag;
	volatile struct eth_tx_descriptor *desc;
	struct eth_tx_descriptor reg_desc;
	struct sk_buff *skb_to_free;
	int byteoff;
	int len;
	struct sw_eg_pkt_header pkth = { 0 };
	avmnet_netdev_priv_t *priv = NULL;
	avmnet_device_t *avm_dev;


	if_id = get_eth_port(dev);
	if (if_id < 0 || if_id >= num_registered_eth_dev ) {
		printk("%s: Improper ethernet device\n", __FUNCTION__);
		return -ENODEV;
	}

	priv = netdev_priv(dev);
	avm_dev = priv->avm_dev;

    if ( g_stop_datapath != 0 )
        goto ETH_XMIT_DROP;

	if (skb->priority >= NUM_ENTITY(g_eth_prio_queue_map[if_id]))
		qid =
				g_eth_prio_queue_map[if_id][NUM_ENTITY(g_eth_prio_queue_map[if_id])
						- 1];
	else
		qid = g_eth_prio_queue_map[if_id][skb->priority];

	pkth.dpid_en = 1;
	pkth.dpid = 0;
	pkth.spid = 2; //  CPU
	pkth.port_map_en = 1;
	pkth.port_map_sel = 1;
	pkth.port_map = (1 << if_id_to_mac_nr(if_id));
	pkth.lrn_dis = 0;
	pkth.class_en = 0;
	pkth.class = 0;

#if 0
	pkth.spid = 2; //  CPU
	pkth.dpid = if_id;//  eth0/eth1
	pkth.lrn_dis = 0;
	pkth.class_en = 1;
	pkth.class = g_wan_queue_class_map[qid];
	if ( pkth.dpid < 2 )
	pkth.dpid_en = g_egress_direct_forward_enable;
#endif

	len = skb->len <= ETH_MIN_TX_PACKET_LENGTH ?
			ETH_MIN_TX_PACKET_LENGTH : skb->len;

	dump_skb(skb, DUMP_SKB_LEN, "eth_qos_hard_start_xmit", if_id, qid, 1, 0);

	/*  reserve space to put pointer in skb */
	byteoff = (((unsigned int) skb->data - sizeof(pkth)) & (DMA_TX_ALIGNMENT - 1))
			+ sizeof(pkth);
	//if ( skb_headroom(skb) < sizeof(struct sk_buff *) + byteoff || skb->end - skb->data < 1564 /* 1518 (max ETH frame) + 22 (reserved for firmware) + 10 (AAL5) + 6 (PPPoE) + 4 x 2 (2 VLAN) */ || skb_shared(skb) || skb_cloned(skb) )
	if (skb_headroom(skb) < sizeof(struct sk_buff *) + byteoff
			|| skb_shared(skb) || skb_cloned(skb)
			|| (byteoff > 31) 
			|| ((unsigned int)skb->end - (unsigned int)skb->data < DMA_PACKET_SIZE )) {
		struct sk_buff *new_skb;



		ASSERT(byteoff < 32, "dma-byteoff too big: %d ", byteoff);

		ASSERT(skb_headroom(skb) >= sizeof(struct sk_buff *) + byteoff,
				"skb_headroom(skb) < sizeof(struct sk_buff *) + byteoff");

		dma_alignment_eth_bad_count++;
		new_skb = alloc_skb_tx(
				sizeof(pkth) + skb->len < DMA_PACKET_SIZE ? DMA_PACKET_SIZE : skb->len); //  may be used by RX fastpath later
		if (!new_skb) {
			err("no memory");
			goto ALLOC_SKB_TX_FAIL;
		}
		skb_put(new_skb, sizeof(pkth) + skb->len);
		my_memcpy(new_skb->data, &pkth, sizeof(pkth));
		my_memcpy(new_skb->data + sizeof(pkth), skb->data, skb->len);
		dev_kfree_skb_any(skb);
		skb = new_skb;
		byteoff = (u32) skb->data & (DMA_TX_ALIGNMENT - 1);
#ifndef CONFIG_MIPS_UNCACHED
		/*  write back to physical memory   */
		dma_cache_wback((u32)skb->data, skb->len);
#endif
	} else {

		dma_alignment_eth_good_count++;

		skb_push(skb, sizeof(pkth));
		byteoff -= sizeof(pkth);
		my_memcpy(skb->data, &pkth, sizeof(pkth));
		*(struct sk_buff **) ((u32) skb->data - byteoff
				- sizeof(struct sk_buff *)) = skb;
#ifndef CONFIG_MIPS_UNCACHED
		/*  write back to physical memory   */
		dma_cache_wback((u32)skb->data - byteoff - sizeof(struct sk_buff *),
				skb->len + byteoff + sizeof(struct sk_buff *));
#endif
	}

	/*  allocate descriptor */
	local_irq_save(sys_flag);
	desc = (volatile struct eth_tx_descriptor *) (CPU_TO_WAN_TX_DESC_BASE
			+ g_cpu_to_wan_tx_desc_pos);
	if (desc->own) //  PPE hold
	{
		local_irq_restore(sys_flag);
		err("PPE hold");
		goto NO_FREE_DESC;
	}
	if (++g_cpu_to_wan_tx_desc_pos == CPU_TO_WAN_TX_DESC_NUM)
		g_cpu_to_wan_tx_desc_pos = 0;
	local_irq_restore(sys_flag);

	/*  load descriptor from memory */
	reg_desc = *desc;

	/*  free previous skb   */
	ASSERT((reg_desc.dataptr & (DMA_TX_ALIGNMENT - 1)) == 0,
			"reg_desc.dataptr (0x%#x) must be 8 DWORDS aligned",
			reg_desc.dataptr);
	skb_to_free = get_skb_pointer(reg_desc.dataptr);
	dev_kfree_skb_any(skb_to_free);

    /*  detach from protocol    */
    skb_to_free = skb;
    skb = skb_break_away_from_protocol(skb);
    dev_kfree_skb_any(skb_to_free);

	put_skb_to_dbg_pool(skb);

#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC
	if (g_mirror_netdev != NULL) {
		struct sk_buff *new_skb = skb_copy(skb, GFP_ATOMIC);

		if (new_skb != NULL) {
			skb_pull(new_skb, sizeof(pkth));
			new_skb->dev = g_mirror_netdev;
			dev_queue_xmit(new_skb);
		}
	}
#endif

	/*  update descriptor   */
	reg_desc.small =
			(unsigned int) skb->end - (unsigned int) skb->data < DMA_PACKET_SIZE ?
					1 : 0;
	reg_desc.dataptr = (u32) skb->data & (0x1FFFFFFF ^ (DMA_TX_ALIGNMENT - 1)); //  byte address
	reg_desc.byteoff = byteoff;
	reg_desc.datalen = len + sizeof(pkth);
	reg_desc.qid = qid;
	reg_desc.own = 1;
	reg_desc.c = 0;
	reg_desc.sop = reg_desc.eop = 1;

	ASSERT(byteoff < 32, "dma-byteoff still too big: %d ", byteoff);

	/*  update MIB  */

	dev->trans_start = jiffies;

	avm_dev->device_stats.tx_packets++;
	avm_dev->device_stats.tx_bytes += len;

	/*  write discriptor to memory and write back cache */
	*((volatile u32 *) desc + 1) = *((u32 *) &reg_desc + 1);
	*(volatile u32 *) desc = *(u32 *) &reg_desc;

	return 0;

NO_FREE_DESC: 
ALLOC_SKB_TX_FAIL: 
ETH_XMIT_DROP:
    dev_kfree_skb_any(skb);

	avm_dev->device_stats.tx_dropped++;
	return 0;
}
#endif

/*------------------------------------------------------------------------------------------*\
 * Set the MAC address
 \*------------------------------------------------------------------------------------------*/


int ifx_ppa_eth_ioctl_a5(struct net_device *dev, struct ifreq *ifr, int cmd) {
	int port;

	port = get_eth_port(dev);
	if (port < 0)
		return -ENODEV;

	switch (cmd) {
	case SIOCETHTOOL:
		return ethtool_ioctl(dev, ifr);
	case SET_VLAN_COS: {
		struct vlan_cos_req vlan_cos_req;

		if (copy_from_user(&vlan_cos_req, ifr->ifr_data, sizeof(struct vlan_cos_req)))
			return -EFAULT;
#if 0   //  TODO
		set_vlan_cos(&vlan_cos_req);
#endif
	}
		break;
	case SET_DSCP_COS: {
		struct dscp_cos_req dscp_cos_req;

		if (copy_from_user(&dscp_cos_req, ifr->ifr_data, sizeof(struct dscp_cos_req)))
			return -EFAULT;
#if 0   //  TODO
		set_dscp_cos(&dscp_cos_req);
#endif
	}
		break;
#if 0   //  TODO
		case ENABLE_VLAN_CLASSIFICATION:
		*ENETS_COS_CFG(port) |= ENETS_COS_CFG_VLAN_SET; break;
		case DISABLE_VLAN_CLASSIFICATION:
		*ENETS_COS_CFG(port) &= ENETS_COS_CFG_VLAN_CLEAR; break;
		case ENABLE_DSCP_CLASSIFICATION:
		*ENETS_COS_CFG(port) |= ENETS_COS_CFG_DSCP_SET; break;
		case DISABLE_DSCP_CLASSIFICATION:
		*ENETS_COS_CFG(port) &= ENETS_COS_CFG_DSCP_CLEAR; break;
		case VLAN_CLASS_FIRST:
		*ENETS_CFG(port) &= ENETS_CFG_FTUC_CLEAR; break;
		case VLAN_CLASS_SECOND:
		*ENETS_CFG(port) |= ENETS_CFG_VL2_SET; break;
		case PASS_UNICAST_PACKETS:
		*ENETS_CFG(port) &= ENETS_CFG_FTUC_CLEAR; break;
		case FILTER_UNICAST_PACKETS:
		*ENETS_CFG(port) |= ENETS_CFG_FTUC_SET; break;
		case KEEP_BROADCAST_PACKETS:
		*ENETS_CFG(port) &= ENETS_CFG_DPBC_CLEAR; break;
		case DROP_BROADCAST_PACKETS:
		*ENETS_CFG(port) |= ENETS_CFG_DPBC_SET; break;
		case KEEP_MULTICAST_PACKETS:
		*ENETS_CFG(port) &= ENETS_CFG_DPMC_CLEAR; break;
		case DROP_MULTICAST_PACKETS:
		*ENETS_CFG(port) |= ENETS_CFG_DPMC_SET; break;
#endif
	case ETH_MAP_PKT_PRIO_TO_Q: {
		struct ppe_prio_q_map cmd;

		if (copy_from_user(&cmd, ifr->ifr_data, sizeof(cmd)))
			return -EFAULT;

		if (cmd.pkt_prio < 0 || cmd.pkt_prio >= (int)NUM_ENTITY(g_eth_prio_queue_map[port]))return -EINVAL;

		if (cmd.qid < 0)
			return -EINVAL;
		if (cmd.qid
				>= ((g_wan_itf & (1 << port)) && (g_eth_wan_mode == 3)
						&& g_wanqos_en ? g_wanqos_en : SWITCH_CLASS_VALUE))
			return -EINVAL;

		g_eth_prio_queue_map[port][cmd.pkt_prio] = cmd.qid;
	}
		break;
	default:
		return -EOPNOTSUPP;
	}

	return 0;
}

void ifx_ppa_eth_tx_timeout_a5(struct net_device *dev) {
	unsigned long sys_flag;

	avmnet_netdev_priv_t *priv = netdev_priv(dev);
	avmnet_device_t *avm_dev = NULL;
    
    if (priv)
    	avm_dev = priv->avm_dev;

    if (avm_dev)
        avm_dev->device_stats.tx_errors++;

	local_irq_save(sys_flag);
	g_dma_device_ppe->tx_chan[2]->disable_irq(g_dma_device_ppe->tx_chan[2]);
	netif_wake_queue(g_eth_net_dev[0]);
	if (!g_wanqos_en)
		netif_wake_queue(g_eth_net_dev[1]);
	local_irq_restore(sys_flag);

	return;
}

static INLINE int eth_xmit(struct sk_buff *skb, unsigned int dpid, int ch,
		int spid __attribute__((unused)), int prio __attribute__((unused)) ) {

	struct sw_eg_pkt_header pkth = { 0 };
	int len;
	unsigned int headroom;
	avmnet_netdev_priv_t *priv = NULL;
	avmnet_device_t *avm_dev = NULL;
    struct skb_shared_info *shinfo ;

    if ( g_stop_datapath != 0 )
        goto ETH_XMIT_DROP;

	if (skb->len <= ETH_MIN_TX_PACKET_LENGTH ){

		/*--- 0byte padding behind skb->len, skb->len is not increased ---*/

		if (skb_pad(skb, ETH_MIN_TX_PACKET_LENGTH - skb->len )){
			skb = NULL;
			goto ETH_XMIT_DROP;
		}

		len = ETH_MIN_TX_PACKET_LENGTH;
		padded_frame_count++;

	} else {
		len = skb->len;
	}

#if defined (CONFIG_AVM_SCATTER_GATHER)
    /* If packet is not checksummed and device does not support
    * checksumming for this protocol, complete checksumming here.
    */
    if (skb->ip_summed == CHECKSUM_PARTIAL) {
        if(unlikely(skb_checksum_help(skb))){
            err(KERN_ERR "skb_checksum fails\n");
            goto ETH_XMIT_DROP;
        }
    }
#endif


#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC
	if (g_mirror_netdev != NULL && (g_wan_itf & (1 << dpid))) {
		struct sk_buff *new_skb = skb_copy(skb, GFP_ATOMIC);

		if (new_skb != NULL) {
			new_skb->dev = g_mirror_netdev;
			dev_queue_xmit(new_skb);
		}
	}
#endif

	dump_skb(skb, DUMP_SKB_LEN, "eth_xmit - raw data", dpid, ch, 1, 0);

	if ( spid == 2 ){ // don't track directpath traffic here
	    if ( dpid < num_registered_eth_dev ) {
	        g_eth_net_dev[dpid]->trans_start = jiffies;

	        priv = netdev_priv(g_eth_net_dev[dpid]);

	        avm_dev = priv->avm_dev;
	        avm_dev->device_stats.tx_packets++;
	        avm_dev->device_stats.tx_bytes += len;
#       if defined(CONFIG_IP_MULTICAST_FASTFORWARD)
	        if(likely(priv->mcfw_used)){
	            mcfw_portset set;
	            mcfw_portset_reset(&set);
	            mcfw_portset_port_add(&set, 0);
	            (void) mcfw_snoop_send(&priv->mcfw_netdrv, set, skb);
	        }
#       endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) ---*/
	    }
	}

	headroom = sizeof(struct sw_eg_pkt_header);

	if (skb_headroom(skb) < headroom) {
		struct sk_buff *new_skb;
        
		dma_alignment_eth_bad_count++;
        new_skb = alloc_skb_tx(len + sizeof(struct sw_eg_pkt_header));
		if (!new_skb)
			goto ETH_XMIT_DROP;
		else {
			skb_put(new_skb, len + sizeof(struct sw_eg_pkt_header));
			memcpy(new_skb->data + sizeof(struct sw_eg_pkt_header), skb->data, len);
			dev_kfree_skb_any(skb);
			skb = new_skb;
		}
	} else {
		skb_push(skb, sizeof(struct sw_eg_pkt_header));
		dma_alignment_eth_good_count++;
		skb_orphan( skb );
	}
	len += sizeof(struct sw_eg_pkt_header);

	if (spid == 2) {

		// for cpu -> ethernet
		pkth.spid = 2; //  CPU
		pkth.dpid_en = 1;
		pkth.dpid = 1; // 0 or 1 for PORT_MAP_MODE
		pkth.port_map_en = 1;
		pkth.port_map_sel = 1;
		pkth.port_map = (1 << if_id_to_mac_nr(dpid));
		pkth.lrn_dis = 0;
		pkth.class_en = 0;
		pkth.class = 0;
		dbg_dma("port_map: %#x", (unsigned int)pkth.port_map);

	} else {

		// for directpath Loop to CPU (might be interrupted by PCE-Rule -> Ethernet , or PPA -> WAN)
		pkth.spid = spid;
		pkth.dpid_en = 0;
		pkth.dpid = 1;          // don_t care
		pkth.port_map_en = 0;   // don_t care
		pkth.port_map_sel = 0;  // don_t care
		pkth.lrn_dis = 1;
		pkth.class_en = 0;
		pkth.class = prio;
		pkth.port_map = (0x3F);

	}

    put_skb_to_dbg_pool(skb);
	
	memcpy(skb->data, &pkth, sizeof(struct sw_eg_pkt_header));
	
	shinfo = skb_shinfo(skb);
#if defined (CONFIG_AVM_SCATTER_GATHER)
    if(shinfo && shinfo->nr_frags > 0 ) {
        unsigned long lock_flags;

        // this skb is scattered
        dbg_trace_ppa_data("send first fragment (nr_frags = %d)",shinfo->nr_frags );

        lock_flags = acquire_channel_lock(g_dma_device_ppe, ch );
        if(unlikely(send_skb_scattered_first(skb, g_dma_device_ppe , ch))) {
            err("sending first fragment failed");
            free_channel_lock(g_dma_device_ppe, ch, lock_flags );
            goto ETH_XMIT_DROP;
        }
        if(unlikely(send_skb_scattered_successors(skb, g_dma_device_ppe, ch))) {
            err("ah we could just send some of the fragmets, this must not happen!");
            // we cannot do any skb free now :-(
            free_channel_lock(g_dma_device_ppe, ch, lock_flags);
            return 0;
        }
        free_channel_lock(g_dma_device_ppe, ch, lock_flags );
    }
#else
    if(unlikely(shinfo && ( shinfo->nr_frags > 0 ))) {
        err("we don't support Scatter Gather, but we got a scattered SKB");
        goto ETH_XMIT_DROP;
    }
#endif
    else {
	    if (dma_device_write(g_dma_device_ppe, skb->data, len, skb, ch) != len)
		    goto ETH_XMIT_DROP;
    }

	return 0;

	ETH_XMIT_DROP:
	dev_kfree_skb_any(skb);
	if (priv && priv->avm_dev) {
		avm_dev->device_stats.tx_dropped++;
	}
	return -1;
}

/*
 *  Description:
 *    Handle ioctl command SIOCETHTOOL.
 *  Input:
 *    dev --- struct net_device *, device responsing to the command.
 *    ifr --- struct ifreq *, interface request structure to pass parameters
 *            or result.
 *  Output:
 *    int --- 0:    Success
 *            else: Error Code (-EFAULT, -EOPNOTSUPP)
 */
static INLINE int ethtool_ioctl(struct net_device *dev __attribute__((unused)), struct ifreq *ifr) {
	struct ethtool_cmd cmd;
#if 0   //  TODO
	int port;
	port = get_eth_port(dev);
#endif

	if (copy_from_user(&cmd, ifr->ifr_data, sizeof(cmd)))
		return -EFAULT;


	switch (cmd.cmd) {
	case ETHTOOL_GSET: /*  get hardware information        */
	{
		memset(&cmd, 0, sizeof(cmd));

#if 0   //  TODO
		cmd.supported = SUPPORTED_Autoneg | SUPPORTED_TP | SUPPORTED_MII |
		SUPPORTED_10baseT_Half | SUPPORTED_10baseT_Full |
		SUPPORTED_100baseT_Half | SUPPORTED_100baseT_Full;
		cmd.port = PORT_MII;
		cmd.transceiver = XCVR_EXTERNAL;
		cmd.phy_address = port;
		cmd.speed = ENET_MAC_CFG_SPEED(port) ? SPEED_100 : SPEED_10;
		cmd.duplex = ENET_MAC_CFG_DUPLEX(port) ? DUPLEX_FULL : DUPLEX_HALF;

		if ( (*ETOP_MDIO_CFG & ETOP_MDIO_CFG_UMM(port, 1)) )
		{
			/*  auto negotiate  */
			cmd.autoneg = AUTONEG_ENABLE;
			cmd.advertising |= ADVERTISED_10baseT_Half | ADVERTISED_10baseT_Full |
			ADVERTISED_100baseT_Half | ADVERTISED_100baseT_Full;
		}
		else
		{
			cmd.autoneg = AUTONEG_DISABLE;
			cmd.advertising &= ~(ADVERTISED_10baseT_Half | ADVERTISED_10baseT_Full |
					ADVERTISED_100baseT_Half | ADVERTISED_100baseT_Full);
		}
#endif

		if (copy_to_user(ifr->ifr_data, &cmd, sizeof(cmd)))
			return -EFAULT;
	}
		break;
	case ETHTOOL_SSET: /*  force the speed and duplex mode */
	{
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;

		if (cmd.autoneg == AUTONEG_ENABLE) {
			/*  set property and start autonegotiation                                  */
			/*  have to set mdio advertisement register and restart autonegotiation     */
			/*  which is a very rare case, put it to future development if necessary.   */
		} else {
#if 0   //  TODO
			/*  set property without autonegotiation    */
			*ETOP_MDIO_CFG &= ~ETOP_MDIO_CFG_UMM(port, 1);

			/*  set speed   */
			if ( cmd.speed == SPEED_10 )
			ENET_MAC_CFG_SPEED_10M(port);
			else if ( cmd.speed == SPEED_100 )
			ENET_MAC_CFG_SPEED_100M(port);

			/*  set duplex  */
			if ( cmd.duplex == DUPLEX_HALF )
			ENET_MAC_CFG_DUPLEX_HALF(port);
			else if ( cmd.duplex == DUPLEX_FULL )
			ENET_MAC_CFG_DUPLEX_FULL(port);

			ENET_MAC_CFG_LINK_OK(port);
#endif
		}
	}
		break;
	case ETHTOOL_GDRVINFO: /*  get driver information          */
	{
		struct ethtool_drvinfo info;
		char str[32];

		memset(&info, 0, sizeof(info));
		strncpy(info.driver, "Danube Eth Driver (A4)", sizeof(info.driver) - 1);
		sprintf(str, "%d.%d.%d.%d.%d.%d", (int) FW_VER_ID->family,
				(int) FW_VER_ID->fwtype, (int) FW_VER_ID->interface,
				(int) FW_VER_ID->fwmode, (int) FW_VER_ID->major,
				(int) FW_VER_ID->minor);
		strncpy(info.fw_version, str, sizeof(info.fw_version) - 1);
		strncpy(info.bus_info, "N/A", sizeof(info.bus_info) - 1);
		info.regdump_len = 0;
		info.eedump_len = 0;
		info.testinfo_len = 0;
		if (copy_to_user(ifr->ifr_data, &info, sizeof(info)))
			return -EFAULT;
	}
		break;
	case ETHTOOL_NWAY_RST: /*  restart auto negotiation        */
#if 0   //  TODO
		*ETOP_MDIO_CFG |= ETOP_MDIO_CFG_SMRST(port) | ETOP_MDIO_CFG_UMM(port, 1);
#endif
		break;
	default:
		return -EOPNOTSUPP;
	}

	return 0;
}

/*
 *  Description:
 *    Add a 32-bit value to 64-bit value, and put result in a 64-bit variable.
 *  Input:
 *    opt1 --- ppe_u64_t, first operand, a 64-bit unsigned integer value
 *    opt2 --- u32, second operand, a 32-bit unsigned integer value
 *    ret  --- ppe_u64_t, pointer to a variable to hold result
 *  Output:
 *    none
 */
static INLINE void u64_add_u32(ppe_u64_t opt1, u32 opt2, ppe_u64_t *ret) {
	ret->l = opt1.l + opt2;
	if (ret->l < opt1.l || ret->l < opt2)
		ret->h++;
}

static INLINE void adsl_led_flash(void) {
#if defined(ENABLE_LED_FRAMEWORK) && ENABLE_LED_FRAMEWORK && defined(CONFIG_IFX_LED)
	if ( g_data_led_trigger != NULL )
	ifx_led_trigger_activate(g_data_led_trigger);
#else
	if (!IS_ERR(&ifx_mei_atm_led_blink) && &ifx_mei_atm_led_blink)
		ifx_mei_atm_led_blink();
#endif
}

//  add one lower priority queue
static int sw_tx_queue_add(int conn) {
	int ret;
	unsigned int new_sw_tx_queue;
	u32 new_sw_tx_queue_table;
	int sw_tx_queue_to_virt_queue_map[ATM_SW_TX_QUEUE_NUMBER] = { 0 };
	int virt_queue_to_sw_tx_queue_map[8] = { 0 };
	unsigned int num_of_virt_queue;
	u32 sw_tx_queue_table;
	int sw_tx_queue;
	int virt_queue;
	struct wtx_queue_config wtx_queue_config;
	int f_got_wtx_queue_config = 0;
	int i;

	if (!in_interrupt())
		down(&g_sem);

	for (new_sw_tx_queue = 0;
			new_sw_tx_queue < g_atm_priv_data.max_sw_tx_queues;
			new_sw_tx_queue++)
		if (!(g_atm_priv_data.sw_tx_queue_table & (1 << new_sw_tx_queue)))
			break;
	if (new_sw_tx_queue == g_atm_priv_data.max_sw_tx_queues) {
		ret = -EINVAL;
		goto SW_TX_QUEUE_ADD_EXIT;
	}

	new_sw_tx_queue_table = g_atm_priv_data.connection[conn].sw_tx_queue_table
			| (1 << new_sw_tx_queue);

	for (sw_tx_queue_table = g_atm_priv_data.connection[conn].sw_tx_queue_table, sw_tx_queue =
			virt_queue = 0; sw_tx_queue_table != 0;
			sw_tx_queue_table >>= 1, sw_tx_queue++)
		if ((sw_tx_queue_table & 0x01)) {
			if (!f_got_wtx_queue_config) {
				wtx_queue_config = *WTX_QUEUE_CONFIG(sw_tx_queue);
				f_got_wtx_queue_config = 1;
			}

			sw_tx_queue_to_virt_queue_map[sw_tx_queue] = virt_queue++;
			WTX_QUEUE_CONFIG(sw_tx_queue)->same_vc_qmap = new_sw_tx_queue_table
					& ~(1 << sw_tx_queue);
		}

	wtx_queue_config.same_vc_qmap = new_sw_tx_queue_table
			& ~(1 << new_sw_tx_queue);
	*WTX_QUEUE_CONFIG(new_sw_tx_queue) = wtx_queue_config;

	for (num_of_virt_queue = 0, sw_tx_queue_table = new_sw_tx_queue_table, sw_tx_queue =
			0;
			num_of_virt_queue < NUM_ENTITY(virt_queue_to_sw_tx_queue_map)
					&& sw_tx_queue_table != 0;
			sw_tx_queue_table >>= 1, sw_tx_queue++)
		if ((sw_tx_queue_table & 0x01))
			virt_queue_to_sw_tx_queue_map[num_of_virt_queue++] = sw_tx_queue;

	g_atm_priv_data.sw_tx_queue_table |= 1 << new_sw_tx_queue;
	g_atm_priv_data.connection[conn].sw_tx_queue_table = new_sw_tx_queue_table;

	for (i = 0; i < 8; i++) {
		sw_tx_queue = g_atm_priv_data.connection[conn].prio_queue_map[i];
		virt_queue = sw_tx_queue_to_virt_queue_map[sw_tx_queue];
		sw_tx_queue = virt_queue_to_sw_tx_queue_map[virt_queue];
		g_atm_priv_data.connection[conn].prio_queue_map[i] = sw_tx_queue;
	}

	ret = 0;

	SW_TX_QUEUE_ADD_EXIT:

	if (!in_interrupt())
		up(&g_sem);

	return ret;
}

//  delete lowest priority queue
static int sw_tx_queue_del(int conn) {
	int ret;
	int rem_sw_tx_queue;
	int new_sw_tx_queue_table;
	int sw_tx_queue_table;
	int sw_tx_queue;
	int i;

	if (!in_interrupt())
		down(&g_sem);

	for (sw_tx_queue_table = g_atm_priv_data.connection[conn].sw_tx_queue_table, 
            rem_sw_tx_queue = g_atm_priv_data.max_sw_tx_queues - 1;
			(sw_tx_queue_table & (1 << (g_atm_priv_data.max_sw_tx_queues - 1))) == 0; 
            sw_tx_queue_table <<= 1, rem_sw_tx_queue--)
		;

	if (g_atm_priv_data.connection[conn].sw_tx_queue_table == (1U << rem_sw_tx_queue)) {
		ret = -EIO;
		goto SW_TX_QUEUE_DEL_EXIT;
	}

	new_sw_tx_queue_table = g_atm_priv_data.connection[conn].sw_tx_queue_table
			& ~(1 << rem_sw_tx_queue);

	for (sw_tx_queue_table = new_sw_tx_queue_table, sw_tx_queue = g_atm_priv_data.max_sw_tx_queues - 1;
			(sw_tx_queue_table & (1 << (g_atm_priv_data.max_sw_tx_queues - 1))) == 0; 
            sw_tx_queue_table <<= 1, sw_tx_queue--)
		;

	for (i = 0; i < 8; i++) {
		if (g_atm_priv_data.connection[conn].prio_queue_map[i] == rem_sw_tx_queue)
			g_atm_priv_data.connection[conn].prio_queue_map[i] = sw_tx_queue;
    }

	g_atm_priv_data.sw_tx_queue_table &= ~(1 << rem_sw_tx_queue);
	g_atm_priv_data.connection[conn].sw_tx_queue_table = new_sw_tx_queue_table;

	for (sw_tx_queue_table = new_sw_tx_queue_table, sw_tx_queue = 0;
			sw_tx_queue_table != 0; sw_tx_queue_table >>= 1, sw_tx_queue++) {
		if ((sw_tx_queue_table & 0x01)) {
			WTX_QUEUE_CONFIG(sw_tx_queue)->same_vc_qmap = new_sw_tx_queue_table & ~(1 << sw_tx_queue);
        }
    }

	ret = 0;

	SW_TX_QUEUE_DEL_EXIT:

	if (!in_interrupt())
		up(&g_sem);

	return ret;
}

static int sw_tx_prio_to_queue(int conn, int prio, int queue,
		int *p_num_of_virt_queue) {
	int virt_queue_to_sw_tx_queue_map[8] = { 0 };
    int num_of_virt_queue;
	u32 sw_tx_queue_table;
	int sw_tx_queue;

	if (conn < 0)
		return -EINVAL;

	if (!in_interrupt())
		down(&g_sem);

	for (num_of_virt_queue = 0, 
            sw_tx_queue_table = g_atm_priv_data.connection[conn].sw_tx_queue_table, 
            sw_tx_queue = 0;
			num_of_virt_queue < NUM_ENTITY(virt_queue_to_sw_tx_queue_map) && sw_tx_queue_table != 0;
			sw_tx_queue_table >>= 1, sw_tx_queue++)
    {
		if ((sw_tx_queue_table & 0x01))
			virt_queue_to_sw_tx_queue_map[num_of_virt_queue++] = sw_tx_queue;
    }

	if (p_num_of_virt_queue != NULL)
		*p_num_of_virt_queue = num_of_virt_queue;

	if (prio < 0 || queue < 0 || queue >= num_of_virt_queue) {
		if (!in_interrupt())
			up(&g_sem);
		return -EINVAL;
	}

	g_atm_priv_data.connection[conn].prio_queue_map[prio] =
			virt_queue_to_sw_tx_queue_map[queue];

	if (!in_interrupt())
		up(&g_sem);

	return 0;
}

/*
 *  Description:
 *    Setup QSB queue.
 *  Input:
 *    vcc        --- struct atm_vcc *, structure of an opened connection
 *    qos        --- struct atm_qos *, QoS parameter of the connection
 *    connection --- unsigned int, QSB queue ID, which is same as connection ID
 *  Output:
 *    none
 */
static void set_qsb(struct atm_vcc *vcc, struct atm_qos *qos,
		unsigned int connection) {
    u32 qsb_clk = get_qsb_clk();
	union qsb_queue_parameter_table qsb_queue_parameter_table = { { 0 } };
	union qsb_queue_vbr_parameter_table qsb_queue_vbr_parameter_table =
			{ { 0 } };
	u32 tmp;

	connection += QSB_QUEUE_NUMBER_BASE; //  qsb qid = firmware qid + 1

#if defined(DEBUG_QOS) && DEBUG_QOS
	if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS)) {
		static char *str_traffic_class[9] = { "ATM_NONE", "ATM_UBR", "ATM_CBR",
				"ATM_VBR", "ATM_ABR", "ATM_ANYCLASS", "ATM_VBR_RT",
				"ATM_UBR_PLUS", "ATM_MAX_PCR" };

		unsigned char traffic_class = qos->txtp.traffic_class;
		int max_pcr = qos->txtp.max_pcr;
		int pcr = qos->txtp.pcr;
		int min_pcr = qos->txtp.min_pcr;
#if !defined(DISABLE_VBR) || !DISABLE_VBR
		int scr = qos->txtp.scr;
		int mbs = qos->txtp.mbs;
		int cdv = qos->txtp.cdv;
#endif

		printk("TX QoS\n");

		printk("  traffic class : ");
		if (traffic_class == (unsigned char) ATM_MAX_PCR)
			printk("ATM_MAX_PCR\n");
		else if (traffic_class > ATM_UBR_PLUS)
			printk("Unknown Class\n");
		else
			printk("%s\n", str_traffic_class[traffic_class]);

		printk("  max pcr       : %d\n", max_pcr);
		printk("  desired pcr   : %d\n", pcr);
		printk("  min pcr       : %d\n", min_pcr);

#if !defined(DISABLE_VBR) || !DISABLE_VBR
		printk("  sustained rate: %d\n", scr);
		printk("  max burst size: %d\n", mbs);
		printk("  cell delay var: %d\n", cdv);
#endif
	}
#endif    //  defined(DEBUG_QOS) && DEBUG_QOS
	/*
	 *  Peak Cell Rate (PCR) Limiter
	 */
	if (qos->txtp.max_pcr == 0)
		qsb_queue_parameter_table.bit.tp = 0; /*  disable PCR limiter */
	else {
		/*  peak cell rate would be slightly lower than requested [maximum_rate / pcr = (qsb_clock / 8) * (time_step / 4) / pcr] */
		tmp = ((qsb_clk * qsb_tstep) >> 5) / qos->txtp.max_pcr + 1;
		/*  check if overflow takes place   */
		qsb_queue_parameter_table.bit.tp =
				tmp > QSB_TP_TS_MAX ? QSB_TP_TS_MAX : tmp;
	}

	//  A funny issue. Create two PVCs, one UBR and one UBR with max_pcr.
	//  Send packets to these two PVCs at same time, it trigger strange behavior.
	//  In A1, RAM from 0x80000000 to 0x0x8007FFFF was corrupted with fixed pattern 0x00000000 0x40000000.
	//  In A4, PPE firmware keep emiting unknown cell and do not respond to driver.
	//  To work around, create UBR always with max_pcr.
	//  If user want to create UBR without max_pcr, we give a default one larger than line-rate.
	if (qos->txtp.traffic_class == ATM_UBR
			&& qsb_queue_parameter_table.bit.tp == 0) {
		int port =
				g_atm_priv_data.connection[connection - QSB_QUEUE_NUMBER_BASE].port;
		unsigned int max_pcr = g_atm_priv_data.port[port].tx_max_cell_rate
				+ 1000;

		tmp = ((qsb_clk * qsb_tstep) >> 5) / max_pcr + 1;
		if (tmp > QSB_TP_TS_MAX)
			tmp = QSB_TP_TS_MAX;
		else if (tmp < 1)
			tmp = 1;
		qsb_queue_parameter_table.bit.tp = tmp;
	}

	/*
	 *  Weighted Fair Queueing Factor (WFQF)
	 */
	switch (qos->txtp.traffic_class) {
	case ATM_CBR:
	case ATM_VBR_RT:
		/*  real time queue gets weighted fair queueing bypass  */
		qsb_queue_parameter_table.bit.wfqf = 0;
		break;
	case ATM_VBR_NRT:
	case ATM_UBR_PLUS:
		/*  WFQF calculation here is based on virtual cell rates, to reduce granularity for high rates  */
		/*  WFQF is maximum cell rate / garenteed cell rate                                             */
		/*  wfqf = qsb_minimum_cell_rate * QSB_WFQ_NONUBR_MAX / requested_minimum_peak_cell_rate        */
		if (qos->txtp.min_pcr == 0)
			qsb_queue_parameter_table.bit.wfqf = QSB_WFQ_NONUBR_MAX;
		else {
			tmp = QSB_GCR_MIN * QSB_WFQ_NONUBR_MAX / qos->txtp.min_pcr;
			if (tmp == 0)
				qsb_queue_parameter_table.bit.wfqf = 1;
			else if (tmp > QSB_WFQ_NONUBR_MAX)
				qsb_queue_parameter_table.bit.wfqf = QSB_WFQ_NONUBR_MAX;
			else
				qsb_queue_parameter_table.bit.wfqf = tmp;
		}
		break;
	default:
	case ATM_UBR:
		qsb_queue_parameter_table.bit.wfqf = QSB_WFQ_UBR_BYPASS;
	}

	/*
	 *  Sustained Cell Rate (SCR) Leaky Bucket Shaper VBR.0/VBR.1
	 */
	if (qos->txtp.traffic_class == ATM_VBR_RT
			|| qos->txtp.traffic_class == ATM_VBR_NRT) {
#if defined(DISABLE_VBR) && DISABLE_VBR
		/*  disable shaper  */
		qsb_queue_vbr_parameter_table.bit.taus = 0;
		qsb_queue_vbr_parameter_table.bit.ts = 0;
#else
		if (qos->txtp.scr == 0) {
			/*  disable shaper  */
			qsb_queue_vbr_parameter_table.bit.taus = 0;
			qsb_queue_vbr_parameter_table.bit.ts = 0;
		} else {
			/*  Cell Loss Priority  (CLP)   */
			if ((vcc->atm_options & ATM_ATMOPT_CLP))
				/*  CLP1    */
				qsb_queue_parameter_table.bit.vbr = 1;
			else
				/*  CLP0    */
				qsb_queue_parameter_table.bit.vbr = 0;
			/*  Rate Shaper Parameter (TS) and Burst Tolerance Parameter for SCR (tauS) */
			tmp = ((qsb_clk * qsb_tstep) >> 5) / qos->txtp.scr + 1;
			qsb_queue_vbr_parameter_table.bit.ts =
					tmp > QSB_TP_TS_MAX ? QSB_TP_TS_MAX : tmp;
			tmp = (qos->txtp.mbs - 1)
					* (qsb_queue_vbr_parameter_table.bit.ts
							- qsb_queue_parameter_table.bit.tp) / 64;
			if (tmp == 0)
				qsb_queue_vbr_parameter_table.bit.taus = 1;
			else if (tmp > QSB_TAUS_MAX)
				qsb_queue_vbr_parameter_table.bit.taus = QSB_TAUS_MAX;
			else
				qsb_queue_vbr_parameter_table.bit.taus = tmp;
		}
#endif
	} else {
		qsb_queue_vbr_parameter_table.bit.taus = 0;
		qsb_queue_vbr_parameter_table.bit.ts = 0;
	}

	/*  Queue Parameter Table (QPT) */
	*QSB_RTM = QSB_RTM_DM_SET(QSB_QPT_SET_MASK);
	*QSB_RTD = QSB_RTD_TTV_SET(qsb_queue_parameter_table.dword);
	*QSB_RAMAC = QSB_RAMAC_RW_SET(QSB_RAMAC_RW_WRITE)
			| QSB_RAMAC_TSEL_SET(QSB_RAMAC_TSEL_QPT)
			| QSB_RAMAC_LH_SET(QSB_RAMAC_LH_LOW)
			| QSB_RAMAC_TESEL_SET(connection);
#if defined(DEBUG_QOS) && DEBUG_QOS
	if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS))
		printk(
				"QPT: QSB_RTM (%08X) = 0x%08X, QSB_RTD (%08X) = 0x%08X, QSB_RAMAC (%08X) = 0x%08X\n",
				(u32) QSB_RTM, *QSB_RTM, (u32) QSB_RTD, *QSB_RTD,
				(u32) QSB_RAMAC, *QSB_RAMAC);
#endif
	/*  Queue VBR Paramter Table (QVPT) */
	*QSB_RTM = QSB_RTM_DM_SET(QSB_QVPT_SET_MASK);
	*QSB_RTD = QSB_RTD_TTV_SET(qsb_queue_vbr_parameter_table.dword);
	*QSB_RAMAC = QSB_RAMAC_RW_SET(QSB_RAMAC_RW_WRITE)
			| QSB_RAMAC_TSEL_SET(QSB_RAMAC_TSEL_VBR)
			| QSB_RAMAC_LH_SET(QSB_RAMAC_LH_LOW)
			| QSB_RAMAC_TESEL_SET(connection);
#if defined(DEBUG_QOS) && DEBUG_QOS
	if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS))
		printk(
				"QVPT: QSB_RTM (%08X) = 0x%08X, QSB_RTD (%08X) = 0x%08X, QSB_RAMAC (%08X) = 0x%08X\n",
				(u32) QSB_RTM, *QSB_RTM, (u32) QSB_RTD, *QSB_RTD,
				(u32) QSB_RAMAC, *QSB_RAMAC);
#endif

#if defined(DEBUG_QOS) && DEBUG_QOS
	if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS)) {
		printk("set_qsb\n");
		printk("  qsb_clk = %lu\n", (unsigned long) qsb_clk);
		printk("  qsb_queue_parameter_table.bit.tp       = %d\n",
				(int) qsb_queue_parameter_table.bit.tp);
		printk("  qsb_queue_parameter_table.bit.wfqf     = %d (0x%08X)\n",
				(int) qsb_queue_parameter_table.bit.wfqf,
				(int) qsb_queue_parameter_table.bit.wfqf);
		printk("  qsb_queue_parameter_table.bit.vbr      = %d\n",
				(int) qsb_queue_parameter_table.bit.vbr);
		printk("  qsb_queue_parameter_table.dword        = 0x%08X\n",
				(int) qsb_queue_parameter_table.dword);
		printk("  qsb_queue_vbr_parameter_table.bit.ts   = %d\n",
				(int) qsb_queue_vbr_parameter_table.bit.ts);
		printk("  qsb_queue_vbr_parameter_table.bit.taus = %d\n",
				(int) qsb_queue_vbr_parameter_table.bit.taus);
		printk("  qsb_queue_vbr_parameter_table.dword    = 0x%08X\n",
				(int) qsb_queue_vbr_parameter_table.dword);
	}
#endif
}

/*
 *  Description:
 *    Add one entry to HTU table.
 *  Input:
 *    vpi        --- unsigned int, virtual path ID
 *    vci        --- unsigned int, virtual channel ID
 *    connection --- unsigned int, connection ID
 *    aal5       --- int, 0 means AAL0, else means AAL5
 *  Output:
 *    none
 */
static INLINE void set_htu_entry(unsigned int vpi, unsigned int vci,
		unsigned int conn, int aal5) {
	struct htu_entry htu_entry = { res1: 0x00, pid
			: g_atm_priv_data.connection[conn].port & 0x01, vpi: vpi, vci: vci,
			pti: 0x00, vld: 0x01 };

	struct htu_mask htu_mask = { set: 0x03, pid_mask: 0x02, vpi_mask: 0x00,
			vci_mask: 0x0000, pti_mask: 0x03, //  0xx, user data
			clear: 0x00 };

	struct htu_result htu_result = { res1: 0x00, cellid: conn, res2: 0x00, type:
			aal5 ? 0x00 : 0x01, ven: 0x01, res3: 0x00, qid: conn };

#if 0
	htu_entry.vld = 1;
	htu_mask.pid_mask = 0x03;
	htu_mask.vpi_mask = 0xFF;
	htu_mask.vci_mask = 0xFFFF;
	htu_mask.pti_mask = 0x7;
	htu_result.cellid = 1;
	htu_result.type = 0;
	htu_result.ven = 0;
	htu_result.qid = conn;
#endif
	*HTU_RESULT(conn + OAM_HTU_ENTRY_NUMBER) = htu_result;
	*HTU_MASK(conn + OAM_HTU_ENTRY_NUMBER) = htu_mask;
	*HTU_ENTRY(conn + OAM_HTU_ENTRY_NUMBER) = htu_entry;

//    printk("Config HTU (%d) for QID %d\n", conn + OAM_HTU_ENTRY_NUMBER, conn);
//    printk("  HTU_RESULT = 0x%08X\n", *(u32*)HTU_RESULT(conn + OAM_HTU_ENTRY_NUMBER));
//    printk("  HTU_MASK   = 0x%08X\n", *(u32*)HTU_MASK(conn + OAM_HTU_ENTRY_NUMBER));
//    printk("  HTU_ENTRY  = 0x%08X\n", *(u32*)HTU_ENTRY(conn + OAM_HTU_ENTRY_NUMBER));
}

/*
 *  Description:
 *    Remove one entry from HTU table.
 *  Input:
 *    conn --- unsigned int, connection ID
 *  Output:
 *    none
 */
static INLINE void clear_htu_entry(unsigned int conn) {
	HTU_ENTRY(conn + OAM_HTU_ENTRY_NUMBER)->vld = 0;
}

/*
 *  Description:
 *    Setup QSB.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static INLINE void qsb_global_set(void) {
    u32 qsb_clk = get_qsb_clk();
	int i;
	u32 tmp1, tmp2, tmp3;

	*QSB_ICDV = QSB_ICDV_TAU_SET(qsb_tau);
	*QSB_SBL = QSB_SBL_SBL_SET(qsb_srvm);
	*QSB_CFG = QSB_CFG_TSTEPC_SET(qsb_tstep >> 1);
#if defined(DEBUG_QOS) && DEBUG_QOS
	if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS)) {
		printk("qsb_clk = %u\n", qsb_clk);
		printk(
				"QSB_ICDV (%08X) = %d (%d), QSB_SBL (%08X) = %d (%d), QSB_CFG (%08X) = %d (%d)\n",
				(u32) QSB_ICDV, *QSB_ICDV, QSB_ICDV_TAU_SET(qsb_tau),
				(u32) QSB_SBL, *QSB_SBL, QSB_SBL_SBL_SET(qsb_srvm),
				(u32) QSB_CFG, *QSB_CFG, QSB_CFG_TSTEPC_SET(qsb_tstep >> 1));
	}
#endif

	/*
	 *  set SCT and SPT per port
	 */
	for (i = 0; i < ATM_PORT_NUMBER; i++)
		if (g_atm_priv_data.port[i].tx_max_cell_rate != 0) {
			tmp1 = ((qsb_clk * qsb_tstep) >> 1)
					/ g_atm_priv_data.port[i].tx_max_cell_rate;
			tmp2 = tmp1 >> 6; /*  integer value of Tsb    */
			tmp3 = (tmp1 & ((1 << 6) - 1)) + 1; /*  fractional part of Tsb  */
			/*  carry over to integer part (?)  */
			if (tmp3 == (1 << 6)) {
				tmp3 = 0;
				tmp2++;
			}
			if (tmp2 == 0)
				tmp2 = tmp3 = 1;
			/*  1. set mask                                 */
			/*  2. write value to data transfer register    */
			/*  3. start the tranfer                        */
			/*  SCT (FracRate)  */
			*QSB_RTM = QSB_RTM_DM_SET(QSB_SET_SCT_MASK);
			*QSB_RTD = QSB_RTD_TTV_SET(tmp3);
			*QSB_RAMAC = QSB_RAMAC_RW_SET(QSB_RAMAC_RW_WRITE)
					| QSB_RAMAC_TSEL_SET(QSB_RAMAC_TSEL_SCT)
					| QSB_RAMAC_LH_SET(QSB_RAMAC_LH_LOW)
					| QSB_RAMAC_TESEL_SET(i & 0x01);
#if defined(DEBUG_QOS) && DEBUG_QOS
			if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS))
				printk(
						"SCT: QSB_RTM (%08X) = 0x%08X, QSB_RTD (%08X) = 0x%08X, QSB_RAMAC (%08X) = 0x%08X\n",
						(u32) QSB_RTM, *QSB_RTM, (u32) QSB_RTD, *QSB_RTD,
						(u32) QSB_RAMAC, *QSB_RAMAC);
#endif
			/*  SPT (SBV + PN + IntRage)    */
			*QSB_RTM = QSB_RTM_DM_SET(QSB_SET_SPT_MASK);
			*QSB_RTD =
					QSB_RTD_TTV_SET(QSB_SPT_SBV_VALID | QSB_SPT_PN_SET(i & 0x01) | QSB_SPT_INTRATE_SET(tmp2));
			*QSB_RAMAC = QSB_RAMAC_RW_SET(QSB_RAMAC_RW_WRITE)
					| QSB_RAMAC_TSEL_SET(QSB_RAMAC_TSEL_SPT)
					| QSB_RAMAC_LH_SET(QSB_RAMAC_LH_LOW)
					| QSB_RAMAC_TESEL_SET(i & 0x01);
#if defined(DEBUG_QOS) && DEBUG_QOS
			if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS))
				printk(
						"SPT: QSB_RTM (%08X) = 0x%08X, QSB_RTD (%08X) = 0x%08X, QSB_RAMAC (%08X) = 0x%08X\n",
						(u32) QSB_RTM, *QSB_RTM, (u32) QSB_RTD, *QSB_RTD,
						(u32) QSB_RAMAC, *QSB_RAMAC);
#endif
		}
}

static INLINE void setup_oam_htu_entry(void) {
	struct htu_entry htu_entry = { 0 };
	struct htu_result htu_result = { 0 };
	struct htu_mask htu_mask = { set: 0x03, pid_mask: 0x00, vpi_mask: 0x00,
			vci_mask: 0x00, pti_mask: 0x00, clear: 0x00 };
	unsigned int i;

	/*
	 *  HTU Tables
	 */
	for (i = 0; i < g_atm_priv_data.max_connections; i++) {
		htu_result.qid = (unsigned int) i;

		*HTU_ENTRY(i + OAM_HTU_ENTRY_NUMBER) = htu_entry;
		*HTU_MASK(i + OAM_HTU_ENTRY_NUMBER) = htu_mask;
		*HTU_RESULT(i + OAM_HTU_ENTRY_NUMBER) = htu_result;
	}
	/*  OAM HTU Entry   */
	htu_entry.vci = 0x03;
	htu_mask.pid_mask = 0x03;
	htu_mask.vpi_mask = 0xFF;
	htu_mask.vci_mask = 0x0000;
	htu_mask.pti_mask = 0x07;
	htu_result.cellid = 0; //  need confirm
	htu_result.type = 1;
	htu_result.ven = 1;
	htu_result.qid = 0; //  need confirm
	*HTU_RESULT(OAM_F4_SEG_HTU_ENTRY) = htu_result;
	*HTU_MASK(OAM_F4_SEG_HTU_ENTRY) = htu_mask;
	*HTU_ENTRY(OAM_F4_SEG_HTU_ENTRY) = htu_entry;
	htu_entry.vci = 0x04;
	htu_result.cellid = 0; //  need confirm
	htu_result.type = 1;
	htu_result.ven = 1;
	htu_result.qid = 0; //  need confirm
	*HTU_RESULT(OAM_F4_TOT_HTU_ENTRY) = htu_result;
	*HTU_MASK(OAM_F4_TOT_HTU_ENTRY) = htu_mask;
	*HTU_ENTRY(OAM_F4_TOT_HTU_ENTRY) = htu_entry;
	htu_entry.vci = 0x00;
	htu_entry.pti = 0x04;
	htu_mask.vci_mask = 0xFFFF;
	htu_mask.pti_mask = 0x01;
	htu_result.cellid = 0; //  need confirm
	htu_result.type = 1;
	htu_result.ven = 1;
	htu_result.qid = 0; //  need confirm
	*HTU_RESULT(OAM_F5_HTU_ENTRY) = htu_result;
	*HTU_MASK(OAM_F5_HTU_ENTRY) = htu_mask;
	*HTU_ENTRY(OAM_F5_HTU_ENTRY) = htu_entry;
}

/*
 *  Description:
 *    Add HTU entries to capture OAM cell.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static INLINE void validate_oam_htu_entry(void) {
	HTU_ENTRY(OAM_F4_SEG_HTU_ENTRY)->vld = 1;
	HTU_ENTRY(OAM_F4_TOT_HTU_ENTRY)->vld = 1;
	HTU_ENTRY(OAM_F5_HTU_ENTRY)->vld = 1;
}

/*
 *  Description:
 *    Remove HTU entries which are used to capture OAM cell.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static INLINE void invalidate_oam_htu_entry(void) {
	register int i;

	HTU_ENTRY(OAM_F4_SEG_HTU_ENTRY)->vld = 0;
	HTU_ENTRY(OAM_F4_TOT_HTU_ENTRY)->vld = 0;
	HTU_ENTRY(OAM_F5_HTU_ENTRY)->vld = 0;
	/*  idle for a while to finish running HTU search   */
	for (i = 0; i < IDLE_CYCLE_NUMBER; i++)
		;
}

/*
 *  Description:
 *    Loop up for connection ID with virtual path ID.
 *  Input:
 *    vpi --- unsigned int, virtual path ID
 *  Output:
 *    int --- negative value: failed
 *            else          : connection ID
 */
static INLINE int find_vpi(int vpi) {
	unsigned int i;
	volatile struct atm_connection *connection = g_atm_priv_data.connection;

	for (i = 0; i < g_atm_priv_data.max_connections; i++)
		if ((g_atm_priv_data.connection_table & (1 << i)) && 
            (connection[i].vcc != NULL) && 
            (vpi == connection[i].vcc->vpi))
        {
			return i;
        }
	return -1;
}

/*
 *  Description:
 *    Loop up for connection ID with virtual path ID and virtual channel ID.
 *  Input:
 *    vpi --- unsigned int, virtual path ID
 *    vci --- unsigned int, virtual channel ID
 *  Output:
 *    int --- negative value: failed
 *            else          : connection ID
 */
static INLINE int find_vpivci(int vpi, int vci) {
	unsigned int i;
	volatile struct atm_connection *connection = g_atm_priv_data.connection;

	for (i = 0; i < g_atm_priv_data.max_connections; i++)
		if ((g_atm_priv_data.connection_table & (1 << i)) && 
            (connection[i].vcc != NULL) && 
            (vpi == connection[i].vcc->vpi) && 
            (vci == connection[i].vcc->vci)) 
        {
			return i;
        }
	return -1;
}

/*
 *  Description:
 *    Loop up for connection ID with atm_vcc structure.
 *  Input:
 *    vcc --- struct atm_vcc *, atm_vcc structure of opened connection
 *  Output:
 *    int --- negative value: failed
 *            else          : connection ID
 */
static INLINE int find_vcc(struct atm_vcc *vcc) {
	unsigned int i;
	volatile struct atm_connection *connection = g_atm_priv_data.connection;

	for (i = 0; i < g_atm_priv_data.max_connections; i++) {
		if ((g_atm_priv_data.connection_table & (1 << i)) && 
            (connection[i].vcc == vcc))
        {
			return i;
        }
    }
	return -1;
}

// ----------------------------  tx_alloc ----------------------------------
static INLINE struct sk_buff *alloc_skb_tx(int len)
{
    struct sk_buff *skb;
    unsigned int align_off;

    skb = dev_alloc_skb(len + DMA_TX_ALIGNMENT);

    if(likely(skb)){
        align_off = DMA_TX_ALIGNMENT - ((u32)skb->data & (DMA_TX_ALIGNMENT - 1));
        skb_reserve(skb, align_off);

        *((u32 *)skb->data - 1) = (u32)skb;
#ifndef CONFIG_MIPS_UNCACHED
        dma_cache_wback((u32)skb->data - sizeof(u32), sizeof(u32));
#endif
    }

    return skb;
}

// ----------------------------  rx_alloc ----------------------------------
static INLINE struct sk_buff *alloc_skb_rx(void)
{
    struct sk_buff *skb;
    unsigned int align_off;

    skb = dev_alloc_skb(DMA_PACKET_SIZE + DMA_RX_ALIGNMENT);
    if(likely(skb)){
        align_off = DMA_RX_ALIGNMENT - ((u32)skb->data & (DMA_RX_ALIGNMENT - 1));
        skb_reserve(skb, align_off);

        *((u32 *)skb->data - 1) = (u32)skb;
#ifndef CONFIG_MIPS_UNCACHED
        dma_cache_wback((u32)skb->data - sizeof(u32), sizeof(u32));
#endif
    }

    return skb;
}

static struct sk_buff* skb_break_away_from_protocol(struct sk_buff *skb)
{
    struct sk_buff *new_skb;

    if ( skb_shared(skb) ) {
        new_skb = skb_copy(skb, GFP_ATOMIC);
        if ( new_skb == NULL )
            return NULL;
    }
    else
        new_skb = skb_get(skb);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
    dst_release(new_skb->dst);
    new_skb->dst = NULL;
#else
    skb_dst_drop(new_skb);
#endif
#ifdef CONFIG_XFRM
    secpath_put(new_skb->sp);
    new_skb->sp = NULL;
#endif
    if ( new_skb->destructor )
    {
        WARN_ON(in_irq());
        new_skb->destructor(new_skb);
        new_skb->destructor = NULL;
    }
#if !defined(CONFIG_IFX_PPA_AVM_USAGE) && (defined(CONFIG_NETFILTER) || LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32))
    nf_conntrack_put(new_skb->nfct);
    new_skb->nfct = NULL;
  #if defined(CONFIG_NF_CONNTRACK) || defined(CONFIG_NF_CONNTRACK_MODULE)
    nf_conntrack_put_reasm(new_skb->nfct_reasm);
    new_skb->nfct_reasm = NULL;
  #endif
  #ifdef CONFIG_BRIDGE_NETFILTER
    nf_bridge_put(new_skb->nf_bridge);
    new_skb->nf_bridge = NULL;
  #endif
#endif

    return new_skb;
}

static INLINE void __get_skb_from_dbg_pool(
		struct sk_buff *skb __attribute__((unused)),
		const char *func_name __attribute__((unused)),
		unsigned int line_num __attribute__((unused)))
{
#if defined(DEBUG_SKB_SWAP) && DEBUG_SKB_SWAP
    int i;

    for ( i = 0; i < NUM_ENTITY(g_dbg_skb_swap_pool) && g_dbg_skb_swap_pool[i] != skb; i++ );
    if ( i == NUM_ENTITY(g_dbg_skb_swap_pool) || (unsigned int)skb < KSEG0 )
    {
        err("%s:%d: skb (0x%08x) not in g_dbg_skb_swap_pool", func_name, line_num, (unsigned int)skb);
    }
    else
        g_dbg_skb_swap_pool[i] = NULL;
#endif
}

static INLINE struct sk_buff *__get_skb_pointer(unsigned int dataptr, const char *func_name, unsigned int line_num)
{
    unsigned int skb_dataptr;
    struct sk_buff *skb;

    //  usually, CPE memory is less than 256M bytes
    //  so NULL means invalid pointer
    if ( dataptr == 0 ) {
        dbg("dataptr is 0, it's supposed to be invalid pointer");
        return NULL;
    }

    skb_dataptr = (dataptr - 4) | KSEG1;
    skb = *(struct sk_buff **)skb_dataptr;

    __get_skb_from_dbg_pool(skb, func_name, line_num);

	ASSERT((unsigned int)skb >= KSEG0, "%s:%d: invalid skb - skb = %#08x, dataptr = %#08x", func_name, line_num, (unsigned int)skb, dataptr);
	ASSERT( (((unsigned int)skb->data & (0x1FFFFFFF ^ (DMA_RX_ALIGNMENT - 1))) | KSEG1) == (dataptr | KSEG1) ||
			(((unsigned int)skb->data & (0x1FFFFFFF ^ (DMA_TX_ALIGNMENT - 1))) | KSEG1) == (dataptr | KSEG1),
			"%s:%d: invalid skb - skb = %#08x, skb->data = %#08x, dataptr = %#08x",
			func_name, line_num, (unsigned int)skb, (unsigned int)skb->data, dataptr
			);

    return skb;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static INLINE void __put_skb_to_dbg_pool(struct sk_buff *skb __attribute__((unused)),
		const char *func_name __attribute__((unused)), unsigned int line_num __attribute__((unused))) {
#if defined(DEBUG_SKB_SWAP) && DEBUG_SKB_SWAP
	int i;

	for ( i = 0; i < NUM_ENTITY(g_dbg_skb_swap_pool) && g_dbg_skb_swap_pool[i] != NULL; i++ );
	if ( i == NUM_ENTITY(g_dbg_skb_swap_pool) )
	{
		err("%s:%d: g_dbg_skb_swap_pool overrun", func_name, line_num);
	}
	else
	g_dbg_skb_swap_pool[i] = skb;
#endif
}

static struct sk_buff* atm_alloc_tx(struct atm_vcc *vcc, unsigned int size) {
	struct sk_buff *skb;

	/*  send buffer overflow    */
	if (atomic_read(&sk_atm(vcc)->sk_wmem_alloc) && !atm_may_send(vcc, size)) {
		err("send buffer overflow");
		return NULL;
	}

	skb = alloc_skb_tx(size < DMA_PACKET_SIZE ? DMA_PACKET_SIZE : size);
	if (skb == NULL) {
		err("sk buffer is used up");
		return NULL;
	}

	atomic_add(skb->truesize, &sk_atm(vcc)->sk_wmem_alloc);

	return skb;
}

static INLINE void atm_free_tx_skb_vcc(struct sk_buff *skb) {
	struct atm_vcc* vcc;

	ASSERT((u32)skb > 0x80000000, "atm_free_tx_skb_vcc: skb = %08X", (u32)skb);

	vcc = ATM_SKB(skb)->vcc;

	if (vcc != NULL && vcc->pop != NULL) {
		ASSERT(atomic_read(&skb->users) != 0,
				"atm_free_tx_skb_vcc(vcc->pop): skb->users == 0, skb = %08X",
				(u32)skb);
		vcc->pop(vcc, skb);
	} else {
		ASSERT(
				atomic_read(&skb->users) != 0,
				"atm_free_tx_skb_vcc(dev_kfree_skb_any): skb->users == 0, skb = %08X",
				(u32)skb);
		dev_kfree_skb_any(skb);
	}
}

static irqreturn_t mailbox0_irq_handler(int irq __attribute__((unused)), void *dev_id __attribute__((unused)))
{
    u32 mailbox_signal = 0;

    mailbox_signal = *MBOX_IGU0_ISR & *MBOX_IGU0_IER;
	dbg("you've got mail in box0 ;)");
    if ( !mailbox_signal )
        return IRQ_HANDLED;
    *MBOX_IGU0_ISRC = mailbox_signal;

    if ( (mailbox_signal & DMA_TX_CH0_SIG) )
    {
        g_dma_device_ppe->tx_chan[0]->open(g_dma_device_ppe->tx_chan[0]);
        *MBOX_IGU0_IER &= DMA_TX_CH1_SIG;
        dbg("DMA_TX_CH0_SIG"); // showtime signal b
    }

    if ( (mailbox_signal & DMA_TX_CH1_SIG) )
    {
        g_dma_device_ppe->tx_chan[1]->open(g_dma_device_ppe->tx_chan[1]);
        *MBOX_IGU0_IER &= DMA_TX_CH0_SIG;
        dbg("DMA_TX_CH1_SIG");
    }

    return IRQ_HANDLED;
}

/*
 *  Description:
 *    Handle IRQ of mailbox and despatch to relative handler.
 *  Input:
 *    irq    --- int, IRQ number
 *    dev_id --- void *, argument passed when registering IRQ handler
 *    regs   --- struct pt_regs *, registers' value before jumping into handler
 *  Output:
 *    none
 */
static irqreturn_t mailbox1_irq_handler(int irq __attribute__((unused)), void *dev_id __attribute__((unused))) {

    u32 mailbox_signal = 0;
    dbg("you've got mail in box1 ;)");
    mailbox_signal = *MBOX_IGU1_ISR & g_mailbox_signal_mask;
    if (!mailbox_signal)
        return IRQ_HANDLED;
    *MBOX_IGU1_ISRC = mailbox_signal;
    if ( (mailbox_signal & WAN_RX_SIG(0)) ) //  AAL5
    {
        g_mailbox_signal_mask &= ~WAN_RX_SIG(0);
        g_dma_device_ppe->tx_chan[1]->open(g_dma_device_ppe->tx_chan[1]);
        *MBOX_IGU1_IER = g_mailbox_signal_mask;
        dbg("WAN_RX_SIG(0) AAL5");
    }

    if ((mailbox_signal & WAN_RX_SIG(1))) //  OAM
    {
        volatile struct rx_descriptor *desc;
        struct sk_buff *skb;
        struct uni_cell_header *header;
        int conn;
        struct atm_vcc *vcc;

        dbg("WAN_RX_SIG(1): OAM");
        do {
            while (1) {
                desc = WAN_RX_DESC_BASE(1) + g_wan_rx1_desc_pos;
                if (desc->own) //  PPE hold
                    break;
                skb = g_wan_rx1_desc_skb[g_wan_rx1_desc_pos];
                if (++g_wan_rx1_desc_pos == WAN_RX_DESC_NUM(1))
                    g_wan_rx1_desc_pos = 0;

#if defined(DEBUG_DUMP_SKB) && DEBUG_DUMP_SKB
                skb->tail = skb->data + CELL_SIZE;
                skb->len = CELL_SIZE;
                dump_skb(skb, DUMP_SKB_LEN, "mailbox1_irq_handler - OAM", 7, 0, 0, 0);
#endif

                header = (struct uni_cell_header *) skb->data;
                if (header->pti == ATM_PTI_SEGF5 || header->pti == ATM_PTI_E2EF5)
                    conn = find_vpivci(header->vpi, header->vci);
                else if (header->vci == 0x03 || header->vci == 0x04)
                    conn = find_vpi(header->vpi);
                else
                    conn = -1;

                if (conn
                        >= 0 && (vcc = g_atm_priv_data.connection[conn].vcc) != NULL) {g_atm_priv_data.connection[conn].access_time = current_kernel_time();

                        if ( vcc->push_oam != NULL )
                            vcc->push_oam(vcc, skb->data);
                        else
                            ifx_push_oam(skb->data);
                }
                else {
                    dbg("OAM cell dropped due to wrong qid/conn");
                    oam_cell_drop_count++;
                }

                desc->c = 0;
                desc->own = 1;
            }
            *MBOX_IGU1_ISRC = WAN_RX_SIG(1);
        } while (!desc->own); //  check after clean, if MIPS hold, repeat
    }

    if ((mailbox_signal & CPU_TO_WAN_SWAP_SIG)) {
        struct sk_buff *new_skb;
        volatile struct tx_descriptor *desc =
                &CPU_TO_WAN_SWAP_DESC_BASE[g_cpu_to_wan_swap_desc_pos];
        struct tx_descriptor reg_desc = { 0 };
        struct sk_buff *skb_to_free;

        dbg("CPU_TO_WAN_SWAP_SIG");
        while (desc->own == 0) {
            new_skb = alloc_skb_rx();
            if (new_skb == NULL)
                break;

#ifndef CONFIG_MIPS_UNCACHED
            /*  invalidate cache    */
            dma_cache_inv((unsigned long)new_skb->data, DMA_PACKET_SIZE);
#endif

            if (++g_cpu_to_wan_swap_desc_pos == CPU_TO_WAN_SWAP_DESC_NUM)
                g_cpu_to_wan_swap_desc_pos = 0;

            /*  free previous skb   */
            skb_to_free = get_skb_pointer(desc->dataptr);
            if (skb_to_free != NULL)
                dev_kfree_skb_any(skb_to_free);
            put_skb_to_dbg_pool(new_skb);

            /*  update descriptor   */
            reg_desc.dataptr = (unsigned int) new_skb->data
                    & (0x1FFFFFFF ^ (DMA_TX_ALIGNMENT - 1));
            reg_desc.own = 1;

            /*  write discriptor to memory  */
            *((volatile unsigned int *) desc + 1) = *((unsigned int *) &reg_desc
                    + 1);
            wmb();
            *(volatile unsigned int *) desc = *(unsigned int *) &reg_desc;

            desc = &CPU_TO_WAN_SWAP_DESC_BASE[g_cpu_to_wan_swap_desc_pos];
        }
    }

    return IRQ_HANDLED;
}

static INLINE void turn_on_dma_rx(int mask) {
	unsigned int i;

	if (!g_f_dma_opened) {
		ASSERT((u32)g_dma_device_ppe >= 0x80000000, "g_dma_device_ppe = 0x%08X",
				(u32)g_dma_device_ppe);

		for (i = 0; i < g_dma_device_ppe->max_rx_chan_num; i++) {
			ASSERT((u32)g_dma_device_ppe->rx_chan[i] >= 0x80000000,
					"g_dma_device_ppe->rx_chan[%d] = 0x%08X",
					i, (u32)g_dma_device_ppe->rx_chan[i]);
			ASSERT(g_dma_device_ppe->rx_chan[i]->control == IFX_DMA_CH_ON,
					"g_dma_device_ppe->rx_chan[i]->control = %d",
					g_dma_device_ppe->rx_chan[i]->control);

			if (g_dma_device_ppe->rx_chan[i]->control == IFX_DMA_CH_ON) {
				ASSERT((u32)g_dma_device_ppe->rx_chan[i]->open >= 0x80000000,
						"g_dma_device_ppe->rx_chan[%d]->open = 0x%08X",
						i, (u32)g_dma_device_ppe->rx_chan[i]->open);

				//  channel 1, 2 is for fast path
/*---
 * Seems to be a lantiq hack
 *  1) put channel-config to tx-mode
 *  2) then open it (no irq is not enabled but channel is activated)
 *  3) put channnel-config to undefined
 * We use normal open and no_cpu_data flag for setting up directpath channels
 */
#if 0
				if (i == 1 || i == 2)
					g_dma_device_ppe->rx_chan[i]->dir = 1;
#endif
				g_dma_device_ppe->rx_chan[i]->open(g_dma_device_ppe->rx_chan[i]);
#if 0
				if (i == 1 || i == 2)
					g_dma_device_ppe->rx_chan[i]->dir = -1;
#endif
			}
		}
	}
	g_f_dma_opened |= 1 << mask;
}

static INLINE void turn_off_dma_rx(int mask) {
	unsigned int i;

	g_f_dma_opened &= ~(1 << mask);
	if (!g_f_dma_opened)
		for (i = 0; i < g_dma_device_ppe->max_rx_chan_num; i++)
			g_dma_device_ppe->rx_chan[i]->close(g_dma_device_ppe->rx_chan[i]);
}

static u8 *dma_buffer_alloc(int len __attribute__((unused)), int *byte_offset, void **opt) {
	u8 *buf;
	struct sk_buff *skb;

	skb = alloc_skb_rx();
	if (!skb)
		return NULL;
    
    put_skb_to_dbg_pool(skb);
	buf = (u8 *) skb->data;
	*(u32 *) opt = (u32) skb;
	*byte_offset = 0;
	return buf;
}

static int dma_buffer_free(u8 *dataptr __attribute__((unused)), void *opt) {
	//  ! be careful
	//    this method makes kernel crash when free this DMA device
	//    since the pointers used by fastpast which is supposed not visible to CPU
	//    mix up the pointers so that "opt" is not align with "dataptr".

//    ASSERT(*(void **)((((u32)dataptr | KSEG1) & ~0x0F) - 4) == opt, "address is not match: dataptr = %08X, opt = %08X, *(void **)((((u32)dataptr | KSEG1) & ~0x0F) - 4) = %08X", (u32)dataptr, (u32)opt, *(u32 *)((((u32)dataptr | KSEG1) & ~0x0F) - 4));

    if ( opt ) {
        get_skb_from_dbg_pool((struct sk_buff *)opt);
        dev_kfree_skb_any((struct sk_buff *)opt);
    }

	return 0;
}

static int dma_int_handler(struct dma_device_info *dma_dev, int status, int chan_no) {
	int ret = 0;
	unsigned long sys_flag;
#if !defined(ENABLE_DIRECTPATH_TX_QUEUE) || !ENABLE_DIRECTPATH_TX_QUEUE
	unsigned int i;
#endif

	switch (status) {
	case RCV_INT:
		ret = dma_rx_int_handler(dma_dev, chan_no);
		break;
	case TX_BUF_FULL_INT:
		if (chan_no == 2) {
			unsigned int dev_nr;
			dbg("eth0/1 TX buffer full!");
			local_irq_save(sys_flag);

            /*
             * TODO vgl. UGW - hier wurden trans_start Zeiten geswapt - warum? -
             * bei mehr als 2 devices ist das nicht mehr trivial -
             */
            for (dev_nr = 0; dev_nr < num_registered_eth_dev; dev_nr++){
                struct net_device *dev = g_eth_net_dev[dev_nr];
                avmnet_netdev_priv_t *priv = netdev_priv(dev);
                avmnet_device_t *avm_dev = avm_dev = priv->avm_dev;
                avm_dev->flags |= AVMNET_DEVICE_FLAG_DMA_TX_QUEUE_FULL;
                netif_stop_queue(dev);
            }

			if (g_dma_device_ppe->tx_chan[2]->control == IFX_DMA_CH_ON)
				g_dma_device_ppe->tx_chan[2]->enable_irq(g_dma_device_ppe->tx_chan[2]);
			local_irq_restore(sys_flag);
		} else if (chan_no == 3) {
			dbg("directpath TX (CPU->PPE) buffer full!");
			local_irq_save(sys_flag);
			if (!g_directpath_dma_full) {
				g_directpath_dma_full = 1;
				if (g_dma_device_ppe->tx_chan[3]->control == IFX_DMA_CH_ON) {
					g_dma_device_ppe->tx_chan[3]->enable_irq( g_dma_device_ppe->tx_chan[3]);
                }
#if !defined(ENABLE_DIRECTPATH_TX_QUEUE) || !ENABLE_DIRECTPATH_TX_QUEUE
				for (i = 0; i < push_through_devs_supported_by_datapath; i++) {
					if ( ppa_virt_rx_devs[i] )
					    avm_pa_rx_channel_suspend( ppa_virt_rx_devs[i]->pid_handle );
                }
#endif
			}
			local_irq_restore(sys_flag);
		} else {
			err( "incorrect DMA TX channel: %d (0 - 1 is reserved for fast path)", chan_no);
		}
		break;
	case TRANSMIT_CPT_INT:
		if (chan_no == 2) {
			unsigned int dev_nr;
			dbg("eth0/1 TX buffer released!");
			local_irq_save(sys_flag);
			g_dma_device_ppe->tx_chan[2]->disable_irq(g_dma_device_ppe->tx_chan[2]);

            for (dev_nr = 0; dev_nr < num_registered_eth_dev; dev_nr++){
                struct net_device *dev = g_eth_net_dev[dev_nr];
                avmnet_netdev_priv_t *priv = netdev_priv(dev);
                avmnet_device_t *avm_dev = avm_dev = priv->avm_dev;
                avm_dev->flags &= ~AVMNET_DEVICE_FLAG_DMA_TX_QUEUE_FULL;
                netif_wake_queue( dev );
            }

			local_irq_restore(sys_flag);
		} else if (chan_no == DMA_CHAN_NR_DIRECTPATH_CPU_TO_PPE) {
			dbg("directpath TX (CPU->PPE) buffer released");
			local_irq_save(sys_flag);
			if (g_directpath_dma_full) {
				g_directpath_dma_full = 0;
				g_dma_device_ppe->tx_chan[3]->disable_irq(g_dma_device_ppe->tx_chan[3]);
#if !defined(ENABLE_DIRECTPATH_TX_QUEUE) || !ENABLE_DIRECTPATH_TX_QUEUE
				for (i = 0; i < push_through_devs_supported_by_datapath; i++) {
					if ( ppa_virt_rx_devs[i] )
					    avm_pa_rx_channel_resume( ppa_virt_rx_devs[i]->pid_handle );
                }
#endif
			}
			local_irq_restore(sys_flag);
#if defined(ENABLE_DIRECTPATH_TX_QUEUE) && ENABLE_DIRECTPATH_TX_QUEUE
			send_directpath_tx_queue();
#endif
		} else {
			err(
					"incorrect DMA TX channel: %d (0 - 1 is reserved for fast path)",
					chan_no);
		}
		break;
	default:
		err("unkown DMA interrupt event - %d", status);
	}

	return ret;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

void dump_atm_tx_desc( struct tx_descriptor *d __attribute__((unused))){
#if defined(CONFIG_AVMNET_DEBUG) 
	if (!(g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_TX))
		return;

	printk(KERN_ERR "[%s] called by %pF", __FUNCTION__, __builtin_return_address(0));
	printk(KERN_ERR "   own:     %d\n", d->own);
	printk(KERN_ERR "   c:       %d\n", d->c);
	printk(KERN_ERR "   sop:     %d\n", d->sop);
	printk(KERN_ERR "   eop:     %d\n", d->eop);
	printk(KERN_ERR "   byteoff: %d\n", d->byteoff);
	printk(KERN_ERR "   mpoa_pt: %d\n", d->mpoa_pt);
	printk(KERN_ERR "   mpoa_ty: %d\n", d->mpoa_type);
	printk(KERN_ERR "   pdu_ty:  %d\n", d->pdu_type);
	printk(KERN_ERR "   qid:     %d\n", d->qid);
	printk(KERN_ERR "   datalen: %d\n", d->datalen);
#endif
}


/*------------------------------------------------------------------------------------------*\
 * wird auschliesslich von der RX-Routine verwendet
\*------------------------------------------------------------------------------------------*/
void ifx_ppa_atm_hard_start_xmit(struct sk_buff *skb,
		struct flag_header *fheader) {
	unsigned long sys_flag;
	volatile struct tx_descriptor *desc;
	struct tx_descriptor reg_desc;
	struct sk_buff *skb_to_free;
	int byteoff;
	int conn;

//	conn = (header_val >> 16) & 0x07;
	conn = fheader->qid & 0x07;
	dbg_trace_ppa_data("atm start xmit");

#if defined(ENABLE_DFE_LOOPBACK_TEST_INDIRECT) && ENABLE_DFE_LOOPBACK_TEST_INDIRECT
	{
		unsigned char *p = (unsigned char *)skb->data;

		if ( WRX_QUEUE_CONFIG(conn)->mpoa_type < 2 ) //  EoA
		{
			unsigned char mac[6];

			//  mac
			memcpy(mac, p, 6);
			memcpy(p, p + 6, 6);
			memcpy(p + 6, mac, 6);
#ifndef CONFIG_MIPS_UNCACHED
			dma_cache_wback((unsigned long)p, 12);
#endif
		}
		if ( fheader->mpoa_type != 2 ) //  EoA or IPoA
		{
			unsigned char ip[4];
			unsigned char port[2];

			//  IP
			p += fheader->ip_offset + 8; //hier nicht ip_inner_offset
			memcpy(ip, p + 4, 4);
			memcpy(p + 4, p + 8, 4);
			memcpy(p + 8, ip, 4);

			//  port
			if ( p[1] == 0x06 || p[1] == 0x11 )//  TCP or UDP
			{
				memcpy(port, p + 12, 2);
				memcpy(p + 12, p + 14, 2);
				memcpy(p + 14, port, 2);
			}
#ifndef CONFIG_MIPS_UNCACHED
			dma_cache_wback((unsigned long)p, 22);
#endif
		}
	}
#endif


	byteoff = (u32) skb->data & (DMA_TX_ALIGNMENT - 1);
	ASSERT(skb_headroom(skb) >= sizeof(struct sk_buff *) + byteoff,
				"   skb_headroom < sizeof(struct sk_buff *) +  byteoff \n");

#if defined(CONFIG_AVMNET_DEBUG) 
	if ( g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_ALIGNMENT )
		printk( KERN_ERR
			"\n"
			"   skb_headroom :   %d \n"
			"   sizeof(struct sk_buff *): %d \n"
			"   byteoff:         %d \n"
			"   skb->head:       %p\n"
			"   skb->data:       %p\n"
			"   in_dev:          %s\n",
				skb_headroom(skb),
				sizeof(struct sk_buff *),
				byteoff,
				skb->head,
				skb->data,
				skb->input_dev?skb->input_dev->name:"no_input_device");
#endif

	*(struct sk_buff **) ((u32) skb->data - byteoff - sizeof(struct sk_buff *)) =
			skb;
#ifndef CONFIG_MIPS_UNCACHED
	/*  write back to physical memory   */
	dma_cache_wback((u32)skb->data - byteoff - sizeof(struct sk_buff *),
			sizeof(struct sk_buff *));
#endif

	/*  allocate descriptor */
	local_irq_save(sys_flag);
	desc = CPU_TO_WAN_TX_DESC_BASE + g_cpu_to_wan_tx_desc_pos;
	if (!desc->own) //  PPE hold
	{
		local_irq_restore(sys_flag);
		err("PPE hold");
		dev_kfree_skb_any(skb);
		return;
	}
	if (++g_cpu_to_wan_tx_desc_pos == CPU_TO_WAN_TX_DESC_NUM)
		g_cpu_to_wan_tx_desc_pos = 0;
	local_irq_restore(sys_flag);

	/*  load descriptor from memory */
	reg_desc = *desc;

	/*  free previous skb   */
	ASSERT((reg_desc.dataptr & (DMA_TX_ALIGNMENT - 1)) == 0,
			"reg_desc.dataptr (0x%#x) must be %d bytes aligned",
			reg_desc.dataptr, DMA_TX_ALIGNMENT);

	if ((reg_desc.dataptr & (DMA_TX_ALIGNMENT - 1)) == 0)
		dma_alignment_atm_good_count++;
	else
		dma_alignment_atm_bad_count++;

	skb_to_free = get_skb_pointer(reg_desc.dataptr);
	dev_kfree_skb_any(skb_to_free);
	put_skb_to_dbg_pool(skb);

	/*  update descriptor   */
	reg_desc.dataptr = (u32) skb->data & (0x1FFFFFFF ^ (DMA_TX_ALIGNMENT - 1)); //  byte address
	reg_desc.byteoff = byteoff;
	reg_desc.datalen = skb->len;
	reg_desc.mpoa_pt = 0; //  firmware apply MPoA
	reg_desc.mpoa_type = WRX_QUEUE_CONFIG(conn)->mpoa_type;
	reg_desc.pdu_type = 0; //  AAL5
	reg_desc.qid = conn;
	reg_desc.own = 0;
	reg_desc.c = 0;
	reg_desc.sop = reg_desc.eop = 1;

	dump_atm_tx_desc (&reg_desc) ;
#if 0
	/*  update MIB  */
	UPDATE_VCC_STAT(conn, tx_packets, 1);
	UPDATE_VCC_STAT(conn, tx_bytes, skb->len);
#endif

#if defined(CONFIG_AVMNET_DEBUG) 
	if (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_TX)
		dump_skb(skb, DUMP_SKB_LEN, "ppe_send", 7, 0, 1, 1);
#endif

	/*  write discriptor to memory and write back cache */
	*((volatile u32 *) desc + 1) = *((u32 *) &reg_desc + 1);
	*(volatile u32 *) desc = *(u32 *) &reg_desc;

	adsl_led_flash();
	return;

}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static INLINE int dma_rx_int_handler(struct dma_device_info* dma_dev, int dma_chan_no)
{
    unsigned char* buf = NULL;
    int len = 0;
    struct sk_buff *skb = NULL;
    struct softnet_data *queue;
    unsigned long flags;

#if defined(TRACE_RX_PATH)
    static int once = 5;
    if (once) {
        dump_stack();
        once--;
    }
#endif


    len = dma_device_read(dma_dev, &buf, (void**)&skb, dma_chan_no);

    dbg_dma( "[%s] dma_device_read len:%d, skb->len:%d\n",__func__,len, skb->len);

    if(unlikely(skb == NULL)){
        err( "%s[%d]: Can't restore the Packet !!!\n", __func__,__LINE__);
        return 0;
    }

    get_skb_from_dbg_pool(skb);

    if(unlikely(g_stop_datapath)){
        dbg("[%s]: datapath to be stopped, dropping skb.\n", __func__);
        dev_kfree_skb_any(skb);
        return 0;
    }

    if(unlikely((len > (skb->end - skb->data)) || (len < 64))){
        err( "%s[%d]: Packet is too large/small (%d)!!!\n", __func__,__LINE__,len);
        dev_kfree_skb_any(skb);
        return 0;
    }

    /* do not remove CRC */
    // len -= 4;
    if(unlikely(len > (skb->end - skb->tail))){
        err( "%s[%d]: len:%d end:%p tail:%p Err!!!\n", __func__, __LINE__, len, skb->end, skb->tail);
        dev_kfree_skb_any(skb);
        return 0;
    }

    skb_put(skb, len);

    if(buf){
        // dbg("rx len:%d\n",len);
        dump_skb(skb, DUMP_SKB_LEN, "dma_rx_int_handler", 0, dma_chan_no, 0, 0);
    }

    local_irq_save(flags);
    queue = &__get_cpu_var(g_7port_eth_queue);

    if(likely(queue->input_pkt_queue.qlen <= 1000)){
        if(unlikely(queue->input_pkt_queue.qlen == 0)){
            napi_schedule(&queue->backlog);
        }

        __skb_queue_tail(&queue->input_pkt_queue, skb);
    } else {
        dev_kfree_skb_any(skb);
    }

    local_irq_restore(flags);

    return 0;
}

static int process_backlog_a5(struct napi_struct *napi, int quota)
{
    struct sk_buff *skb = NULL;
    struct softnet_data *queue = &__get_cpu_var(g_7port_eth_queue);
    unsigned long start_time = jiffies;

    int work = 0;
    unsigned long flags;

    napi->weight = 64;

    do{
        local_irq_save(flags);
        skb = __skb_dequeue(&queue->input_pkt_queue);

        if (!skb){
            __napi_complete(napi);
            local_irq_restore(flags);
            break;
        }
        local_irq_restore(flags);

        handle_skb_a5(skb);
    }while (++work < quota && jiffies == start_time);

    return work;
}

/*------------------------------------------------------------------------------------------*\
 * merged handler
\*------------------------------------------------------------------------------------------*/

static INLINE int handle_skb_a5(struct sk_buff *skb)
{
	struct flag_header *fheader;
	struct pmac_header *pheader;
	avmnet_device_t *avm_dev = NULL;
	int conn;
	int dump_enforce __attribute__((unused)) = 0;
    unsigned int exp_len, hw_dev_id;

	dbg_trace_ppa_data("[%s] s\n", __FUNCTION__);

	dump_skb(skb, DUMP_SKB_LEN, "handle_skb_a5 raw data", 0, 0, 0, dump_enforce);
	dump_flag_header(g_fwcode, (struct flag_header *) skb->data, __FUNCTION__, 0);

    // fetch flag_header from skb
    fheader = (struct flag_header *) skb->data;

    // fetch pmac_header from skb
    pheader = (struct pmac_header *) (skb->data + fheader->pl_byteoff);

    dump_pmac_header(pheader, __FUNCTION__);

   /* 
    * in indirect fast path pheader->len and fheader->len doesn't contain valid packet length, 
    * so we can do PMAC header check only in unaccelerated packets.
    */
    if ( ! fheader->acc_done ) {
       
       /*
        * calculate expected SKB len.
        * Payload offset from fheader + packet length from PMAC header + 4 bytes for frame crc
        */
        exp_len = fheader->pl_byteoff + (pheader->PKT_LEN_HI << 8) + pheader->PKT_LEN_LO + 4u;

        if(exp_len != skb->len ){
            printk(KERN_ERR "[%s] Malformed PMAC header: exp_len=%d, real_len=%d\n", __func__,exp_len,skb->len );
            dump_skb( skb, DUMP_SKB_LEN, "handle_skb_e5", 0, -1, 0, 1);
            dev_kfree_skb_any(skb);
            return 0;
        }
    }
    
    skb_pull(skb, fheader->pl_byteoff + sizeof(struct pmac_header));

#if defined(CONFIG_AVMNET_DEBUG) 
	if ((g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_RX) && (( fheader->src_itf == 6 ) || ( fheader->src_itf == 7 ) ))
		dump_enforce = 1;

	dump_skb(skb, DUMP_SKB_LEN, "handle_skb_a5 packet data", fheader->src_itf, 0, 0, dump_enforce);
#endif

	//  implement indirect fast path
	if (fheader->acc_done && fheader->dest_list) //  acc_done == 1 && dest_list != 0
			{
		//  Current implementation send packet to highest priority dest only (0 -> 7, 0 is highest)
		//  2 is CPU, so we ignore it here
		if ((fheader->dest_list & (1 << 0))) //  0 - eth0
		{
			eth_xmit(skb, 0, DMA_CHAN_NR_DEFAULT_CPU_TX, 2, fheader->qid);
			return 0;
		} else if ((fheader->dest_list & (1 << 1))) //  1 - eth1
		{
			eth_xmit(skb, 1, DMA_CHAN_NR_DEFAULT_CPU_TX, 2, fheader->qid);
			return 0;
		} else if ((fheader->dest_list & (1 << 7))) //  7 - DSL
		{
			ifx_ppa_atm_hard_start_xmit(skb, fheader);
			return 0;
		} else {
			//  fill MAC info into skb, in case the driver hooked to direct path need such info

			int dst_id;

			for (dst_id = 3; dst_id < 7; dst_id++) {
				if ((fheader->dest_list & (1 << dst_id))) //  3...6 (einschliesslich 3 u. 6)
				{
					int if_id = HW_DEV_ID(dst_id);

					if ( ifx_lookup_routing_entry && ppa_virt_tx_devs[ if_id ] ){
					    avm_session_handle avm_pa_session;

						//TODO ab hier ist mit adsl dazugekommen
						//  detect ATM port and remove the tag having PVC info
						if (fheader->src_itf == 7 && fheader->is_udp == 0
								&& fheader->is_tcp == 0 && skb->data[12] == 0x81
								&& skb->data[13] == 0x00
								&& (skb->data[14] & 0x0F) == 0x0F
								&& (skb->data[15] & 0xF0) == 0xF0) {
							int i;
							dbg_trace_ppa_data("remove outer VLAN tag");
							//  remove outer VLAN tag
							for (i = 11; i >= 0; i--)
								skb->data[i + 4] = skb->data[i];
							skb_pull(skb, 4);
						}
						//TODO bis hier ist mit adsl dazugekommen
			            
						dbg_trace_ppa_data("Achtung ungesteter Code!\n");

    					dbg_trace_ppa_data("ACC-Done: ifid=%d, avm_rout_index=%d, is_mc=%d, is_lan=%d",
    					        if_id,
    					        fheader->avm_rout_index,
    					        fheader->avm_is_mc,
    					        fheader->avm_is_lan );
    					ppa_virt_tx_devs[ if_id ]->hw_pkt_tx ++;

                      	avm_pa_session = ifx_lookup_routing_entry(fheader->avm_rout_index, fheader->avm_is_lan, fheader->avm_is_mc );

                    	if (avm_pa_session != INVALID_SESSION ){
						    avm_pa_tx_channel_accelerated_packet( ppa_virt_tx_devs[ if_id ]->pid_handle, avm_pa_session, skb );
                    	} else {
                    	    dev_kfree_skb_any(skb);
                    	    ppa_virt_tx_devs[ if_id ]->hw_pkt_tx_session_lookup_failed ++;
                    	}

						return 0;
					}
				}
			}
			dev_kfree_skb_any(skb);
			return 0;
		}
	}
	switch (fheader->src_itf) {
	case 0: //  incomming LAN IF ( vom Switch klassifizerit )
	case 1: //  incomming WAN IF ( vom Switch klassifizerit )
		avm_dev = mac_nr_to_avm_dev(pheader->SPPID);

		// TKL: FIXME workaround gegen Haenger beim Laden der Wasp-FW, check auf carrier_ok
		if (avm_dev && avm_dev->device && netif_running(avm_dev->device) && netif_carrier_ok(avm_dev->device)) {
			//  do not need to remove CRC
			//skb->tail -= ETH_CRC_LENGTH;
			//skb->len  -= ETH_CRC_LENGTH;
			unsigned int keep_len = 0;

			dbg_trace_ppa_data("got receive if: avm_name:%s ifname:%s\n",
					avm_dev->device_name, avm_dev->device->name);
			skb->dev = avm_dev->device;
#           if defined(CONFIG_IP_MULTICAST_FASTFORWARD)
            {
                avmnet_netdev_priv_t *priv = netdev_priv(avm_dev->device);

                if(likely(priv->mcfw_used)) {
                    mcfw_snoop_recv(&priv->mcfw_netdrv, 0, skb);
                }
            }
#           endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) ---*/
			skb->protocol = eth_type_trans(skb, avm_dev->device);

			if (skb->len > 0) {
				keep_len = skb->len;
            } else {
				dbg_trace_ppa_data(KERN_ERR "[%s] skb->len <= 0: %d", __FUNCTION__, skb->len);
            }

			if(likely(avm_pa_dev_receive(AVM_PA_DEVINFO(avm_dev->device), skb) == AVM_PA_RX_ACCELERATED)) {
                avm_dev->device->last_rx = jiffies;
                return 0;
            }

			avm_dev->device_stats.rx_packets++;
			avm_dev->device_stats.rx_bytes += keep_len;
			netif_receive_skb(skb);
			return 0;
		} else {
			dbg_trace_ppa_data(KERN_ERR "[%s] netif %d down drop packet\n", __FUNCTION__, fheader->src_itf);
		}
		break;
	case 6: // ATM (IPoA, PPPoA)
	case 7: // ATM (EoA)

		if (fheader->src_itf == 6) {
			conn = skb->data[11] & 0x3F; //  QID: 6 bits
			dbg_trace_ppa_data("src_itf = 6; conn = %d", conn);
		} else {
			unsigned int i;
            int j;

			ASSERT(skb->data[12] == 0x81 && skb->data[13] == 0x00, "EoA traffic: no outer VLAN for PVCs differentiation");
			//  outer VLAN
#if defined(ENABLE_CONFIGURABLE_DSL_VLAN) && ENABLE_CONFIGURABLE_DSL_VLAN

			if (skb->data[14] != 0x0F || (skb->data[15] & 0xF0) != 0xF0) {
				unsigned short vlan_id = ((unsigned short) skb->data[14] << 8) | (unsigned short) skb->data[15];
				for (i = 0; i < NUM_ENTITY(g_dsl_vlan_qid_map); i += 2) {
					if (g_dsl_vlan_qid_map[i] == vlan_id) {
						skb->data[15] = (unsigned char) g_dsl_vlan_qid_map[i + 1];
						dbg_trace_ppa_data("setup vlan id");
						break;
					}
                }
				ASSERT(i < NUM_ENTITY(g_dsl_vlan_qid_map), "search DSL VLAN table fail");
			}
#endif
			conn = skb->data[15] & 0x0F;
			//  remove outer VLAN tag (die ersten 12 Byte werden um 4 byte nach hinten geschoben (ueber skb_data[12;15])
			dbg_trace_ppa_data("remove outer VLAN tag II");
			for (j = 11; j >= 0; j--)
				skb->data[j + 4] = skb->data[j];
			skb_pull(skb, 4);
		}
		if (conn < 0 || conn >= ATM_QUEUE_NUMBER) {
			dbg("DMA channel %d, ATM conn (%d) is out of range 0 - %d", fheader->src_itf, conn, ATM_QUEUE_NUMBER - 1);
			break;
		}
		if ( g_atm_priv_data.connection[conn].vcc != NULL ) {
			dbg_trace_ppa_data("conn = %d, vcc = %p", conn, g_atm_priv_data.connection[conn].vcc );
			if (g_atm_priv_data.connection[conn].vcc && atm_charge( g_atm_priv_data.connection[conn].vcc, skb->truesize)) {
				//err("reminder: how to handle raw packet without flag aal5_raw?");
				//ASSERT(vcc->qos.aal == ATM_AAL5 || (header_val & (1 << 23)), "ETH packet comes in from ATM port non-AAL5 VC");

				ATM_SKB(skb)->vcc = g_atm_priv_data.connection[conn].vcc;

				//if ( (header_val & (1 << 23)) )
				//{
				//    /*  AAL 5 raw packet    */
				//    //skb->tail -= ETH_CRC_LENGTH;
				//    //skb->len  -= ETH_CRC_LENGTH;
				//}
				//else
				{
					/*  ETH packet, need recover ATM encapsulation  */
					if (WRX_QUEUE_CONFIG(conn)->mpoa_mode) {
						unsigned int proto_type;
						dbg_trace_ppa_data("mpoa_mode, type=%d", WRX_QUEUE_CONFIG(conn)->mpoa_type);

						/*  LLC */
						switch (WRX_QUEUE_CONFIG(conn)->mpoa_type) {

						case 0: //  EoA w/o FCS
							ASSERT(
									skb_headroom(skb) >= 10,
									"not enough skb headroom (LLC EoA w/o FCS)");
							ASSERT(
									((u32)skb->data & 0x03) == 2,
									"not aligned data pointer (LLC EoA w/o FCS)");
							skb->data -= 10;
							skb->len += 10;
							//skb->tail -= ETH_CRC_LENGTH;
							//skb->len  += 10 - ETH_CRC_LENGTH;
							((u32 *) skb->data)[0] = 0xAAAA0300;
							((u32 *) skb->data)[1] = 0x80C20007;
							((u16 *) skb->data)[4] = 0x0000;
							break;
						case 1: //  EoA w FCS
							ASSERT(skb_headroom(skb) >= 10,
									"not enough skb headroom (LLC EoA w FCS)");
							ASSERT(((u32)skb->data & 0x03) == 2,
									"not aligned data pointer (LLC EoA w FCS)");
							skb->data -= 10;
							skb->len += 10;
							((u32 *) skb->data)[0] = 0xAAAA0300;
							((u32 *) skb->data)[1] = 0x80C20001;
							((u16 *) skb->data)[4] = 0x0000;
							break;
						case 2: //  PPPoA
							switch ((proto_type = ntohs(
									*(u16 *) ( skb->data + 12 )))) {
							case 0x0800: // EtherType=IPv4-Tag
								dbg_trace_ppa_data("proto_type %#x", proto_type);
								proto_type = 0x0021;
								break;
							case 0x86DD: // EtherType=IPv6-Tag
								dbg_trace_ppa_data("proto_type %#x", proto_type);
								proto_type = 0x0057;
								break;
							}
							dbg_trace_ppa_data("final proto_type %#x", proto_type);
							ASSERT(
									fheader->ip_inner_offset - 6 >= 0 || skb_headroom(skb) >= 6 - fheader->ip_inner_offset,
									"not enough skb headroom (LLC PPPoA)");
							skb->data += 0x0E - 6;
							skb->len -= 0x0E - 6;
							//skb->tail -= ETH_CRC_LENGTH;
							//skb->len  -= 0x0E - 6 + ETH_CRC_LENGTH;
							ASSERT(((u32)skb->data & 0x03) == 2,
									"not aligned data pointer (LLC PPPoA)");
							((u32 *) skb->data)[0] = 0xFEFE03CF;
							((u16 *) skb->data)[2] = (u16) proto_type;
							break;
						case 3: //  IPoA
							ASSERT(
									fheader->ip_inner_offset - 8 >= 0 || skb_headroom(skb) >= 8 - fheader->ip_inner_offset,
									"not enough skb headroom (LLC IPoA)");
							skb->data += fheader->ip_inner_offset - 8;
							skb->len -= fheader->ip_inner_offset - 8;
							//skb->tail -= ETH_CRC_LENGTH;
							//skb->len  -= header->ip_inner_offset - 8 + ETH_CRC_LENGTH;
							ASSERT(((u32)skb->data & 0x03) == 0,
									"not aligned data pointer (LLC IPoA)");
							((u32 *) skb->data)[0] = 0xAAAA0300;
							//((u32 *)skb->data)[1] = header->is_ipv4 ? 0x00000800 : 0x000086DD;
							((u16 *) skb->data)[2] = 0x0000;
							break;
						}
					} else {
						unsigned int proto_type;

						dbg_trace_ppa_data("no mpoa_mode, type=%d, ip_inner_offset=%d",
								WRX_QUEUE_CONFIG(conn)->mpoa_type, fheader->ip_inner_offset);
						switch (WRX_QUEUE_CONFIG(conn)->mpoa_type) {
						case 0: //  EoA w/o FCS
							ASSERT(
									skb_headroom(skb) >= 2,
									"not enough skb headroom (VC-mux EoA w/o FCS)");
							ASSERT(
									((u32)skb->data & 0x03) == 2,
									"not aligned data pointer (VC-mux EoA w/o FCS)");
							skb->data -= 2;
							skb->len += 2;
							//skb->tail -= ETH_CRC_LENGTH;
							//skb->len  += 2 - ETH_CRC_LENGTH;
							*(u16 *) skb->data = 0x0000;
							break;
						case 1: //  EoA w FCS
							ASSERT(
									skb_headroom(skb) >= 2,
									"not enough skb headroom (VC-mux EoA w FCS)");
							ASSERT(
									((u32)skb->data & 0x03) == 2,
									"not aligned data pointer (VC-mux EoA w FCS)");
							skb->data -= 2;
							skb->len += 2;
							*(u16 *) skb->data = 0x0000;
							break;
						case 2: //  PPPoA
							dbg_trace_ppa_data("got PPPoA");
							switch ((proto_type = ntohs(
									*(u16 *) (skb->data + 12)))) {
							case 0x0800: // EtherType=IPv4-Tag
								proto_type = 0x0021;
								break;
							case 0x86DD: // EtherType=IPv6-Tag
								proto_type = 0x0057;
								break;
							}
							dbg_trace_ppa_data("got PPPoA: proto_type=%#x", proto_type);
							ASSERT(
									fheader->ip_inner_offset - 2 >= 0 || skb_headroom(skb) >= 2 - fheader->ip_inner_offset,
									"not enough skb headroom (VC-mux PPPoA)");
							skb->data += 0x0E - 2;
							skb->len -= 0x0E - 2;
							// skb->tail -= ETH_CRC_LENGTH;
							// skb->len  -= 0x0E - 2 + ETH_CRC_LENGTH;
							ASSERT(((u32)skb->data & 0x03) == 2,
									"not aligned data pointer (VC-mux PPPoA)");
							*(u16 *) skb->data = (u16) proto_type;
							break;
						case 3: //  IPoA oder RAW-Packet fuer kdsld-probe
							if ( fheader->ip_inner_offset ) {
								skb->data += fheader->ip_inner_offset ;
								skb->len -= fheader->ip_inner_offset ;
								//skb->tail -= ETH_CRC_LENGTH;
								//skb->len  -= header->ip_inner_offset + ETH_CRC_LENGTH;
							} else {
								int default_raw_offset = 14;
								skb->data += default_raw_offset;
								skb->len -= default_raw_offset;
								ASSERT( (skb->len == fheader->pkt_len - 4), "raw offset does not fit to packet length" );
								dbg_trace_ppa_data("seems to be raw packet, default_raw_offset=%d (current_ip_inner_offset = %d)", default_raw_offset, fheader->ip_inner_offset);
							}
							break;
						}
					}
					dump_skb(skb, DUMP_SKB_LEN, "dma_rx_int_handler AAL5 encapsulated", fheader->src_itf, 0, 0, dump_enforce);
				}

				g_atm_priv_data.connection[conn].access_time =
						current_kernel_time();

				adsl_led_flash();

				if (g_atm_priv_data.connection[conn].vcc && g_atm_priv_data.connection[conn].vcc->stats)
					atomic_inc(&g_atm_priv_data.connection[conn].vcc->stats->rx);


				if (g_atm_priv_data.connection[conn].vcc && g_atm_priv_data.connection[conn].vcc->qos.aal == ATM_AAL5)
					g_atm_priv_data.wrx_pdu++;

				UPDATE_VCC_STAT(conn, rx_packets, 1);
				UPDATE_VCC_STAT(conn, rx_bytes, skb->len);

				if (g_atm_priv_data.connection[conn].vcc && g_atm_priv_data.connection[conn].vcc->push) {
                    dbg_trace_ppa_data("push, %pF", g_atm_priv_data.connection[conn].vcc->push);
    				g_atm_priv_data.connection[conn].vcc->push(g_atm_priv_data.connection[conn].vcc, skb);
                } else {
                    dbg_trace_ppa_data("dropping packet -> vcc seesm to be closed");
	                dev_kfree_skb_any(skb);
                }

				return 0;
			} else {
				dbg_trace_ppa_data("inactive qid %d", conn);

				if ( g_atm_priv_data.connection[conn].vcc && g_atm_priv_data.connection[conn].vcc->stats)
					atomic_inc( &g_atm_priv_data.connection[conn].vcc->stats->rx_drop );

				if ( g_atm_priv_data.connection[conn].vcc && g_atm_priv_data.connection[conn].vcc->qos.aal == ATM_AAL5)
					g_atm_priv_data.wrx_drop_pdu++;

				UPDATE_VCC_STAT(conn, rx_dropped, 1);
			}
		}

		break;
	case 3: //  WLAN
	default: //  other interface receive
        hw_dev_id = HW_DEV_ID(fheader->src_itf);

        if (hw_dev_id < ARRAY_SIZE(ppa_virt_rx_devs) && ppa_virt_rx_devs[hw_dev_id]){
			dbg_trace_ppa_data("PPE_DIRECTPATH_DATA_RX_ENABLE: src_itf=%d", fheader->src_itf );

            ppa_virt_rx_devs[hw_dev_id]->hw_pkt_slow_cnt ++;
            avm_pa_rx_channel_packet_not_accelerated(ppa_virt_rx_devs[hw_dev_id]->pid_handle, skb);
			return 0;
		}else {
			dbg_trace_ppa_data("!PPE_DIRECTPATH_DATA_RX_ENABLE: src_itf=%d", fheader->src_itf );
		}
	}

	if (avm_dev)
		avm_dev->device_stats.rx_dropped++;
	dev_kfree_skb_any(skb);
	return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if 1
static int pseudo_dma_int_handler(struct dma_device_info *dma_dev, int status, int chan_no) {
    int ret = 0;
    struct sk_buff *skb = NULL;
    u8 *buf;

    switch ( status )
    {
    case RCV_INT:
        dma_device_read(dma_dev, &buf, (void **)&skb, chan_no);
        if ( skb )
            dev_kfree_skb_any(skb);
    }

    return ret;
}

static INLINE void reset_ppe(void) {
    struct dma_device_info *dma_device = NULL;
    unsigned long org_jiffies;
    unsigned int i;

    //if ( (*DM_RXCFG & 0xFF) == 0x30 )
    //    return;

    dma_device = dma_device_reserve("PPE");
    if ( !dma_device )
    {
        err("Reserve DMA for PPE failed");
        return;
    }

    dma_device->buffer_alloc    = dma_buffer_alloc;
    dma_device->buffer_free     = dma_buffer_free;
    dma_device->intr_handler    = pseudo_dma_int_handler;
    dma_device->tx_burst_len    = 8;
    dma_device->rx_burst_len    = 8;
    dma_device->port_packet_drop_enable = 1;

    for ( i = 0; i < dma_device->max_rx_chan_num; i++ )
    {
        dma_device->rx_chan[i]->packet_size = DMA_PACKET_SIZE;
        dma_device->rx_chan[i]->control     = IFX_DMA_CH_ON;
    }
    for ( i = 0; i < dma_device->max_tx_chan_num; i++ )
        dma_device->tx_chan[i]->control     = IFX_DMA_CH_ON;

    dma_device_register(dma_device);

    for ( i = 0; i < dma_device->max_rx_chan_num; i++ )
        dma_device->rx_chan[i]->open(dma_device->rx_chan[i]);

    org_jiffies = jiffies;
    while ( (IFX_REG_R32(DM_RXPKTCNT) & 0x0FF) != 0 && jiffies - org_jiffies < HZ / 10 + 1 )
        schedule();

    //  reset PPE
	ifx_rcu_rst(IFX_RCU_DOMAIN_DSLDFE, IFX_RCU_MODULE_PPA);
	udelay(1000);
    ifx_rcu_rst(IFX_RCU_DOMAIN_DSLDSP, IFX_RCU_MODULE_PPA);
    udelay(1000);
    ifx_rcu_rst(IFX_RCU_DOMAIN_DSLTC, IFX_RCU_MODULE_PPA);
    udelay(1000);
    ifx_rcu_rst(IFX_RCU_DOMAIN_PPE, IFX_RCU_MODULE_PPA);
    udelay(1000);
    *PP32_SRST &= ~0x000303CF;
    udelay(1000);
    *PP32_SRST |= 0x000303CF;
    udelay(1000);

    for ( i = 0; i < dma_device->max_rx_chan_num; i++ )
        dma_device->rx_chan[i]->reset(dma_device->rx_chan[i]);

    dma_device_unregister(dma_device);
    dma_device_release(dma_device);
    dma_device = NULL;
}
#else // VR9 Style

static INLINE void reset_ppe(void) {
	//  reset PPE
	ifx_rcu_rst(IFX_RCU_DOMAIN_DSLDFE, IFX_RCU_MODULE_PPA);
	udelay(1000);
	ifx_rcu_rst(IFX_RCU_DOMAIN_DSLTC, IFX_RCU_MODULE_PPA);
	udelay(1000);
	ifx_rcu_rst(IFX_RCU_DOMAIN_PPE, IFX_RCU_MODULE_PPA);
	udelay(1000);
	*PP32_SRST &= ~0x000303CF;
	udelay(1000);
	*PP32_SRST |= 0x000303CF;
	udelay(1000);
}
#endif

static INLINE void init_pmu(void) {
	//   5 - DMA, should be enabled in DMA driver
	//   9 - DSL DFE/AFE
	//  15 - AHB
	//  19 - SLL
	//  21 - PPE TC
	//  22 - PPE EMA
	//  23 - PPE DPLUS Master
	//  24 - PPE DPLUS Slave
	//  28 - SWITCH
	//  29 - PPE_TOP

	DSL_DFE_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_TPE_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_SLL01_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_TC_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_EMA_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_DPLUS_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_DPLUSS_PMU_SETUP(IFX_PMU_ENABLE);
	SWITCH_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_TOP_PMU_SETUP(IFX_PMU_ENABLE);
	PPE_QSB_PMU_SETUP(IFX_PMU_ENABLE);
}

static INLINE uint32_t get_qsb_clk(void)
{
    uint32_t fpi_dvsn;

    fpi_dvsn = ((*AMAZON_S_CGU_SYS)>>28) & 0x03;
    return (ifx_get_fpi_hz() >> fpi_dvsn);
}


static INLINE void start_cpu_port(void) {
	int i;

	*DS_RXCFG |= 0x00008000; //  enable DSRX
	*DM_RXCFG |= 0x80000000; //  enable DMRX
	for (i = 0; i < 7; i++) {
		IFX_REG_W32_MASK(0, 1, FDMA_PCTRL_REG(i));
		//  start port 0 - 6
		IFX_REG_W32_MASK(0, 1, SDMA_PCTRL_REG(i));
		//  start port 0 - 6
	}
	*PCE_PCTRL_REG(6, 0) |= 0x07; //  clear_dplus

	disable_cpuport_flow_control(); // has been in enable_fdma_sdma before
}

#if 0
static INLINE void stop_datapath(void)
{
    unsigned long org_jiffies;
    int i;

    g_stop_datapath = 1;    //  going to be unloaded

    //  disable switch external port
    for ( i = 0; i < 6; i++ ) {
        IFX_REG_W32_MASK(1, 0, SDMA_PCTRL_REG(i));  //  stop port 0 - 5
        IFX_REG_W32_MASK(0, 1, FDMA_PCTRL_REG(i));
    }

    //  turn on DMA RX channels
    turn_on_dma_rx(31);

    //  clean ingress datapath
    org_jiffies = jiffies;
    while ( ((IFX_REG_R32(DM_RXPGCNT) & 0x00FF) != 0 || (IFX_REG_R32(DM_RXPKTCNT) & 0x00FF) != 0 || (IFX_REG_R32(DS_RXPGCNT) & 0x00FF) != 0) && jiffies - org_jiffies < HZ / 10 + 1 )
        schedule();
    if ( jiffies - org_jiffies > HZ / 10 )
        err("Can not clear DM_RXPGCNT/DM_RXPKTCNT/DS_RXPGCNT");

    //  turn off DMA RX channels (including loopback RX channels)
    while ( (i = clz(g_f_dma_opened)) >= 0 )
        turn_off_dma_rx(i);

    //  turn off DMA Loopback TX channels
    g_dma_device_ppe->tx_chan[0]->close(g_dma_device_ppe->tx_chan[0]);
    g_dma_device_ppe->tx_chan[1]->close(g_dma_device_ppe->tx_chan[1]);

    *PMAC_HD_CTL_REG = 0x0000074C;  //  recover to default value
    *FDMA_PCTRL_REG(6) &= ~0x02;
}
#endif


static INLINE void init_internal_tantos_qos_setting(void) {
}

static INLINE void init_internal_tantos(void) {
	int i;

	*GLOB_CTRL_REG = 1 << 15; //  enable Switch

	for (i = 0; i < 7; i++)
		sw_clr_rmon_counter(i);

	//*FDMA_PCTRL_REG(6) &= ~0x01;//  disable port (FDMA)
	*FDMA_PCTRL_REG(6) |= 0x02; //  insert special tag

	*PMAC_HD_CTL_REG = 0x01DC; //  PMAC Head

	for (i = 6; i < 12; i++)
		*PCE_PCTRL_REG(i, 0) |= 1 << 11; //  ingress special tag

	*MAC_FLEN_REG = 1518 + 8 + 4 * 2; //  MAC frame + 8-byte special tag + 4-byte VLAN tag * 2
	*MAC_CTRL_REG(6, 2) |= 1 << 3; //  enable jumbo frame
    for(i = 0; i < 7; i ++){
        *MAC_CTRL_REG(i,0) |= 1 << 7;   //set  FCS generation Enable for each port
    }

//    for ( i = 0; i < 7; i++ )
//        *PCE_PCTRL_REG(i, 2) |= 3;  //  highest class, mapped to RX channel 0

#if 0
	for ( i = 0; i < 12; i++ )
	*PCE_PCTRL_REG(i, 0) |= 1 << 5; //  VLAN transparent mode

	//  VLAN active table
	*PCE_TBL_KEY(0) = 0x0055;
	*PCE_TBL_VAL(0) = 0x0000;
	*PCE_TBL_ADDR = 0x0000;
	*PCE_TBL_CTRL = 0x9021;

	//  VLAN mapping table
	*PCE_TBL_VAL(0) = 0x0055;
	*PCE_TBL_VAL(1) = 0x0FFF;
	*PCE_TBL_VAL(2) = 0x0000;
	*PCE_TBL_ADDR = 0x0000;
	*PCE_TBL_CTRL = 0x9022;
#endif
#if 0
	*PCE_GCTRL_REG(0) |= 1 << 14; //  VLAN-aware Switching

	for ( i = 0; i < 12; i++ )
	*PCE_PCTRL_REG(i, 5) |= 1 << 0;//  VLAN unmatch rule    TO BE TESTED
#endif

#if 0
	//  xuliang: MFE test
	for ( i = 0; i < 6; i++ )
	*PCE_PCTRL_REG(i, 3) |= 1 << 11;//  enable egress re-direct to monitoring port for physical port 0-5
	*PCE_PMAP_REG(1) = 0x40;//  monitoring
#else
	*PCE_PMAP_REG(1) = 0x7F; //  monitoring
#endif
	//*PCE_PMAP_REG(2) = 0x7F & ~g_pmac_ewan; //  broadcast and unknown multicast
	*PCE_PMAP_REG(2) = 0x7F; //  broadcast and unknown multicast
	//*PCE_PMAP_REG(3) = 0x7F & ~g_pmac_ewan; //  unknown uni-cast
	*PCE_PMAP_REG(3) = 0x7F; //  unknown uni-cast

	*PMAC_EWAN_REG = g_pmac_ewan;

	*PMAC_RX_IPG_REG = 0x8F;

	//IFX_AR10_Switch_PCE_Micro_Code_Int();

	init_internal_tantos_qos_setting();
}

static INLINE void init_dplus(void)
{
    *DM_RXCFG &= ~0x80000000;   //  disable DMRX
    *DS_RXCFG &= ~0x00008000;   //  disable DSRX
    *DM_RXPGCNT = 0x00020000;   //  clear page pointer & counter
    *DS_RXPGCNT = 0x00040000;   //  clear page pointer & counter

    *DM_RXDB    = 0x1450;       //  Set Dplus Master Data base to 0x1450 which would be mapped to 0x3450
    *DM_RXCB    = 0x143C;       //  Set Dplus Master Control Base address to 0x143C
    *DM_RXPKTCNT= 0x00000200;   //  Set Dplus Master Packet counter in FW mode
    *DM_RXCFG   = 0x00F87014;
    *DS_RXCFG   = 0x7014;
}

static INLINE void init_ema(void)
{
    *EMA_CMDCFG  = (EMA_CMD_BUF_LEN << 16) | (EMA_CMD_BASE_ADDR >> 2);
    *EMA_DATACFG = (EMA_DATA_BUF_LEN << 16) | (EMA_DATA_BASE_ADDR >> 2);
    *EMA_IER     = 0x000000FF;
    *EMA_CFG     = EMA_READ_BURST | (EMA_WRITE_BURST << 2);
}

static INLINE void init_mailbox(void) {
	*MBOX_IGU1_ISRC = 0xFFFFFFFF;
	*MBOX_IGU1_IER = 0x00000000; //  Don't need to enable RX interrupt, EMA driver handle RX path.
	dbg("MBOX_IGU1_IER = 0x%08X", *MBOX_IGU1_IER);
}

static INLINE void clear_share_buffer(void)
{
    volatile u32 *p = SB_RAM0_ADDR(0);
    unsigned int i;

    //  write all zeros only
    for ( i = 0; i < SB_RAM0_DWLEN + SB_RAM1_DWLEN + SB_RAM2_DWLEN + SB_RAM3_DWLEN + SB_RAM4_DWLEN; i++ )
        *p++ = 0;

    //  Configure share buffer master selection
    *SB_MST_PRI0 = 1;
    *SB_MST_PRI1 = 1;
}

static INLINE void clear_cdm(void)
{
    volatile u32 *dest;
    int i;

    //  Both blocks are set to code memory
    *CDM_CFG = CDM_CFG_RAM1_SET(0x01) | CDM_CFG_RAM0_SET(0x00);
    mdelay(1);

    dest = CDM_CODE_MEMORY_RAM0_ADDR(0);
    for ( i = 0; i < CDM_CODE_MEMORY_RAM0_DWLEN + CDM_CODE_MEMORY_RAM1_DWLEN; i++ )
        *dest++ = 0;
}


static INLINE void board_init(void) {
}

static INLINE void init_tc_hw(void)
{
     int i;

    *SFSM_CFG0          = 0x00010014;
    for (i=0;i<10;i++);

    *SFSM_DBA0          = 0x1800;
    *SFSM_CBA0          = 0x1918;
    for (i=0;i<10;i++);

    *SFSM_CFG0          = 0x00014014;
    *FFSM_DBA0          = 0x17AC;
    *FFSM_CFG0          = 0x00010006;

    *SFSM_CFG1          = 0x00010014;
    for (i=0;i<10;i++);

    *SFSM_DBA1          = 0x192C;
    *SFSM_CBA1          = 0x1A44;
    *FFSM_DBA1          = 0x1758;
    *FFSM_CFG1          = 0x00010006;

    //*FFSM_IDLE_HEAD_BC0 = 0x00000001;
}


static INLINE void setup_ppe_conf(void)
{
    // Disable PPE and MIPS access to DFE memory
    *AMAZON_S_RCU_PPE_CONF = *AMAZON_S_RCU_PPE_CONF & 0x7FFFFFFF;
}

static inline void setup_ppe_250(void)
{
    *AMAZON_S_CGU_SYS = ((*AMAZON_S_CGU_SYS & ~(0x70000)) | 0x10000);
}


static INLINE void init_hw(void) {
	setup_ppe_250();

	init_pmu();

	reset_ppe();

	setup_ppe_conf();

	init_internal_tantos();

	init_dplus();

	init_tc_hw();

	init_ema();

	init_mailbox();

	clear_share_buffer();

	clear_cdm();

	board_init();
}

/*
 *  Description:
 *    Download PPE firmware binary code.
 *  Input:
 *    src       --- u32 *, binary code buffer
 *    dword_len --- unsigned int, binary code length in DWORD (32-bit)
 *  Output:
 *    int       --- 0:    Success
 *                  else: Error Code
 */
static INLINE int pp32_download_code(const u32 *code_src, unsigned int code_dword_len, const u32 *data_src, unsigned int data_dword_len)
{
    volatile u32 *dest;
    int i;

    if ( code_src == 0 || ((unsigned long)code_src & 0x03) != 0
        || data_src == 0 || ((unsigned long)data_src & 0x03) != 0
        || (code_dword_len > 0x1000 && data_dword_len > 0) )
        return -EINVAL;

    /*  set PPE code memory to FPI bus access mode  */
    if ( code_dword_len <= 0x1000 )
        *CDM_CFG = CDM_CFG_RAM1_SET(0x00) | CDM_CFG_RAM0_SET(0x00);
    else
        *CDM_CFG = CDM_CFG_RAM1_SET(0x01) | CDM_CFG_RAM0_SET(0x00);
    for ( i = 0; i < 1000; i++ );

    dbg("code_dword_len = 0x%X, data_dword_len = 0x%X", code_dword_len, data_dword_len);

    /*  copy code   */
    dest = CDM_CODE_MEMORY_RAM0_ADDR(0);
    while ( code_dword_len-- > 0 )
        *dest++ = *code_src++;

    /*  copy data   */
    dest = PP32_DATA_MEMORY_RAM1_ADDR(0);
    while ( data_dword_len-- > 0 )
        *dest++ = *data_src++;

    return 0;
}

/*
 *  Description:
 *    Do PP32 specific initialization.
 *  Input:
 *    data --- void *, specific parameter passed in.
 *  Output:
 *    int  --- 0:    Success
 *             else: Error Code
 */
static INLINE int pp32_specific_init(int fwcode __attribute__((unused)), void *data __attribute__((unused)))
{
    return 0;
}

/*
 *  Description:
 *    Initialize and start up PP32.
 *  Input:
 *    none
 *  Output:
 *    int  --- 0:    Success
 *             else: Error Code
 */
static INLINE int pp32_start(int fwcode)
{
    int ret;

    /*  download firmware   */
    ret = pp32_download_code(ar10_a5_binary_code, sizeof(ar10_a5_binary_code) / sizeof(*ar10_a5_binary_code), ar10_a5_binary_data, sizeof(ar10_a5_binary_data) / sizeof(*ar10_a5_binary_data));
    if ( ret )
        return ret;

    /*  firmware specific initialization    */
    ret = pp32_specific_init(fwcode, NULL);
    if ( ret )
        return ret;

    /*  run PP32    */
    *PP32_DBG_CTRL = DBG_CTRL_RESTART;
    /*  idle for a while to let PP32 init itself    */
    udelay(200);

    return 0;
}

/*
 *  Description:
 *    Halt PP32.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static INLINE void pp32_stop(void)
{
    /*  halt PP32   */
    *PP32_DBG_CTRL = DBG_CTRL_STOP;
}

static INLINE int init_local_variables(void) {
	unsigned int i, j;

	/*  initialize semaphore used by open and close */
	sema_init(&g_sem, 1);

	if (ethwan == 1)
		g_eth_wan_mode = 2;
	else if (ethwan == 2)
		g_eth_wan_mode = 3;

	if (wanitf == ~0 || wanitf == 0) {
		switch (g_eth_wan_mode) {
		case 0: /*  DSL WAN     */
			g_wan_itf = 3 << 6;
			break;
		case 2: /*  Mixed Mode  */
			g_wan_itf = 1 << 0;
			break;
		case 3: /*  ETH WAN     */
			g_wan_itf = 1 << 1;
			break;
		}
	} else {
		g_wan_itf = wanitf;
		switch (g_eth_wan_mode) {
		case 0: /*  DSL WAN     */
			g_wan_itf &= 0xF8;
			g_wan_itf |= 3 << 6;
			break; //  ETH0/1 can not be WAN
		case 2: /*  Mixed Mode  */
			g_wan_itf &= 0x03; //  only ETH0/1 support mixed mode
			if (g_wan_itf == 0x03 || g_wan_itf == 0) { //both ETH0/1 in mixed mode, reset to eth0 mixed mode
				g_wan_itf = 1;
			}
			break;
		case 3: /*  ETH WAN     */
			g_wan_itf &= 0x3B; //  DSL disabled in ETH WAN mode
			if ((g_wan_itf & 0x03) == 0x03) { //both ETH0/1 in WAN mode, remove eth0 wan setting
				g_wan_itf &= ~0x01;
			}
			if ((g_wan_itf & 0x03) == 0) {
				g_wan_itf |= 1 << 1;
			}
			break;
		}

	}
	g_ipv6_acc_en = ipv6_acc_en ? 1 : 0;
    g_ipv4_lan_collision_rout_fwdc_table_base = __IPV4_LAN_COLLISION_ROUT_FWDC_TABLE_BASE;
    g_ipv4_wan_collision_rout_fwda_table_base = __IPV4_WAN_COLLISION_ROUT_FWDA_TABLE_BASE;


	g_wanqos_en =
			g_eth_wan_mode == 3 && (g_wan_itf & 0x03) != 0x03 && wanqos_en ?
					wanqos_en : 0;

    g_mailbox_signal_mask = g_wanqos_en ? CPU_TO_WAN_SWAP_SIG : 0;
    if ( g_eth_wan_mode == 0 )
        g_mailbox_signal_mask |= WAN_RX_SIG(0) | WAN_RX_SIG(1);

	memset(&g_atm_priv_data, 0, sizeof(g_atm_priv_data));
	for (i = 0; i < ATM_PORT_NUMBER; i++) {
		g_atm_priv_data.port[i].tx_max_cell_rate = (u32) port_cell_rate_up[i];
	}
	g_atm_priv_data.max_connections = ATM_QUEUE_NUMBER;
	g_atm_priv_data.max_sw_tx_queues = ATM_SW_TX_QUEUE_NUMBER;


	for (i = 0; i < 2; i++) {
		unsigned int max_packet_priority = NUM_ENTITY(g_eth_prio_queue_map[i]); //  assume only 8 levels are used in Linux
		unsigned int tx_num_q;
		unsigned int q_step, q_accum, p_step;

		tx_num_q =
				(g_wan_itf & (1 << i)) && g_wanqos_en ?
						__ETH_WAN_TX_QUEUE_NUM : SWITCH_CLASS_VALUE;
		q_step = tx_num_q - 1;
		p_step = max_packet_priority - 1;
		for (j = 0, q_accum = 0; j < max_packet_priority;
				j++, q_accum += q_step)
			g_eth_prio_queue_map[i][j] = (q_accum + (p_step >> 1)) / p_step;

		if ((g_wan_itf & (1 << i)) && g_wanqos_en)
			for (j = 0; j < max_packet_priority; j++)
				g_eth_prio_queue_map[i][j] = q_step
						- g_eth_prio_queue_map[i][j];
	}
    for ( i = 0; i < NUM_ENTITY(g_wan_queue_class_map); i++ )
        g_wan_queue_class_map[i] = i;
    if ( (g_wan_itf & 0x03) && g_wanqos_en )
    {
        for ( i = 0; i < NUM_ENTITY(g_wan_queue_class_map); i++ )
            g_wan_queue_class_map[i] = NUM_ENTITY(g_wan_queue_class_map) - 1 - i;
    }

	return 0;
}

static INLINE void clear_local_variables(void) {
}

#if defined(ENABLE_FW_MODULE_TEST) && ENABLE_FW_MODULE_TEST

#include "cfg_arrays_a5.h"

unsigned int ppe_fw_mode = 0; // 0: normal mode, 1: mix mode
unsigned int ppe_wan_hash_algo = 0;// 0: using algo 0, 1: using algo 1
unsigned int acc_proto = 0;// 0: UDP, 1:TCP

void setup_acc_action_tbl(void)
{
	unsigned int idx;

	if (acc_proto == 0) {
		// clear bit16 of dw3 of each action entry
		unsigned long udp_mask = ~ (1 << 16);

		// setup for Hash entry
		idx = 3;
		while (idx < sizeof(lan_uc_act_tbl_normal_mode_cfg)/sizeof(unsigned long)) {
			lan_uc_act_tbl_normal_mode_cfg[idx] &= udp_mask;
			lan_uc_act_tbl_mixed_mode_cfg[idx] &= udp_mask;

			wan_uc_act_tbl_alo_0_cfg[idx] &= udp_mask;
			wan_uc_act_tbl_alo_1_cfg[idx] &= udp_mask;
			idx += 6;
		}

		// setup for Collsion entry
		idx = 3;
		while (idx < sizeof(lan_uc_col_act_tbl_normal_mode_cfg)/sizeof(unsigned long)) {
			lan_uc_col_act_tbl_normal_mode_cfg[idx] &= udp_mask;
			lan_uc_col_act_tbl_mixed_mode_cfg[idx] &= udp_mask;
			wan_uc_col_act_tbl_cfg[idx] &= udp_mask;
			idx += 6;
		}
	}
	else {
		// set bit16 of dw3 of each action entry
		unsigned long tcp_mask = (1 << 16);

		// setup for Hash entry
		idx = 3;
		while (idx < sizeof(lan_uc_act_tbl_normal_mode_cfg)/sizeof(unsigned long)) {
			lan_uc_act_tbl_normal_mode_cfg[idx] |= tcp_mask;
			lan_uc_act_tbl_mixed_mode_cfg[idx] |= tcp_mask;

			wan_uc_act_tbl_alo_0_cfg[idx] |= tcp_mask;
			wan_uc_act_tbl_alo_1_cfg[idx] |= tcp_mask;
			idx += 6;
		}

		// setup for Collsion entry
		idx = 3;
		while (idx < sizeof(lan_uc_col_act_tbl_normal_mode_cfg)/sizeof(unsigned long)) {
			lan_uc_col_act_tbl_normal_mode_cfg[idx] |= tcp_mask;
			lan_uc_col_act_tbl_mixed_mode_cfg[idx] |= tcp_mask;
			wan_uc_col_act_tbl_cfg[idx] |= tcp_mask;
			idx += 6;
		}
	}
}

void init_acc_tables(void)
{
	setup_acc_action_tbl();

	// init MAC table
	memcpy((void *)ROUT_MAC_CFG_TBL(0), rout_mac_cfg, sizeof(rout_mac_cfg));

	// PPPoE table
	memcpy((void *)PPPOE_CFG_TBL(0), pppoe_cfg, sizeof(pppoe_cfg));

	// Outer VLAN Config
	memcpy((void *)OUTER_VLAN_TBL(0), outer_vlan_cfg, sizeof(outer_vlan_cfg));

	// lan uc_cmp_tbl_cfg (Hash) and collision
	memcpy((void *)ROUT_LAN_HASH_CMP_TBL(0), lan_uc_cmp_tbl_cfg, sizeof(lan_uc_cmp_tbl_cfg));
	memcpy((void *)ROUT_LAN_COLL_CMP_TBL(0), lan_uc_col_cmp_tbl_cfg, sizeof(lan_uc_col_cmp_tbl_cfg));

	// lan action
	if(ppe_fw_mode == 0) {
		// normal mode
		memcpy((void *)ROUT_LAN_HASH_ACT_TBL(0), lan_uc_act_tbl_normal_mode_cfg, sizeof(lan_uc_act_tbl_normal_mode_cfg));
		memcpy((void *)ROUT_LAN_COLL_ACT_TBL(0), lan_uc_col_act_tbl_normal_mode_cfg, sizeof(lan_uc_col_act_tbl_normal_mode_cfg));
	}
	else {
		// mix mode
		memcpy((void *)ROUT_LAN_HASH_ACT_TBL(0), lan_uc_act_tbl_mixed_mode_cfg, sizeof(lan_uc_act_tbl_mixed_mode_cfg));
		memcpy((void *)ROUT_LAN_COLL_ACT_TBL(0), lan_uc_col_act_tbl_mixed_mode_cfg, sizeof(lan_uc_col_act_tbl_mixed_mode_cfg));
	}

	// wan hash cmp anc act table
	if ( ppe_wan_hash_algo == 0) {
		// WAN algo 0
		memcpy((void *)ROUT_WAN_HASH_CMP_TBL(0), wan_uc_cmp_tbl_alo_0_cfg, sizeof(wan_uc_cmp_tbl_alo_0_cfg));
		memcpy((void *)ROUT_WAN_HASH_ACT_TBL(0), wan_uc_act_tbl_alo_0_cfg, sizeof(wan_uc_act_tbl_alo_0_cfg));
	}
	else {
		// WAN algo 1
		memcpy((void *)ROUT_WAN_HASH_CMP_TBL(0), wan_uc_cmp_tbl_alo_1_cfg, sizeof(wan_uc_cmp_tbl_alo_1_cfg));
		memcpy((void *)ROUT_WAN_HASH_ACT_TBL(0), wan_uc_act_tbl_alo_1_cfg, sizeof(wan_uc_act_tbl_alo_1_cfg));
	}

	// wan collision cmp and act table
	memcpy((void *)ROUT_WAN_COLL_CMP_TBL(0), wan_uc_col_cmp_tbl_cfg, sizeof(wan_uc_col_cmp_tbl_cfg));
	memcpy((void *)ROUT_WAN_COLL_ACT_TBL(0), wan_uc_col_act_tbl_cfg, sizeof(wan_uc_col_act_tbl_cfg));

	// wan multicast cmp and act table
	memcpy((void *)ROUT_WAN_MC_CMP_TBL(0), wan_mc_cmp_tbl_cfg, sizeof(wan_mc_cmp_tbl_cfg));
	memcpy((void *)ROUT_WAN_MC_ACT_TBL(0), wan_mc_act_tbl_cfg, sizeof(wan_mc_act_tbl_cfg));
}

#endif  //  #if defined(ENABLE_FW_MODULE_TEST) && ENABLE_FW_MODULE_TEST
static INLINE void init_communication_data_structures(int fwcode  __attribute__((unused))) {
	struct eth_ports_cfg eth_ports_cfg = { 0 };
	struct rout_tbl_cfg lan_rout_tbl_cfg = { 0 };
	struct rout_tbl_cfg wan_rout_tbl_cfg = { 0 };
	struct gen_mode_cfg1 gen_mode_cfg1 = { 0 };
	struct wrx_queue_config wrx_queue_config = { 0 };
	struct wtx_port_config wtx_port_config = { 0 };
	struct wtx_queue_config wtx_queue_config = { 0 };
	unsigned int i,j;

	*CDM_CFG = CDM_CFG_RAM1_SET(0x00) | CDM_CFG_RAM0_SET(0x00);

	for (i = 0; i < 1000; i++)
		;

	*PSEUDO_IPv4_BASE_ADDR = 0xFFFFFF00;

	*CFG_WRX_HTUTS = g_atm_priv_data.max_connections + OAM_HTU_ENTRY_NUMBER;
	*CFG_WRDES_DELAY = write_descriptor_delay;
	*WRX_EMACH_ON = 0x03;
	*WTX_EMACH_ON = 0x7FFF;
	*WRX_HUNT_BITTH = DEFAULT_RX_HUNT_BITTH;

	eth_ports_cfg.wan_vlanid_hi = 0;
	eth_ports_cfg.wan_vlanid_lo = 0;
	eth_ports_cfg.eth0_type = 0; //  lan
	*ETH_PORTS_CFG = eth_ports_cfg;

	lan_rout_tbl_cfg.rout_num = MAX_COLLISION_ROUTING_ENTRIES;
	lan_rout_tbl_cfg.ttl_en = 1;
	lan_rout_tbl_cfg.rout_drop = 0;
	*LAN_ROUT_TBL_CFG = lan_rout_tbl_cfg;

	wan_rout_tbl_cfg.rout_num = MAX_COLLISION_ROUTING_ENTRIES;
	wan_rout_tbl_cfg.wan_rout_mc_num = MAX_WAN_MC_ENTRIES;
	wan_rout_tbl_cfg.ttl_en = 1;
	wan_rout_tbl_cfg.wan_rout_mc_drop = 0;
	wan_rout_tbl_cfg.rout_drop = 0;
	*WAN_ROUT_TBL_CFG = wan_rout_tbl_cfg;

	gen_mode_cfg1.app2_indirect = 1; //  0: means DMA RX ch 3 is dedicated for CPU1 process
	gen_mode_cfg1.us_indirect = 0;
	gen_mode_cfg1.us_early_discard_en = 1;
#if defined(CONFIG_IFX_PPA_MFE) && CONFIG_IFX_PPA_MFE
	gen_mode_cfg1.classification_num = MAX_CLASSIFICATION_ENTRIES;
#else
	gen_mode_cfg1.classification_num = 0;
#endif
	gen_mode_cfg1.ipv6_rout_num = g_ipv6_acc_en ? 2 : 2;
//	gen_mode_cfg1.ipv6_rout_num = g_ipv6_acc_en ? 2 : 0;
	gen_mode_cfg1.session_ctl_en = 0;
	gen_mode_cfg1.wan_hash_alg = 0;
#if defined(CONFIG_IFX_PPA_MFE) && CONFIG_IFX_PPA_MFE
	gen_mode_cfg1.brg_class_en = 1;
#else
	gen_mode_cfg1.brg_class_en = 0;
#endif
	gen_mode_cfg1.ipv4_mc_acc_mode = 0;
	gen_mode_cfg1.ipv6_acc_en = g_ipv6_acc_en;
	gen_mode_cfg1.wan_acc_en = 1;
	gen_mode_cfg1.lan_acc_en = 1;
	gen_mode_cfg1.sw_iso_mode = 1;
	gen_mode_cfg1.sys_cfg = g_eth_wan_mode; //  3 - ETH1 WAN, 2 - ETH0 mixed, 0 - DSL WAN
	*GEN_MODE_CFG1 = gen_mode_cfg1;
	//GEN_MODE_CFG2->itf_outer_vlan_vld   = 1 << 7;   //  DSL EoA port use outer VLAN for PVCs differentiation
	GEN_MODE_CFG2->itf_outer_vlan_vld |= 0xFF; //  Turn on outer VLAN for all interfaces;

    *CFG_FEATURES = 0;  //  clock gating is disabled

    *CFG_CLASS2QID_MAP(0) = 0;
    *CFG_CLASS2QID_MAP(1) = 0;
    *CFG_QID2CLASS_MAP(0) = 0;
    *CFG_QID2CLASS_MAP(0) = 0;
    i = (g_wan_itf & 0x01) ? 0 : 1;
    for ( j = 0; j < NUM_ENTITY(g_eth_prio_queue_map[i]); j++ )
        *CFG_CLASS2QID_MAP(j >> 3) |= g_eth_prio_queue_map[i][j] << ((j & 0x07) * 4);
    for ( j = 0; j < NUM_ENTITY(g_wan_queue_class_map); j++ )
        *CFG_QID2CLASS_MAP(j >> 3) |= g_wan_queue_class_map[j] << ((j & 0x07) * 4);

	//  fake setting, when open ATM VC, it will be set with the real value
	wrx_queue_config.new_vlan = 0;
	wrx_queue_config.vlan_ins = 0;
	wrx_queue_config.mpoa_type = 3; //  IPoA
	wrx_queue_config.ip_ver = 0;
	wrx_queue_config.mpoa_mode = 0; //TODO 0: VC mux war das orginal
	wrx_queue_config.oversize = DMA_PACKET_SIZE;
	wrx_queue_config.undersize = 0;
	wrx_queue_config.mfs = DMA_PACKET_SIZE;
	wrx_queue_config.uumask = 0xFF;
	wrx_queue_config.cpimask = 0xFF;
	wrx_queue_config.uuexp = 0;
	wrx_queue_config.cpiexp = 0;
	for (i = 0; i < 16; i++)
		*WRX_QUEUE_CONFIG(i) = wrx_queue_config;

	wtx_port_config.qid = 0;
	wtx_port_config.qsben = 1;
	for (i = 0; i < 2; i++)
		*WTX_PORT_CONFIG(i) = wtx_port_config;

	//  fake setting, when open ATM VC, it will be set with the real value
	wtx_queue_config.uu = 0;
	wtx_queue_config.cpi = 0;
	wtx_queue_config.same_vc_qmap = 0;
	wtx_queue_config.sbid = 0;
	wtx_queue_config.mpoa_mode = 0; //TODO 0: VC mux war das orginal
	wtx_queue_config.qsben = 1;
	wtx_queue_config.atm_header = 0;
	for (i = 0; i < 15; i++) {
		wtx_queue_config.qsb_vcid = i + QSB_QUEUE_NUMBER_BASE;
		*WTX_QUEUE_CONFIG(i) = wtx_queue_config;
	}

	//*MTU_CFG_TBL(0) = DMA_PACKET_SIZE;          //  for ATM
	*MTU_CFG_TBL(0) = ETH_MAX_DATA_LENGTH;
	for (i = 1; i < 8; i++)
		*MTU_CFG_TBL(i) = ETH_MAX_DATA_LENGTH; //  for ETH

	if (g_eth_wan_mode == 0)
		setup_oam_htu_entry();

	if (g_wanqos_en) {
		struct wtx_qos_q_desc_cfg wtx_qos_q_desc_cfg = { 0 };
		struct tx_qos_cfg tx_qos_cfg = { 0 };
		struct cfg_std_data_len cfg_std_data_len = { 0 };
        int ii;

		for (ii = 0; ii < __ETH_WAN_TX_QUEUE_NUM; ii++) {
			wtx_qos_q_desc_cfg.length = ETH_WAN_TX_DESC_NUM(ii);
			wtx_qos_q_desc_cfg.addr = __ETH_WAN_TX_DESC_BASE(ii);
			*WTX_QOS_Q_DESC_CFG(ii) = wtx_qos_q_desc_cfg;
		}

		tx_qos_cfg.time_tick = cgu_get_pp32_clock() / 62500; //  16 * (cgu_get_pp32_clock() / 1000000)
		tx_qos_cfg.eth1_eg_qnum = __ETH_WAN_TX_QUEUE_NUM;
		tx_qos_cfg.eth1_qss = 1;
		tx_qos_cfg.eth1_burst_chk = 1;
		*TX_QOS_CFG = tx_qos_cfg;

		cfg_std_data_len.byte_off = 0;
		cfg_std_data_len.data_len = DMA_PACKET_SIZE;

		*CFG_STD_DATA_LEN = cfg_std_data_len;
	}

	*CFG_WAN_PORTMAP = g_wan_itf;
	*CFG_MIXED_PORTMAP = g_eth_wan_mode == 2 /* Mixed Mode */? g_wan_itf : 0;

    *TUNNEL_MAX_ID = -1;
#if defined(ENABLE_FW_MODULE_TEST) && ENABLE_FW_MODULE_TEST
	init_acc_tables();
#endif
}

static INLINE int alloc_dma(void) {
	int ret;
	unsigned int pre_alloc_desc_total_num;
	struct sk_buff **skb_pool;
	struct sk_buff **pskb;
	volatile u32 *p;
	unsigned int i;

    printk(KERN_ERR "[%s] g_eth_wan_mode = %d, g_wanqos_en=%d\n", __func__, g_eth_wan_mode, g_wanqos_en );

#if defined(ENABLE_P2P_COPY_WITHOUT_DESC) && ENABLE_P2P_COPY_WITHOUT_DESC
	pre_alloc_desc_total_num = 0;
#else
	pre_alloc_desc_total_num = DMA_RX_CH2_DESC_NUM;
	if ( g_eth_wan_mode != 0 && g_wanqos_en == 0 ) //  Mix mode or ETH1 WAN without QoS
	pre_alloc_desc_total_num += DMA_RX_CH1_DESC_NUM;
#endif
	if (g_eth_wan_mode == 0) //  DSL WAN
		pre_alloc_desc_total_num += DMA_RX_CH1_DESC_NUM + CPU_TO_WAN_TX_DESC_NUM
				+ WAN_TX_DESC_NUM_TOTAL + WAN_RX_DESC_NUM(0) * 2;
	else if (g_wanqos_en)
		pre_alloc_desc_total_num += DMA_RX_CH1_DESC_NUM + CPU_TO_WAN_TX_DESC_NUM
				+ WAN_TX_DESC_NUM_TOTAL
				/* + CPU_TO_WAN_SWAP_DESC_NUM , das war falsch, denn es ist der gleiche Speicherbereich, wie CPU_TO_WAN_TX_DESC */
				+ DMA_TX_CH1_DESC_NUM;

	skb_pool = (struct sk_buff **) kmalloc(
			pre_alloc_desc_total_num * sizeof(*skb_pool), GFP_KERNEL);
	if (skb_pool == NULL) {
		ret = -ENOMEM;
		goto ALLOC_SKB_POOL_FAIL;
	}

	for (i = 0; i < pre_alloc_desc_total_num; i++) {
		skb_pool[i] = alloc_skb_rx();
		if (!skb_pool[i]) {
			ret = -ENOMEM;
			goto ALLOC_SKB_FAIL;
		}
#ifndef CONFIG_MIPS_UNCACHED
		/*  invalidate cache    */
		dma_cache_inv((unsigned long)skb_pool[i]->data, DMA_PACKET_SIZE);
#endif
#if defined(DEBUG_SKB_SWAP) && DEBUG_SKB_SWAP
		g_dbg_skb_swap_pool[i] = skb_pool[i];
#endif
	}

	g_dma_device_ppe = dma_device_reserve("PPE");
	if (!g_dma_device_ppe) {
		ret = -EIO;
		goto RESERVE_DMA_FAIL;
	}

#if defined(ENABLE_NAPI) && ENABLE_NAPI
    g_dma_device_ppe->activate_poll = dma_activate_poll;
    g_dma_device_ppe->inactivate_poll = dma_inactivate_poll;
#endif

	g_dma_device_ppe->buffer_alloc = dma_buffer_alloc;
	g_dma_device_ppe->buffer_free = dma_buffer_free;
	g_dma_device_ppe->intr_handler = dma_int_handler;
	//g_dma_device_ppe->max_rx_chan_num = 6;
	//g_dma_device_ppe->max_tx_chan_num = 4;
	g_dma_device_ppe->tx_burst_len = 8;
	g_dma_device_ppe->rx_burst_len = 8;

	for (i = 0; i < g_dma_device_ppe->max_rx_chan_num; i++) {
		g_dma_device_ppe->rx_chan[i]->packet_size = DMA_PACKET_SIZE;
		g_dma_device_ppe->rx_chan[i]->control = IFX_DMA_CH_ON;
		//g_dma_device_ppe->rx_chan[i]->channel_packet_drop_enable    = 1;
	}
	// PPA -> DSL
	g_dma_device_ppe->rx_chan[1]->desc_base = (int) DMA_RX_CH1_DESC_BASE;
	g_dma_device_ppe->rx_chan[1]->desc_len = DMA_RX_CH1_DESC_NUM;
	g_dma_device_ppe->rx_chan[1]->req_irq_to_free = g_dma_device_ppe->rx_chan[1]->irq;
	g_dma_device_ppe->rx_chan[1]->no_cpu_data = 1;

	/* RX2 -> TX0: (PPA -> Switch(port6))*/
	g_dma_device_ppe->rx_chan[2]->desc_base = (int) DMA_RX_CH2_DESC_BASE;
	g_dma_device_ppe->rx_chan[2]->desc_len = DMA_RX_CH2_DESC_NUM;
	g_dma_device_ppe->rx_chan[2]->req_irq_to_free = g_dma_device_ppe->rx_chan[2]->irq;
	g_dma_device_ppe->rx_chan[2]->no_cpu_data = 1;

#if defined(ENABLE_P2P_COPY_WITHOUT_DESC) && ENABLE_P2P_COPY_WITHOUT_DESC
	if (g_eth_wan_mode != 0 && g_wanqos_en == 0) {
		g_dma_device_ppe->rx_chan[1]->loopback_enable = 1;
		g_dma_device_ppe->rx_chan[1]->loopback_channel_number = 1 * 2 + 1;
	}
	/* RX2 -> TX0: (PPA -> Switch(port6))*/
	g_dma_device_ppe->rx_chan[2]->loopback_enable = 1;
	g_dma_device_ppe->rx_chan[2]->loopback_channel_number = 0 * 2 + 1;
#endif

	for (i = 0; i < g_dma_device_ppe->max_tx_chan_num; i++) {
		g_dma_device_ppe->tx_chan[i]->control = IFX_DMA_CH_ON;
#if defined (CONFIG_AVM_SCATTER_GATHER)
        g_dma_device_ppe->tx_chan[i]->scatter_gather_channel = 1;
#endif
	}
	g_dma_device_ppe->tx_chan[0]->desc_base = (int) DMA_TX_CH0_DESC_BASE;
	g_dma_device_ppe->tx_chan[0]->desc_len = DMA_TX_CH0_DESC_NUM;
	g_dma_device_ppe->tx_chan[1]->desc_base = (int) DMA_TX_CH1_DESC_BASE;
	g_dma_device_ppe->tx_chan[1]->desc_len = DMA_TX_CH1_DESC_NUM;
#if !defined(ENABLE_P2P_COPY_WITHOUT_DESC) || !ENABLE_P2P_COPY_WITHOUT_DESC
	g_dma_device_ppe->tx_chan[0]->global_buffer_len = DMA_PACKET_SIZE;
	g_dma_device_ppe->tx_chan[0]->peri_to_peri = 1;
	g_dma_device_ppe->tx_chan[1]->global_buffer_len = DMA_PACKET_SIZE;
	g_dma_device_ppe->tx_chan[1]->peri_to_peri = 1;
#else
	g_dma_device_ppe->tx_chan[0]->loopback_enable = 1;
	if (g_eth_wan_mode == 0 || g_wanqos_en != 0) {

		// DSL -> Switch(port11)
		g_dma_device_ppe->tx_chan[1]->global_buffer_len = DMA_PACKET_SIZE;
		g_dma_device_ppe->tx_chan[1]->peri_to_peri = 1;
		g_dma_device_ppe->tx_chan[1]->req_irq_to_free = g_dma_device_ppe->tx_chan[1]->irq;
		g_dma_device_ppe->tx_chan[1]->no_cpu_data = 1;

		// PPA -> Switch(port 6)
		g_dma_device_ppe->tx_chan[0]->req_irq_to_free = g_dma_device_ppe->tx_chan[0]->irq;
		g_dma_device_ppe->tx_chan[0]->no_cpu_data = 1;

	} else
		g_dma_device_ppe->tx_chan[1]->loopback_enable = 1;
#endif

	g_dma_device_ppe->port_packet_drop_enable = 1;

	if (dma_device_register(g_dma_device_ppe) != IFX_SUCCESS) {
		err("failed in \"dma_device_register\"");
		ret = -EIO;
		goto DMA_DEVICE_REGISTER_FAIL;
	}

	pskb = skb_pool;

    if ( g_eth_wan_mode == 0 || g_wanqos_en ) {
        p = (volatile u32 *)CPU_TO_WAN_TX_DESC_BASE;
        for ( i = 0; i < CPU_TO_WAN_TX_DESC_NUM; i++ )
        {
            ASSERT(((u32)(*pskb)->data & (DMA_TX_ALIGNMENT - 1)) == 0, "CPU to WAN TX data pointer (0x%#x) must be 8 DWORDS aligned", (u32)(*pskb)->data);
            *p++ = g_wanqos_en ? 0x30000000 : 0xB0000000;
            if ( g_eth_wan_mode == 0 || g_wanqos_en )
                *p++ = (u32)(*pskb++)->data & 0x1FFFFFE0;
            else
                *p++ = 0;
        }

        p = (volatile u32 *)DSL_WAN_TX_DESC_BASE(0);
        for ( i = 0; i < WAN_TX_DESC_NUM_TOTAL; i++ )
        {
            ASSERT(((u32)(*pskb)->data & (DMA_TX_ALIGNMENT - 1)) == 0, "WAN (ATM/ETH) TX data pointer (0x%#x) must be 8 DWORDS aligned", (u32)(*pskb)->data);
            *p++ = 0x30000000;
            if ( g_eth_wan_mode == 0 )
                *p++ = ((u32)(*pskb++)->data & 0x1FFFFFE0) >> 2;
            else if ( g_wanqos_en )
                *p++ = (u32)(*pskb++)->data & 0x1FFFFFE0;
            else
                *p++ = 0;
        }
    }

    // VORSICHT: Der folgende Code ist falsch! - AR9 / AR10 benutzen den Gleichen Speicher fuer
    // CPU_TO_WAN_SWAP_DESC_BASE und DSL_WAN_TX_DESC_BASE(0)
    // Somit darf nur einmal Alloc bzw. Free aufgerufen werden
#if 0
	if (g_wanqos_en) {
		p = (volatile u32 *) CPU_TO_WAN_SWAP_DESC_BASE;
		for (i = 0; i < CPU_TO_WAN_SWAP_DESC_NUM; i++) {
			ASSERT(
					((u32)(*pskb)->data & (DMA_TX_ALIGNMENT - 1)) == 0,
					"WAN (ETH1) Swap data pointer (0x%08x) must be 8 DWORDS aligned",
					(u32)(*pskb)->data);
			*p++ = 0xB0000000;
			*p++ = (u32) (*pskb++)->data & 0x1FFFFFE0;
		}
	}
#endif

	if (g_eth_wan_mode == 0) {
		p = (volatile u32 *) WAN_RX_DESC_BASE(0);
		for (i = 0; i < WAN_RX_DESC_NUM(0); i++) {
			ASSERT(
					((u32)(*pskb)->data & (DMA_TX_ALIGNMENT - 1)) == 0,
					"WAN (ATM) RX channel 0 (AAL) data pointer (0x%#x) must be 8 DWORDS aligned",
					(u32)(*pskb)->data);
			*p++ = 0xB0000000 | DMA_PACKET_SIZE;
			*p++ = ((u32) (*pskb++)->data & 0x1FFFFFE0);
		}

		p = (volatile u32 *) WAN_RX_DESC_BASE(1);
		for (i = 0; i < WAN_RX_DESC_NUM(1); i++) {
			ASSERT(
					((u32)(*pskb)->data & 3) == 0,
					"WAN (ATM) RX channel 1 (OAM) data pointer (0x%#x) must be DWORD aligned",
					(u32)(*pskb)->data);
			g_wan_rx1_desc_skb[i] = *pskb;
			*p++ = 0xB0000000 | DMA_PACKET_SIZE;
			*p++ = ((u32) (*pskb++)->data & 0x1FFFFFFC) >> 2;
		}
	}

	p = (volatile u32 *) DMA_RX_CH1_DESC_BASE;
	for (i = 0; i < DMA_RX_CH1_DESC_NUM; i++) {
#if !defined(ENABLE_P2P_COPY_WITHOUT_DESC) || !ENABLE_P2P_COPY_WITHOUT_DESC
		ASSERT(((u32)(*pskb)->data & (DMA_RX_ALIGNMENT - 1)) == 0, "DMA RX channel 1 data pointer (0x%#x) must be 8 DWORDS aligned", (u32)(*pskb)->data);
#else
		if (g_eth_wan_mode == 0 || g_wanqos_en) {
			ASSERT(
					((u32)(*pskb)->data & (DMA_RX_ALIGNMENT - 1)) == 0,
					"DMA RX channel 1 data pointer (0x%#x) must be 8 DWORDS aligned",
					(u32)(*pskb)->data);
        }
#endif
		*p++ = 0x80000000 | DMA_PACKET_SIZE;
#if !defined(ENABLE_P2P_COPY_WITHOUT_DESC) || !ENABLE_P2P_COPY_WITHOUT_DESC
		*p++ = (u32)(*pskb++)->data & 0x1FFFFFE0;
#else
		if (g_eth_wan_mode == 0 || g_wanqos_en)
			*p++ = (u32) (*pskb++)->data & 0x1FFFFFE0;
		else
			*p++ = 0x00000000;
#endif
	}

	p = (volatile u32 *) DMA_RX_CH2_DESC_BASE;
	for (i = 0; i < DMA_RX_CH2_DESC_NUM; i++) {
#if !defined(ENABLE_P2P_COPY_WITHOUT_DESC) || !ENABLE_P2P_COPY_WITHOUT_DESC
		ASSERT(((u32)(*pskb)->data & (DMA_TX_ALIGNMENT - 1)) == 0, "DMA RX channel 2 data pointer (0x%#x) must be 8 DWORDS aligned", (u32)(*pskb)->data);
#endif
		*p++ = 0xB0000000 | DMA_PACKET_SIZE;
#if !defined(ENABLE_P2P_COPY_WITHOUT_DESC) || !ENABLE_P2P_COPY_WITHOUT_DESC
		*p++ = (u32)(*pskb++)->data & 0x1FFFFFE0;
#else
		*p++ = 0x00000000;
#endif
	}

	if (g_wanqos_en) {
		p = (volatile u32 *) DMA_TX_CH1_DESC_BASE;
		for (i = 0; i < DMA_TX_CH1_DESC_NUM; i++) {
			ASSERT(
					((u32)(*pskb)->data & (DMA_TX_ALIGNMENT - 1)) == 0,
					"DMA TX channel 1 data pointer (0x%#x) must be 8 DWORDS aligned",
					(u32)(*pskb)->data);
			*p++ = 0xB0000000;
			*p++ = (u32) (*pskb++)->data & 0x1FFFFFE0;
		}
	}

	g_f_dma_opened = 0;

	kfree(skb_pool);

	return 0;

	DMA_DEVICE_REGISTER_FAIL: dma_device_release(g_dma_device_ppe);
	g_dma_device_ppe = NULL;
	RESERVE_DMA_FAIL: i = pre_alloc_desc_total_num;
	ALLOC_SKB_FAIL: while (i--)
		dev_kfree_skb_any(skb_pool[i]);
	kfree(skb_pool);
	ALLOC_SKB_POOL_FAIL: return ret;
}

static INLINE void free_dma(void) {
    volatile struct tx_descriptor *p;
    struct sk_buff *skb_to_free;
    unsigned int i;

    printk(KERN_ERR "[%s] g_eth_wan_mode = %d, g_wanqos_en=%d\n", __func__, g_eth_wan_mode, g_wanqos_en );
    if ( g_eth_wan_mode == 0 || g_wanqos_en )
    {
        // can be cleared here (CPU -> Shared-Mem)
        printk(KERN_ERR "[%s] free cpu -> shared mem \n", __func__);
        p = (volatile struct tx_descriptor *)CPU_TO_WAN_TX_DESC_BASE;
        for ( i = 0; i < CPU_TO_WAN_TX_DESC_NUM; i++ )
        {
            skb_to_free = get_skb_pointer(p->dataptr);
            dev_kfree_skb_any(skb_to_free);
            p++;
        }

        // should be cleared here? not handled by DMA-HW
        printk(KERN_ERR "[%s] free dsl wan tx \n", __func__);
        p = (volatile struct tx_descriptor *)DSL_WAN_TX_DESC_BASE(0);
        for ( i = 0; i < WAN_TX_DESC_NUM_TOTAL; i++ )
        {
            skb_to_free = get_skb_pointer(g_eth_wan_mode == 0 ? p->dataptr << 2 : p->dataptr);
            dev_kfree_skb_any(skb_to_free);
            p++;
        }
    }

    // VORSICHT: Der folgende Code ist falsch! - AR9 / AR10 benutzen den Gleichen Speicher fuer
    // CPU_TO_WAN_SWAP_DESC_BASE und DSL_WAN_TX_DESC_BASE(0)
    // Somit darf nur einmal Alloc bzw. Free aufgerufen werden
#if 0
    if ( g_wanqos_en )
    {
        // should be cleared here? used in MBOX irq
        printk(KERN_ERR "[%s] free wan swap \n", __func__);
        p = (volatile struct tx_descriptor *)CPU_TO_WAN_SWAP_DESC_BASE;
        for ( i = 0; i < CPU_TO_WAN_SWAP_DESC_NUM; i++ )
        {
            skb_to_free = get_skb_pointer(p->dataptr);
            dev_kfree_skb_any(skb_to_free);
            p++;
        }
    }
#endif

    if ( g_eth_wan_mode == 0 )
    {
        // should be cleared here? used in MBOX (OAM) irq
        printk(KERN_ERR "[%s] free mbox oam0 \n", __func__);
        p = (volatile struct tx_descriptor *)WAN_RX_DESC_BASE(0);
        for ( i = 0; i < WAN_RX_DESC_NUM(0); i++ )
        {
            skb_to_free = get_skb_pointer(p->dataptr);
            dev_kfree_skb_any(skb_to_free);
            p++;
        }

        // should be cleared here? used in MBOX (OAM) irq
        printk(KERN_ERR "[%s] free mbox oam1 \n", __func__);
        p = (volatile struct tx_descriptor *)WAN_RX_DESC_BASE(1);
        for ( i = 0; i < WAN_RX_DESC_NUM(1); i++ )
        {
            skb_to_free = get_skb_pointer(p->dataptr << 2);
            dev_kfree_skb_any(skb_to_free);
            p++;
        }
    }

#if 0
	// PPA -> DSL cleard by DMA unregister (is real dma rx-chan with no_cpu_data)
    printk(KERN_ERR "[%s] free ppa -> dsl \n", __func__);
    p = (volatile struct tx_descriptor *)DMA_RX_CH1_DESC_BASE;
    for ( i = 0; i < DMA_RX_CH1_DESC_NUM; i++ )
    {
#if defined(ENABLE_P2P_COPY_WITHOUT_DESC) && ENABLE_P2P_COPY_WITHOUT_DESC
        if ( g_eth_wan_mode != 0 && !g_wanqos_en )
            break;
#endif
        skb_to_free = get_skb_pointer(p->dataptr);
        dev_kfree_skb_any(skb_to_free);
        p++;
    }

#if !defined(ENABLE_P2P_COPY_WITHOUT_DESC) || !ENABLE_P2P_COPY_WITHOUT_DESC
	// RX2 -> TX0: (PPA -> Switch(port6)) cleard by DMA unregister (is real dma rx-chan with no_cpu_data)
    printk(KERN_ERR "[%s] free rx2 -> tx0 \n", __func__);
    p = (volatile struct tx_descriptor *)DMA_RX_CH2_DESC_BASE;
    for ( i = 0; i < DMA_RX_CH2_DESC_NUM; i++ )
    {
        skb_to_free = get_skb_pointer(p->dataptr);
        dev_kfree_skb_any(skb_to_free);
        p++;
    }
#endif
#endif

    // DSL -> Switch(port11) not cleared by DMA unregister (is not a real dma rx-chan)
    if ( g_wanqos_en )
    {
        printk(KERN_ERR "[%s] free dsl -> switch \n", __func__);
        p = (volatile struct tx_descriptor *)DMA_TX_CH1_DESC_BASE;
        for ( i = 0; i < DMA_TX_CH1_DESC_NUM; i++ )
        {
            skb_to_free = get_skb_pointer(p->dataptr);
            dev_kfree_skb_any(skb_to_free);
            p++;
        }
    }

	dma_device_unregister(g_dma_device_ppe);
	dma_device_release(g_dma_device_ppe);
	g_dma_device_ppe = NULL;

#if defined(DEBUG_SKB_SWAP) && DEBUG_SKB_SWAP
    for ( i = 0; i < NUM_ENTITY(g_dbg_skb_swap_pool); i++ )
        if ( g_dbg_skb_swap_pool[i] != NULL )
        {
            err("skb swap pool is not clean: %d - %08x\n", i, (unsigned int)g_dbg_skb_swap_pool[i]);
        }
#endif
}

static void dsl_led_polling(unsigned long arg  __attribute__ ((unused)) ) {
	if (DSL_WAN_MIB_TABLE->wrx_correct_pdu != g_dsl_wrx_correct_pdu
			|| DSL_WAN_MIB_TABLE->wtx_total_pdu != g_dsl_wtx_total_pdu) {
		g_dsl_wrx_correct_pdu = DSL_WAN_MIB_TABLE->wrx_correct_pdu;
		g_dsl_wtx_total_pdu = DSL_WAN_MIB_TABLE->wtx_total_pdu;

		adsl_led_flash();
	}

	if (ifx_atm_alloc_tx != NULL) {
		g_dsl_led_polling_timer.expires = jiffies + HZ / 10; //  100ms
		add_timer(&g_dsl_led_polling_timer);
	}
}

#if defined(ENABLE_MY_MEMCPY) && ENABLE_MY_MEMCPY
static INLINE void *my_memcpy(unsigned char *dest, const unsigned char *src, unsigned int count)
{
	char *d = (char *)dest, *s = (char *)src;

	if (count >= 32) {
		int i = 8 - (((unsigned long) d) & 0x7);

		if (i != 8)
		while (i-- && count--) {
			*d++ = *s++;
		}

		if (((((unsigned long) d) & 0x7) == 0) &&
				((((unsigned long) s) & 0x7) == 0)) {
			while (count >= 32) {
				unsigned long long t1, t2, t3, t4;
				t1 = *(unsigned long long *) (s);
				t2 = *(unsigned long long *) (s + 8);
				t3 = *(unsigned long long *) (s + 16);
				t4 = *(unsigned long long *) (s + 24);
				*(unsigned long long *) (d) = t1;
				*(unsigned long long *) (d + 8) = t2;
				*(unsigned long long *) (d + 16) = t3;
				*(unsigned long long *) (d + 24) = t4;
				d += 32;
				s += 32;
				count -= 32;
			}
			while (count >= 8) {
				*(unsigned long long *) d =
				*(unsigned long long *) s;
				d += 8;
				s += 8;
				count -= 8;
			}
		}

		if (((((unsigned long) d) & 0x3) == 0) &&
				((((unsigned long) s) & 0x3) == 0)) {
			while (count >= 4) {
				*(unsigned long *) d = *(unsigned long *) s;
				d += 4;
				s += 4;
				count -= 4;
			}
		}

		if (((((unsigned long) d) & 0x1) == 0) &&
				((((unsigned long) s) & 0x1) == 0)) {
			while (count >= 2) {
				*(unsigned short *) d = *(unsigned short *) s;
				d += 2;
				s += 2;
				count -= 2;
			}
		}
	}

	while (count--) {
		*d++ = *s++;
	}

	return d;
}
#endif


static int print_fw_ver(char *buf, int buf_len) {
    static char * fw_ver_family_str[] = {
        "reserved - 0",
        "Danube",
        "Twinpass",
        "Amazon-SE",
        "reserved - 4",
        "AR9",
        "GR9",
        "VR9",
        "AR10"
    };
    static char * fw_ver_type_str[] = {
        "reserved - 0",
        "Standard",
        "Acceleration",
        "VDSL2 Bonding"
    };
    static char * fw_ver_interface_str[] = {
        "MII0",
        "MII0 + MII1",
        "MII0 + ATM",
        "MII0 + PTM",
        "MII0/1 + ATM",
        "MII0/1 + PTM",
        "reserved - 6",
        "MII0/1 + PTM Bonding"
    };
    static char * fw_ver_mode_str[] = {
        "reserved - 0",
        "PTM (E1)",
        "reserved - 2",
        "Routing",
        "reserved - 4",
        "Bridging",
        "Bridging + IPv4 Routing",
        "Bridging + IPv4/6 Routing",
        "6rd/DS-lite"
    };

    int len = 0;

    len += snprintf(buf + len, buf_len - len, "PPE firmware info:\n");
    len += snprintf(buf + len, buf_len - len, "  Version ID: %d.%d.%d.%d.%d.%d\n", (int)FW_VER_ID->family, (int)FW_VER_ID->fwtype, (int)FW_VER_ID->interface, (int)FW_VER_ID->fwmode, (int)FW_VER_ID->major, (int)FW_VER_ID->minor);
    if ( FW_VER_ID->family >= NUM_ENTITY(fw_ver_family_str) )
        len += snprintf(buf + len, buf_len - len, "  Family    : reserved - %d\n", (int)FW_VER_ID->family);
    else
        len += snprintf(buf + len, buf_len - len, "  Family    : %s\n", fw_ver_family_str[FW_VER_ID->family]);
    if ( FW_VER_ID->fwtype >= NUM_ENTITY(fw_ver_type_str) )
        len += snprintf(buf + len, buf_len - len, "  FW Type   : reserved - %d\n", (int)FW_VER_ID->fwtype);
    else
        len += snprintf(buf + len, buf_len - len, "  FW Type   : %s\n", fw_ver_type_str[FW_VER_ID->fwtype]);
    if ( FW_VER_ID->interface >= NUM_ENTITY(fw_ver_interface_str) )
        len += snprintf(buf + len, buf_len - len, "  Interface : reserved - %d\n", (int)FW_VER_ID->interface);
    else
        len += snprintf(buf + len, buf_len - len, "  Interface : %s\n", fw_ver_interface_str[FW_VER_ID->interface]);
    if ( FW_VER_ID->fwmode >= NUM_ENTITY(fw_ver_mode_str) )
        len += snprintf(buf + len, buf_len - len, "  Mode      : reserved - %d\n", (int)FW_VER_ID->fwmode);
    else
        len += snprintf(buf + len, buf_len - len, "  Mode      : %s\n", fw_ver_mode_str[FW_VER_ID->fwmode]);
    len += snprintf(buf + len, buf_len - len, "  Release   : %d.%d\n", (int)FW_VER_ID->major, (int)FW_VER_ID->minor);

    return len;
}

static int print_driver_ver(char *buf, int buf_len, char *title, unsigned int family, unsigned int type, unsigned int itf, unsigned int mode, unsigned int major, unsigned int mid, unsigned int minor)
{
    static char * dr_ver_family_str[] = {
        NULL,
        "Danube",
        "Twinpass",
        "Amazon-SE",
        NULL,
        "AR9",
        "VR9",
        "AR10",
        NULL,
    };
    static char * dr_ver_type_str[] = {
        "Normal Data Path",
        "Indirect-Fast Path",
        "HAL",
        "Hook",
        "OS Adatpion Layer",
        "PPA API"
    };
    static char * dr_ver_interface_str[] = {
        "MII0",
        "MII1",
        "ATM",
        "PTM"
    };
    static char * dr_ver_accmode_str[] = {
        "Routing",
        "Bridging",
    };

    int len = 0;
    unsigned int bit;
    unsigned int i, j;

    len += snprintf(buf + len, buf_len - len, "%s:\n", title);
    len += snprintf(buf + len, buf_len - len, "  Version ID: %d.%d.%d.%d.%d.%d.%d\n", family, type, itf, mode, major, mid, minor);
    len += snprintf(buf + len, buf_len - len, "  Family    : ");
    for ( bit = family, i = j = 0; bit != 0 && i < NUM_ENTITY(dr_ver_family_str); bit >>= 1, i++ )
        if ( (bit & 0x01) && dr_ver_family_str[i] != NULL )
        {
            if ( j )
                len += snprintf(buf + len, buf_len - len, " | %s", dr_ver_family_str[i]);
            else
                len += snprintf(buf + len, buf_len - len, dr_ver_family_str[i]);
            j++;
        }
    if ( j )
        len += snprintf(buf + len, buf_len - len, "\n");
    else
        len += snprintf(buf + len, buf_len - len, "N/A\n");
    len += snprintf(buf + len, buf_len - len, "  DR Type   : ");
    for ( bit = type, i = j = 0; bit != 0 && i < NUM_ENTITY(dr_ver_type_str); bit >>= 1, i++ )
        if ( (bit & 0x01) && dr_ver_type_str[i] != NULL )
        {
            if ( j )
                len += snprintf(buf + len, buf_len - len, " | %s", dr_ver_type_str[i]);
            else
                len += snprintf(buf + len, buf_len - len, dr_ver_type_str[i]);
            j++;
        }
    if ( j )
        len += snprintf(buf + len, buf_len - len, "\n");
    else
        len += snprintf(buf + len, buf_len - len, "N/A\n");
    len += snprintf(buf + len, buf_len - len, "  Interface : ");
    for ( bit = itf, i = j = 0; bit != 0 && i < NUM_ENTITY(dr_ver_interface_str); bit >>= 1, i++ )
        if ( (bit & 0x01) && dr_ver_interface_str[i] != NULL )
        {
            if ( j )
                len += snprintf(buf + len, buf_len - len, " | %s", dr_ver_interface_str[i]);
            else
                len += snprintf(buf + len, buf_len - len, dr_ver_interface_str[i]);
            j++;
        }
    if ( j )
        len += snprintf(buf + len, buf_len - len, "\n");
    else
        len += snprintf(buf + len, buf_len - len, "N/A\n");
    len += snprintf(buf + len, buf_len - len, "  Mode      : ");
    for ( bit = mode, i = j = 0; bit != 0 && i < NUM_ENTITY(dr_ver_accmode_str); bit >>= 1, i++ )
        if ( (bit & 0x01) && dr_ver_accmode_str[i] != NULL )
        {
            if ( j )
                len += snprintf(buf + len, buf_len - len, " | %s", dr_ver_accmode_str[i]);
            else
                len += snprintf(buf + len, buf_len - len, dr_ver_accmode_str[i]);
            j++;
        }
    if ( j )
        len += snprintf(buf + len, buf_len - len, "\n");
    else
        len += snprintf(buf + len, buf_len - len, "N/A\n");
    len += snprintf(buf + len, buf_len - len, "  Release   : %d.%d.%d\n", major, mid, minor);

    return len;
}

#if 0
static int print_dsl_dma_desc(char *buf, volatile u32 *desc_base, int desc_pos) {
    int len = 0;
    u32 *p;

    p = (u32*)desc_base;
    if ( desc_pos == 0 ) {
        len += sprintf(buf + len,"DSL Tx descriptor list:\n" );

        len += sprintf(buf + len," no  address        data pointer command bits (Own, Complete, SoP, EoP, Offset) \n");
        len += sprintf(buf + len,"---------------------------------------------------------------------------------\n");
    }
    len += sprintf(buf + len,"%3d  ",desc_pos);
    len += sprintf(buf + len,"0x%08x     ", (u32)(p + (desc_pos * 2)));
    len += sprintf(buf + len,"%08x     ", *(p + (desc_pos * 2 + 1)));
    len += sprintf(buf + len,"%08x     ", *(p + (desc_pos * 2)));

    if(*(p + (desc_pos * 2)) & 0x80000000)
        len += sprintf(buf + len, "C "); // ist bei A5 Treiber invers: 1 == CPU
    else
    	len += sprintf(buf + len, "D "); // ist bei A5 Treiber invers: 0 == DMA
    if(*(p + (desc_pos * 2)) & 0x40000000)
        len += sprintf(buf + len, "C ");
    else
        len += sprintf(buf + len, "c ");
    if(*(p + (desc_pos * 2)) & 0x20000000)
        len += sprintf(buf + len, "S ");
    else
        len += sprintf(buf + len, "s ");
    if(*(p + (desc_pos * 2)) & 0x10000000)
        len += sprintf(buf + len, "E ");
    else
        len += sprintf(buf + len, "e ");
    len += sprintf(buf + len,"%02x ", (*(p + (desc_pos * 2)) & 0x0F800000) >> 23);
    if( g_cpu_to_wan_tx_desc_pos  == desc_pos )
        len += sprintf(buf + len,"<- CURR");
    len += sprintf(buf + len,"\n");

    return len;
}
#endif

#if 0
static int proc_read_dsl_tx_dma_descr(IFX_char_t *page, IFX_char_t **start, off_t off __attribute__((unused)),
		int count, int *eof, void *data __attribute__((unused)))
{
    int len = 0;
    int len_max = off + count;
    char *pstr;
    char *str;
    int llen;
    int j;

    pstr = *start = page;

    str = vmalloc(1024);
    if(!str) {
        return -ENOMEM;
    }

    __sync();

	for (j = 0; j < CPU_TO_WAN_TX_DESC_NUM; j++) {
		llen = print_dsl_dma_desc(str,  (volatile u32 *)CPU_TO_WAN_TX_DESC_BASE, j);
		if ( len <= off && len + llen > off ) {
			memcpy(pstr, str + off - len, len + llen - off);
			pstr += len + llen - off;
		}
		else if ( len > off ) {
			memcpy(pstr, str, llen);
			pstr += llen;
		}
		len += llen;
		if ( len >= len_max )
			goto PROC_READ_DESC_LIST_OVERRUN_END;
	}

    *eof = 1;

    vfree(str);

    return len - off;

PROC_READ_DESC_LIST_OVERRUN_END:
    vfree(str);

    return len - llen - off;
}
#endif

static INLINE void proc_file_create(void) {
	struct proc_dir_entry *res;

//	g_eth_proc_dir = proc_mkdir("eth", NULL);  wird im comman Treiber angelegt

	create_proc_read_entry("ver", 0, g_eth_proc_dir, proc_read_ver, NULL);


	res = create_proc_entry("mib", 0, g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_mib;
        res->write_proc = proc_write_mib;
    }

#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC
    res = create_proc_entry("mpoa", 0, g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_mpoa;
        res->write_proc = proc_write_mpoa;
    }

	res = create_proc_entry("dsl_tx_dma_descr", 0, g_eth_proc_dir);
	if (res) {
		res->read_proc = proc_read_dsl_tx_dma_descr;
	}


#if defined(PROC_WRITE_ROUTE) && PROC_WRITE_ROUTE
    res = create_proc_entry("route",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_route;
        res->write_proc = proc_write_route;
    }
#endif

    create_proc_read_entry("mc",
                           0,
                           g_eth_proc_dir,
                           proc_read_mc,
                           NULL);

    res = create_proc_entry("genconf",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_genconf;
        res->write_proc = proc_write_genconf;
    }

    res = create_proc_entry("queue",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_queue;
        res->write_proc = proc_write_queue;
    }

    create_proc_read_entry("pppoe",
                           0,
                           g_eth_proc_dir,
                           proc_read_pppoe,
                           NULL);

    create_proc_read_entry("mtu",
                           0,
                           g_eth_proc_dir,
                           proc_read_mtu,
                           NULL);

    res = create_proc_entry("hit",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_hit;
        res->write_proc = proc_write_hit;
    }

    create_proc_read_entry("mac",
                           0,
                           g_eth_proc_dir,
                           proc_read_mac,
                           NULL);

    create_proc_read_entry("out_vlan",
                            0,
                            g_eth_proc_dir,
                            proc_read_out_vlan,
                            NULL);

    if ( g_ipv6_acc_en )
    {
        create_proc_read_entry("ipv6_ip",
                            0,
                            g_eth_proc_dir,
                            proc_read_ipv6_ip,
                            NULL);
    }

#endif

    
#if defined(DEBUG_HTU_PROC) && DEBUG_HTU_PROC
    create_proc_read_entry("htu",
                           0,
                           g_eth_proc_dir,
                           proc_read_htu,
                           NULL);
#endif

#if defined(DEBUG_MEM_PROC) && DEBUG_MEM_PROC
    res = create_proc_entry("mem",
                            0,
                            g_eth_proc_dir);
    if ( res )
        res->write_proc = proc_write_mem;
#endif


    res = create_proc_entry("burstlen",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_burstlen;
        res->write_proc = proc_write_burstlen;
    }

#if defined(DEBUG_FW_PROC) && DEBUG_FW_PROC
    res = create_proc_entry("fw",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_fw;
        res->write_proc = proc_write_fw;
    }

    res = create_proc_read_entry("tfwdbg",
                                   0,
                                   g_eth_proc_dir,
                                   NULL,
                                   NULL);
    if(res)
        res->write_proc = proc_write_fwdbg;
#endif

#if defined(DEBUG_SEND_PROC) && DEBUG_SEND_PROC
    res = create_proc_entry("send",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_send;
        res->write_proc = proc_write_send;
    }
#endif

    res = create_proc_entry("port0mib",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = port0_counters_proc_read;
        res->write_proc = port_counters_proc_write;
    }

    res = create_proc_entry("port1mib",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = port1_counters_proc_read;
        res->write_proc = port_counters_proc_write;
    }

    res = create_proc_entry("port2mib",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = port2_counters_proc_read;
        res->write_proc = port_counters_proc_write;
    }

    res = create_proc_entry("port3mib",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = port3_counters_proc_read;
        res->write_proc = port_counters_proc_write;
    }

    res = create_proc_entry("port4mib",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = port4_counters_proc_read;
        res->write_proc = port_counters_proc_write;
    }

    res = create_proc_entry("port5mib",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = port5_counters_proc_read;
        res->write_proc = port_counters_proc_write;
    }

    res = create_proc_entry("port6mib",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = port6_counters_proc_read;
        res->write_proc = port_counters_proc_write;
    }

    res = create_proc_entry("direct_forwarding",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_directforwarding;
        res->write_proc = proc_write_directforwarding;
    }

    res = create_proc_entry("clk",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_clk;
        res->write_proc = proc_write_clk;
    }

    res = create_proc_entry("prio",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_prio;
        res->write_proc = proc_write_prio;
    }

    res = create_proc_entry("ewan",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_ewan;
        res->write_proc = proc_write_ewan;
    }

#ifdef CONFIG_IFX_PPA_QOS
    if ( g_wanqos_en )
    {
        create_proc_read_entry("qos",
                               0,
                               g_eth_proc_dir,
                               proc_read_qos,
                               NULL);
    }
#endif

#if defined(ENABLE_CONFIGURABLE_DSL_VLAN) && ENABLE_CONFIGURABLE_DSL_VLAN
    res = create_proc_entry("dsl_vlan",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_dsl_vlan;
        res->write_proc = proc_write_dsl_vlan;
    }
#endif

#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC
    res = create_proc_entry("mirror",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_mirror;
        res->write_proc = proc_write_mirror;
    }
#endif
    res = create_proc_entry("class",
                            0,
                            g_eth_proc_dir);
    if ( res )
    {
        res->read_proc  = proc_read_class;
        res->write_proc = proc_write_class;
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static INLINE void proc_file_delete(void)
{
    remove_proc_entry("class",
                      g_eth_proc_dir);
#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC
    remove_proc_entry("mirror",
                      g_eth_proc_dir);
#endif

#if defined(ENABLE_CONFIGURABLE_DSL_VLAN) && ENABLE_CONFIGURABLE_DSL_VLAN
    remove_proc_entry("dsl_vlan",
                      g_eth_proc_dir);
#endif

#ifdef CONFIG_IFX_PPA_QOS
    if ( g_wanqos_en )
    {
        remove_proc_entry("qos",
                          g_eth_proc_dir);
    }
#endif

    remove_proc_entry("ewan",
                      g_eth_proc_dir);

    remove_proc_entry("prio",
                      g_eth_proc_dir);

    remove_proc_entry("clk",
                      g_eth_proc_dir);

    remove_proc_entry("direct_forwarding",
                      g_eth_proc_dir);

    remove_proc_entry("port6mib",
                      g_eth_proc_dir);

    remove_proc_entry("port5mib",
                      g_eth_proc_dir);

    remove_proc_entry("port4mib",
                      g_eth_proc_dir);

    remove_proc_entry("port3mib",
                      g_eth_proc_dir);

    remove_proc_entry("port2mib",
                      g_eth_proc_dir);

    remove_proc_entry("port1mib",
                      g_eth_proc_dir);

    remove_proc_entry("port0mib",
                      g_eth_proc_dir);

#if defined(DEBUG_SEND_PROC) && DEBUG_SEND_PROC
    remove_proc_entry("send",
                      g_eth_proc_dir);
#endif

#if defined(DEBUG_FW_PROC) && DEBUG_FW_PROC
    remove_proc_entry("tfwdbg",
                      g_eth_proc_dir);

    remove_proc_entry("fw",
                      g_eth_proc_dir);
#endif

    remove_proc_entry("burstlen",
                      g_eth_proc_dir);


#if defined(DEBUG_MEM_PROC) && DEBUG_MEM_PROC
    remove_proc_entry("mem",
                      g_eth_proc_dir);
#endif

#if defined(DEBUG_HTU_PROC) && DEBUG_HTU_PROC
    remove_proc_entry("htu",
                      g_eth_proc_dir);
#endif

#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC

    if ( g_ipv6_acc_en )
    {
        remove_proc_entry("ipv6_ip",
                          g_eth_proc_dir);
    }

    remove_proc_entry("out_vlan",
                      g_eth_proc_dir);

    remove_proc_entry("mac",
                      g_eth_proc_dir);

    remove_proc_entry("hit",
                      g_eth_proc_dir);

    remove_proc_entry("mtu",
                      g_eth_proc_dir);

    remove_proc_entry("pppoe",
                      g_eth_proc_dir);

    remove_proc_entry("queue",
                      g_eth_proc_dir);

    remove_proc_entry("genconf",
                      g_eth_proc_dir);

    remove_proc_entry("mc",
                      g_eth_proc_dir);

    remove_proc_entry("route",
                      g_eth_proc_dir);
#endif

    remove_proc_entry("dsl_tx_dma_descr", g_eth_proc_dir);

    remove_proc_entry("mib",
                      g_eth_proc_dir);

    remove_proc_entry("mpoa",
                      g_eth_proc_dir);

    remove_proc_entry("ver",
                      g_eth_proc_dir);

    // remove_proc_entry("eth", NULL); wird im common treiber angelegt
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static int proc_read_ver(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count,
		int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += print_driver_ver(page + len, count - len,
			"PPE datapath driver info", VER_FAMILY, VER_DRTYPE, VER_INTERFACE,
			VER_ACCMODE, VER_MAJOR, VER_MID, VER_MINOR);
	len += print_fw_ver(page + len, count - len);

	*eof = 1;

	return len;
}



static int proc_read_mib(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)), int *eof, void *data __attribute__((unused)))
{
    int len = 0;
#if defined(SETUP_OWN_NETDEVS)
    struct eth_priv_data *priv0, *priv1;
#endif
    char *itf_name[] = {"eth0", "eth1", "cpu", "ext_int1", "ext_int2", "ext_int3", "", "atm"};
    char *row_name[] = {"ig_fast_brg_pkts", "ig_fast_brg_bytes", "ig_fast_rt_ipv4_udp_pkts", "ig_fast_rt_ipv4_tcp_pkts", "ig_fast_rt_ipv4_mc_pkts", "ig_fast_rt_ipv4_bytes", "ig_fast_rt_ipv6_udp_pkts", "ig_fast_rt_ipv6_tcp_pkts", NULL, "ig_fast_rt_ipv6_bytes", NULL, "ig_cpu_pkts", "ig_cpu_bytes", "ig_drop_pkts", "ig_drop_bytes", "eg_fast_pkts"};
    unsigned int i, j, k, h;

#if defined(SETUP_OWN_NETDEVS)
    eth_get_stats(g_eth_net_dev[0]);
    eth_get_stats(g_eth_net_dev[1]);
    priv0 = ifx_eth_fw_netdev_priv(g_eth_net_dev[0]);
    priv1 = ifx_eth_fw_netdev_priv(g_eth_net_dev[1]);
#endif

    len += sprintf(page + len, "Ethernet\n");

    len += sprintf(page + len, "  Firmware  (");
    for ( i = h = 0; i < NUM_ENTITY(itf_name); i++ )
        if ( (g_mib_itf & (1 << i)) )
            len += sprintf(page + len, h++ == 0 ? "%s" : ", %s", itf_name[i]);
    len += sprintf(page + len, ")\n");
    for ( i = k = 0; i < NUM_ENTITY(row_name); i++ )
        if ( row_name[i] != NULL && k < (j = strlen(row_name[i])) )
            k = j;
    k += 6;
    for ( i = 0; i < NUM_ENTITY(row_name); i++ ) {
        if ( row_name[i] == NULL )
            continue;

        len += (j = sprintf(page + len, "    %s: ", row_name[i]));
        for ( j = k - j; j > 0; j-- )
            page[len++] = ' ';
        for ( j = h = 0; j < 8; j++ )
            if ( (g_mib_itf & (1 << j)) )
                len += sprintf(page + len, h++ == 0 ? "%10u" : ", %10u", ((volatile unsigned int *)ITF_MIB_TBL(j))[i]);
        len += sprintf(page + len, "\n");
    }

#if defined(SETUP_OWN_NETDEVS)
    len += sprintf(page + len, "Driver (ETH0, ETH1, ATM)\n");
    len += sprintf(page + len, "  rx_packets: %10u, %10u, %10u\n",       priv0->rx_packets,  priv1->rx_packets,  g_atm_priv_data.wrx_pdu);
    len += sprintf(page + len, "  rx_bytes:   %10u, %10u,        N/A\n", priv0->rx_bytes,    priv1->rx_bytes);
    len += sprintf(page + len, "  rx_dropped: %10u, %10u, %10u\n",       priv0->rx_dropped,  priv1->rx_dropped,  g_atm_priv_data.wrx_drop_pdu);
    len += sprintf(page + len, "  tx_packets: %10u, %10u, %10u\n",       priv0->tx_packets,  priv1->tx_packets,  g_atm_priv_data.wtx_pdu);
    len += sprintf(page + len, "  tx_bytes:   %10u, %10u,        N/A\n", priv0->tx_bytes,    priv1->tx_bytes);
    len += sprintf(page + len, "  tx_errors:  %10u, %10u, %10u\n",       priv0->tx_errors,   priv1->tx_errors,   g_atm_priv_data.wtx_err_pdu);
    len += sprintf(page + len, "  tx_dropped: %10u, %10u, %10u\n",       priv0->tx_dropped,  priv1->tx_dropped,  g_atm_priv_data.wtx_drop_pdu);
#endif

    len += sprintf(page + len, "DSL WAN MIB\n");
    len += sprintf(page + len, "  wrx_drophtu_cell: %10u\n", DSL_WAN_MIB_TABLE->wrx_drophtu_cell);
    len += sprintf(page + len, "  wrx_dropdes_pdu:  %10u\n", DSL_WAN_MIB_TABLE->wrx_dropdes_pdu);
    len += sprintf(page + len, "  wrx_correct_pdu:  %10u\n", DSL_WAN_MIB_TABLE->wrx_correct_pdu);
    len += sprintf(page + len, "  wrx_err_pdu:      %10u\n", DSL_WAN_MIB_TABLE->wrx_err_pdu);
    len += sprintf(page + len, "  wrx_dropdes_cell: %10u\n", DSL_WAN_MIB_TABLE->wrx_dropdes_cell);
    len += sprintf(page + len, "  wrx_correct_cell: %10u\n", DSL_WAN_MIB_TABLE->wrx_correct_cell);
    len += sprintf(page + len, "  wrx_err_cell:     %10u\n", DSL_WAN_MIB_TABLE->wrx_err_cell);
    len += sprintf(page + len, "  wrx_total_byte:   %10u\n", DSL_WAN_MIB_TABLE->wrx_total_byte);
    len += sprintf(page + len, "  wtx_total_pdu:    %10u\n", DSL_WAN_MIB_TABLE->wtx_total_pdu);
    len += sprintf(page + len, "  wtx_total_cell:   %10u\n", DSL_WAN_MIB_TABLE->wtx_total_cell);
    len += sprintf(page + len, "  wtx_total_byte:   %10u\n", DSL_WAN_MIB_TABLE->wtx_total_byte);

#if 0  //proc stack overflow so remove it
    if ( g_eth_wan_mode == 0 )
    {
        len += sprintf(page + len, "DSL Queue DROP MIB\n");
        for ( i = 0; i < ATM_QUEUE_NUMBER; i++ )
        {
            len += sprintf(page + len, "  queue%02d: %10u", i, DSL_QUEUE_TX_DROP_MIB_TABLE(i)->pdu);
            if ( (i % 4) == 3 )
                len += sprintf(page + len, "\n");
        }
        len += sprintf(page + len, "\n");
    }
#endif

	len += sprintf(page + len, "dma_alignment_atm_good_count: %d\n", dma_alignment_atm_good_count );
	len += sprintf(page + len, "dma_alignment_atm_bad_count: %d\n", dma_alignment_atm_bad_count   );
	len += sprintf(page + len, "dma_alignment_eth_good_count: %d\n", dma_alignment_eth_good_count );
	len += sprintf(page + len, "dma_alignment_eth_bad_count: %d\n", dma_alignment_eth_bad_count   );
	len += sprintf(page + len, "oam_cell_drop_count: %u\n", oam_cell_drop_count);
	len += sprintf(page + len, "padded_frame_count : %d\n", padded_frame_count );


    *eof = 1;

    ASSERT(len <= 4096, "proc read buffer overflow");
    return len;
}

static int proc_write_mib(struct file *file __attribute__((unused)), const char *buf, unsigned long count, void *data __attribute__((unused)))
{
    char *p1, *p2;
    int len;
    int colon;
    // char local_buf[1024];

#if defined(SETUP_OWN_NETDEVS)
    struct eth_priv_data *priv;
#endif
    unsigned int mib_itf = 0;
    char *itf_name[] = {"eth0", "eth1", "cpu", "ext_int1", "ext_int2", "ext_int3", "", "atm"};
    unsigned int i;

    len = sizeof(common_proc_read_buff) < count ? sizeof(common_proc_read_buff) - 1 : count;
    len = len - copy_from_user(common_proc_read_buff, buf, len);
    common_proc_read_buff[len] = 0;

    p1 = common_proc_read_buff;
    p2 = NULL;
    colon = 1;
    while ( ifx_get_token(&p1, &p2, &len, &colon) )
    {
        if ( ifx_stricmp(p1, "queue") == 0 )
        {
            for ( i = 0; i < 16; i++ )
                printk("Qid %2d: rx_pdu = %10u, rx_bytes = %10u, tx_pdu = %10u, tx_bytes = %10u\n", i,
                        DSL_QUEUE_RX_MIB_TABLE(i)->pdu, DSL_QUEUE_RX_MIB_TABLE(i)->bytes,
                        DSL_QUEUE_TX_MIB_TABLE(i)->pdu, DSL_QUEUE_TX_MIB_TABLE(i)->bytes);
        }
        else if ( ifx_stricmp(p1, "pvc") == 0 )
        {
            int vpi, vci;
            int conn;

            ifx_ignore_space(&p2, &len);
            vpi = ifx_get_number(&p2, &len, 0);
            ifx_ignore_space(&p2, &len);
            vci = ifx_get_number(&p2, &len, 0);

            conn = find_vpivci(vpi, vci);

            printk("vpi = %u, vci = %u, conn = %d\n", vpi, vci, conn);

            if ( conn >= 0 )
            {
                printk("rx_packets  = %u\n", g_atm_priv_data.connection[conn].rx_packets);
                printk("rx_bytes    = %u\n", g_atm_priv_data.connection[conn].rx_bytes);
                printk("rx_errors   = %u\n", g_atm_priv_data.connection[conn].rx_errors);
                printk("rx_dropped  = %u\n", g_atm_priv_data.connection[conn].rx_dropped);
                printk("tx_packets  = %u\n", g_atm_priv_data.connection[conn].tx_packets);
                printk("tx_bytes    = %u\n", g_atm_priv_data.connection[conn].tx_bytes);
                printk("tx_errors   = %u\n", g_atm_priv_data.connection[conn].tx_errors);
                printk("tx_dropped  = %u\n", g_atm_priv_data.connection[conn].tx_dropped);
                printk("prio_tx_packets:");
                for ( i = 0; i < 8; i++ )
                    printk(" %u", g_atm_priv_data.connection[conn].prio_tx_packets[i]);
                printk("\n");
            }
        }
        else if ( ifx_stricmp(p1, "on") == 0 || ifx_stricmp(p1, "enable") == 0 )
            mib_itf = 0x40000000;
        else if ( ifx_stricmp(p1, "off") == 0 || ifx_stricmp(p1, "disable") == 0 )
            mib_itf = 0x20000000;
        else if ( ifx_stricmp(p1, "clear") == 0 || ifx_stricmp(p1, "clean") == 0 )
            mib_itf = 0x80000000;
        else if ( (mib_itf & 0xE0000000) )
        {
            if ( ifx_stricmp(p1, "all") == 0 )
                mib_itf |= 0xFF;
            else if ( strlen(p1) == 1 && *p1 >= '0' && *p1 <= '7' )
                mib_itf |= 1 << (*p1 - '0');
            else
            {
                for ( i = 0; i < NUM_ENTITY(itf_name); i++ )
                    if ( ifx_stricmp(p1, itf_name[i]) == 0 )
                    {
                        if ( i < 6 )
                            mib_itf |= 1 << i;
                        else
                            mib_itf |= 3 << 6;
                        break;
                    }
            }
        }

        p1 = p2;
        colon = 1;
    }

    if ( (mib_itf & 0x40000000) )
        g_mib_itf |= mib_itf & 0xFF;
    else if ( (mib_itf & 0x20000000) )
        g_mib_itf &= ~(mib_itf & 0xFF);
    else if ( (mib_itf & 0x80000000) )
    {
        if ( mib_itf == 0x80000000 )
            mib_itf = 0x800000FF;

#if defined(SETUP_OWN_NETDEVS)
        if ( (mib_itf & 0x01) )
        {
            eth_get_stats(g_eth_net_dev[0]);
            priv = ifx_eth_fw_netdev_priv(g_eth_net_dev[0]);
            memset(priv, 0, (u32)&priv->tx_dropped - (u32)priv + sizeof(u32));
        }
        if ( (mib_itf & 0x02) )
        {
            eth_get_stats(g_eth_net_dev[1]);
            priv = ifx_eth_fw_netdev_priv(g_eth_net_dev[1]);
            memset(priv, 0, (u32)&priv->tx_dropped - (u32)priv + sizeof(u32));
        }
#endif
        if ( (mib_itf & 0xC0) )
        {
            volatile struct dsl_queue_drop_mib *dsl_drop_mib;

            memset(&g_atm_priv_data.wrx_total_byte, 0, (u32)&g_atm_priv_data.prev_mib - (u32)&g_atm_priv_data.wrx_total_byte);
            for ( i = 0; i < ATM_QUEUE_NUMBER; i++ )
                memset(&g_atm_priv_data.connection[i].rx_packets, 0, (u32)&g_atm_priv_data.connection[i].port - (u32)&g_atm_priv_data.connection[i].rx_packets);
            memset((void *)DSL_WAN_MIB_TABLE, 0, sizeof(struct dsl_wan_mib_table));
//            for ( i = 0; i < 15; i++ )
//                memset((void *)DSL_WAN_VC_MIB_TABLE(i), 0, sizeof(struct dsl_wan_vc_mib_table));
            memset((void *)DSL_QUEUE_RX_MIB_TABLE(0), 0, sizeof(*DSL_QUEUE_RX_MIB_TABLE(0)) * 2);
            memset((void *)DSL_QUEUE_TX_MIB_TABLE(0), 0, sizeof(*DSL_QUEUE_TX_MIB_TABLE(0)) * 15);
            for ( i = 0; i < ATM_QUEUE_NUMBER; i++ )
            {
                memset(&g_atm_priv_data.connection[i].prio_tx_packets, 0, 8 * sizeof(u32));
                if ( g_eth_wan_mode == 0 )
                {
                    dsl_drop_mib = DSL_QUEUE_TX_DROP_MIB_TABLE(i);
                    memset((void *)dsl_drop_mib, 0, sizeof(DSL_QUEUE_TX_DROP_MIB_TABLE(i)));
                }
            }
        }
        for ( i = 0; i < 8; i++ )
            if ( (mib_itf & (1 << i)) )
            {
                void *ptmp = (void *)ITF_MIB_TBL(i);
                memset(ptmp, 0, sizeof(struct itf_mib));
            }
    }
    else
    {
        printk("echo <on|off|clear> <all|eth0|eth1|atm|cpu|ext_int?> > /proc/eth/mib\n");
        printk("echo queue > /proc/eth/mib\n");
        printk("echo pvc <vpi>.<vci> > /proc/eth/mib\n");
    }

    return count;
}


#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC
#if defined(PROC_WRITE_ROUTE) && PROC_WRITE_ROUTE
static int proc_read_route(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    struct rout_forward_compare_tbl *pcompare;
    struct rout_forward_action_tbl *paction;
    int len = 0;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;
    int i;

    pstr = *start = page;

    __sync();

    llen = sprintf(common_proc_read_buff, "Wan Routing Table\n");
    proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen);

    pcompare = (struct rout_forward_compare_tbl *)ROUT_WAN_HASH_CMP_TBL(0);
    paction = (struct rout_forward_action_tbl *)ROUT_WAN_HASH_ACT_TBL(0);
    for ( i = 0; i < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK; i++ )
    {
        if ( *(u32*)pcompare && *((u32*)pcompare + 1) )
        {
            llen = print_route(common_proc_read_buff, i, 0, pcompare, paction);
            if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                goto PROC_READ_ROUTE_OVERRUN_END;
        }

        pcompare++;
        paction++;
    }

    pcompare = (struct rout_forward_compare_tbl *)ROUT_WAN_COLL_CMP_TBL(0);
    paction = (struct rout_forward_action_tbl *)ROUT_WAN_COLL_ACT_TBL(0);
    for ( i = 0; i < WAN_ROUT_TBL_CFG->rout_num; i++ )
    {
        if ( *(u32*)pcompare && *((u32*)pcompare + 1) )
        {
            llen = print_route(common_proc_read_buff, i + MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK, 0, pcompare, paction);
            if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                goto PROC_READ_ROUTE_OVERRUN_END;
        }

        pcompare++;
        paction++;
    }

    llen = sprintf(common_proc_read_buff, "Lan Routing Table\n");
    if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
        goto PROC_READ_ROUTE_OVERRUN_END;

    pcompare = (struct rout_forward_compare_tbl *)ROUT_LAN_HASH_CMP_TBL(0);
    paction = (struct rout_forward_action_tbl *)ROUT_LAN_HASH_ACT_TBL(0);
    for ( i = 0; i < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK; i++ )
    {
        if ( *(u32*)pcompare && *((u32*)pcompare + 1) )
        {
            llen = print_route(common_proc_read_buff, i, 1, pcompare, paction);
            if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                goto PROC_READ_ROUTE_OVERRUN_END;
        }

        pcompare++;
        paction++;
    }

    pcompare = (struct rout_forward_compare_tbl *)ROUT_LAN_COLL_CMP_TBL(0);
    paction = (struct rout_forward_action_tbl *)ROUT_LAN_COLL_ACT_TBL(0);
    for ( i = 0; i < LAN_ROUT_TBL_CFG->rout_num; i++ )
    {
        if ( *(u32*)pcompare && *((u32*)pcompare + 1) )
        {
            llen = print_route(common_proc_read_buff, i + MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK, 1, pcompare, paction);
            if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                goto PROC_READ_ROUTE_OVERRUN_END;
        }

        pcompare++;
        paction++;
    }

    *eof = 1;

PROC_READ_ROUTE_OVERRUN_END:
    return len - off;
}

static int proc_write_route(struct file *file __attribute__((unused)), const char *buf, unsigned long count, void *data __attribute__((unused)))
{
	static const char *command[] = {
			"add",      //  0
			"del",      //  1
			"LAN",      //  2
			"WAN",      //  3
			"new",      //  4
			"src",      //  5
			"dest",     //  6
			"MAC",      //  7
			"route",    //  8
			"type",     //  9
			"DSCP",     //  10
			"MTU",      //  11
			"index",    //  12
			"VLAN",     //  13
			"insert",   //  14
			"remove",   //  15
			"list",     //  16
			"PPPoE",    //  17
			"mode",     //  18
			"ch",       //  19
			"id",       //  20
			"delete",   //  21
			"disable",  //  22
			"enable",   //  23
			"transparent",  //  24
			"termination",  //  25
			"NULL",     //  26
			"IPv4",     //  27
			"NAT",      //  28
			"NAPT",     //  29
			"entry",    //  30
			"tcp",      //  31
			"help",     //  32
			"vci",      //  33
			"yes",      //  34
			"no",       //  35
			"qid",      //  36
			"outer",    //  37
			"IP",       //  38
			"IPv6",     //  39
		};


	static const char *dest_list[] = {"ETH0", "ETH1", "CPU0", "EXT_INT1", "EXT_INT2", "EXT_INT3", "EXT_INT4", "ATM"};
	static const int dest_list_strlen[] = {4, 4, 4, 8, 8, 8, 8, 3};

    int state;              //  1: new,
    int prev_cmd;
    int operation;          //  1: add, 2: delete
    int type;               //  1: LAN, 2: WAN, 0: auto detect
    int entry;
    struct rout_forward_compare_tbl compare_tbl;
    struct rout_forward_action_tbl action_tbl;
    uint32_t ipv6_src_ip[4] = {0}, ipv6_dst_ip[4] = {0};
    int is_ipv6 = 0;
    unsigned int val[20];
//    char common_proc_read_buff[1024];
    int len;
    char *p1, *p2;
    int colon;
    int i, j;
    u32 mask;
    u32 bit;
    u32 *pu1, *pu2, *pu3;

    len = sizeof(common_proc_read_buff) < count ? sizeof(common_proc_read_buff) - 1 : count;
    len = len - copy_from_user(common_proc_read_buff, buf, len);
    common_proc_read_buff[len] = 0;

    state = 0;
    prev_cmd = 0;
    operation = 0;
    type = 0;
    entry = -1;

    memset(&compare_tbl, 0, sizeof(compare_tbl));
    memset(&action_tbl, 0, sizeof(action_tbl));

    p1 = common_proc_read_buff;
    colon = 1;
    while ( ifx_get_token(&p1, &p2, &len, &colon) )
    {
        for ( i = 0; i < sizeof(command) / sizeof(*command); i++ )
            if ( ifx_stricmp(p1, command[i]) == 0 )
            {
                switch ( i )
                {
                case 0:
                    if ( !state && !operation )
                    {
                        operation = 1;
//                      printk("add\n");
                    }
                    state = prev_cmd = 0;
                    break;
                case 1:
                case 21:
                    if ( !state && !operation )
                    {
                        operation = 2;
//                      printk("delete\n");
                    }
                    state = prev_cmd = 0;
                    break;
                case 2:
                    if ( !state && !type )
                    {
                        type = 1;
//                      printk("lan\n");
                    }
                    state = prev_cmd = 0;
                    break;
                case 3:
                    if ( !state && !type )
                    {
                        type = 2;
//                      printk("wan\n");
                    }
                    state = prev_cmd = 0;
                    break;
                case 4:
                    if ( !state )
                    {
                        state = 1;
                        prev_cmd = 4;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 5:
                    if ( state == 1 )
                    {
                        if ( prev_cmd == 4 )
                        {
                            //  new src
                            if ( !type )
                                type = 1;

                            //  check for "new src mac index"
                            ifx_ignore_space(&p2, &len);
                            if ( ifx_strincmp(p2, "mac ", 4) == 0 )
                            {
                                state = 2;
                                prev_cmd = 5;
                                break;
                            }
                            else
                            {
                                get_ip_port(&p2, &len, val);
//                              printk("new src: %d.%d.%d.%d:%d\n", val[0], val[1], val[2], val[3], val[4]);
                                action_tbl.new_ip = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                                action_tbl.new_port = val[4];
                            }
                        }
                        else
                            state = 0;
                    }
                    if ( state == 0 )
                    {
                        //  src
                        get_ip_port(&p2, &len, val);
//                      printk("src: %d.%d.%d.%d:%d\n", val[0], val[1], val[2], val[3], val[4]);
                        if ( val[5] == 4 )
                        {
                            if ( is_ipv6 )
                                ipv6_src_ip[0] = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                            else
                                compare_tbl.src_ip = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                        }
                        else
                        {
                            is_ipv6 = 1;
                            ipv6_src_ip[0] = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                            ipv6_src_ip[1] = (val[6] << 24) | (val[7] << 16) | (val[8] << 8) | val[9];
                            ipv6_src_ip[2] = (val[10] << 24) | (val[11] << 16) | (val[12] << 8) | val[13];
                            ipv6_src_ip[3] = (val[14] << 24) | (val[15] << 16) | (val[16] << 8) | val[17];
                        }
                        compare_tbl.src_port = val[4];
                    }
                    state = prev_cmd = 0;
                    break;
                case 6:
                    if ( state == 1 )
                    {
                        if ( prev_cmd == 4 )
                        {
                            //  new dest
                            if ( !type )
                                type = 2;

                            get_ip_port(&p2, &len, val);
//                          printk("new dest: %d.%d.%d.%d:%d\n", val[0], val[1], val[2], val[3], val[4]);
                            action_tbl.new_ip = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                            action_tbl.new_port = val[4];
                        }
                        else
                            state = 0;
                    }
                    if ( state == 0 )
                    {
                        if ( !colon )
                        {
                            int llen;

                            llen = len;
                            p1 = p2;
                            while ( llen && *p1 <= ' ' )
                            {
                                llen--;
                                p1++;
                            }
                            if ( llen && (*p1 == ':' || (*p1 >= '0' && *p1 <= '9')) )
                                colon = 1;
                        }
                        if ( colon )
                        {
                            //  dest
                            get_ip_port(&p2, &len, val);
//                          printk("dest: %d.%d.%d.%d:%d\n", val[0], val[1], val[2], val[3], val[4]);
                            if ( val[5] == 4 )
                            {
                                if ( is_ipv6 )
                                    ipv6_dst_ip[0] = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                                else
                                    compare_tbl.dest_ip = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                            }
                            else
                            {
                                is_ipv6 = 1;
                                ipv6_dst_ip[0] = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
                                ipv6_dst_ip[1] = (val[6] << 24) | (val[7] << 16) | (val[8] << 8) | val[9];
                                ipv6_dst_ip[2] = (val[10] << 24) | (val[11] << 16) | (val[12] << 8) | val[13];
                                ipv6_dst_ip[3] = (val[14] << 24) | (val[15] << 16) | (val[16] << 8) | val[17];
                            }
                            compare_tbl.dest_port = val[4];
                        }
                        else
                        {
                            state = 1;
                            prev_cmd = 6;
                            break;
                        }
                    }
                    state = prev_cmd = 0;
                    break;
                case 7:
                    if ( state == 1 && prev_cmd == 4 )
                    {
                        //  new MAC
                        get_mac(&p2, &len, val);
//                      printk("new MAC: %02X.%02X.%02X.%02X:%02X:%02X\n", val[0], val[1], val[2], val[3], val[4], val[5]);
                        action_tbl.new_dest_mac54 = (val[0] << 8) | val[1];
                        action_tbl.new_dest_mac30 = (val[2] << 24) | (val[3] << 16) | (val[4] << 8) | val[5];
                    }
                    else if ( state == 2 && prev_cmd == 5 )
                    {
                        state = 3;
                        prev_cmd = 7;
                        break;
                    }
                    state = prev_cmd = 0;
                    break;
                case 8:
                    if ( !state )
                    {
                        state = 1;
                        prev_cmd = 8;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 9:
                    if ( state == 1 && prev_cmd == 8 )
                    {
                        state = 2;
                        prev_cmd = 9;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 10:
                    if ( state == 1 && prev_cmd == 4 )
                    {
                        ifx_ignore_space(&p2, &len);
                        if ( len && *p2 >= '0' && *p2 <= '9' )
                        {
                            //  new DSCP
                            val[0] = ifx_get_number(&p2, &len, 0);
//                          printk("new DSCP: %d\n", val[0]);
                            if ( !action_tbl.new_dscp_en )
                            {
                                action_tbl.new_dscp_en = 1;
                                action_tbl.new_dscp = val[0];
                            }
                        }
                        else if ( (len == 8 || (len > 8 && (p2[8] <= ' ' || p2[8] == ','))) && ifx_strincmp(p2, "original", 8) == 0 )
                        {
                            p2 += 8;
                            len -= 8;
                            //  new DSCP original
//                          printk("new DSCP: original\n");
                            //  the reset value is 0, so don't need to do anything
                        }
                    }
                    state = prev_cmd = 0;
                    break;
                case 11:
                    if ( !state )
                    {
                        state = 1;
                        prev_cmd = 11;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 12:
                    if ( state == 1 )
                    {
                        if ( prev_cmd == 11 )
                        {
                            //  MTU index
                            ifx_ignore_space(&p2, &len);
                            val[0] = ifx_get_number(&p2, &len, 0);
//                          printk("MTU index: %d\n", val[0]);
                            action_tbl.mtu_ix = val[0];
                        }
                        else if ( prev_cmd == 13 )
                        {
                            //  VLAN insert enable
                            //  VLAN index
                            ifx_ignore_space(&p2, &len);
                            val[0] = ifx_get_number(&p2, &len, 0);
//                          printk("VLAN insert: enable, index %d\n", val[0]);
                            if ( !action_tbl.in_vlan_ins )
                            {
                                action_tbl.in_vlan_ins = 1;
                                action_tbl.new_in_vci = val[0];
                            }
                        }
                        else if ( prev_cmd == 17 )
                        {
                            //  PPPoE index
                            ifx_ignore_space(&p2, &len);
                            val[0] = ifx_get_number(&p2, &len, 0);
//                          printk("PPPoE index: %d\n", val[0]);
                            action_tbl.pppoe_ix = val[0];
                        }
                    }
                    else if ( state == 3 && prev_cmd == 7 )
                    {
                        //  new src mac index
                        ifx_ignore_space(&p2, &len);
                        val[0] = ifx_get_number(&p2, &len, 0);
                        action_tbl.new_src_mac_ix = val[0];
                    }
                    else if ( state == 2 && prev_cmd == 13 )
                    {
                        //  outer VLAN enable
                        //  outer VLAN index
                        ifx_ignore_space(&p2, &len);
                        val[0] = ifx_get_number(&p2, &len, 0);
//                      printk("outer VLAN insert: enable, index %d\n", val[0]);
                        if ( !action_tbl.out_vlan_ins )
                        {
                            action_tbl.out_vlan_ins = 1;
                            action_tbl.out_vlan_ix = val[0];
                        }
                    }
                    state = prev_cmd = 0;
                    break;
                case 13:
                    if ( !state )
                    {
                        state = 1;
                        prev_cmd = 13;
                    }
                    else if ( state == 1 && prev_cmd == 37 )
                    {
                        state = 2;
                        prev_cmd = 13;
                        printk("outer vlan\n");
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 14:
                    if ( (state == 1 || state == 2) && prev_cmd == 13 )
                    {
                        state++;
                        prev_cmd = 14;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 15:
                    if ( (state == 1 || state == 2) && prev_cmd == 13 )
                    {
                        state++;
                        prev_cmd = 15;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 16:
                    if ( state == 1 && prev_cmd == 6 )
                    {
                        mask = 0;
                        do
                        {
                            ifx_ignore_space(&p2, &len);
                            if ( !len )
                                break;
                            for ( j = 0, bit = 1; j < sizeof(dest_list) / sizeof(*dest_list); j++, bit <<= 1 )
                                if ( (len == dest_list_strlen[j] || (len > dest_list_strlen[j] && (p2[dest_list_strlen[j]] <= ' ' || p2[dest_list_strlen[j]] == ','))) && ifx_strincmp(p2, dest_list[j], dest_list_strlen[j]) == 0 )
                                {
                                    p2 += dest_list_strlen[j];
                                    len -= dest_list_strlen[j];
                                    mask |= bit;
                                    break;
                                }
                        } while ( j < sizeof(dest_list) / sizeof(*dest_list) );
//                      if ( mask )
//                      {
//                          //  dest list
//                          printk("dest list:");
//                          for ( j = 0, bit = 1; j < sizeof(dest_list) / sizeof(*dest_list); j++, bit <<= 1 )
//                              if ( (mask & bit) )
//                              {
//                                  printk(" %s", dest_list[j]);
//                              }
//                          printk("\n");
//                      }
//                      else
//                          printk("dest list: none\n");
                        action_tbl.dest_list = mask;
                    }
                    state = prev_cmd = 0;
                    break;
                case 17:
                    if ( !state )
                    {
                        state = 1;
                        prev_cmd = 17;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 18:
                    if ( state == 1 && prev_cmd == 17 )
                    {
                        state = 2;
                        prev_cmd = 18;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 19:
                    if ( state == 1 && prev_cmd == 6 )
                    {
                        state = 2;
                        prev_cmd = 19;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 20:
                    if ( state == 2 && prev_cmd == 19 )
                    {
                        //  dest ch id
                        ifx_ignore_space(&p2, &len);
                        val[0] = ifx_get_number(&p2, &len, 0);
//                      printk("dest ch id: %d\n", val[0]);
                        //action_tbl.dest_chid = val[0];
                    }
                    state = prev_cmd = 0;
                    break;
                case 22:
                case 23:
                    if ( state == 2 )
                    {
                        if ( prev_cmd == 14 )
                        {
                            //  VLAN insert
//                          printk("VLAN insert: %s (%d)", command[i], i - 22);
                            if ( (i - 22) )
                            {
                                ifx_ignore_space(&p2, &len);
                                if ( len > 5 && (p2[5] <= ' ' || p2[5] == ':') && ifx_strincmp(p2, "index", 5) == 0 )
                                {
                                    p2 += 6;
                                    len -= 6;
                                    //  VLAN index
                                    ifx_ignore_space(&p2, &len);
                                    val[0] = ifx_get_number(&p2, &len, 0);
//                                  printk(", index %d", val[0]);
                                    if ( !action_tbl.in_vlan_ins )
                                    {
                                        action_tbl.in_vlan_ins = 1;
                                        //action_tbl.vlan_ix = val[0];
                                        action_tbl.new_in_vci = val[0];
                                    }
                                }
                                else if ( len > 3 && (p2[3] <= ' ' || p2[3] == ':') && ifx_strincmp(p2, "vci", 3) == 0 )
                                {
                                    p2 += 4;
                                    len -= 4;
                                    //  New VCI
                                    ifx_ignore_space(&p2, &len);
                                    val[0] = ifx_get_number(&p2, &len, 1);
//                                  printk(", vci 0x%04X", val[0]);
                                    if ( !action_tbl.in_vlan_ins )
                                    {
                                        action_tbl.in_vlan_ins = 1;
                                        action_tbl.new_in_vci = val[0];
                                    }
                                }
                            }
                            else
                            {
                                action_tbl.in_vlan_ins = 0;
                                action_tbl.new_in_vci = 0;
                            }
//                          printk("\n");
                        }
                        else if ( prev_cmd == 15 )
                        {
                            //  VLAN remove
//                          printk("VLAN remove: %s (%d)\n", command[i], i - 22);
                            action_tbl.in_vlan_rm = i - 22;
                        }
                    }
                    else if ( state == 3 )
                    {
                        if ( prev_cmd == 14 )
                        {
                            //  outer vlan insert
//                          printk("outer VLAN insert: %s (%d)", command[i], i - 22);
                            if ( (i - 22) )
                            {
                                ifx_ignore_space(&p2, &len);
                                if ( len > 5 && (p2[5] <= ' ' || p2[5] == ':') && ifx_strincmp(p2, "index", 5) == 0 )
                                {
                                    p2 += 6;
                                    len -= 6;
                                    //  outer VLAN index
                                    ifx_ignore_space(&p2, &len);
                                    val[0] = ifx_get_number(&p2, &len, 0);
//                                  printk(", index %d", val[0]);
                                    if ( !action_tbl.out_vlan_ins )
                                    {
                                        action_tbl.out_vlan_ins = 1;
                                        action_tbl.out_vlan_ix = val[0];
                                    }
                                }
                            }
                            else
                            {
                                action_tbl.out_vlan_ins = 0;
                                action_tbl.out_vlan_ix = 0;
                            }
//                          printk("\n");
                        }
                        else if ( prev_cmd == 15 )
                        {
                            //  outer VLAN remove
//                          printk("outer VLAN remove: %s (%d)\n", command[i], i - 22);
                            action_tbl.out_vlan_rm = i - 22;
                        }
                    }
                    state = prev_cmd = 0;
                    break;
                case 24:
                case 25:
                    if ( state == 2 && prev_cmd == 18 )
                    {
                        //  PPPoE mode
//                      printk("PPPoE mode: %s (%d)\n", command[i], i - 24);
                        action_tbl.pppoe_mode = i - 24;
                    }
                    state = prev_cmd = 0;
                    break;
                case 38:
                case 39:
                    i = 27;
                case 26:
                case 27:
                case 28:
                case 29:
                    if ( state == 2 && prev_cmd == 9 )
                    {
                        //  route type
//                      printk("route type: %s (%d)\n", command[i], i - 26);
                        action_tbl.rout_type = i - 26;
                    }
                    state = prev_cmd = 0;
                    break;
                case 30:
                    if ( !state )
                    {
                        if ( entry < 0 )
                        {
                            ifx_ignore_space(&p2, &len);
                            if ( len && *p2 >= '0' && *p2 <= '9' )
                            {
                                entry = ifx_get_number(&p2, &len, 0);
                                //  entry
//                              printk("entry: %d\n", entry);
                            }
                        }
                    }
                    state = prev_cmd = 0;
                    break;
                case 31:  // if this flag is not presented, it's UDP by default
                    if ( !state )
                    {
                        state = 1;
                        prev_cmd = 31;
                        action_tbl.protocol = 1;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                case 32:
                    printk("add\n");
                    printk("  LAN/WAN entry ???\n");
                    printk("    compare\n");
                    printk("      src:  ???.???.???.???:????\n");
                    printk("      dest: ???.???.???.???:????\n");
                    printk("    action\n");
                    printk("      new src/dest:      ???.???.???.???:????\n");
                    printk("      new MAC:           ??:??:??:??:??:?? (HEX)\n");
                    printk("      route type:        NULL/IP/NAT/NAPT\n");
                    printk("      new DSCP:          original/??\n");
                    printk("      MTU index:         ??\n");
                    printk("      VLAN insert:       disable/enable, VCI ???? (HEX)\n");
                    printk("      VLAN remove:       disable/enable\n");
                    printk("      dest list:         ETH0/ETH1/CPU0/EXT_INT?\n");
                    printk("      PPPoE mode:        transparent/termination\n");
                    printk("      PPPoE index:       ??\n");
                    printk("      new src MAC index: ??\n");
                    printk("      outer VLAN insert: disable/enable, index ??\n");
                    printk("      outer VLAN remove: disable/enable\n");
                    printk("      tcp:               yes/no\n");
                    printk("      dest qid:          ??\n");
                    printk("\n");
                    printk("delete\n");
                    printk("  LAN/WAN entry ???\n");
                    printk("    compare\n");
                    printk("      src:  ???.???.???.???:????\n");
                    printk("      dest: ???.???.???.???:????\n");
                    printk("\n");
                    state = prev_cmd = 0;
                    break;
                case 33:
                    if ( state == 1 && prev_cmd == 13 )
                    {
                        //  vlan vci
                        ifx_ignore_space(&p2, &len);
                        val[0] = ifx_get_number(&p2, &len, 1);
                        if ( !action_tbl.in_vlan_ins )
                        {
                            action_tbl.in_vlan_ins = 1;
                            action_tbl.new_in_vci = val[0];
                        }
                    }
                    state = prev_cmd = 0;
                    break;
                case 34:
                    if ( state == 1 && prev_cmd == 31 )
                        //  tcp yes
                        action_tbl.protocol = 1;
                    state = prev_cmd = 0;
                    break;
                case 35:
                    if ( state == 1 && prev_cmd == 31 )
                        //  tcp no
                        action_tbl.protocol = 0;
                    state = prev_cmd = 0;
                    break;
                case 36:
                    if ( state == 1 && prev_cmd == 6 )
                    {
                        //  dest qid
                        ifx_ignore_space(&p2, &len);
                        val[0] = ifx_get_number(&p2, &len, 0);
                        action_tbl.dest_qid = val[0];
                    }
                    state = prev_cmd = 0;
                    break;
                case 37:
                    if ( !state )
                    {
                        state = 1;
                        prev_cmd = 37;
                    }
                    else
                        state = prev_cmd = 0;
                    break;
                default:
                    state = prev_cmd = 0;
                }

                break;
            }

        if ( i == sizeof(command) / sizeof(*command) )
            state = prev_cmd = 0;

        p1 = p2;
        colon = 1;
    }

    if ( operation == 2 )
    {
        u32 is_lan = 0;

        //  delete
        pu1 = (u32*)&compare_tbl;
        pu2 = NULL;
        pu3 = NULL;
        if ( entry < 0 )
        {
            //  search the entry number
            if ( *pu1 && pu1[1] )
            {
                if ( (!type || type == 1) )
                {
                    //  LAN
                    for ( entry = 0; entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK; entry++ )
                    {
                        pu2 = (u32*)ROUT_LAN_HASH_CMP_TBL(entry);
                        if ( *pu2 == *pu1 && pu2[1] == pu1[1] && pu2[2] == pu1[2] )
                        {
                            pu3 = (u32*)ROUT_LAN_HASH_ACT_TBL(entry);
                            break;
                        }
                    }
                    if ( entry == MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK )
                    {
                        for ( entry = 0; entry < LAN_ROUT_TBL_CFG->rout_num; entry++ )
                        {
                            pu2 = (u32*)ROUT_LAN_COLL_CMP_TBL(entry);
                            if ( *pu2 == *pu1 && pu2[1] == pu1[1] && pu2[2] == pu1[2] )
                            {
                                pu3 = (u32*)ROUT_LAN_COLL_ACT_TBL(entry);
                                break;
                            }
                        }
                        if ( entry == LAN_ROUT_TBL_CFG->rout_num )
                            pu2 = NULL;
                        else
                            entry += MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK;
                    }
                    if ( pu3 != NULL )
                        is_lan = 1;
                }
                if ( (!type && !pu2) || type == 2 )
                {
                    //  WAN
                    for ( entry = 0; entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK; entry++ )
                    {
                        pu2 = (u32*)ROUT_WAN_HASH_CMP_TBL(entry);
                        if ( *pu2 == *pu1 && pu2[1] == pu1[1] && pu2[2] == pu1[2] )
                        {
                            pu3 = (u32*)ROUT_WAN_HASH_ACT_TBL(entry);
                            break;
                        }
                    }
                    if ( entry == MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK )
                    {
                        for ( entry = 0; entry < WAN_ROUT_TBL_CFG->rout_num; entry++ )
                        {
                            pu2 = (u32*)ROUT_WAN_COLL_CMP_TBL(entry);
                            if ( *pu2 == *pu1 && pu2[1] == pu1[1] && pu2[2] == pu1[2] )
                            {
                                pu3 = (u32*)ROUT_WAN_COLL_ACT_TBL(entry);
                                break;
                            }
                        }
                        if ( entry == WAN_ROUT_TBL_CFG->rout_num )
                            pu2 = NULL;
                        else
                            entry += MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK;
                    }
                    if ( pu3 != NULL )
                        is_lan = 0;
                }
            }
        }
        else
        {
            if ( *pu1 && pu1[1] )
            {
                pu3 = NULL;
                //  check compare
                if ( !type || type == 1 )
                {
                    //  LAN
                    if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK )
                    {
                        pu2 = (u32*)ROUT_LAN_HASH_CMP_TBL(entry);
                        if ( *pu2 != *pu1 || pu2[1] != pu1[1] || pu2[2] != pu1[2] )
                            pu2 = NULL;
                        else
                            pu3 = (u32*)ROUT_LAN_HASH_ACT_TBL(entry);
                    }
                    else if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK + LAN_ROUT_TBL_CFG->rout_num )
                    {
                        pu2 = (u32*)ROUT_LAN_COLL_CMP_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                        if ( *pu2 != *pu1 || pu2[1] != pu1[1] || pu2[2] != pu1[2] )
                            pu2 = NULL;
                        else
                            pu3 = (u32*)ROUT_LAN_COLL_ACT_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                    }
                    if ( pu3 != NULL )
                        is_lan = 1;
                }
                if ( !type || type == 2 )
                {
                    //  WAN
                    if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK )
                    {
                        pu2 = (u32*)ROUT_WAN_HASH_CMP_TBL(entry);
                        if ( *pu2 != *pu1 || pu2[1] != pu1[1] || pu2[2] != pu1[2] )
                            pu2 = NULL;
                        else
                            pu3 = (u32*)ROUT_WAN_HASH_ACT_TBL(entry);
                    }
                    else if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK + WAN_ROUT_TBL_CFG->rout_num )
                    {
                        pu2 = (u32*)ROUT_WAN_COLL_CMP_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                        if ( *pu2 != *pu1 || pu2[1] != pu1[1] || pu2[2] != pu1[2] )
                            pu2 = NULL;
                        else
                            pu3 = (u32*)ROUT_WAN_COLL_ACT_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                    }
                    if ( pu3 != NULL )
                        is_lan = 0;
                }
            }
            else if ( !*pu1 && !pu1[1] )
            {
                if ( type == 1 )
                {
                    if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK )
                    {
                        pu2 = (u32*)ROUT_LAN_HASH_CMP_TBL(entry);
                        pu3 = (u32*)ROUT_LAN_HASH_ACT_TBL(entry);
                    }
                    else if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK + LAN_ROUT_TBL_CFG->rout_num )
                    {
                        pu2 = (u32*)ROUT_LAN_COLL_CMP_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                        pu3 = (u32*)ROUT_LAN_COLL_ACT_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                    }
                    if ( pu3 != NULL )
                        is_lan = 1;
                }
                else if ( type == 2 )
                {
                    if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK )
                    {
                        pu2 = (u32*)ROUT_WAN_HASH_CMP_TBL(entry);
                        pu3 = (u32*)ROUT_WAN_HASH_ACT_TBL(entry);
                    }
                    else if ( entry < MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK + WAN_ROUT_TBL_CFG->rout_num )
                    {
                        pu2 = (u32*)ROUT_WAN_COLL_CMP_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                        pu3 = (u32*)ROUT_WAN_COLL_ACT_TBL(entry - MAX_ROUTING_ENTRIES_PER_HASH_BLOCK * MAX_HASH_BLOCK);
                    }
                    if ( pu3 != NULL )
                        is_lan = 0;
                }
            }
        }
        if ( pu2 && pu3 )
        {
            is_lan <<= 31;

            if ( ifx_ppa_drv_hal_generic_hook == NULL )
            {
                printk("HAL is not loaded - verify only, no firmware table operation\n");
                printk("del_routing_entry(%d - %08x)\n", entry, (u32)entry | is_lan);
            }
            else
            {
                PPE_ROUTING_INFO route = {0};
                int ret;

                route.entry = (u32)entry | is_lan;
                ret = ifx_ppa_drv_hal_generic_hook(PPA_GENERIC_HAL_DEL_ROUTE_ENTRY, &route, 0);
                printk("%s entry %d deleted %s\n", is_lan ? "LAN" : "WAN", entry, ret == IFX_SUCCESS ? "successfully" : "fail");
            }
        }
    }
    else if ( operation == 1 && type && ((!is_ipv6 && *(u32*)&compare_tbl && *((u32*)&compare_tbl + 1)) || (is_ipv6 && *((u32*)&compare_tbl + 2))) )
    {
        if ( ifx_ppa_drv_hal_generic_hook == NULL )
        {
            char str_to_print[1024];

            printk("HAL is not loaded - verify only, no firmware table operation\n");

            if ( is_ipv6 )
            {
                printk("ipv6 src ip:  %d.%d.%d.%d", ipv6_src_ip[0] >> 24, (ipv6_src_ip[0] >> 16) & 0xFF, (ipv6_src_ip[0] >> 8) & 0xFF, ipv6_src_ip[0] & 0xFF);
                printk(".%d.%d.%d.%d", ipv6_src_ip[1] >> 24, (ipv6_src_ip[1] >> 16) & 0xFF, (ipv6_src_ip[1] >> 8) & 0xFF, ipv6_src_ip[1] & 0xFF);
                printk(".%d.%d.%d.%d", ipv6_src_ip[2] >> 24, (ipv6_src_ip[2] >> 16) & 0xFF, (ipv6_src_ip[2] >> 8) & 0xFF, ipv6_src_ip[2] & 0xFF);
                printk(".%d.%d.%d.%d", ipv6_src_ip[3] >> 24, (ipv6_src_ip[3] >> 16) & 0xFF, (ipv6_src_ip[3] >> 8) & 0xFF, ipv6_src_ip[3] & 0xFF);
                printk("\n");
                printk("ipv6 dest ip: %d.%d.%d.%d", ipv6_dst_ip[0] >> 24, (ipv6_dst_ip[0] >> 16) & 0xFF, (ipv6_dst_ip[0] >> 8) & 0xFF, ipv6_dst_ip[0] & 0xFF);
                printk(".%d.%d.%d.%d", ipv6_dst_ip[1] >> 24, (ipv6_dst_ip[1] >> 16) & 0xFF, (ipv6_dst_ip[1] >> 8) & 0xFF, ipv6_dst_ip[1] & 0xFF);
                printk(".%d.%d.%d.%d", ipv6_dst_ip[2] >> 24, (ipv6_dst_ip[2] >> 16) & 0xFF, (ipv6_dst_ip[2] >> 8) & 0xFF, ipv6_dst_ip[2] & 0xFF);
                printk(".%d.%d.%d.%d", ipv6_dst_ip[3] >> 24, (ipv6_dst_ip[3] >> 16) & 0xFF, (ipv6_dst_ip[3] >> 8) & 0xFF, ipv6_dst_ip[3] & 0xFF);
                printk("\n");
            }
            print_route(str_to_print, 0, type == 1 ? 1 : 0, &compare_tbl, &action_tbl);
            printk(str_to_print);
        }
        else
        {
            PPE_ROUTING_INFO route = {0};
            u8 test_mac[PPA_ETH_ALEN] = {0};

            test_mac[0] = (action_tbl.new_dest_mac54 >> 8) & 0xFF;
            test_mac[1] = action_tbl.new_dest_mac54 & 0xFF;
            test_mac[2] = (action_tbl.new_dest_mac30 >> 24) & 0xFF;
            test_mac[3] = (action_tbl.new_dest_mac30 >> 16) & 0xFF;
            test_mac[4] = (action_tbl.new_dest_mac30 >> 8) & 0xFF;
            test_mac[5] = action_tbl.new_dest_mac30 & 0xFF;

            route.f_is_lan = type == 1 ? 1 : 0;
            if ( is_ipv6 )
            {
                route.src_ip.f_ipv6 = 1;
                memcpy(route.src_ip.ip.ip6, ipv6_src_ip, 16);
                route.dst_ip.f_ipv6 = 1;
                memcpy(route.dst_ip.ip.ip6, ipv6_dst_ip, 16);
            }
            else
            {
                route.src_ip.f_ipv6 = 0;
                route.src_ip.ip.ip = compare_tbl.src_ip;
                route.dst_ip.f_ipv6 = 0;
                route.dst_ip.ip.ip = compare_tbl.dest_ip;
                route.new_ip.f_ipv6 = 0;
                route.new_ip.ip.ip = action_tbl.new_ip;
            }
            route.src_port              = compare_tbl.src_port;
            route.dst_port              = compare_tbl.dest_port;
            route.new_port              = action_tbl.new_port;
            route.f_is_tcp              = action_tbl.protocol;
            route.route_type            = action_tbl.rout_type;
            memcpy(route.new_dst_mac, test_mac, 6);
            route.src_mac.mac_ix        = action_tbl.new_src_mac_ix;
            route.mtu_info.mtu_ix       = action_tbl.mtu_ix;
            route.pppoe_mode            = action_tbl.pppoe_mode;
            route.pppoe_info.pppoe_ix   = action_tbl.pppoe_ix;
            route.f_new_dscp_enable     = action_tbl.new_dscp_en;
            route.new_dscp              = action_tbl.new_dscp;
            route.f_vlan_ins_enable     = action_tbl.in_vlan_ins;
            route.new_vci               = action_tbl.new_in_vci;
            route.f_vlan_rm_enable      = action_tbl.in_vlan_rm;
            route.f_out_vlan_ins_enable = action_tbl.out_vlan_ins;
            route.out_vlan_info.vlan_entry = action_tbl.out_vlan_ix;
            route.f_out_vlan_rm_enable  = action_tbl.out_vlan_rm;
            route.dslwan_qid            = action_tbl.dest_qid;
            route.dest_list             = action_tbl.dest_list;

            if ( ifx_ppa_drv_hal_generic_hook(PPA_GENERIC_HAL_ADD_ROUTE_ENTRY, &route, 0) == IFX_SUCCESS )
                printk("%s entry add successfully: entry = %d\n", (route.entry & (1 << 31)) ? "LAN" : "WAN", route.entry & ~(1 << 31));
            else
                printk("%s entry add fail\n", type == 1 ? "LAN" : "WAN");
        }
    }
    else
        printk("No operation: operation = %d, type = %d, is_ipv6 = %d\n", operation, type, is_ipv6);

    return count;
}
#endif

static int proc_read_mc(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    volatile struct wan_rout_multicast_cmp_tbl *pcompare;
    volatile struct wan_rout_multicast_act_tbl *paction;
    int len = 0;
    char *pstr;
    int llen;
    int i;

    pstr = *start = page;

    __sync();

    llen = sprintf(common_proc_read_buff, "Multicast Table\n");
    proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen);

    pcompare = ROUT_WAN_MC_CMP_TBL(0);
    paction = ROUT_WAN_MC_ACT_TBL(0);
    for ( i = 0; i < MAX_WAN_MC_ENTRIES; i++ )
    {
        if ( pcompare->wan_dest_ip )
        {
            llen = print_mc(common_proc_read_buff, i, (struct wan_rout_multicast_cmp_tbl *)pcompare, (struct wan_rout_multicast_act_tbl *)paction);
            if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                goto PROC_READ_MC_OVERRUN_END;
        }

        pcompare++;
        paction++;
    }

    *eof = 1;

PROC_READ_MC_OVERRUN_END:
    return len - off;
}

static int proc_read_genconf(char *page, char **start, off_t off, int count __attribute__((unused)), int *eof, void *data __attribute__((unused)))
{
    char *str;
    int llen = 0;
    int i;
    unsigned long bit;

    str = *start = page;

    __sync();

    llen += sprintf(str + llen, "CFG_WRX_HTUTS    (0x%08X): %d\n", (u32)CFG_WRX_HTUTS, *CFG_WRX_HTUTS);
    llen += sprintf(str + llen, "CFG_WRDES_DELAY  (0x%08X): %d\n", (u32)CFG_WRDES_DELAY, *CFG_WRDES_DELAY);
    llen += sprintf(str + llen, "WRX_EMACH_ON     (0x%08X): AAL - %s, OAM - %s\n", (u32)WRX_EMACH_ON, (*WRX_EMACH_ON & 0x01) ? "on" : "off",
                                                                                                      (*WRX_EMACH_ON & 0x02) ? "on" : "off");
    llen += sprintf(str + llen, "WTX_EMACH_ON     (0x%08X):", (u32)WTX_EMACH_ON);
    for ( i = 0, bit = 1; i < 15; i++, bit <<= 1 )
    {
        llen += sprintf(str + llen, " %d - %s", i, (*WTX_EMACH_ON & bit) ? "on " : "off");
        if ( i == 3 || i == 7 || i == 11 )
            llen += sprintf(str + llen, "\n                               ");
    }
    llen += sprintf(str + llen, "\n");
    llen += sprintf(str + llen, "WRX_HUNT_BITTH   (0x%08X): %d\n", (u32)WRX_HUNT_BITTH, *WRX_HUNT_BITTH);

    llen += sprintf(str + llen, "ETH_PORTS_CFG    (0x%08X): WAN VLAN ID Hi - %d, WAN VLAN ID Lo - %d\n", (u32)ETH_PORTS_CFG, ETH_PORTS_CFG->wan_vlanid_hi, ETH_PORTS_CFG->wan_vlanid_lo);

    llen += sprintf(str + llen, "LAN_ROUT_TBL_CFG (0x%08X): num - %d, TTL check - %s, no hit drop - %s\n",
                                                            (u32)LAN_ROUT_TBL_CFG,
                                                            LAN_ROUT_TBL_CFG->rout_num,
                                                            LAN_ROUT_TBL_CFG->ttl_en ? "on" : "off",
                                                            LAN_ROUT_TBL_CFG->rout_drop  ? "on" : "off"
                                                            );
    llen += sprintf(str + llen, "                               IP checksum - %s, TCP/UDP checksum - %s, checksum error drop - %s\n",
                                                            LAN_ROUT_TBL_CFG->ip_ver_en ? "on" : "off",
                                                            LAN_ROUT_TBL_CFG->tcpdup_ver_en ? "on" : "off",
                                                            LAN_ROUT_TBL_CFG->iptcpudperr_drop ? "on" : "off"
                                                            );

    llen += sprintf(str + llen, "WAN_ROUT_TBL_CFG (0x%08X): num - %d, TTL check - %s, no hit drop - %s\n",
                                                            (u32)WAN_ROUT_TBL_CFG,
                                                            WAN_ROUT_TBL_CFG->rout_num,
                                                            WAN_ROUT_TBL_CFG->ttl_en ? "on" : "off",
                                                            WAN_ROUT_TBL_CFG->rout_drop  ? "on" : "off"
                                                            );
    llen += sprintf(str + llen, "                               MC num - %d, MC drop - %s\n",
                                                            WAN_ROUT_TBL_CFG->wan_rout_mc_num,
                                                            WAN_ROUT_TBL_CFG->wan_rout_mc_drop ? "on" : "off"
                                                            );
    llen += sprintf(str + llen, "                               IP checksum - %s, TCP/UDP checksum - %s, checksum error drop - %s\n",
                                                            WAN_ROUT_TBL_CFG->ip_ver_en ? "on" : "off",
                                                            WAN_ROUT_TBL_CFG->tcpdup_ver_en ? "on" : "off",
                                                            WAN_ROUT_TBL_CFG->iptcpudperr_drop ? "on" : "off"
                                                            );

    llen += sprintf(str + llen, "GEN_MODE_CFG1    (0x%08X): App2 - %s, U/S - %s, U/S early discard - %s\n"
                                "                               classification table entry - %d, IPv6 route entry - %d\n"
                                "                               session based rate control - %s, IPv4 WAN ingress hash alg - %s\n"
                                "                               multiple field based classification and VLAN assignment for bridging - %s\n"
                                "                               D/S IPv4 multicast acceleration - %s, IPv6 acceleration - %s\n"
                                "                               WAN acceleration - %s, LAN acceleration - %s, switch isolation - %s\n",
                                                            (u32)GEN_MODE_CFG1,
                                                            GEN_MODE_CFG1->app2_indirect ? "indirect" : "direct",
                                                            GEN_MODE_CFG1->us_indirect ? "indirect" : "direct",
                                                            GEN_MODE_CFG1->us_early_discard_en ? "enable" : "disable",
                                                            GEN_MODE_CFG1->classification_num,
                                                            GEN_MODE_CFG1->ipv6_rout_num,
                                                            GEN_MODE_CFG1->session_ctl_en ? "enable" : "disable",
                                                            GEN_MODE_CFG1->wan_hash_alg ? "src IP" : "dest port",
                                                            GEN_MODE_CFG1->brg_class_en ? "enable" : "disable",
                                                            GEN_MODE_CFG1->ipv4_mc_acc_mode ? "IP/port pairs" : "dst IP only",
                                                            GEN_MODE_CFG1->ipv6_acc_en ? "enable" : "disable",
                                                            GEN_MODE_CFG1->wan_acc_en ? "enable" : "disable",
                                                            GEN_MODE_CFG1->lan_acc_en ? "enable" : "disable",
                                                            GEN_MODE_CFG1->sw_iso_mode ? "isolated" : "not isolated"
                                                            );
    llen += sprintf(str + llen, "WAN Interfaces:                %s \n", get_wanitf(1));
    llen += sprintf(str + llen, "LAN Interfaces:                %s \n", get_wanitf(0));
    llen += sprintf(str + llen, "GEN_MODE_CFG2    (0x%08X):", (u32)GEN_MODE_CFG2);
    for ( i = 0, bit = 1; i < 8; i++, bit <<= 1 )
    {
        llen += sprintf(str + llen, " %d - %s", i, (GEN_MODE_CFG2->itf_outer_vlan_vld & bit) ? "on " : "off");
        if ( i == 3 )
            llen += sprintf(str + llen, "\n                              ");
    }
    llen += sprintf(str + llen, "\n");

    llen += sprintf(str + llen, "TX_QOS_CFG       (0x%08X): time_tick - %d, overhd_bytes - %d, eth1_eg_qnum - %d\n"
                                "                               eth1_burst_chk - %s, eth1 - %s, rate shaping - %s, WFQ - %s\n",
                                                            (u32)TX_QOS_CFG,
                                                            TX_QOS_CFG->time_tick,
                                                            TX_QOS_CFG->overhd_bytes,
                                                            TX_QOS_CFG->eth1_eg_qnum,
                                                            TX_QOS_CFG->eth1_burst_chk ? "on" : "off",
                                                            TX_QOS_CFG->eth1_qss ? "FW QoS" : "HW QoS",
                                                            TX_QOS_CFG->shape_en ? "on" : "off",
                                                            TX_QOS_CFG->wfq_en ? "on" : "off");

    llen += sprintf(str + llen, "KEY_SEL_n        (0x%08X): %08x %08x %08x %08x\n",
                                                            (u32)KEY_SEL_n(0),
                                                            *KEY_SEL_n(0), *KEY_SEL_n(1), *KEY_SEL_n(2), *KEY_SEL_n(3));

    *eof = 1;

    return llen - off;
}

static int proc_write_genconf(struct file *file __attribute__((unused)), const char *buf, unsigned long count, void *data __attribute__((unused)))
{
//    char common_proc_read_buff[2048];
    char *p;
    int len, rlen;

    int f_wan_hi = 0;

    len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
    rlen = len - copy_from_user(common_proc_read_buff, buf, len);
    while ( rlen && common_proc_read_buff[rlen - 1] <= ' ' )
        rlen--;
    common_proc_read_buff[rlen] = 0;
    for ( p = common_proc_read_buff; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        return 0;
    }

    if ( ifx_strincmp(p, "wan hi ", 7) == 0 )
    {
        p += 7;
        rlen -= 7;
        f_wan_hi = 1;
    }
    else if ( ifx_strincmp(p, "wan high ", 9) == 0 )
    {
        p += 9;
        rlen -= 9;
        f_wan_hi = 1;
    }
    else if ( ifx_strincmp(p, "wan lo ", 7) == 0 )
    {
        p += 7;
        rlen -= 7;
        f_wan_hi = -1;
    }
    else if ( ifx_strincmp(p, "wan low ", 8) == 0 )
    {
        p += 8;
        rlen -= 8;
        f_wan_hi = -1;
    }
    else if ( ifx_strincmp(p, "eth0 type ", 10) == 0 )
    {
        p += 10;
        rlen -= 10;
        if ( ifx_stricmp(p, "lan") == 0 )
            ETH_PORTS_CFG->eth0_type = 0;
        else if ( ifx_stricmp(p, "wan") == 0 )
            ETH_PORTS_CFG->eth0_type = 1;
        else if ( ifx_stricmp(p, "mix") == 0 )
            ETH_PORTS_CFG->eth0_type = 2;
    }

    if ( f_wan_hi )
    {
        int num;

        num = ifx_get_number(&p, &rlen, 0);
        if ( f_wan_hi > 0 )
            ETH_PORTS_CFG->wan_vlanid_hi = num;
        else
            ETH_PORTS_CFG->wan_vlanid_lo = num;
    }

    return count;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/


static int proc_read_queue(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    static const char *mpoa_type_str[] = {"EoA w/o FCS", "EoA w FCS", "PPPoA", "IPoA"};

    int len = 0;
//    int len_max = off + count;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;
    struct wrx_queue_config rx;
    struct wtx_queue_config tx;
    char qmap_str[64];
    char qmap_flag;
    int qmap_str_len;
    int i, k;
    unsigned int bit;

    pstr = *start = page;

    __sync();

    llen = sprintf(common_proc_read_buff, "RX Queue Config (0x%08X):\n", (u32)WRX_QUEUE_CONFIG(0));
    if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
        goto PROC_READ_QUEUE_OVERRUN_END;

    for ( i = 0; i < 16; i++ )
    {
        rx = *WRX_QUEUE_CONFIG(i);
        llen  = sprintf(common_proc_read_buff       ,  "  %d: MPoA type - %s, MPoA mode - %s, IP version %d\n", i, mpoa_type_str[rx.mpoa_type], rx.mpoa_mode ? "LLC" : "VC mux", rx.ip_ver ? 6 : 4);
        llen += sprintf(common_proc_read_buff + llen,  "     Oversize - %d, Undersize - %d, Max Frame size - %d\n", rx.oversize, rx.undersize, rx.mfs);
        llen += sprintf(common_proc_read_buff + llen,  "     uu mask - 0x%02X, cpi mask - 0x%02X, uu exp - 0x%02X, cpi exp - 0x%02X\n", rx.uumask, rx.cpimask, rx.uuexp, rx.cpiexp);
        if ( rx.vlan_ins )
            llen += sprintf(common_proc_read_buff + llen, "     new_vlan = 0x%08X\n", rx.new_vlan);

        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_QUEUE_OVERRUN_END;
    }

    llen = sprintf(common_proc_read_buff, "TX Queue Config (0x%08X):\n", (u32)WTX_QUEUE_CONFIG(0));
    if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
        goto PROC_READ_QUEUE_OVERRUN_END;

    for ( i = 0; i < 15; i++ )
    {
        tx = *WTX_QUEUE_CONFIG(i);
        qmap_flag = 0;
        qmap_str_len = 0;
        for ( k = 0, bit = 1; k < 8; k++, bit <<= 1 )
            if ( (tx.same_vc_qmap & bit) )
            {
                if ( qmap_flag++ )
                    qmap_str_len += sprintf(qmap_str + qmap_str_len, ", ");
                qmap_str_len += sprintf(qmap_str + qmap_str_len, "%d", k);
            }
        llen  = sprintf(common_proc_read_buff       , "  %d: uu - 0x%02X, cpi - 0x%02X, same VC queue map - %s\n", i, tx.uu, tx.cpi, qmap_flag ? qmap_str : "null");
        llen += sprintf(common_proc_read_buff + llen,  "     bearer channel - %d, QSB ID - %d, MPoA mode - %s\n", tx.sbid, tx.qsb_vcid, tx.mpoa_mode ? "LLC" : "VC mux");
        llen += sprintf(common_proc_read_buff + llen,  "     ATM header - 0x%08X\n", tx.atm_header);

        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_QUEUE_OVERRUN_END;
    }

    *eof = 1;

PROC_READ_QUEUE_OVERRUN_END:
    return len - off;
}

static int proc_write_queue(struct file *file __attribute__((unused)), const char *buf __attribute__((unused)), unsigned long count, void *data)
{
    return count;
}

static int proc_read_pppoe(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    int len = 0;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;
    int i;

    pstr = *start = page;

    __sync();

    llen = sprintf(pstr, "PPPoE Table (0x%08X) - Session ID:\n", (u32)PPPOE_CFG_TBL(0));
    pstr += llen;
    len += llen;

    for ( i = 0; i < 8; i++ )
    {
        llen = sprintf(common_proc_read_buff, "  %d: %d\n", i, *PPPOE_CFG_TBL(i));
        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_PPPOE_OVERRUN_END;
    }

    *eof = 1;

PROC_READ_PPPOE_OVERRUN_END:
    return len - off;
}

static int proc_read_mtu(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    int len = 0;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;
    int i;

    pstr = *start = page;

    __sync();

    llen = sprintf(pstr, "MTU Table (0x%08X):\n", (u32)MTU_CFG_TBL(0));
    pstr += llen;
    len += llen;

    for ( i = 0; i < 8; i++ )
    {
        llen = sprintf(common_proc_read_buff, "  %d: %d\n", i, *MTU_CFG_TBL(i));
        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_MTU_OVERRUN_END;
    }

    *eof = 1;

PROC_READ_MTU_OVERRUN_END:
    return len - off;
}

static int proc_read_mpoa(char *page __attribute__((unused)), char **start __attribute__((unused)), off_t off, int count __attribute__((unused)), int *eof, void *data __attribute__((unused)))
{
	int len = 0;
    *eof = 1;

// PROC_READ_HIT_OVERRUN_END:
    return len - off;
}

static int proc_write_mpoa(struct file *file __attribute__((unused)), const char *buf __attribute__((unused)), unsigned long count, void *data __attribute__((unused)))
{
    int len ;
    int conn = -1;
    int mpoa_type = -1;
    int mpoa_mode = -1;

    len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
    common_proc_read_buff[len] = 0;

    sscanf(common_proc_read_buff, "%d %d %d", &conn, &mpoa_type, &mpoa_mode);
    printk(KERN_ERR "got con=%d type=%d mode=%d\n", conn, mpoa_type, mpoa_mode);
    if ( conn >= 0  && mpoa_type >= 0 && mpoa_mode >= 0)
    	mpoa_setup_conn(conn, mpoa_type, mpoa_mode);

    return count;
}

static int proc_read_hit(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    int len = 0;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;
    int i;
    int n;
    unsigned long bit;

    pstr = *start = page;

    __sync();

    llen = sprintf(pstr, "Unicast Routing Hit Table (0x%08X):\n", (u32)ROUT_LAN_HASH_HIT_STAT_TBL(0));
    pstr += llen;
    len += llen;

    llen = sprintf(pstr, "             1 2 3 4 5 6 7 8 9 10\n");
    pstr += llen;
    len += llen;

    n = 1;
    for ( i = 0; i < 8; i++ )
        for ( bit = 0x80000000; bit; bit >>= 1 )
        {
            if ( n % 10 == 1 )
                llen = sprintf(common_proc_read_buff, "  %3d - %3d:", n, n + 9);

            llen += sprintf(common_proc_read_buff + llen, " %d", (*ROUT_LAN_HASH_HIT_STAT_TBL(i) & bit) ? 1 : 0);

            if ( n++ % 10 == 0 )
            {
                llen += sprintf(common_proc_read_buff + llen, "\n");

                if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                    goto PROC_READ_HIT_OVERRUN_END;
            }
        }

    if ( n % 10 != 0 )
    {
        llen += sprintf(common_proc_read_buff + llen, "\n");

        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_HIT_OVERRUN_END;
    }

    llen = sprintf(common_proc_read_buff, "Multicast Routing Hit Table (0x%08X):\n", (u32)ROUT_WAN_MC_HIT_STAT_TBL(0));
    llen += sprintf(common_proc_read_buff + llen, "             1 2 3 4 5 6 7 8 9 10\n");
    if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
        goto PROC_READ_HIT_OVERRUN_END;

    n = 1;
    for ( i = 0; i < 1; i++ )
        for ( bit = 0x80000000; bit; bit >>= 1 )
        {
            if ( n % 10 == 1 )
                llen = sprintf(common_proc_read_buff, "  %3d - %3d:", n, n + 9);

            llen += sprintf(common_proc_read_buff + llen, " %d", (*ROUT_WAN_MC_HIT_STAT_TBL(i) & bit) ? 1 : 0);

            if ( n++ % 10 == 0 )
            {
                llen += sprintf(common_proc_read_buff + llen, "\n");

                if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                    goto PROC_READ_HIT_OVERRUN_END;
            }
        }

    if ( n % 10 != 0 )
    {
        llen += sprintf(common_proc_read_buff + llen, "\n");

        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_HIT_OVERRUN_END;
    }

    *eof = 1;

PROC_READ_HIT_OVERRUN_END:
    return len - off;
}

static int proc_write_hit(struct file *file __attribute__((unused)), const char *buf, unsigned long count, void *data __attribute__((unused)))
{
//    char common_proc_read_buff[2048];
    char *p;
    int len, rlen;

    len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
    rlen = len - copy_from_user(common_proc_read_buff, buf, len);
    while ( rlen && common_proc_read_buff[rlen - 1] <= ' ' )
        rlen--;
    common_proc_read_buff[rlen] = 0;
    for ( p = common_proc_read_buff; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        return 0;
    }

    if ( ifx_stricmp(p, "clear") == 0 || ifx_stricmp(p, "clean") == 0 )
    {
        ppa_memset((void*)ROUT_LAN_HASH_HIT_STAT_TBL(0), 0, 64);
        ppa_memset((void*)ROUT_LAN_COLL_HIT_STAT_TBL(0), 0, 2);
        ppa_memset((void*)ROUT_WAN_HASH_HIT_STAT_TBL(0), 0, 64);
        ppa_memset((void*)ROUT_WAN_COLL_HIT_STAT_TBL(0), 0, 2);
        ppa_memset((void*)ROUT_WAN_MC_HIT_STAT_TBL(0), 0, 2);
    }

    return count;
}

static int proc_read_mac(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    int len = 0;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;
    int i;
    unsigned int mac52, mac10;

    pstr = *start = page;

    __sync();

    llen = sprintf(pstr, "MAC Table:\n");
    pstr += llen;
    len += llen;

    llen = sprintf(pstr, "  ROUT_MAC_CFG_TBL (0x%08X)\n", (u32)ROUT_MAC_CFG_TBL(0));
    pstr += llen;
    len += llen;

    for ( i = 0; i < 16; i++ )
    {
        mac52 = *ROUT_MAC_CFG_TBL(i);
        mac10 = *(ROUT_MAC_CFG_TBL(i) + 1);

        llen = sprintf(common_proc_read_buff, "    %2d: %02X:%02X:%02X:%02X:%02X:%02X\n", i + 1, mac52 >> 24, (mac52 >> 16) & 0xFF, (mac52 >> 8) & 0xFF, mac52 & 0xFF, mac10 >> 24, (mac10 >> 16) & 0xFF);
        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_MAC_OVERRUN_END;
    }

    *eof = 1;

PROC_READ_MAC_OVERRUN_END:
    return len - off;
}

static int proc_read_out_vlan(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    int len = 0;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;
    int i;

    pstr = *start = page;

    __sync();

    llen = sprintf(pstr, "Outer VLAN Table (0x%08X):\n", (u32)OUTER_VLAN_TBL(0));
    pstr += llen;
    len += llen;

    for ( i = 0; i < 32; i++ )
    {
        llen = sprintf(common_proc_read_buff, "  %d: 0x%08X\n", i, *OUTER_VLAN_TBL(i));
        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
            goto PROC_READ_OUT_VLAN_OVERRUN_END;
    }

    *eof = 1;

PROC_READ_OUT_VLAN_OVERRUN_END:
    return len - off;
}

static int proc_read_ipv6_ip(char *page, char **start, off_t off, int count, int *eof, void *data __attribute__((unused)))
{
    int len = 0;
    char *pstr;
//    char common_proc_read_buff[2048];
    int llen;

    int i, j, x;
    volatile char *p;

    pstr = *start = page;

    __sync();

    llen = sprintf(common_proc_read_buff, "IPv6 IP Table:\n");
    proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen);

    for ( x = 0; x < MAX_IPV6_IP_ENTRIES_BLOCK; x++ )
        for ( i = 0; i < MAX_IPV6_IP_ENTRIES_PER_BLOCK; i++ )
            if ( IPv6_IP_IDX_TBL(x, i)[0] != 0 || IPv6_IP_IDX_TBL(x, i)[1] != 0 || IPv6_IP_IDX_TBL(x, i)[2] != 0 || IPv6_IP_IDX_TBL(x, i)[3] != 0 )
            {
                p = (volatile char *)IPv6_IP_IDX_TBL(x, i);
                llen = sprintf(common_proc_read_buff, "%3d - %u", x * MAX_IPV6_IP_ENTRIES_PER_BLOCK + i, (unsigned int)p[0]);
                for ( j = 1; j < 16; j++ )
                    llen += sprintf(common_proc_read_buff + llen, ".%u", (unsigned int)p[j]);
                llen += sprintf(common_proc_read_buff + llen, "\n");

                if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
                    goto PROC_READ_IPv6_IP_OVERRUN_END;
            }

    *eof = 1;

PROC_READ_IPv6_IP_OVERRUN_END:
    return len - off;
}

#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

#if defined(DEBUG_HTU_PROC) && DEBUG_HTU_PROC
static INLINE int print_htu(char *buf, int i) {
	int len = 0;

	if (HTU_ENTRY(i)->vld) {
		len += sprintf(buf + len, "%2d. valid\n", i);
		len += sprintf(buf + len,
				"    entry  0x%08x - pid %01x vpi %02x vci %04x pti %01x\n",
				*(u32*) HTU_ENTRY(i), HTU_ENTRY(i)->pid, HTU_ENTRY(i)->vpi,
				HTU_ENTRY(i)->vci, HTU_ENTRY(i)->pti);
		len += sprintf(buf + len,
				"    mask   0x%08x - pid %01x vpi %02x vci %04x pti %01x\n",
				*(u32*) HTU_MASK(i), HTU_MASK(i)->pid_mask,
				HTU_MASK(i)->vpi_mask, HTU_MASK(i)->vci_mask,
				HTU_MASK(i)->pti_mask);
		len += sprintf(buf + len, "    result 0x%08x - type: %s, qid: %d",
				*(u32*) HTU_RESULT(i), HTU_RESULT(i)->type ? "cell" : "AAL5",
				HTU_RESULT(i)->qid);
		if (HTU_RESULT(i)->type)
			len += sprintf(buf + len, ", cell id: %d, verification: %s",
					HTU_RESULT(i)->cellid, HTU_RESULT(i)->ven ? "on" : "off");
		len += sprintf(buf + len, "\n");
	} else
		len += sprintf(buf + len, "%2d. invalid\n", i);

	return len;
}

static int proc_read_htu(char *page, char **start, off_t off, int count,
		int *eof, void *data __attribute__((unused))) {
	int len = 0;
//	int len_max = off + count;
	char *pstr;
//	char common_proc_read_buff[1024];
	int llen;

	int htuts = *CFG_WRX_HTUTS;
	int i;

	pstr = *start = page;

	__sync();

	for (i = 0; i < htuts; i++) {
		llen = print_htu(common_proc_read_buff, i);
        if ( proc_buf_copy(&pstr, count, off, &len, common_proc_read_buff, llen) )
			goto PROC_READ_HTU_OVERRUN_END;
	}

	*eof = 1;

	return len - off;

PROC_READ_HTU_OVERRUN_END:
	return len -  off;
}
#endif


/*------------------------------------------------------------------------------------------*\
 * proc_read_dbg u. proc_write_dbg sind bei uns im common teil
\*------------------------------------------------------------------------------------------*/


#define PP32_REG_ADDR_BEGIN     0x0
#define PP32_REG_ADDR_END       0x1FFF
#define PP32_SB_ADDR_BEGIN      0x2000
#define PP32_SB_ADDR_END        0x4BFF

static inline unsigned long sb_addr_to_fpi_addr_convert(unsigned long sb_addr) {
	if (sb_addr <= PP32_SB_ADDR_END) {
		return (unsigned long) SB_BUFFER(sb_addr);
	} else {
		return sb_addr;
	}
}

#if defined(DEBUG_MEM_PROC) && DEBUG_MEM_PROC
static int proc_write_mem(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
	char *p1, *p2;
	int len;
	int colon;
	unsigned long *p;
    unsigned long dword;

//	char common_proc_read_buff[1024];
	int i, n, l;

	len = sizeof(common_proc_read_buff) < count ? sizeof(common_proc_read_buff) - 1 : count;
	len = len - copy_from_user(common_proc_read_buff, buf, len);
	common_proc_read_buff[len] = 0;

	p1 = common_proc_read_buff;
    p2 = NULL;
	colon = 1;
	while (ifx_get_token(&p1, &p2, &len, &colon)) {
		if (ifx_stricmp(p1, "w") == 0 || ifx_stricmp(p1, "write") == 0
				|| ifx_stricmp(p1, "r") == 0 || ifx_stricmp(p1, "read") == 0)
			break;

		p1 = p2;
		colon = 1;
	}

	if (*p1 == 'w') {
		ifx_ignore_space(&p2, &len);
		if (p2[0] == 's' && p2[1] == 'w' && (p2[2] == ' ' || p2[2] == '\t')) {
			unsigned long temp;

			p2 += 3;
			len -= 3;
			ifx_ignore_space(&p2, &len);
			temp = ifx_get_number(&p2, &len, 1);
			p = (unsigned long *) VR9_SWIP_MACRO_REG(temp);
		} else {
			p = (unsigned long *) ifx_get_number(&p2, &len, 1);
			p = (unsigned long *) sb_addr_to_fpi_addr_convert(
					(unsigned long) p);
		}

		if ((u32) p >= KSEG0)
			while (1) {
				ifx_ignore_space(&p2, &len);
				if (!len
						|| !((*p2 >= '0' && *p2 <= '9')
								|| (*p2 >= 'a' && *p2 <= 'f')
								|| (*p2 >= 'A' && *p2 <= 'F')))
					break;

				*p++ = (u32) ifx_get_number(&p2, &len, 1);
			}
	} else if (*p1 == 'r') {
		ifx_ignore_space(&p2, &len);
        if ( p2[0] == 's' && p2[1] == 'w' && (p2[2] == ' ' || p2[2] == '\t') )
        {
            unsigned long temp;

            p2 += 3;
            len -= 3;
            ifx_ignore_space(&p2, &len);
            temp = ifx_get_number(&p2, &len, 1);
            p = (unsigned long *)VR9_SWIP_MACRO_REG(temp);
        }
        else
        {
            p = (unsigned long *)ifx_get_number(&p2, &len, 1);
            p = (unsigned long *)sb_addr_to_fpi_addr_convert( (unsigned long) p);
        }
		if ((u32) p >= KSEG0) {
			ifx_ignore_space(&p2, &len);
			n = (int) ifx_get_number(&p2, &len, 0);
			if (n) {
				char str[32] = { 0 };
				char *pch = str;
				int k;
				char c;

				n += (l = ((int) p >> 2) & 0x03);
				p = (unsigned long *) ((u32) p & ~0x0F);
				for (i = 0; i < n; i++) {
					if ((i & 0x03) == 0) {
						printk("%08X:", (u32) p);
						pch = str;
					}
					if (i < l) {
						printk("         ");
						sprintf(pch, "    ");
					} else {
						dword = (u32) *p;
						printk(" %08X", (unsigned int)dword);
						for (k = 0; k < 4; k++) {
							c = ((char*) &dword)[k];
							pch[k] = c < ' ' ? '.' : c;
						}
					}
					p++;
					pch += 4;
					if ((i & 0x03) == 0x03) {
						pch[0] = 0;
						printk(" ; %s\n", str);
					}
				}
				if ((n & 0x03) != 0x00) {
					for (k = 4 - (n & 0x03); k > 0; k--)
						printk("         ");
					pch[0] = 0;
					printk(" ; %s\n", str);
				}
			}
		}
	}

	return count;
}
#endif

static int proc_read_burstlen(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)),
		int *eof, void *data __attribute__((unused))) {
	int len = 0;
	u32 dma_ps;
	u32 dma_pctrl;
	unsigned long sys_flag;

	local_irq_save(sys_flag);
	dma_ps = *(volatile u32 *) 0xBE104140;
	*(volatile u32 *) 0xBE104140 = 0;
	dma_pctrl = *(volatile u32 *) 0xBE104144;
	*(volatile u32 *) 0xBE104140 = dma_ps;
	local_irq_restore(sys_flag);

	len += sprintf(page + len, "DMA-PPE burst length: Rx %d, Tx %d\n",
			1 << ((dma_pctrl >> 2) & 0x03), 1 << ((dma_pctrl >> 4) & 0x03));

	*eof = 1;

	return len;
}

static int proc_write_burstlen(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
//	char common_proc_read_buff[2048];
	int len;
	char *p1, *p2;
	int colon = 0;
	int mask = 0x3C; //  rx: 0x0C, tx: 0x30, all: 0x3C
	int burstlen = 0;
	int burstlen_mask = 0;
	int f_help = 0;
	u32 dma_ps;
	unsigned long sys_flag;

	len = sizeof(common_proc_read_buff) < count ? sizeof(common_proc_read_buff) - 1 : count;
	len = len - copy_from_user(common_proc_read_buff, buf, len);
	common_proc_read_buff[len] = 0;

	p1 = common_proc_read_buff;
	while (ifx_get_token(&p1, &p2, &len, &colon)) {
		if (ifx_stricmp(p1, "rx") == 0)
			burstlen_mask |= (mask = 0x0C);
		else if (ifx_stricmp(p1, "tx") == 0)
			burstlen_mask |= (mask = 0x30);
		else if (ifx_stricmp(p1, "all") == 0)
			burstlen_mask |= (mask = 0x3C);
		else if (strcmp(p1, "2") == 0)
			burstlen = 0x14 & mask;
		else if (strcmp(p1, "4") == 0)
			burstlen = 0x28 & mask;
		else if (strcmp(p1, "8") == 0)
			burstlen = 0x3C & mask;
		else {
			f_help = 1;
			break;
		}

		p1 = p2;
	}

	if (!burstlen_mask && burstlen)
		burstlen_mask = 0x3C;

	if (!burstlen_mask || !burstlen)
		f_help = 1;

	if (!f_help) {
		local_irq_save(sys_flag);
		dma_ps = *(volatile u32 *) 0xBE104140;
		*(volatile u32 *) 0xBE104140 = 0;
		*(volatile u32 *) 0xBE104144 = (*(volatile u32 *) 0xBE104144
				& ~burstlen_mask) | burstlen;
		*(volatile u32 *) 0xBE104140 = dma_ps;
		local_irq_restore(sys_flag);
	} else {
		printk("echo [rx/tx/all] <2/4/8> > /proc/eth/burstlen\n");
	}

	return count;
}

#if defined(DEBUG_FW_PROC) && DEBUG_FW_PROC

static int proc_read_fw(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)), int *eof, void *data __attribute__((unused)))
{
    int len = 0;

//    u32 sll_state;
    u32 tir4;
    u32 tir4_sll_state;
    u32 tir4_dplus_rx_on;

    len += sprintf(page + len, "Firmware\n");

    tir4 = *TIR(4);
    tir4_sll_state = (tir4 >> 21) & 7;
    tir4_dplus_rx_on = (tir4 >> 20) & 1;

    len += sprintf(page + len, "\nQoS:\n");
    len += sprintf(page + len, "  DSLWAN_TX_SWAP_RDPTR = %04X, %04X, %04X, %04X, %04X, %04X, %04X, %04X\n %04X, %04X, %04X, %04X, %04X, %04X, %04X\n",
                                            * __DSLWAN_TXDES_SWAP_PTR(0), * __DSLWAN_TXDES_SWAP_PTR(1),
                                            * __DSLWAN_TXDES_SWAP_PTR(2), * __DSLWAN_TXDES_SWAP_PTR(3),
                                            * __DSLWAN_TXDES_SWAP_PTR(4), * __DSLWAN_TXDES_SWAP_PTR(5),
                                            * __DSLWAN_TXDES_SWAP_PTR(6), * __DSLWAN_TXDES_SWAP_PTR(7),
                                            * __DSLWAN_TXDES_SWAP_PTR(8), * __DSLWAN_TXDES_SWAP_PTR(9),
                                            * __DSLWAN_TXDES_SWAP_PTR(10), * __DSLWAN_TXDES_SWAP_PTR(11),
                                            * __DSLWAN_TXDES_SWAP_PTR(12), * __DSLWAN_TXDES_SWAP_PTR(13),
                                            * __DSLWAN_TXDES_SWAP_PTR(14));

    len += sprintf(page + len, "  DSLWAN_TX_PKT_CNT    = %04X, %04X, %04X, %04X, %04X, %04X, %04X, %04X\n %04X, %04X, %04X, %04X, %04X, %04X, %04X\n ",
                                            * __DSLWAN_TX_PKT_CNT(0), * __DSLWAN_TX_PKT_CNT(1),
                                            * __DSLWAN_TX_PKT_CNT(2), * __DSLWAN_TX_PKT_CNT(3),
                                            * __DSLWAN_TX_PKT_CNT(4), * __DSLWAN_TX_PKT_CNT(5),
                                            * __DSLWAN_TX_PKT_CNT(6), * __DSLWAN_TX_PKT_CNT(7),
                                            * __DSLWAN_TX_PKT_CNT(8), * __DSLWAN_TX_PKT_CNT(9),
                                            * __DSLWAN_TX_PKT_CNT(10), * __DSLWAN_TX_PKT_CNT(11),
                                            * __DSLWAN_TX_PKT_CNT(12), * __DSLWAN_TX_PKT_CNT(13),
                                            * __DSLWAN_TX_PKT_CNT(14));


    len += sprintf(page + len, "  ENETS PGNUM             = %d\n",  __ENETS_PGNUM);
    len += sprintf(page + len, "  ENETS DBUF BASE ADDR     = %08X\n",  (unsigned int)__ERX_DBUF_BASE);
    len += sprintf(page + len, "  ENETS CBUF BASE ADDR     = %08X\n",  (unsigned int)__ERX_CBUF_BASE);

    len += sprintf(page + len, "  QOSD_DPLUS_RDPTR     = %04X\n",
                                            * __DSLWAN_FP_RXDES_SWAP_RDPTR);

    len += sprintf(page + len, "  QOSD_CPUTX_RDPTR     = %04X\n",
                                            * __CPU_TXDES_SWAP_RDPTR);

    len += sprintf(page + len, "  DPLUS_RXDES_RDPTR    = %04X\n",
                                            * __DSLWAN_FP_RXDES_DPLUS_WRPTR);


    len += sprintf(page + len, "  pre_dplus_ptr    = %04X\n", *PRE_DPLUS_PTR);

    len += sprintf(page + len, "  pre_dplus_cnt    = %04X\n", (*DM_RXPKTCNT) & 0xff);

    len += sprintf(page + len, "  dplus_ptr        = %04X\n", *DPLUS_PTR);

    len += sprintf(page + len, "  dplus_cnt        = %04X\n", *DPLUS_CNT);

    len += sprintf(page + len, "  DPLUS_RX_ON      = %d\n",  *DPLUS_RX_ON);

    len += sprintf(page + len, "  ISR_IS           = %08X\n ",  *ISR_IS);

    len += sprintf(page + len, "\nQoS Mib:\n");
    len += sprintf(page + len, "  cputx_pkts:             %u\n",   *__CPU_TO_DSLWAN_TX_PKTS);
    len += sprintf(page + len, "  cputx_bytes:            %u\n",   *__CPU_TO_DSLWAN_TX_BYTES);
    len += sprintf(page + len, "  cputx_drop_pkts:        %u\n",   *__CPU_TX_SWAPPER_DROP_PKTS);
    len += sprintf(page + len, "  cputx_drop_bytess:      %u\n",   *__CPU_TX_SWAPPER_DROP_BYTES);

    len += sprintf(page + len, "  dslwan_fp_drop_pkts:    %u\n",   *__DSLWAN_FP_SWAPPER_DROP_PKTS );
    len += sprintf(page + len, "  dslwan_fp_drop_bytes:   %u\n",   *__DSLWAN_FP_SWAPPER_DROP_BYTES );

    len += sprintf(page + len, "  dslwan_tx_qf_drop_pkts:  (%u, %u, %u, %u, %u, %u, %u, %u ,%u, %u, %u, %u, %u, %u, %u)\n",
                 *__DSLWAN_TX_THRES_DROP_PKT_CNT(0),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(1),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(2) ,  *__DSLWAN_TX_THRES_DROP_PKT_CNT(3) ,
                 *__DSLWAN_TX_THRES_DROP_PKT_CNT(4),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(5),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(6) ,  *__DSLWAN_TX_THRES_DROP_PKT_CNT(7),
                 *__DSLWAN_TX_THRES_DROP_PKT_CNT(8),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(9),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(10),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(11),
                 *__DSLWAN_TX_THRES_DROP_PKT_CNT(12),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(13),  *__DSLWAN_TX_THRES_DROP_PKT_CNT(14));

    *eof = 1;

    return len;
}


static int proc_write_fw(struct file __attribute__((unused)) *file, const char *buf, unsigned long count, void *data __attribute__((unused)))
{
//    char common_proc_read_buff[2048];
    char *p;
    int len, rlen;
    int i;

    len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
    rlen = len - copy_from_user(common_proc_read_buff, buf, len);
    while ( rlen && common_proc_read_buff[rlen - 1] <= ' ' )
        rlen--;
    common_proc_read_buff[rlen] = 0;
    for ( p = common_proc_read_buff; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        return 0;
    }

    if ( ifx_stricmp(p, "clear") == 0 || ifx_stricmp(p, "clean") == 0 )
    {
       for( i=0;i<15;i++)
               * __DSLWAN_TX_THRES_DROP_PKT_CNT(i) = 0;

        * __CPU_TO_DSLWAN_TX_PKTS  = 0;
        * __CPU_TO_DSLWAN_TX_BYTES = 0;


        * __CPU_TX_SWAPPER_DROP_PKTS    = 0;
        * __CPU_TX_SWAPPER_DROP_BYTES   = 0;
        * __DSLWAN_FP_SWAPPER_DROP_PKTS = 0;
        * __DSLWAN_FP_SWAPPER_DROP_PKTS = 0;
    }

    return count;
}

static int INLINE is_char_in_set(char ch, char * set)
{

    while(*set) {
        if (ch == *set)
            return 1;
        set ++;
    }

    return 0;
}

static int tokenize(char * in, char * sep, char **tokens, int max_token_num, int * finished)
{
    int token_num = 0;

    * finished = 0;

    while (*in && token_num < max_token_num) {

        // escape all seperators
        while (*in && is_char_in_set(*in, sep)) {
            in ++;
        }

        // break if no more chars left
        if(! *in) {
            break;
        }

        // found a new token, remember start position
        tokens[token_num ++] = in;

        // search end of token
        in ++;
        while (*in && ! is_char_in_set(*in, sep)) {
            in ++;
        }

        if(! *in) {
            // break if no more chars left
            break;
        }
        else {
            // tokenize
            *in = '\0';
            in ++;
        }

    }

    if ( ! *in )
        * finished = 1;

    return token_num;
}

static void fwdbg_read_ethbuf(char **tokens, int token_num)
{
    unsigned int start_pg = 0;
    unsigned int print_pg_num;
    unsigned int num;
    unsigned int pg_size;
    volatile unsigned int *dbase, *cbase;
    unsigned int pnum, i;
    unsigned int *cw;


    dbase = (volatile unsigned int *)SB_BUFFER( ( ( struct dmrx_dba * )DM_RXDB)->dbase + 0x2000 );
    cbase = (volatile unsigned int *)SB_BUFFER( ( ( struct dmrx_cba * )DM_RXCB)->cbase + 0x2000);
    pnum = ( ( struct dmrx_cfg * )DM_RXCFG)->pnum;

    i = 0;
    start_pg = i < token_num  ? simple_strtoul(tokens[i ++ ], NULL, 0) : start_pg;
    print_pg_num = i < token_num ? simple_strtoul(tokens[i ++ ], NULL, 0) : pnum;

    start_pg %= pnum;
    print_pg_num = print_pg_num > pnum ? pnum : print_pg_num;

    pg_size =  ((( struct  dmrx_cfg * )DM_RXCFG)->psize + 1) * 32 / 4;


    printk("Share buffer data/ctrl buffer:\n\n");
    for(i = start_pg, num = 0 ; num < print_pg_num ; num ++, i = (i + 1) % pnum ) {

        struct ctrl_dmrx_2_fw * p_ctrl_dmrx;
        struct ctrl_fw_2_dsrx * p_ctrl_dsrx;
        int j = 0;

        cw = (unsigned int *)dbase + i * pg_size;

        p_ctrl_dmrx = (struct ctrl_dmrx_2_fw *) ( &cbase[i]);
        p_ctrl_dsrx = (struct ctrl_fw_2_dsrx *) ( &cbase[i]);


        for(j =0; j < pg_size; j=j+4) {

            if(j==0) {

                printk("Pg_id %2d: %08x %08x %08x %08x ", i, cw[0], cw[1], cw[2], cw[3]);
                printk("pg_val: %x, byte_off: %x, cos: %x \n", p_ctrl_dmrx->pg_val, p_ctrl_dmrx->byte_off, p_ctrl_dmrx->cos);
            }


            else if(j==4) {
                printk("          %08x %08x %08x %08x ",  cw[j], cw[j+1], cw[j+2], cw[j+3]);
                printk("bytes_cnt: %x, eof: %x \n", p_ctrl_dmrx->bytes_cnt, p_ctrl_dmrx->eof);
            }

            else if(j==12) {
                printk("          %08x %08x %08x %08x ",  cw[j], cw[j+1], cw[j+2], cw[j+3]);
                printk("pg_val: %x, byte_off: %x, acc_sel: %x\n", p_ctrl_dsrx->pg_val, p_ctrl_dsrx->byte_off,p_ctrl_dsrx->acc_sel );
            }

            else if(j==16) {
                printk("          %08x %08x %08x %08x ",  cw[j], cw[j+1], cw[j+2], cw[j+3]);
                printk("cos: %x, bytes_cnt: %x, eof: %x\n", p_ctrl_dsrx->cos, p_ctrl_dsrx->bytes_cnt,p_ctrl_dsrx->eof );
            }

            else {
                printk("          %08x %08x %08x %08x \n",  cw[j], cw[j+1], cw[j+2], cw[j+3]);
            }

        }

            printk("\n");

    }

    return;
}



static void fwdbg_clear_ethbuf(char **tokens, int token_num)
{
    uint32_t i = 0;
    uint32_t pnum, pg_num, pg_size, start_pg = 0;
    volatile unsigned int *dbase, *cbase;

    dbase = (volatile unsigned int *)SB_BUFFER( ( ( struct dmrx_dba * )DM_RXDB)->dbase + 0x2000 );
    cbase = (volatile unsigned int *)SB_BUFFER( ( ( struct dmrx_cba * )DM_RXCB)->cbase + 0x2000);
    pnum = ((struct dmrx_cfg * )DM_RXCFG)->pnum;
    pg_num = pnum;
    pg_size =  ((( struct  dmrx_cfg * )DM_RXCFG)->psize + 1) * 32 / 4;

    start_pg = i < token_num ? simple_strtoul(tokens[i++], NULL, 0) : start_pg;
    pg_num = i < token_num ? simple_strtoul(tokens[i++], NULL, 0) : pg_num;
    start_pg %= pnum;
    pg_num = pg_num > pnum ? pnum : pg_num;


    dbase = dbase + start_pg * pg_size;
    cbase = cbase + start_pg * pg_size;

    ppa_memset((void *)dbase, 0, pg_num * pg_size * sizeof(uint32_t));
    ppa_memset((void *)cbase, 0, pg_num * sizeof(uint32_t));

    return;

}

static void fwdbg_dump_sthdr(char **tokens, int token_num)
{
    //uint32_t i = 0;
    uint32_t pnum,pg_size,pg_idx;
    volatile unsigned int *dbase;
    struct flag_header *pfhdr;
    unsigned char *paddr;

    dbase = (volatile unsigned int *)SB_BUFFER( ( ( struct dmrx_dba * )DM_RXDB)->dbase + 0x2000 );

    if(token_num < 1){
        fw_dbg_start("help dump_st_hdr");
        return;
    }

    pnum = ((struct dmrx_cfg * )DM_RXCFG)->pnum;
    pg_size = ((( struct  dmrx_cfg * )DM_RXCFG)->psize + 1) * 32 / 4;
    pg_idx = simple_strtoul(tokens[0], NULL, 0);
    if(pg_idx >= pnum){
        printk("Error : page NO %d is equal or bigger than total page number: %d\n", pg_idx, pnum);
        return;
    }

    pfhdr = (struct flag_header *)(dbase + pg_idx * pg_size);
    dump_flag_header(g_fwcode, pfhdr, __FUNCTION__, 0);

    if(token_num > 1 && ifx_strincmp(tokens[1], "iphdr", strlen(tokens[1])) == 0){
        if(pfhdr->is_inner_ipv4){
            printk("First header (IPv4)\n");
            paddr = (unsigned char *)pfhdr;
            paddr += 8 + pfhdr->pl_byteoff + pfhdr->ip_inner_offset;
            dump_ipv4_hdr((struct iphdr *)paddr);

        }else if(pfhdr->is_inner_ipv6){
            printk("First header (IPv6)\n");
            paddr = (unsigned char *)pfhdr;
            paddr += 8 + pfhdr->pl_byteoff + pfhdr->ip_inner_offset;
            dump_ipv6_hdr((struct ipv6hdr *)paddr);
        }

        printk("------------------------------------\n");
        if(pfhdr->is_outer_ipv4){
            printk("Second header (IPv4)\n");
            paddr = (unsigned char *)pfhdr;
            paddr += 8 + pfhdr->pl_byteoff + pfhdr->ip_outer_offset;
            dump_ipv4_hdr((struct iphdr *)paddr);
        }else if(pfhdr->is_outer_ipv6){
            printk("Second header (IPv6)\n");
            paddr = (unsigned char *)pfhdr;
            paddr += 8 + pfhdr->pl_byteoff + pfhdr->ip_outer_offset;
            dump_ipv6_hdr((struct ipv6hdr *)paddr);
        }

        printk("------------------------------------\n");
    }
    return;

}

//echo dump_rg [dmrx | dsrx | SLL]
static void fwdbg_dump_rg(char **tokens, int token_num)
{
    uint32_t i;

    if(token_num < 1){
        fw_dbg_start("help dump_rg");
        return;
    }

    if(ifx_strincmp(tokens[0],"dmrx", strlen(tokens[0])) == 0){

        printk("DMRX_DBA(0x612): 0x%x\n", ((struct dmrx_dba * )DM_RXDB)->dbase);
        printk("DMRX_CBA(0x613): 0x%x\n", (( struct dmrx_cba * )DM_RXCB)->cbase);
        printk("DMRX_CFG(0x614): \n");
            printk("\tSEN: \t0x%x\n", ((struct dmrx_cfg * )DM_RXCFG)->sen);
            printk("\tTRLPG: \t0x%x\n", ((struct dmrx_cfg * )DM_RXCFG)->trlpg);
            printk("\tHDLEN:\t0x%x\n", ((struct dmrx_cfg * )DM_RXCFG)->hdlen);
            printk("\tENDIAN:\t0x%x\n", ((struct dmrx_cfg * )DM_RXCFG)->endian);
            printk("\tPSIZE:\t0x%x\n", ((struct dmrx_cfg * )DM_RXCFG)->psize);
            printk("\tPNUM:\t0x%x\n", ((struct dmrx_cfg * )DM_RXCFG)->pnum);
        printk("DMRX_PGCNT(0x615): \n");
            printk("\tPGPTR: \t0x%x\n", ((struct dmrx_pgcnt * )DM_RXPGCNT)->pgptr);
            printk("\tDSRC: \t0x%x\n", ((struct dmrx_pgcnt * )DM_RXPGCNT)->dsrc);
            printk("\tDVAL: \t0x%x\n", ((struct dmrx_pgcnt * )DM_RXPGCNT)->dval);
            printk("\tDCMD: \t0x%x\n", ((struct dmrx_pgcnt * )DM_RXPGCNT)->dcmd);
            printk("\tUPAGE: \t0x%x\n", ((struct dmrx_pgcnt * )DM_RXPGCNT)->upage);
        printk("DMRX_PKTCNT(0x616): \n");
            printk("\tDSRC: \t0x%x\n", ((struct dmrx_pktcnt * )DM_RXPKTCNT)->dsrc);
            printk("\tDCMD: \t0x%x\n", ((struct dmrx_pktcnt * )DM_RXPKTCNT)->dcmd);
            printk("\tUPKT: \t0x%x\n", ((struct dmrx_pktcnt * )DM_RXPKTCNT)->upkt);

    }else if(ifx_strincmp(tokens[0],"dsrx", strlen(tokens[0])) == 0){
        printk("DSRX_DB(0x710): 0x%x\n", ((struct dsrx_dba * )DS_RXDB)->dbase);
        printk("DSRX_DB(0x711): 0x%x\n", ((struct dsrx_cba * )DS_RXCB)->cbase);
        printk("DSRX_CFG(0x712): \n");
            printk("\tDREN: \t0x%x\n", ((struct dsrx_cfg * )DS_RXCFG)->dren);
            printk("\tENDIAN:\t0x%x\n", ((struct dsrx_cfg * )DS_RXCFG)->endian);
            printk("\tPSIZE:\t0x%x\n", ((struct dsrx_cfg * )DS_RXCFG)->psize);
            printk("\tPNUM:\t0x%x\n", ((struct dsrx_cfg * )DS_RXCFG)->pnum);
        printk("DSRX_PGCNT(0x713): \n");
            printk("\tPGPTR: \t0x%x\n", ((struct dsrx_pgcnt* )DS_RXPGCNT)->pgptr);
            printk("\tISRC: \t0x%x\n", ((struct dsrx_pgcnt* )DS_RXPGCNT)->isrc);
            printk("\tIVAL: \t0x%x\n", ((struct dsrx_pgcnt* )DS_RXPGCNT)->ival);
            printk("\tICMD: \t0x%x\n", ((struct dsrx_pgcnt* )DS_RXPGCNT)->icmd);
            printk("\tVPAGE: \t0x%x\n", ((struct dsrx_pgcnt* )DS_RXPGCNT)->upage);

    }else if(ifx_strincmp(tokens[0],"sll", strlen(tokens[0])) == 0){
        unsigned int keynum = ((struct sll_cmd1 *)SLL_CMD1)->ksize;

        printk("SLL_CMD0(0x901):\n");
            printk("\tCMD: \t0x%x\n", ((struct sll_cmd0* )SLL_CMD0)->cmd);
            printk("\tENUM: \t0x%x\n", ((struct sll_cmd0* )SLL_CMD0)->eynum);
            printk("\tEYBASE: \t0x%x\n", ((struct sll_cmd0* )SLL_CMD0)->eybase);
        printk("Key size: 0x%x\n", keynum);
        for(i = 0; i < keynum; i ++){
            printk("Key[%d]:    0x%x\n", i, *(uint32_t*)SLL_KEY(i));
        }
        printk("SLL_RESULT(0x920) \n");
            printk("\tVLD: \t0x%x\n", ((struct sll_result* )SLL_RESULT)->vld);
            printk("\tFOUND: \t0x%x\n", ((struct sll_result* )SLL_RESULT)->fo);
            printk("\tINDEX: \t0x%x\n", ((struct sll_result* )SLL_RESULT)->index);
    }else{
        fw_dbg_start("help dump_rg");
    }

    return;
}

static void fwdbg_route(char **tokens, int token_num)
{
    unsigned long addr, *paddr;
    int i;
    volatile struct rout_forward_action_tbl *action_tbl = NULL;
    int val[16], len;

    if(! token_num ){
        goto print_help;
    }

    if(ifx_strincmp(tokens[0], "help", strlen(tokens[0])) == 0){
            goto print_help;;
    }else{
            addr = simple_strtoul(tokens[0], NULL, 16);
            if(!addr){
                goto print_help;
            }
            paddr = (unsigned long *)sb_addr_to_fpi_addr_convert(addr);
            action_tbl = (struct rout_forward_action_tbl*)paddr;
    }

    for(i = 1; i < token_num - 1; i += 2){
        if(ifx_strincmp(tokens[i], "newMAC", strlen(tokens[i])) == 0){
            len = strlen(tokens[i + 1]);
            get_mac(&tokens[i + 1], &len, val);
//                      printk("new MAC: %02X.%02X.%02X.%02X:%02X:%02X\n", val[0], val[1], val[2], val[3], val[4], val[5]);
            action_tbl->new_dest_mac54 = (val[0] << 8) | val[1];
            action_tbl->new_dest_mac30 = (val[2] << 24) | (val[3] << 16) | (val[4] << 8) | val[5];

        }else if(ifx_strincmp(tokens[i], "newip", strlen(tokens[i])) == 0){
            len = strlen(tokens[i + 1]);
            get_ip_port(&tokens[i + 1], &len, val);
//                          printk("new dest: %d.%d.%d.%d:%d\n", val[0], val[1], val[2], val[3], val[4]);
            action_tbl->new_ip = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
            action_tbl->new_port = val[4];
        }else if(ifx_strincmp(tokens[i], "routetype", strlen(tokens[i])) == 0){
            if(ifx_strincmp(tokens[i + 1], "NULL", strlen(tokens[i + 1])) == 0){
                action_tbl->rout_type = 0;

            }else if(ifx_strincmp(tokens[i + 1], "IP", strlen(tokens[i + 1])) == 0
                || ifx_strincmp(tokens[i + 1], "ROUTE", strlen(tokens[i + 1])) == 0){
                action_tbl->rout_type = 1;

            }if(ifx_strincmp(tokens[i + 1], "NAT", strlen(tokens[i + 1])) == 0){
                action_tbl->rout_type = 2;
            }if(ifx_strincmp(tokens[i + 1], "NAPT", strlen(tokens[i + 1])) == 0){
                action_tbl->rout_type = 3;
            }
        }else if(ifx_strincmp(tokens[i], "newDSCP", strlen(tokens[i])) == 0){
            action_tbl->new_dscp = simple_strtoul(tokens[i + 1], NULL, 0);
            //action_tbl->new_dscp_en = 1;

        }else if(ifx_strincmp(tokens[i], "MTUidx", strlen(tokens[i])) == 0){
            action_tbl->mtu_ix = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "VLANins", strlen(tokens[i])) == 0){
            action_tbl->in_vlan_ins = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "VLANrm", strlen(tokens[i])) == 0){
            action_tbl->in_vlan_rm = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "newDSCPen", strlen(tokens[i])) == 0){
            action_tbl->new_dscp_en = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "entry_vld", strlen(tokens[i])) == 0){
            action_tbl->entry_vld = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "TCP", strlen(tokens[i])) == 0){
            action_tbl->protocol = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "destlist", strlen(tokens[i])) == 0){
            if(ifx_strincmp(tokens[i + 1], "ETH0", strlen(tokens[i + 1])) == 0){
                action_tbl->dest_list |= 1 << 0;
            }else if(ifx_strincmp(tokens[i + 1], "ETH1", strlen(tokens[i + 1])) == 0){
                action_tbl->dest_list |= 1 << 1;
            }else if(ifx_strincmp(tokens[i + 1], "CPU0", strlen(tokens[i + 1])) == 0){
                action_tbl->dest_list |= 1 << 2;
            }else if(ifx_strincmp(tokens[i + 1], "EXT_INT1", strlen(tokens[i + 1])) == 0){
                action_tbl->dest_list |= 1 << 3;
            }else if(ifx_strincmp(tokens[i + 1], "EXT_INT2", strlen(tokens[i + 1])) == 0){
                action_tbl->dest_list |= 1 << 4;
            }else if(ifx_strincmp(tokens[i + 1], "EXT_INT3", strlen(tokens[i + 1])) == 0){
                action_tbl->dest_list |= 1 << 5;
            }else if(ifx_strincmp(tokens[i + 1], "DSL_WAN", strlen(tokens[i + 1])) == 0){
                action_tbl->dest_list |= 1 << 7;
            }
        }else if(ifx_strincmp(tokens[i], "PPPoEmode", strlen(tokens[i])) == 0){
            action_tbl->pppoe_mode = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "PPPoEidx", strlen(tokens[i])) == 0){
            action_tbl->pppoe_ix = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "newMACidx", strlen(tokens[i])) == 0){
            action_tbl->new_src_mac_ix = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "VCI", strlen(tokens[i])) == 0){
            action_tbl->new_in_vci = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "outvlanins", strlen(tokens[i])) == 0){
            action_tbl->out_vlan_ins = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "outvlanrm", strlen(tokens[i])) == 0){
            action_tbl->out_vlan_rm = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "outvlanidx", strlen(tokens[i])) == 0){
            action_tbl->out_vlan_ix = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "mpoatype", strlen(tokens[i])) == 0){
            action_tbl->mpoa_type = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "destqid", strlen(tokens[i])) == 0){
            action_tbl->dest_qid = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "tunnelidx", strlen(tokens[i])) == 0){
            action_tbl->tnnl_hdr_idx = simple_strtoul(tokens[i + 1], NULL, 0);

        }else if(ifx_strincmp(tokens[i], "tunnelen", strlen(tokens[i])) == 0){
            action_tbl->encap_tunnel = simple_strtoul(tokens[i + 1], NULL, 0);

        }
    }

    return;

print_help:

    printk("route address [name value] ... \n");
    printk("      newIP(PORT)       ??:??:??:??:????\n");
    printk("      newMAC:           ??:??:??:??:??:?? (HEX)\n");
    printk("      routetype:        NULL/IP/NAT/NAPT\n");
    printk("      newDSCP:          ??\n");
    printk("      newDSCPen:        0|1(disable|enable)\n");
    printk("      MTUidx:           ??\n");
    printk("      VLANins:          0|1(disable/enable)\n");
    printk("      VCI:              ???? (HEX)\n");
    printk("      VLANrm:           0|1(disable/enable)\n");
    printk("      outvlanins:       0|1(disable/enable)\n");
    printk("      outvlanidx:       ??\n");
    printk("      outvlanrm:        0|1(disable/enable)\n");
    printk("      destlist:         ETH0/ETH1/CPU0/EXT_INT/DSL_WAN?\n");
    printk("      PPPoEmode:        0|1(transparent/termination)\n");
    printk("      PPPoEidx:         ??\n");
    printk("      newMACidx:        ??\n");
    printk("      tcp:              0|1(no(UDP)/yes)\n");
    printk("      destqid:          ??\n");
    printk("      entryvld:         0|1(invalid/valid)\n");
    printk("      mpoatype:         ??\n");
    printk("      tunnelidx         ??\n");
    printk("      tunnelen          0|1(disable | enable)\n");

    return;
}

static void dump_ipv4_hdr(struct iphdr * iph)
{
    ASSERT(iph != NULL, "ipv4 hdr point is NULL");

    printk("Version         :%d\n", iph->version);
    printk("Header len      :%d\n", iph->ihl);
    printk("TOS             :%d\n", iph->tos);
    printk("Total len       :%d\n", iph->tot_len);
    printk("Id              :%d\n", iph->id);
    printk("Flags              \n");
    printk("  Don't Fragment:%d\n", iph->frag_off & (1 << 14));
    printk("  More Fragments:%d\n", iph->frag_off & (1 << 13));
    printk("Fragment offset :%d\n", iph->frag_off & ~0x0700);
    printk("TTL             :%d\n", iph->ttl);
    printk("Protocol        :%d\n", iph->protocol);
    printk("Header checksum :%d\n", iph->check);
    printk("Src IP address  :%d.%d.%d.%d\n", iph->saddr >> 24,iph->saddr >> 16 & 0xFF,
                                             iph->saddr >> 8 & 0xFF, iph->saddr & 0xFF);
    printk("Dst IP address  :%d.%d.%d.%d\n", iph->daddr >> 24,iph->daddr >> 16 & 0xFF,
                                             iph->daddr >> 8 & 0xFF, iph->daddr & 0xFF);

    return;
}

static void dump_ipv6_hdr(struct ipv6hdr *iph6)
{
    int i;
    ASSERT(iph6 != NULL, "ipv6 hdr point is NULL");

    printk("Version         :%d\n", iph6->version);
    printk("Traffic clase   :%d\n", iph6->priority << 4 | (iph6->flow_lbl[0] & 0xFF00));
    printk("Flow label      :%d\n", (iph6->flow_lbl[0] & 0xFF)<<16 | (iph6->flow_lbl[1] << 8) | iph6->flow_lbl[2]);
    printk("Payload Len     :%d\n", iph6->payload_len);
    printk("Next Header     :%d\n", iph6->nexthdr);
    printk("Hop limit       :%d\n", iph6->hop_limit);
    printk("Src Address     \n");
    for( i  = 0; i < 16; i ++){
        printk("%s%x", i == 0 ? "    " : i % 4 == 0 ? "." : "", (unsigned int)iph6->saddr.s6_addr + i);
    }
    printk("\n");
    printk("Dst Address     \n");
    for( i  = 0; i < 16; i ++){
        printk("%s%x", i == 0 ? "    " : i % 4 == 0 ? "." : "", (unsigned int)iph6->daddr.s6_addr + i);
    }
    printk("\n");

    return;
}

static void fwdbg_dump_iphdr(char **tokens, int token_num)
{
    unsigned long addr;
    unsigned long *paddr;

    if(token_num < 2){
        goto print_help;
    }

    addr = simple_strtoul(tokens[1], NULL, 16);
    if(!addr){
        goto print_help;
    }
    paddr = (unsigned long *)sb_addr_to_fpi_addr_convert(addr);

    if(ifx_strincmp(tokens[0], "ipv4", strlen(tokens[0])) == 0){
        dump_ipv4_hdr((struct iphdr *)paddr);
    }else if(ifx_strincmp(tokens[0], "ipv6", strlen(tokens[0])) == 0){
        dump_ipv6_hdr((struct ipv6hdr *)paddr);
    }else{
        goto print_help;
    }

    return;
print_help:
    fw_dbg_start("help dump_ip_hdr");
    return;
}

static void fwdbg_help(char **tokens, int token_num)
{
    const char *proc_file = "/proc/eth/tfwdbg";
    const char *cmds[] = {
            "help","route",
            "read_ethbuf", "read_cb",
            "clear_cb", "clear_ethbuf",
            "dump_st_hdr", "dump_ip_hdr","dump_rg",
            NULL,
    };

    int i;

    if(!token_num){//print commands only
        for(i = 0; cmds[i] != NULL; i ++){
            printk(cmds[i]);
            printk("\t");
            if(i % 3 == 0)
                printk("\n");
        }

        printk("\n\n");
        printk("echo help cmd > %s for details\n", proc_file);
        return;
    }

    if(ifx_strincmp(tokens[0], "read_ethbuf", strlen(tokens[0])) == 0){
        printk("echo read_ethbuf [start_pg] [print_pg_num] > %s \n", proc_file);
        printk("   :clear eth rx data and control buffer  \n");

    }else if(ifx_strincmp(tokens[0], "read_cb", strlen(tokens[0])) == 0){
        printk("echo read_rx_cb [rx|tx] [bc0|bc1] [start_pg] [print_pg_num] > %s \n", proc_file);
        printk("   :read rx or tx (default is both rx & tx, bare channel 0 & 1) cell/codeword buffer for bc0 or bc1  \n");

    }else if(ifx_strincmp(tokens[0], "clear_cb", strlen(tokens[0])) == 0){
        printk("echo clear_cb [rx|tx] [bc0|bc1] [start_pg] [clear_pg_num] > %s \n", proc_file);
        printk("   :clear rx or tx (or both) cell/codeword buffer for bc0 or bc1  \n");

    }else if(ifx_strincmp(tokens[0], "clear_ethbuf", strlen(tokens[0])) == 0){
        printk("echo clear_ethbuf [start_pg] [pg_num] > %s \n", proc_file);
        printk("   :clear eth rx data and control buffer  \n");

    }else if(ifx_strincmp(tokens[0], "dump_st_hdr", strlen(tokens[0])) == 0){
        printk("echo dump_st_hdr start_pg(Hex) [iphdr]> %s \n", proc_file);
        printk("   :dump status header of the packet in the provided page and dump ipv4/ipv6 header indicated by flag header  \n");

    }else if(ifx_strincmp(tokens[0], "dump_ip_hdr", strlen(tokens[0])) == 0){
        printk("echo dump_ip_hdr [ipv4 | ipv6] [share_buffer_address] [idx] > %s \n", proc_file);
        printk("   :dump ipv4 or ipv6 header in the given share buffer  \n");

    }else if(ifx_strincmp(tokens[0], "dump_rg", strlen(tokens[0])) == 0){
        printk("echo dump_rg [dmrx | dsrx | sll] %s \n", proc_file);
        printk("   :dump given registers  \n");

    }else if(ifx_strincmp(tokens[0], "route", strlen(tokens[0])) == 0){
        printk("echo route address [name value] %s \n",proc_file);
        printk("   :change route entry's contents \n");
        printk("   :echo route help for details \n");
    }else{
        fwdbg_help((char **)NULL, 0);
    }

}

static void fw_dbg_start(char *cmdbuf)
{
    char * tokens[32];
    int finished, token_num;
    int i;

    fw_dbg_t cmds[]={{"help",fwdbg_help},
                     {"read_ethbuf", fwdbg_read_ethbuf},
                     {"clear_ethbuf",fwdbg_clear_ethbuf},
                     {"dump_st_hdr", fwdbg_dump_sthdr},
                     {"dump_rg",fwdbg_dump_rg},
                     {"route", fwdbg_route},
                     {"dump_ip_hdr", fwdbg_dump_iphdr},
                     {NULL, NULL}};


    token_num = tokenize(cmdbuf, " \t", tokens, 32, &finished);
    if (token_num <= 0) {
        return;
    }

    for(i = 0; cmds[i].cmd != NULL; i ++){
        if(ifx_strincmp(tokens[0], cmds[i].cmd, strlen(tokens[0])) == 0){
            cmds[i].pfunc(&tokens[1],token_num - 1);
            break;
        }
    }

    if(cmds[i].cmd == NULL){
        fw_dbg_start("help");
    }

    return;
}
static int proc_write_fwdbg(struct file *file __attribute__((unused)), const char *buf, unsigned long count, void *data __attribute__((unused)))
{
//    char common_proc_read_buff[2048];
    char *p;
    int len, rlen;

    len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
    rlen = len - copy_from_user(common_proc_read_buff, buf, len);
    while ( rlen && common_proc_read_buff[rlen - 1] <= ' ' )
        rlen--;
    common_proc_read_buff[rlen] = 0;
    for ( p = common_proc_read_buff; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        // MOD_DEC_USE_COUNT;
        return 0;
    }

    ASSERT(p < common_proc_read_buff + 2048, "data point out of range");
    fw_dbg_start(p);

    return count;
}
#endif

#if defined(DEBUG_SEND_PROC) && DEBUG_SEND_PROC
static int ppe_send_cell(int conn, void *cell) {
	struct sk_buff *skb;
	unsigned long sys_flag;
	volatile struct tx_descriptor *desc;
	struct tx_descriptor reg_desc;
	struct sk_buff *skb_to_free;

	/*  allocate sk_buff    */
	skb = alloc_skb_tx(DMA_PACKET_SIZE);
	if (skb == NULL)
		return -ENOMEM;
	ATM_SKB(skb)->vcc = NULL;

	/*  copy data   */
	skb_put(skb, CELL_SIZE);
	my_memcpy(skb->data, cell, CELL_SIZE);

#ifndef CONFIG_MIPS_UNCACHED
	/*  write back to physical memory   */
	dma_cache_wback((u32)skb->data, skb->len);
#endif

	/*  allocate descriptor */
	local_irq_save(sys_flag);
	desc = CPU_TO_WAN_TX_DESC_BASE + g_cpu_to_wan_tx_desc_pos;
	if (!desc->own) //  PPE hold
	{
		local_irq_restore(sys_flag);
		err("Cell not alloc tx connection");
		return -EIO;
	}
	if (++g_cpu_to_wan_tx_desc_pos == CPU_TO_WAN_TX_DESC_NUM)
		g_cpu_to_wan_tx_desc_pos = 0;
	local_irq_restore(sys_flag);

	/*  load descriptor from memory */
	reg_desc = *desc;

	/*  free previous skb   */
	ASSERT((reg_desc.dataptr & (DMA_TX_ALIGNMENT - 1)) == 0,
			"reg_desc.dataptr (0x%#x) must be 8 DWORDS aligned",
			reg_desc.dataptr);
	skb_to_free = get_skb_pointer(reg_desc.dataptr);
	dev_kfree_skb_any(skb_to_free);
	put_skb_to_dbg_pool(skb);

	/*  setup descriptor    */
	reg_desc.dataptr = (u32) skb->data & (0x1FFFFFFF ^ (DMA_TX_ALIGNMENT - 1)); //  byte address
	reg_desc.byteoff = (u32) skb->data & (DMA_TX_ALIGNMENT - 1);
	reg_desc.datalen = skb->len;
	reg_desc.mpoa_pt = 1;
	reg_desc.mpoa_type = 0; //  this flag should be ignored by firmware
	reg_desc.pdu_type = 1; //  cell
	reg_desc.qid = conn;
	reg_desc.own = 0;
	reg_desc.c = 0;
	reg_desc.sop = reg_desc.eop = 1;

	dump_atm_tx_desc(&reg_desc);
	if (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_TX)
		dump_skb(skb, DUMP_SKB_LEN, "ppe_send_cell", 7, 0, 1, 1);

	/*  write discriptor to memory and write back cache */
	*((volatile u32 *) desc + 1) = *((u32 *) &reg_desc + 1);
	*(volatile u32 *) desc = *(u32 *) &reg_desc;

	adsl_led_flash();

	return 0;
}

static int proc_read_send(char *page, char **start __attribute__((unused)), off_t off, int count __attribute__((unused)),
		int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len +=
			sprintf(
					page + len,
					"echo \"[<qid> qid] [<pvc> vpi.vci] <dword0> <dword1> ...... <dword12>\" > /proc/eth/send\n");

	*eof = 1;

	return len;
}

static int proc_write_send(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
//	char common_proc_read_buff[2048];
	char *p;

	int len, rlen;
	int conn = -1;
	int dword_counter;
	unsigned int cell[16];
	int ret;
	int vpi = 0, vci = 0;
	int f_enable_irq = 0;
	int f_continue_send = 0;

	len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
	rlen = len - copy_from_user(common_proc_read_buff, buf, len);
	while (rlen && common_proc_read_buff[rlen - 1] <= ' ')
		rlen--;
	common_proc_read_buff[rlen] = 0;
	for (p = common_proc_read_buff; *p && *p <= ' '; p++, rlen--)
		;
	if (!*p) {
		return 0;
	}

	if (ifx_strincmp(p, "qid ", 4) == 0) {
		p += 4;
		rlen -= 4;
		ifx_ignore_space(&p, &rlen);

		conn = ifx_get_number(&p, &rlen, 0);
	} else if (ifx_strincmp(p, "pvc ", 4) == 0) {
		p += 4;
		rlen -= 4;
		ifx_ignore_space(&p, &rlen);

		vpi = ifx_get_number(&p, &rlen, 0);
		if (rlen && *p == '.') {
			p++;
			rlen--;
		} else
			ifx_ignore_space(&p, &rlen);
		vci = ifx_get_number(&p, &rlen, 0);
		conn = find_vpivci(vpi, vci);
		if (conn < 0)
			printk("no such pvc %d.%d\n", vpi, vci);
	}

	if (!f_continue_send) {
		if (conn >= 0 && conn <= 14) {
			for (dword_counter = 0;
					dword_counter < sizeof(cell) / sizeof(cell[0]);
					dword_counter++) {
				ifx_ignore_space(&p, &rlen);
				if (rlen <= 0)
					break;
				cell[dword_counter] = ifx_get_number(&p, &rlen, 1);
			}
			if (dword_counter < 13)
				printk("not enough data to compose cell\n");
			else {
				if (!g_atm_priv_data.connection[conn].vcc) {
					struct wrx_queue_config wrx_queue_config = { 0 };
					struct wtx_queue_config wtx_queue_config = { 0 };
					struct uni_cell_header *cell_header;

					printk("need open queue by self\n");
					/*  setup RX queue cfg and TX queue cfg */
					wrx_queue_config.mpoa_type = 3; //  set IPoA as default
					wrx_queue_config.ip_ver = 0; //  set IPv4 as default
					wrx_queue_config.mpoa_mode = 0; //  set VC-mux as default
					wrx_queue_config.oversize = aal5r_max_packet_size;
					wrx_queue_config.undersize = aal5r_min_packet_size;
					wrx_queue_config.mfs = aal5r_max_packet_size;
					wtx_queue_config.same_vc_qmap = 0x00; //  up to now, one VC with multiple TX queue has not been implemented
					wtx_queue_config.sbid =
							g_atm_priv_data.connection[conn].port;
					wtx_queue_config.qsb_vcid = conn + QSB_QUEUE_NUMBER_BASE; //  qsb qid = firmware qid + 1
					wtx_queue_config.mpoa_mode = 0; //  set VC-mux as default
					wtx_queue_config.qsben = 1; //  reserved in A4, however put 1 for backward compatible
					cell_header =
							(struct uni_cell_header *) ((u32 *) &wtx_queue_config
									+ 1);
					cell_header->clp = 0;
					cell_header->pti = ATM_PTI_US0;
					cell_header->vci = 1;
					cell_header->vpi = 4;
					cell_header->gfc = 0;
					*WRX_QUEUE_CONFIG(conn) = wrx_queue_config;
					*WTX_QUEUE_CONFIG(conn) = wtx_queue_config;
					if (g_atm_priv_data.connection_table == 0) {
						f_enable_irq = 1;
						// turn_on_dma_rx(31);
					}
				}

				if (dword_counter > 13)
					printk("too much data, last %d dwords ignored\n",
							dword_counter - 13);
				if ((ret = ppe_send_cell(conn, cell)))
					printk("ppe_send_cell(%d, %08x) return %d\n", conn,
							(unsigned int) cell, ret);

				if (f_enable_irq) {
					volatile int j;
					for (j = 0; j < 100000; j++)
						;
					//turn_off_dma_rx(31); dma laueft immer
				}
			}
		} else
			printk("the queue (%d) is wrong\n", conn);
	}

	return count;
}

#endif

#define RX_TOTAL_PCKNT                  0x1F
#define RX_UNICAST_PCKNT                0x23
#define RX_MULTICAST_PCKNT              0x22
#define RX_CRC_ERR_PCKNT                0x21
#define RX_UNDERSIZE_GOOD_PCKNT         0x1D
#define RX_OVER_SIZE_GOOD_PCKNT         0x1B
#define RX_UNDERSIZE_ERR_PCKNT          0x1E
#define RX_GOOD_PAUSE_PCKNT             0x20
#define RX_OVER_SIZE_ERR_PCKNT          0x1C
#define RX_ALLIGN_ERR_PCKNT             0x1A
#define RX_FILTERED_PCKNT               0x19
#define RX_DISCARD_CONGESTION_PCKNT     0x11

#define RX_SIZE_64B_PCKNT               0x12
#define RX_SIZE_65_127B_PCKNT           0x13
#define RX_SIZE_128_255B_PCKNT          0x14
#define RX_SIZE_256_511B_PCKNT          0x15
#define RX_SIZE_512_1023B_PCKNT         0x16
#define RX_SIZE_1024B_MAX_PCKNT         0x17

#define TX_TOTAL_PCKNT                  0x0C
#define TX_UNICAST_PCKNT                0x06
#define TX_MULTICAST_PCKNT              0x07

#define TX_SINGLE_COLLISION_CNT         0x08
#define TX_MULTIPLE_COLLISION_CNT       0x09
#define TX_LATE_COLLISION_CNT           0x0A
#define TX_EXCESSIVE_COLLISION_CNT      0x0B
#define TX_PAUSE_PCKNT                  0x0D

#define TX_SIZE_64B_PCKNT               0x00
#define TX_SIZE_65_127B_PCKNT           0x01
#define TX_SIZE_128_255B_PCKNT          0x02
#define TX_SIZE_256_511B_PCKNT          0x03
#define TX_SIZE_512_1023B_PCKNT         0x04
#define TX_SIZE_1024B_MAX_PCKNT         0x05

#define TX_DROP_PCKNT                   0x10
#define RX_DROP_PCKNT                   0x18

#define RX_GOOD_BYTE_CNT_LOW            0x24
#define RX_GOOD_BYTE_CNT_HIGH           0x25

#define RX_BAD_BYTE_CNT_LOW             0x26
#define RX_BAD_BYTE_CNT_HIGH            0x27

#define TX_GOOD_BYTE_CNT_LOW            0x0E
#define TX_GOOD_BYTE_CNT_HIGH           0x0F

static INLINE unsigned int sw_get_rmon_counter(int port, int addr) {
	*ETHSW_BM_RAM_ADDR_REG = addr;
	*ETHSW_BM_RAM_CTRL_REG = 0x8000 | port;
	while ((*ETHSW_BM_RAM_CTRL_REG & 0x8000))
		;

	return (*ETHSW_BM_RAM_VAL_1_REG << 16) | (*ETHSW_BM_RAM_VAL_0_REG & 0xFFFF);
}

static INLINE void sw_clr_rmon_counter(int port) {
	int i;

	if (port >= 0 && port < 7) {
		*ETHSW_BM_PCFG_REG(port) = 0;
		*ETHSW_BM_RMON_CTRL_REG(port) = 3;
		for (i = 1000; (*ETHSW_BM_RMON_CTRL_REG(port) & 3) != 0 && i > 0; i--)
			;
		if (i == 0)
			*ETHSW_BM_RMON_CTRL_REG(port) = 0;
		*ETHSW_BM_PCFG_REG(port) = 1;
	}
}

static INLINE int read_port_counters(int port, char *buf) {
	int len = 0;
	unsigned int counter;
	unsigned long long byteCnt;

	len += sprintf(buf + len, "\n\tPort [%d] counters\n\n", port);

	len += sprintf(buf + len, "Rx Total PckCnt     :");
	counter = sw_get_rmon_counter(port, RX_TOTAL_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Rx Unicast PckCnt   :");
	counter = sw_get_rmon_counter(port, RX_UNICAST_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Rx Multicast PckCnt :");
	counter = sw_get_rmon_counter(port, RX_MULTICAST_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Tx Total PckCnt     :");
	counter = sw_get_rmon_counter(port, TX_TOTAL_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Tx Unicase PckCnt   :");
	counter = sw_get_rmon_counter(port, TX_UNICAST_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Tx Multicase PckCnt :");
	counter = sw_get_rmon_counter(port, TX_MULTICAST_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Rx 64B PckCnt       :");
	counter = sw_get_rmon_counter(port, RX_SIZE_64B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Rx [65-127B] PckCnt :");
	counter = sw_get_rmon_counter(port, RX_SIZE_65_127B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Rx [128~255B] PckCnt:");
	counter = sw_get_rmon_counter(port, RX_SIZE_128_255B_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);
	len += sprintf(buf + len, "Rx [256~511B] PckCnt:");
	counter = sw_get_rmon_counter(port, RX_SIZE_256_511B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Rx [512~1023B]PckCnt:");
	counter = sw_get_rmon_counter(port, RX_SIZE_512_1023B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Rx [ >1024B] PckCnt :");
	counter = sw_get_rmon_counter(port, RX_SIZE_1024B_MAX_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Tx [64B] PckCnt     :");
	counter = sw_get_rmon_counter(port, TX_SIZE_64B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Tx [65~127B] PckCnt :");
	counter = sw_get_rmon_counter(port, TX_SIZE_65_127B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Tx [128~255B] PckCnt:");
	counter = sw_get_rmon_counter(port, TX_SIZE_128_255B_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);
	len += sprintf(buf + len, "Tx [256~511B] PckCnt:");
	counter = sw_get_rmon_counter(port, TX_SIZE_256_511B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Tx [512~1023B]PckCnt:");
	counter = sw_get_rmon_counter(port, TX_SIZE_512_1023B_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);
	len += sprintf(buf + len, "Tx [>1024B] PckCnt  :");
	counter = sw_get_rmon_counter(port, TX_SIZE_1024B_MAX_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Rx CRC err PckCnt   :");
	counter = sw_get_rmon_counter(port, RX_CRC_ERR_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Rx Unsize err PCkCnt:");
	counter = sw_get_rmon_counter(port, RX_UNDERSIZE_ERR_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Rx Ovsize err PckCnt:");
	counter = sw_get_rmon_counter(port, RX_OVER_SIZE_ERR_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Rx UnsizeGood PckCnt:");
	counter = sw_get_rmon_counter(port, RX_UNDERSIZE_GOOD_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Rx OvsizeGood PckCnt:");
	counter = sw_get_rmon_counter(port, RX_OVER_SIZE_GOOD_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Rx Good Pause PckCnt:");
	counter = sw_get_rmon_counter(port, RX_GOOD_PAUSE_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Rx Align err PckCnt :");
	counter = sw_get_rmon_counter(port, RX_ALLIGN_ERR_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Rx filterd errPckCnt:");
	counter = sw_get_rmon_counter(port, RX_FILTERED_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Tx Single col Cnt   :");
	counter = sw_get_rmon_counter(port, TX_SINGLE_COLLISION_CNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Tx Multiple col Cnt :");
	counter = sw_get_rmon_counter(port, TX_MULTIPLE_COLLISION_CNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Tx Late col  Cnt    :");
	counter = sw_get_rmon_counter(port, TX_LATE_COLLISION_CNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Tx Excessive col Cnt:");
	counter = sw_get_rmon_counter(port, TX_EXCESSIVE_COLLISION_CNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Tx  Pause Cnt       :");
	counter = sw_get_rmon_counter(port, TX_PAUSE_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Tx Drop PckCnt      :");
	counter = sw_get_rmon_counter(port, TX_DROP_PCKNT);
	len += sprintf(buf + len, "0x%08x\t", counter);

	len += sprintf(buf + len, "Rx Drop PckCnt      :");
	counter = sw_get_rmon_counter(port, RX_DROP_PCKNT);
	len += sprintf(buf + len, "0x%08x\n", counter);

	len += sprintf(buf + len, "Rx Good Byte Cnt    :");
	byteCnt = sw_get_rmon_counter(port, RX_GOOD_BYTE_CNT_HIGH);
	byteCnt <<= 32;
	byteCnt += sw_get_rmon_counter(port, RX_GOOD_BYTE_CNT_LOW);
	len += sprintf(buf + len, "0x%llx \t", byteCnt);

	len += sprintf(buf + len, "Tx Good Byte Cnt    :");
	byteCnt = sw_get_rmon_counter(port, TX_GOOD_BYTE_CNT_HIGH);
	byteCnt <<= 32;
	byteCnt += sw_get_rmon_counter(port, TX_GOOD_BYTE_CNT_LOW);
	len += sprintf(buf + len, "0x%llx \n", byteCnt);

	len += sprintf(buf + len, "Rx Bad Byte Cnt     :");
	byteCnt = sw_get_rmon_counter(port, RX_BAD_BYTE_CNT_HIGH);
	byteCnt <<= 32;
	byteCnt += sw_get_rmon_counter(port, RX_BAD_BYTE_CNT_LOW);
	len += sprintf(buf + len, "0x%llx \t", byteCnt);

	len += sprintf(buf + len, "Rx Discard Cong Cnt :");
	byteCnt = sw_get_rmon_counter(port, RX_DISCARD_CONGESTION_PCKNT);
	len += sprintf(buf + len, "0x%llx \n", byteCnt);

	return len;
}

static int port0_counters_proc_read(char *buf, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += read_port_counters(0, buf + len);

	*eof = 1;

	return len;
}

static int port1_counters_proc_read(char *buf, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += read_port_counters(1, buf + len);

	*eof = 1;

	return len;
}

static int port2_counters_proc_read(char *buf, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += read_port_counters(2, buf + len);

	*eof = 1;

	return len;
}

static int port3_counters_proc_read(char *buf, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += read_port_counters(3, buf + len);

	*eof = 1;

	return len;
}

static int port4_counters_proc_read(char *buf, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += read_port_counters(4, buf + len);

	*eof = 1;

	return len;
}

static int port5_counters_proc_read(char *buf, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += read_port_counters(5, buf + len);

	*eof = 1;

	return len;
}

static int port6_counters_proc_read(char *buf, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += read_port_counters(6, buf + len);

	*eof = 1;

	return len;
}
static int port_counters_proc_write(struct file *file, const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
//	char common_proc_read_buff[2048];
	char *p;
	int len, rlen;

	int port;

	len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
	rlen = len - copy_from_user(common_proc_read_buff, buf, len);
	while (rlen && common_proc_read_buff[rlen - 1] <= ' ')
		rlen--;
	common_proc_read_buff[rlen] = 0;
	for (p = common_proc_read_buff; *p && *p <= ' '; p++, rlen--)
		;
	if (!*p) {
		return 0;
	}

	if (ifx_stricmp(p, "clean") == 0 || ifx_stricmp(p, "clear") == 0) {
		strcpy(common_proc_read_buff, file->f_dentry->d_iname);
		port = common_proc_read_buff[4] - '0';
		common_proc_read_buff[4] = ' ';
		if (ifx_stricmp(common_proc_read_buff, "port mib") == 0 && port >= 0 && port < 7)
			sw_clr_rmon_counter(port);
		else
			printk("unknown entry - %s\n", file->f_dentry->d_iname);
	} else if (ifx_stricmp(p, "clean all") == 0 || ifx_stricmp(p, "clear all") == 0) {
		for (port = 0; port < 7; port++)
			sw_clr_rmon_counter(port);
	}

	return count;
}


static int proc_read_directforwarding(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)),
		int count __attribute__((unused)), int *eof, void *data __attribute__((unused))) {
	int len = 0;

	len += sprintf(page + len, "ingress direct forwarding - %s\n",
			g_ingress_direct_forward_enable ? "enable" : "disable");
	len += sprintf(page + len, "egress direct forwarding  - %s\n",
			g_egress_direct_forward_enable ? "enable" : "disable");

	*eof = 1;

	return len;
}
static int proc_write_directforwarding(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
//	char common_proc_read_buff[2048];
	char *p;
	int len, rlen;

	len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
	rlen = len - copy_from_user(common_proc_read_buff, buf, len);
	while (rlen && common_proc_read_buff[rlen - 1] <= ' ')
		rlen--;
	common_proc_read_buff[rlen] = 0;
	for (p = common_proc_read_buff; *p && *p <= ' '; p++, rlen--)
		;
	if (!*p) {
		return 0;
	}

	if (ifx_stricmp(p, "ingress enable") == 0) {
		if (!g_ingress_direct_forward_enable) {
			g_ingress_direct_forward_enable = 1;
			//egress is enabled in eth_xmit
		}
	} else if (ifx_stricmp(p, "egress enable") == 0) {
		if (!g_egress_direct_forward_enable) {
			g_egress_direct_forward_enable = 1;
			//egress is enabled in eth_xmit
		}
	} else if (ifx_stricmp(p, "ingress disable") == 0) {
		if (g_ingress_direct_forward_enable) {
			g_ingress_direct_forward_enable = 0;
		}
	} else if (ifx_stricmp(p, "egress disable") == 0) {
		if (g_egress_direct_forward_enable) {
			g_egress_direct_forward_enable = 0;
			//egress is enabled in eth_xmit
		}
	} else {
		printk(
				"echo <ingress/egress> <enable/disable> > /proc/eth/direct_forwarding\n");
		printk(
				"enable  : pure routing configuration, switch forward packets between port0/1 and cpu port directly w/o learning\n");
		printk(
				"disable : bridging/mix configuration, switch learn MAC address and make decision on which port packet forward to\n");
	}

	return count;
}

static int proc_read_clk(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)),
		int *eof, void *data __attribute__((unused))) {
	int len = 0;
	unsigned int clk;
	int valid = 0;

	switch ((*AR10_CGU_CLKFSR >> 16) & 0x07) {
	case 1:
		clk = 250;
		valid = 1;
		break;
	case 4:
		clk = 400;
		valid = 1;
		break;
	default:
		clk = 0;
	}

	if (valid)
		len += sprintf(page + len, "PPE clock - %uM\n", clk);
	else
		len += sprintf(page + len,
				"PPE clock - %uM (unknown option %u)\n", clk,
				(*AR10_CGU_CLKFSR >> 16) & 0x07);

	*eof = 1;

	return len;
}
static int proc_write_clk(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {

//	char common_proc_read_buff[2048];
	char *p;
	int len, rlen;

	len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
	rlen = len - copy_from_user(common_proc_read_buff, buf, len);
	while (rlen && common_proc_read_buff[rlen - 1] <= ' ')
		rlen--;
	common_proc_read_buff[rlen] = 0;
	for (p = common_proc_read_buff; *p && *p <= ' '; p++, rlen--)
		;
	if (!*p) {
		return 0;
	}

	if (ifx_stricmp(p, "500M") == 0 || ifx_stricmp(p, "500") == 0) {
		*AR10_CGU_CLKFSR &= ~(0x07 << 16);
	} else if (ifx_stricmp(p, "432M") == 0 || ifx_stricmp(p, "432") == 0) {
		*AR10_CGU_CLKFSR = (*AR10_CGU_CLKFSR & ~(0x07 << 16)) | (0x01 << 16);
	} else if (ifx_stricmp(p, "288M") == 0 || ifx_stricmp(p, "288") == 0) {
		*AR10_CGU_CLKFSR = (*AR10_CGU_CLKFSR & ~(0x07 << 16)) | (0x02 << 16);
	} else {
		printk("echo <500/432/288M> > /proc/eth/clk\n");
		printk("  500M - PPE clock 500MHz\n");
		printk("  432M - PPE clock 432MHz\n");
		printk("  288M - PPE clock 288MHz\n");
	}

	return count;
}


static int proc_read_prio(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)),
		int *eof, void *data __attribute__((unused))) {
	int len = 0;
	unsigned int i, j;

	len += sprintf(page + len, "Priority to Queue Map:\n");
	len += sprintf(page + len, "  prio\t\t:  0  1  2  3  4  5  6  7\n");
	for (i = 0; i < 2; i++) {
		len += sprintf(page + len, "  eth%d\t\t:", i);
		for (j = 0; j < NUM_ENTITY(g_eth_prio_queue_map[i]); j++)
			len += sprintf(page + len, "  %2d",
					g_eth_prio_queue_map[i][j]);
		len += sprintf(page + len, "\n");
	}
	for (i = 0; i < ATM_QUEUE_NUMBER; i++) {
		if (g_atm_priv_data.connection[i].vcc != NULL) {
			int sw_tx_queue_to_virt_queue_map[ATM_SW_TX_QUEUE_NUMBER] = { 0 };
			u32 sw_tx_queue_table;
			int sw_tx_queue;
			int virt_queue;

			for (sw_tx_queue_table =
					g_atm_priv_data.connection[i].sw_tx_queue_table, sw_tx_queue =
					virt_queue = 0; sw_tx_queue_table != 0;
					sw_tx_queue_table >>= 1, sw_tx_queue++)
				if ((sw_tx_queue_table & 0x01))
					sw_tx_queue_to_virt_queue_map[sw_tx_queue] = virt_queue++;

			len += sprintf(page + len, "  pvc %d.%d \t:",
					(int) g_atm_priv_data.connection[i].vcc->vpi,
					(int) g_atm_priv_data.connection[i].vcc->vci);
			for (j = 0; j < 8; j++)
				len +=
						sprintf(
								page + len,
								" %2d",
								sw_tx_queue_to_virt_queue_map[g_atm_priv_data.connection[i].prio_queue_map[j]]);
			len += sprintf(page + len, "\n");
			len += sprintf(page + len, "      phys qid\t:");
			for (j = 0; j < 8; j++)
				len += sprintf(page + len, " %2d",
						g_atm_priv_data.connection[i].prio_queue_map[j]);
			len += sprintf(page + len, "\n");
		}
	}
    if ( (g_wan_itf & 0x03) )
    {
        len += sprintf(page + len,    "Queue to Class Map:\n");
        len += sprintf(page + len,    "  queue    :  0  1  2  3  4  5  6  7\n");
        len += sprintf(page + len,    "  eth%d     :", (g_wan_itf & 0x01) ? 0 : 1);
        for ( j = 0; j < NUM_ENTITY(g_wan_queue_class_map); j++ )
            len += sprintf(page + len," %2d", g_wan_queue_class_map[j]);
        len += sprintf(page + len,    "\n");
    }
	*eof = 1;

	return len;
}

static int proc_write_prio(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
	//char common_proc_read_buff[2048];
    int len;
    char *p1, *p2;
    int colon = 1;
    int port = -1;
    int conn = -1;
    int vpi, vci;
    int prio = -1;
    int queue = -1;

    len = sizeof(common_proc_read_buff) < count ? sizeof(common_proc_read_buff) - 1 : count;
    len = len - copy_from_user(common_proc_read_buff, buf, len);
    common_proc_read_buff[len] = 0;

    p1 = common_proc_read_buff;
    while ( ifx_get_token(&p1, &p2, &len, &colon) )
    {
        if ( ifx_stricmp(p1, "help") == 0 )
        {
            printk("echo <eth0/eth1/pvc vpi.vci> prio xx queue xx [prio xx queue xx] > /proc/eth/prio\n");
            printk("echo pvc vpi.vci <add/del> > /proc/eth/prio\n");
            break;
        }
        else if ( ifx_stricmp(p1, "eth0") == 0 )
        {
            port = 0;
            prio = queue = -1;
            dbg("port = 0");
        }
        else if ( ifx_stricmp(p1, "eth1") == 0 )
        {
            port = 1;
            prio = queue = -1;
            dbg("port = 1");
        }
        else if ( ifx_stricmp(p1, "pvc") == 0 )
        {
            ifx_ignore_space(&p2, &len);
            vpi = ifx_get_number(&p2, &len, 0);
            ifx_ignore_space(&p2, &len);
            vci = ifx_get_number(&p2, &len, 0);
            conn = find_vpivci(vpi, vci);
            dbg("vpi = %u, vci = %u, conn = %d", vpi, vci, conn);
            prio = queue = -1;
        }
        else if ( (port != -1) || conn >= 0 )
        {
            if ( ifx_stricmp(p1, "p") == 0 || ifx_stricmp(p1, "prio") == 0 )
            {
                ifx_ignore_space(&p2, &len);
                prio = ifx_get_number(&p2, &len, 0);
                dbg("prio = %d", prio);
                if ( ((port == 0) || (port == 1)) && (prio >= 0) && (prio < NUM_ENTITY(g_eth_prio_queue_map[port])) )
                {
                    if ( queue >= 0 )
                        g_eth_prio_queue_map[port][prio] = queue;
                }
                else if ( conn >= 0 && prio >= 0 && prio < 8 )
                {
                    sw_tx_prio_to_queue(conn, prio, queue, NULL);
                }
                else
                {
                    err("prio (%d) is out of range 0 - %d", prio, conn >= 0 ? 7 : NUM_ENTITY(g_eth_prio_queue_map[port]) - 1);
                }
            }
            else if ( ifx_stricmp(p1, "q") == 0 || ifx_stricmp(p1, "queue") == 0 )
            {
                ifx_ignore_space(&p2, &len);
                queue = ifx_get_number(&p2, &len, 0);
                dbg("queue = %d", queue);
                if ( port >= 0 && port <= 1 )
                {
                    if ( queue >= 0 && queue < ((g_wan_itf & (1 << port)) && g_wanqos_en ? g_wanqos_en : SWITCH_CLASS_VALUE) )
                    {
                        if ( prio >= 0 )
                            g_eth_prio_queue_map[port][prio] = queue;
                    }
                    else
                    {
                        err("queue (%d) is out of range 0 - %d", queue, (g_wan_itf & (1 << port)) && g_wanqos_en ? g_wanqos_en - 1 : SWITCH_CLASS_VALUE - 1);
                    }
                }
                else if ( conn >= 0 )
                {
                    int num_of_virt_queue;

                    if ( sw_tx_prio_to_queue(conn, prio, queue, &num_of_virt_queue) != 0 )
                    {
                        err("queue (%d) is out of range 0 - %d", queue, num_of_virt_queue - 1);
                    }
                }
            }
            else if ( conn >= 0 && ifx_stricmp(p1, "add") == 0 )
            {
                int ret = sw_tx_queue_add(conn);

                if ( ret == 0 )
                {
                    dbg("add tx queue successfully");
                }
                else
                {
                    err("failed in adding tx queue: %d", ret);
                }
            }
            else if ( conn >= 0 && (ifx_stricmp(p1, "del") == 0 || ifx_stricmp(p1, "rem") == 0) )
            {
                int ret = sw_tx_queue_del(conn);

                if ( ret == 0 )
                {
                    dbg("remove tx queue successfully");
                }
                else
                {
                    err("failed in removing tx queue: %d", ret);
                }
            }
            else
            {
                err("unknown command (%s)", p1);
            }
        }
        else
        {
            err("unknown command (%s)", p1);
        }

        p1 = p2;
        colon = 1;
    }

    return count;
}

static int proc_read_ewan(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)),
		int *eof, void *data __attribute__((unused))) {
	int len = 0;
	int i, j;
	unsigned int bit;

	g_pmac_ewan = *PMAC_EWAN_REG;

	len += sprintf(page + len, "EWAN: ");
	for (i = j = 0, bit = 1; i < 7; i++, bit <<= 1)
		if ((g_pmac_ewan & bit)) {
			len += sprintf(page + len,
					(const char *) (j ? " | P%d" : "P%d"), i);
			j++;
		}
	len += sprintf(page + len, (const char *) (j ? "\n" : "null\n"));

	*eof = 1;

	return len;
}

static int proc_write_ewan(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
//	char common_proc_read_buff[2048];
	int len;
	char *p1, *p2;
	int colon = 1;

	unsigned int got_ports = 0;
	unsigned int ports = 0;
	char *ports_name[] = { "p0", "p1", "p2", "p3", "p4", "p5" };
	unsigned int i;

	len = sizeof(common_proc_read_buff) < count ? sizeof(common_proc_read_buff) - 1 : count;
	len = len - copy_from_user(common_proc_read_buff, buf, len);
	common_proc_read_buff[len] = 0;

	p1 = common_proc_read_buff;
	while (ifx_get_token(&p1, &p2, &len, &colon)) {
		if (ifx_stricmp(p1, "help") == 0) {
			printk("echo <p0|p1|p2|p3|p4|p5> > /proc/eth/ewan\n");
			break;
		} else {
			for (i = 0; i < NUM_ENTITY(ports_name); i++)
				if (ifx_stricmp(p1, ports_name[i]) == 0) {
					ports |= 1 << i;
					got_ports++;
					break;
				}
		}

		p1 = p2;
		colon = 1;
	}

	if (got_ports)
		*PMAC_EWAN_REG = g_pmac_ewan = ports;

	return count;
}

#ifdef CONFIG_IFX_PPA_QOS
static int proc_read_qos(char *page, char **start, off_t off __attribute__((unused)) , int count,
		int *eof, void *data __attribute__((unused))) {
    int len = 0;
    char *pstr;
    char str[200];
    int llen;
    int i;
    struct wtx_eg_q_shaping_cfg qos_cfg;
    struct eth_wan_mib_table qos_queue_mib;
    volatile struct tx_qos_cfg tx_qos_cfg = *TX_QOS_CFG;
    volatile struct wtx_qos_q_desc_cfg qos_q_desc_cfg;
    unsigned int portid;

    pstr = *start = page;

    __sync();

    llen = sprintf(pstr, "\n  qos         : %s\n  wfq         : %s\n  Rate shaping: %s\n\n",
                    tx_qos_cfg.eth1_qss ?"enabled":"disabled",
                    tx_qos_cfg.wfq_en?"enabled":"disabled",
                    tx_qos_cfg.shape_en ?"enabled":"disabled");
    pstr += llen;
    len += llen;

    llen = sprintf(pstr, "  Ticks  =%u,    overhd    =%u,       qnum=%u  @%p\n", (u32)tx_qos_cfg.time_tick, (u32)tx_qos_cfg.overhd_bytes, (u32)tx_qos_cfg.eth1_eg_qnum, TX_QOS_CFG );
    pstr += llen;
    len += llen;

    llen = sprintf(pstr, "  PPE clk=%u MHz, basic tick=%u\n", (u32)cgu_get_pp32_clock() / 1000000, TX_QOS_CFG->time_tick / (cgu_get_pp32_clock() / 1000000));
    pstr += llen;
    len += llen;

    if ( tx_qos_cfg.eth1_eg_qnum )
    {
        if ( ifx_ppa_drv_hal_generic_hook != NULL )
        {
            PPE_QOS_RATE_SHAPING_CFG rate_cfg = {0};
            PPE_QOS_WFQ_CFG wfq_cfg = {0};

            if ( g_eth_wan_mode == 0 )
                portid = 7;
            else if ( (g_wan_itf & 0x02) )
                portid = 1;
            else if ( (g_wan_itf & 0x01) )
                portid = 0;
            else
                portid = ~0;

            llen = sprintf(pstr, "\n  Cfg :    T     R     S -->  Bit-rate(kbps)      Weight --> Level       Address       d/w      tick_cnt   b/S\n");
            pstr += llen;
            len += llen;
            for ( i = 0; i < 8; i++ )
            {
  #ifdef CONFIG_IFX_PPA_QOS_RATE_SHAPING
                rate_cfg.portid  = portid;
                rate_cfg.queueid = i;
                ifx_ppa_drv_hal_generic_hook(PPA_GENERIC_HAL_GET_QOS_RATE_SHAPING_CFG, &rate_cfg, 0);
  #endif
  #ifdef CONFIG_IFX_PPA_QOS_WFQ
                wfq_cfg.portid  = portid;
                wfq_cfg.queueid = i;
                ifx_ppa_drv_hal_generic_hook(PPA_GENERIC_HAL_GET_QOS_WFQ_CFG, &rate_cfg, 0);
  #endif
                qos_cfg = *WTX_EG_Q_SHAPING_CFG(i);

                llen = sprintf(str, "\n      %2u:  %03u  %05u  %05u   %07u            %08u   %03u        @0x%p   %08u    %03u     %05u\n", i, qos_cfg.t, qos_cfg.r, qos_cfg.s, rate_cfg.rate_in_kbps, qos_cfg.w, wfq_cfg.weight_level, WTX_EG_Q_SHAPING_CFG(i), qos_cfg.d, qos_cfg.tick_cnt, qos_cfg.b);
                if ( proc_buf_copy(&pstr, count, off, &len, str, llen) )
                    goto PROC_READ_QOS_OVERRUN_END;
            }

            if ( portid < 2 )
            {
                //  PTM WAN mode does not have port based rate shaping, only ETH WAN mode has.

  #ifdef CONFIG_IFX_PPA_QOS_RATE_SHAPING
                rate_cfg.portid  = portid;
                rate_cfg.queueid = i;
                ifx_ppa_drv_hal_generic_hook(PPA_GENERIC_HAL_GET_QOS_RATE_SHAPING_CFG, &rate_cfg, 0);
  #endif
                qos_cfg = *WTX_EG_Q_PORT_SHAPING_CFG(0);

                llen = sprintf(str, "\n    port:  %03u  %05u  %05u   %07u                                  @0x%p   %08u    %03u     %05u\n", qos_cfg.t, qos_cfg.r, qos_cfg.s, rate_cfg.rate_in_kbps, WTX_EG_Q_PORT_SHAPING_CFG(0), qos_cfg.d, qos_cfg.tick_cnt, qos_cfg.b);
                if ( proc_buf_copy(&pstr, count, off, &len, str, llen) )
                    goto PROC_READ_QOS_OVERRUN_END;
            }
        }

        llen = sprintf(pstr, "\n  MIB : rx_pkt/rx_bytes         tx_pkt/tx_bytes       cpu_small_drop/cpu_drop  fast_small_drop/fast_drop_cnt\n");
        pstr += llen;
        len += llen;
        for ( i = 0; i < 8; i++ )
        {
            qos_queue_mib = *ETH_WAN_TX_MIB_TABLE(i);

            llen = sprintf(str, "    %2u: %010u/%010u  %010u/%010u  %010u/%010u  %010u/%010u  @0x%p\n", i,
                qos_queue_mib.wrx_total_pdu, qos_queue_mib.wrx_total_bytes,
                qos_queue_mib.wtx_total_pdu, qos_queue_mib.wtx_total_bytes,
                qos_queue_mib.wtx_cpu_drop_small_pdu, qos_queue_mib.wtx_cpu_drop_pdu,
                qos_queue_mib.wtx_fast_drop_small_pdu, qos_queue_mib.wtx_fast_drop_pdu,
                ETH_WAN_TX_MIB_TABLE(i));

            if ( proc_buf_copy(&pstr, count, off, &len, str, llen) )
                goto PROC_READ_QOS_OVERRUN_END;
        }

        //QOS queue descriptor
        llen = sprintf(pstr, "\n  Desc: threshold  num    base_addr  rd_ptr   wr_ptr\n");
        pstr += llen;
        len += llen;
        for ( i = 0; i < 8; i++ )
        {
            qos_q_desc_cfg = *WTX_QOS_Q_DESC_CFG(i);

            llen = sprintf(pstr, "    %2u: 0x%02x       0x%02x   0x%04x     0x%04x   0x%04x  @0x%p\n", i,
                qos_q_desc_cfg.threshold,
                qos_q_desc_cfg.length,
                qos_q_desc_cfg.addr,
                qos_q_desc_cfg.rd_ptr,
                qos_q_desc_cfg.wr_ptr,
                WTX_QOS_Q_DESC_CFG(i) );

            pstr += llen;
            len += llen;
        }
    }

    *eof = 1;

PROC_READ_QOS_OVERRUN_END:
    return len - off;
}
#endif

#if defined(ENABLE_CONFIGURABLE_DSL_VLAN) && ENABLE_CONFIGURABLE_DSL_VLAN
static int proc_read_dsl_vlan(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)),
		int *eof, void *data __attribute__((unused))) {
	int len = 0;
	unsigned int i;

	len += sprintf(page + len, "PVC to DSL VLAN map:\n");
	for (i = 0; i < ATM_QUEUE_NUMBER; i++) {
		if (HTU_ENTRY(i + OAM_HTU_ENTRY_NUMBER)->vld == 0) {
			len += sprintf(page + len, "  %2d: invalid\n", i);
			continue;
		}

		len += sprintf(page + len, "  %2d: PVC %u.%u\tVLAN 0x%04X\n", i,
				(unsigned int) g_atm_priv_data.connection[i].vcc->vpi,
				(unsigned int) g_atm_priv_data.connection[i].vcc->vci,
				(unsigned int) WRX_QUEUE_CONFIG(i)->new_vlan);
	}
	len += sprintf(page + len, "DSL VLAN to PVC map:\n");
	for (i = 0; i < NUM_ENTITY(g_dsl_vlan_qid_map); i += 2) {
		if (g_dsl_vlan_qid_map[i] == 0) {
			len += sprintf(page + len, "  %2d: invalid\n", i);
			continue;
		}
		len += sprintf(page + len, "  %2d: VLAN 0x%04X\tQID %d\n", i >> 1,
				(unsigned int) g_dsl_vlan_qid_map[i],
				(unsigned int) g_dsl_vlan_qid_map[i + 1]);
	}

	*eof = 1;

	return len;
}

static int proc_write_dsl_vlan(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
//	char common_proc_read_buff[2048];
	int len;
	char *p1, *p2;
	int colon = 1;
	int conn = -1;
	unsigned int vpi, vci;
	unsigned int vlan_tag = ~0U;
	int eth0 = 0, eth1 = 0;
	int bypass = 0, enable = 0;

	len = sizeof(common_proc_read_buff) < count ? sizeof(common_proc_read_buff) - 1 : count;
	len = len - copy_from_user(common_proc_read_buff, buf, len);
	common_proc_read_buff[len] = 0;

	p1 = common_proc_read_buff;
	while (ifx_get_token(&p1, &p2, &len, &colon)) {
		if (ifx_stricmp(p1, "help") == 0) {
			printk(
					"echo <pvc vpi.vci> <vlan xx> [bypass [enable|disbable]] > /proc/eth/dsl_vlan\n");
			printk(
					"  vlan   - 16-bit vlan tag, including PCP, CFI, and VID.\n");
			printk(
					"  bypass - traffic from ATM to MII0 is done by internal switch, no CPU involvement.\n");
			conn = -1;
			break;
		} else if (bypass == 1) {
			if (ifx_stricmp(p1, "enable") == 0)
				enable = 1;
			else if (ifx_stricmp(p1, "disable") == 0)
				enable = -1;
			else if (ifx_stricmp(p1, "eth0") == 0) {
				if (eth0 == 0)
					eth0 = enable == 0 ? 1 : enable;
			} else if (ifx_stricmp(p1, "eth1") == 0) {
				if (eth1 == 0)
					eth1 = enable == 0 ? 1 : enable;
			} else {
				printk("unknown command (%s)\n", p1);
				conn = -1;
				break;
			}
		} else {
			if (bypass == 1 && eth0 == 0)
				eth0 = enable >= 0 ? 1 : -1;

			bypass = 0;
			enable = 0;

			if (ifx_stricmp(p1, "pvc") == 0) {
				ifx_ignore_space(&p2, &len);
				vpi = ifx_get_number(&p2, &len, 0);
				ifx_ignore_space(&p2, &len);
				vci = ifx_get_number(&p2, &len, 0);
				conn = find_vpivci(vpi, vci);
				printk("vpi = %u, vci = %u, conn = %d\n", vpi, vci, conn);
			} else if (vlan_tag == ~0U && ifx_stricmp(p1, "vlan") == 0) {
				ifx_ignore_space(&p2, &len);
				vlan_tag = ifx_get_number(&p2, &len, 0) & 0xFFFF;
				printk("VLAN tag = 0x%04X\n", vlan_tag);
			} else if (ifx_stricmp(p1, "bypass") == 0)
				bypass = 1;
			else {
				printk("unknown command (%s)\n", p1);
				conn = -1;
				break;
			}
		}

		p1 = p2;
		colon = 1;
	}
	if (bypass == 1 && eth0 == 0)
		eth0 = enable >= 0 ? 1 : -1;

	if (conn >= 0 && vlan_tag != ~0U) {
		if (ppe_set_vlan(conn, (unsigned short) vlan_tag) != 0)
			conn = -1;
	}
	if (conn >= 0) {
		vlan_tag = (unsigned int) WRX_QUEUE_CONFIG(conn)->new_vlan;
		if (eth0 != 0)
			sw_set_vlan_bypass(0, vlan_tag, eth0 < 0 ? 1 : 0);
		if (eth1 != 0)
			sw_set_vlan_bypass(1, vlan_tag, eth0 < 0 ? 1 : 0);
	}

	return count;
}

static INLINE int ppe_set_vlan(int conn, unsigned short vlan_tag) {
	int candidate = -1;
	unsigned int i;

	for (i = 0; i < NUM_ENTITY(g_dsl_vlan_qid_map); i += 2) {
		if (candidate < 0 && g_dsl_vlan_qid_map[i] == 0)
			candidate = i;
		else if (g_dsl_vlan_qid_map[i] == (unsigned short) vlan_tag) {
			if (conn != g_dsl_vlan_qid_map[i + 1]) {
				err("VLAN tag 0x%04X is in use", vlan_tag);
				return -1; //  quit (error), VLAN tag is used by other PVC
			}
			return 0; //  quit (success), VLAN tag is assigned already
		}
	}
	//  candidate must be valid, because size of g_dsl_vlan_qid_map is equal to max number of PVCs
	g_dsl_vlan_qid_map[candidate + 1] = conn;
	g_dsl_vlan_qid_map[candidate] = vlan_tag;
	WRX_QUEUE_CONFIG(conn)->new_vlan = vlan_tag;

	return 0;
}

static INLINE int sw_set_vlan_bypass(int itf __attribute__((unused)), unsigned int vlan_tag __attribute__((unused)), int is_del __attribute__((unused))) {
	if (itf == 1) {
		err("does not support eth1 VLAN bypass!");
		return -1;
	}
	return 0;
}
#endif

#if defined(DEBUG_MIRROR_PROC) && DEBUG_MIRROR_PROC

static int proc_read_mirror(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)),
		int *eof, void *data __attribute__((unused))) {
	int len = 0;
	char devname[IFNAMSIZ + 1] = { 0 };

	if (g_mirror_netdev == NULL)
		strcpy(devname, "NONE");
	else
		strncpy(devname, g_mirror_netdev->name, IFNAMSIZ);

	len += sprintf(page + len, "mirror: %s\n", devname);

	*eof = 1;

	return len;
}


static int proc_write_mirror(struct file *file __attribute__((unused)), const char *buf,
		unsigned long count, void *data __attribute__((unused))) {
//	char common_proc_read_buff[2048];
	char *p;
	int len, rlen;

	len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
	rlen = len - copy_from_user(common_proc_read_buff, buf, len);
	while (rlen && common_proc_read_buff[rlen - 1] <= ' ')
		rlen--;
	common_proc_read_buff[rlen] = 0;
	for (p = common_proc_read_buff; *p && *p <= ' '; p++, rlen--)
		;
	if (!*p)
		return count;

	if (g_mirror_netdev != NULL) {
		dev_put(g_mirror_netdev);
		g_mirror_netdev = NULL;
	}
	if (ifx_stricmp(p, "none") == 0)
		printk("disable mirror\n");
	else {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
		g_mirror_netdev = dev_get_by_name(p);
#else
		g_mirror_netdev = dev_get_by_name(&init_net, p);
#endif
		if (g_mirror_netdev != NULL)
			printk("mirror: %s\n", p);
		else
			printk("mirror: none, can't find device \"%s\"\n", p);
	}

	return count;
}

#endif

static int proc_read_class(char *page, char **start __attribute__((unused)), off_t off __attribute__((unused)), int count __attribute__((unused)), int *eof, void *data __attribute__((unused)))
{
    int len = 0;

    len += sprintf(page + len, "CFG_CLASS2QID_MAP (0x%08x): %08x %08x %08x %08x\n", (unsigned int)CFG_CLASS2QID_MAP(0), *CFG_CLASS2QID_MAP(0), *CFG_CLASS2QID_MAP(1), *CFG_CLASS2QID_MAP(2), *CFG_CLASS2QID_MAP(3));
    len += sprintf(page + len, "CFG_QID2CLASS_MAP (0x%08x): %08x %08x %08x %08x\n", (unsigned int)CFG_QID2CLASS_MAP(0), *CFG_QID2CLASS_MAP(0), *CFG_QID2CLASS_MAP(1), *CFG_QID2CLASS_MAP(2), *CFG_QID2CLASS_MAP(3));

    *eof = 1;

    return len;
}

static int proc_write_class(struct file *file __attribute__((unused)), const char *buf, unsigned long count, void *data __attribute__((unused)))
{
    char local_buf[2048];
    int len;
    char *p1, *p2;
    int colon = 1;

    int i, j;

    len = sizeof(local_buf) < count ? sizeof(local_buf) - 1 : count;
    len = len - copy_from_user(local_buf, buf, len);
    local_buf[len] = 0;

    p1 = local_buf;
    while ( ifx_get_token(&p1, &p2, &len, &colon) )
    {
        if ( ifx_stricmp(p1, "CFG_CLASS2QID_MAP") == 0 || ifx_stricmp(p1, "CLASS2QID") == 0 )
        {
            for ( i = 0; i < 4; i++ )
            {
                ifx_ignore_space(&p2, &len);
                if ( !len )
                    break;
                *CFG_CLASS2QID_MAP(i) = ifx_get_number(&p2, &len, 1);
            }
            break;
        }
        else if ( ifx_stricmp(p1, "CFG_QID2CLASS_MAP") == 0 || ifx_stricmp(p1, "QID2CLASS") == 0 )
        {
            for ( i = 0; i < 4; i++ )
            {
                ifx_ignore_space(&p2, &len);
                if ( !len )
                    break;
                *CFG_QID2CLASS_MAP(i) = ifx_get_number(&p2, &len, 1);
                for ( j = 0; j < 8; j++ )
                    g_wan_queue_class_map[i * 8 + j] = (*CFG_QID2CLASS_MAP(i) >> (j * 4)) & 0x0F;
            }
            break;
        }
        else
        {
            printk("echo <CLASS2QID | QID2CLASS> [hex numbers] > /proc/eth/class\n");
            break;
        }

        p1 = p2;
        colon = 1;
    }

    return count;
}
static INLINE int proc_buf_copy(char **pbuf, int size, off_t off, int *ppos, const char *str, int len)
{
    if ( *ppos <= off && *ppos + len > off )
    {
        my_memcpy(*pbuf, str + off - *ppos, *ppos + len - off);
        *pbuf += *ppos + len - off;
    }
    else if ( *ppos > off )
    {
        my_memcpy(*pbuf, str, len);
        *pbuf += len;
    }
    *ppos += len;
    return *ppos >= off + size;
}




#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC
static INLINE void get_ip_port(char **p, int *len, unsigned int *val)
{
	int i;
	unsigned int tmp_val[17] = {0};
	int is_ipv6 = 0;

	memset(val, 0, sizeof(*val) * 6);

	for ( i = 0; i < 4; i++ )
	{
		ifx_ignore_space(p, len);
		if ( !*len )
		break;
		val[i] = ifx_get_number(p, len, 0);
	}

	if ( **p == '.' )
	{
		is_ipv6 = 1;
		for ( i = 0; i < 16 - 4; i++ )
		{
			ifx_ignore_space(p, len);
			if ( !*len )
			break;
			tmp_val[i] = ifx_get_number(p, len, 0);
		}
	}

	ifx_ignore_space(p, len);
	if ( *len )
	val[4] = ifx_get_number(p, len, 0);

	if ( is_ipv6 )
	{
		val[5] = 6;
		for ( i = 0; i < 16 - 4; i++ )
		val[i + 6] = tmp_val[i];
	}
	else
	val[5] = 4;
}

static INLINE void get_mac(char **p, int *len, unsigned int *val)
{
	int i;

	memset(val, 0, sizeof(*val) * 6);

	for ( i = 0; i < 6; i++ )
	{
		ifx_ignore_space(p, len);
		if ( !*len )
		break;
		val[i] = ifx_get_number(p, len, 1);
	}
}
#endif

#if defined(DEBUG_FIRMWARE_TABLES_PROC) && DEBUG_FIRMWARE_TABLES_PROC
static INLINE char *get_wanitf(int iswan)
{
    static char itfs[64];
    char *allitfs[] = {
        "ETH0",
        "ETH1",
        "", //reserve for CPU
        "EXT1",
        "EXT2",
        "EXT3",
        "EXT4",
        "EXT5",
        NULL
    };

    int wan_itf = *CFG_WAN_PORTMAP;
    int wan_mixmap = *CFG_MIXED_PORTMAP;
    int i, len = 0;

    int max_exts_idx = 3 + 2;

    memset(itfs, 0, sizeof(itfs));
    if(wan_mixmap != 0 && !iswan){ //mix mode dont have lan
        return itfs;
    }

    for(i = 0; allitfs[i] != NULL ; i ++){
        if(i == 2) continue; //skip CPU port

        if((iswan && (wan_itf & (1 << i))) || (!iswan && !(wan_itf & (1 << i)))){
            if(i > max_exts_idx){
                if(!iswan){
                    break; //DSL cannot be lan
                }
                else{
                    len += sprintf(itfs + len, "%s ", "DSL");
                    break; //DSL only listed once although in A5 we have 2 ports support DSL
                }
            }

            len += sprintf(itfs + len, "%s ", allitfs[i]);

            if(iswan && (wan_mixmap &  (1 << i))){
                len += sprintf(itfs + len, "(Mixed) ");
            }
        }
    }

    return itfs;
}


#if defined(PROC_WRITE_ROUTE) && PROC_WRITE_ROUTE
static INLINE int print_route(char *buf, int i, int f_is_lan, struct rout_forward_compare_tbl *pcompare, struct rout_forward_action_tbl *pwaction)
{
	static const char *dest_list[] = {"ETH0", "ETH1", "CPU0", "EXT_INT1", "EXT_INT2", "EXT_INT3", "res", "ATM"};
	static const char *mpoa_type[] = {"EoA without FCS", "EoA with FCS", "PPPoA", "IPoA"};

	int len = 0;
	u32 bit;
	int j, k;

	len += sprintf(buf + len, "  entry %d\n", i);
	len += sprintf(buf + len, "    compare (0x%08X)\n", (u32)pcompare);
	len += sprintf(buf + len, "      src:  %d.%d.%d.%d:%d\n", pcompare->src_ip >> 24, (pcompare->src_ip >> 16) & 0xFF, (pcompare->src_ip >> 8) & 0xFF, pcompare->src_ip & 0xFF, pcompare->src_port);
	len += sprintf(buf + len, "      dest: %d.%d.%d.%d:%d\n", pcompare->dest_ip >> 24, (pcompare->dest_ip >> 16) & 0xFF, (pcompare->dest_ip >> 8) & 0xFF, pcompare->dest_ip & 0xFF, pcompare->dest_port);
	len += sprintf(buf + len, "    action  (0x%08X)\n", (u32)pwaction);
	len += sprintf(buf + len, "      new %s    %d.%d.%d.%d:%d\n", f_is_lan ? "src :" : "dest:", pwaction->new_ip >> 24, (pwaction->new_ip >> 16) & 0xFF, (pwaction->new_ip >> 8) & 0xFF, pwaction->new_ip & 0xFF, pwaction->new_port);
	len += sprintf(buf + len, "      new MAC :    %02X:%02X:%02X:%02X:%02X:%02X\n", (pwaction->new_dest_mac54 >> 8) & 0xFF, pwaction->new_dest_mac54 & 0xFF, pwaction->new_dest_mac30 >> 24, (pwaction->new_dest_mac30 >> 16) & 0xFF, (pwaction->new_dest_mac30 >> 8) & 0xFF, pwaction->new_dest_mac30 & 0xFF);
	switch ( pwaction->rout_type )
	{
		case 1: len += sprintf(buf + len, "      route type:  IP\n"); break;
		case 2: len += sprintf(buf + len, "      route type:  NAT\n"); break;
		case 3: len += sprintf(buf + len, "      route type:  NAPT\n"); break;
		default: len += sprintf(buf + len, "      route type:  NULL\n");
	}
	if ( pwaction->new_dscp_en )
	len += sprintf(buf + len, "      new DSCP:    %d\n", pwaction->new_dscp);
	else
	len += sprintf(buf + len, "      new DSCP:    original (not modified)\n");
	len += sprintf(buf + len, "      MTU index:   %d\n", pwaction->mtu_ix);
	if ( pwaction->in_vlan_ins )
	len += sprintf(buf + len, "      VLAN insert: enable, VCI 0x%04x\n", pwaction->new_in_vci);
	else
	len += sprintf(buf + len, "      VLAN insert: disable\n");
	len += sprintf(buf + len, "      VLAN remove: %s\n", pwaction->in_vlan_rm ? "enable" : "disable");
	if ( !pwaction->dest_list )
	len += sprintf(buf + len, "      dest list:   none\n");
	else
	{
		len += sprintf(buf + len, "      dest list:   ");
		for ( bit = 1, j = k = 0; bit < 1 << 8; bit <<= 1, j++ )
		if ( (pwaction->dest_list & bit) )
		{
			if ( k )
			len += sprintf(buf + len, ", ");
			len += sprintf(buf + len, dest_list[j]);
			k = 1;
		}
		len += sprintf(buf + len, "\n");
	}
	if ( pwaction->pppoe_mode )
	{
		len += sprintf(buf + len, "      PPPoE mode:  termination\n");
		if ( f_is_lan )
		len += sprintf(buf + len, "      PPPoE index: %d\n", pwaction->pppoe_ix);
	}
	else
	len += sprintf(buf + len, "      PPPoE mode:  transparent\n");
	len += sprintf(buf + len, "      new src MAC index: %d\n", pwaction->new_src_mac_ix);
    if(f_is_lan){
        len += sprintf(buf + len,      "      encap tunnel: %s\n", pwaction->encap_tunnel ? "yes":"no");

        len += sprintf(buf + len,      "      tunnel iphdr idx: %d\n", pwaction->tnnl_hdr_idx);
    }else{
        len += sprintf(buf + len,      "      decap tunnel: %s\n", pwaction->encap_tunnel ? "yes":"no");
    }
	if ( pwaction->out_vlan_ins )
	len += sprintf(buf + len, "      outer VLAN insert: enable, index %d\n", pwaction->out_vlan_ix);
	else
	len += sprintf(buf + len, "      outer VLAN insert: disable\n");
	if ( f_is_lan || GEN_MODE_CFG1->sys_cfg != 0 /* not DSL WAN mode */)
	len += sprintf(buf + len, "      outer VLAN remove: %s\n", pwaction->out_vlan_rm ? "enable" : "disable");
	else
	len += sprintf(buf + len, "      outer VLAN remove: %s (forced to be enable for ATM PVC differentiation)\n", pwaction->out_vlan_rm ? "enable" : "disable");
	len += sprintf(buf + len, "      tcp:         %s\n", pwaction->protocol ? "yes" : "no (UDP)");
	len += sprintf(buf + len, "      entry valid: %s\n", pwaction->entry_vld ? "yes" : "no");
	if ( f_is_lan )
	{
		len += sprintf(buf + len, "      mpoa type:   %s\n", mpoa_type[pwaction->mpoa_type]);
        len += sprintf(buf + len, "      dslwan qid:  %d (conn %d)\n", pwaction->dest_qid & 0xFF, (pwaction->dest_qid >> 8) & 0xFF);
	}
	else
        len += sprintf(buf + len,      "      dest qid (dslwan qid): %d\n", pwaction->dest_qid);
    len += sprintf(buf + len,          "      accl bytes: %u\n", pwaction->bytes);
	return len;
}
#endif

static INLINE int print_mc(char *buf, int i, struct wan_rout_multicast_cmp_tbl *pcompare, struct wan_rout_multicast_act_tbl *paction)
{
    static const char *rout_type[] = {"no IP level editing", "TTL editing", "reserved - 2", "reserved - 3"};
    static const char *dest_list[] = {"ETH0", "ETH1", "CPU0", "EXT_INT1", "EXT_INT2", "EXT_INT3", "res", "ATM"};

    int len = 0;
    u32 bit;
    int j, k;

    len += sprintf(buf + len,          "  entry %d - %s\n", i, paction->entry_vld ? "valid" : "invalid");
    len += sprintf(buf + len,          "    compare (0x%08X)\n", (u32)pcompare);
    len += sprintf(buf + len,          "      wan_dest_ip:  %d.%d.%d.%d\n", (pcompare->wan_dest_ip >> 24) & 0xFF,  (pcompare->wan_dest_ip >> 16) & 0xFF,  (pcompare->wan_dest_ip >> 8) & 0xFF,  pcompare->wan_dest_ip & 0xFF);
    len += sprintf(buf + len,          "    action  (0x%08X)\n", (u32)paction);
    len += sprintf(buf + len,          "      rout_type:    %s\n", rout_type[paction->rout_type]);
    if ( paction->new_dscp_en )
        len += sprintf(buf + len,      "      new DSCP:     %d\n", paction->new_dscp);
    else
        len += sprintf(buf + len,      "      new DSCP:     original (not modified)\n");
    if ( paction->in_vlan_ins )
        len += sprintf(buf + len,      "      VLAN insert:  enable, VCI 0x%04x\n", paction->new_in_vci);
    else
        len += sprintf(buf + len,      "      VLAN insert:  disable\n");
    len += sprintf(buf + len,          "      VLAN remove:  %s\n", paction->in_vlan_rm ? "enable" : "disable");
    if ( !paction->dest_list )
        len += sprintf(buf + len,      "      dest list:    none\n");
    else
    {
        len += sprintf(buf + len,      "      dest list:   ");
        for ( bit = 1, j = k = 0; bit < 1 << 8; bit <<= 1, j++ )
            if ( (paction->dest_list & bit) )
            {
                if ( k )
                    len += sprintf(buf + len, ", ");
                len += sprintf(buf + len, dest_list[j]);
                k = 1;
            }
        len += sprintf(buf + len, "\n");
    }
    len += sprintf(buf + len,          "      PPPoE mode:   %s\n", paction->pppoe_mode ? "termination" : "transparent");
    if ( paction->new_src_mac_en )
        len += sprintf(buf + len,      "      new src MAC index: %d\n", paction->new_src_mac_ix);
    else
        len += sprintf(buf + len,      "      new src MAC index: disabled\n");
    if ( paction->out_vlan_ins )
        len += sprintf(buf + len,      "      outer VLAN insert: enable, index %d\n", paction->out_vlan_ix);
    else
        len += sprintf(buf + len,      "      outer VLAN insert: disable\n");
    len += sprintf(buf + len,          "      outer VLAN remove: %s\n", paction->out_vlan_rm ? "enable" : "disable");
    len += sprintf(buf + len,          "      dest_qid:          %d\n", paction->dest_qid);

    return len;
}
#endif


#if defined(DEBUG_DUMP_INIT) && DEBUG_DUMP_INIT
static INLINE void dump_init(void)
{
    int i;

	if ( !(g_dbg_datapath & DBG_ENABLE_MASK_DUMP_INIT) )
    	return;
    printk("Share Buffer Conf:\n");
    printk("  SB_MST_SEL(%08X) = 0x%08X\n", (u32)SB_MST_SEL, *SB_MST_SEL);

    printk("ETOP:\n");
    printk("  ETOP_MDIO_CFG        = 0x%08X\n", *ETOP_MDIO_CFG);
    printk("  ETOP_MDIO_ACC        = 0x%08X\n", *ETOP_MDIO_ACC);
    printk("  ETOP_CFG             = 0x%08X\n", *ETOP_CFG);
    printk("  ETOP_IG_VLAN_COS     = 0x%08X\n", *ETOP_IG_VLAN_COS);
    printk("  ETOP_IG_DSCP_COSx(0) = 0x%08X\n", *ETOP_IG_DSCP_COSx(0));
    printk("  ETOP_IG_DSCP_COSx(1) = 0x%08X\n", *ETOP_IG_DSCP_COSx(1));
    printk("  ETOP_IG_DSCP_COSx(2) = 0x%08X\n", *ETOP_IG_DSCP_COSx(2));
    printk("  ETOP_IG_DSCP_COSx(3) = 0x%08X\n", *ETOP_IG_DSCP_COSx(3));
    printk("  ETOP_IG_PLEN_CTRL0   = 0x%08X\n", *ETOP_IG_PLEN_CTRL0);
    printk("  ETOP_ISR             = 0x%08X\n", *ETOP_ISR);
    printk("  ETOP_IER             = 0x%08X\n", *ETOP_IER);
    printk("  ETOP_VPID            = 0x%08X\n", *ETOP_VPID);

    for ( i = 0; i < 2; i++ )
    {
        printk("ENET%d:\n", i);
        printk("  ENET_MAC_CFG(%d)      = 0x%08X\n", i, *ENET_MAC_CFG(i));
        printk("  ENETS_DBA(%d)         = 0x%08X\n", i, *ENETS_DBA(i));
        printk("  ENETS_CBA(%d)         = 0x%08X\n", i, *ENETS_CBA(i));
        printk("  ENETS_CFG(%d)         = 0x%08X\n", i, *ENETS_CFG(i));
        printk("  ENETS_PGCNT(%d)       = 0x%08X\n", i, *ENETS_PGCNT(i));
        printk("  ENETS_PKTCNT(%d)      = 0x%08X\n", i, *ENETS_PKTCNT(i));
        printk("  ENETS_BUF_CTRL(%d)    = 0x%08X\n", i, *ENETS_BUF_CTRL(i));
        printk("  ENETS_COS_CFG(%d)     = 0x%08X\n", i, *ENETS_COS_CFG(i));
        printk("  ENETS_IGDROP(%d)      = 0x%08X\n", i, *ENETS_IGDROP(i));
        printk("  ENETS_IGERR(%d)       = 0x%08X\n", i, *ENETS_IGERR(i));
        printk("  ENETS_MAC_DA0(%d)     = 0x%08X\n", i, *ENETS_MAC_DA0(i));
        printk("  ENETS_MAC_DA1(%d)     = 0x%08X\n", i, *ENETS_MAC_DA1(i));
        printk("  ENETF_DBA(%d)         = 0x%08X\n", i, *ENETF_DBA(i));
        printk("  ENETF_CBA(%d)         = 0x%08X\n", i, *ENETF_CBA(i));
        printk("  ENETF_CFG(%d)         = 0x%08X\n", i, *ENETF_CFG(i));
        printk("  ENETF_PGCNT(%d)       = 0x%08X\n", i, *ENETF_PGCNT(i));
        printk("  ENETF_PKTCNT(%d)      = 0x%08X\n", i, *ENETF_PKTCNT(i));
        printk("  ENETF_HFCTRL(%d)      = 0x%08X\n", i, *ENETF_HFCTRL(i));
        printk("  ENETF_TXCTRL(%d)      = 0x%08X\n", i, *ENETF_TXCTRL(i));
        printk("  ENETF_VLCOS0(%d)      = 0x%08X\n", i, *ENETF_VLCOS0(i));
        printk("  ENETF_VLCOS1(%d)      = 0x%08X\n", i, *ENETF_VLCOS1(i));
        printk("  ENETF_VLCOS2(%d)      = 0x%08X\n", i, *ENETF_VLCOS2(i));
        printk("  ENETF_VLCOS3(%d)      = 0x%08X\n", i, *ENETF_VLCOS3(i));
        printk("  ENETF_EGCOL(%d)       = 0x%08X\n", i, *ENETF_EGCOL(i));
        printk("  ENETF_EGDROP(%d)      = 0x%08X\n", i, *ENETF_EGDROP(i));
    }

    printk("DPLUS:\n");
    printk("  DPLUS_TXDB           = 0x%08X\n", *DPLUS_TXDB);
    printk("  DPLUS_TXCB           = 0x%08X\n", *DPLUS_TXCB);
    printk("  DPLUS_TXCFG          = 0x%08X\n", *DPLUS_TXCFG);
    printk("  DPLUS_TXPGCNT        = 0x%08X\n", *DPLUS_TXPGCNT);
    printk("  DPLUS_RXDB           = 0x%08X\n", *DPLUS_RXDB);
    printk("  DPLUS_RXCB           = 0x%08X\n", *DPLUS_RXCB);
    printk("  DPLUS_RXCFG          = 0x%08X\n", *DPLUS_RXCFG);
    printk("  DPLUS_RXPGCNT        = 0x%08X\n", *DPLUS_RXPGCNT);

	printk("Communication:\n");
	printk("  FW_VER_ID(%08X)  = 0x%08X\n", (u32)FW_VER_ID, *(u32*)FW_VER_ID);
}
#endif

#if defined(ENABLE_DIRECTPATH_TX_QUEUE) && ENABLE_DIRECTPATH_TX_QUEUE
static int send_directpath_tx_queue(void)
{
	unsigned long sys_flag;
	struct sk_buff *skb;
	int len;
	int i;

	for ( i = 0; !g_directpath_dma_full && i < NUM_ENTITY(ppa_pushthrough_data); i++ )
	{
		while ( !g_directpath_dma_full && ppa_pushthrough_data[i].skb_list )
		{
			local_irq_save(sys_flag);

			skb = skb_get(ppa_pushthrough_data[i].skb_list);

			len = skb->len;
			if ( len < ETH_MIN_TX_PACKET_LENGTH + sizeof(struct sw_eg_pkt_header) )
			len = ETH_MIN_TX_PACKET_LENGTH + sizeof(struct sw_eg_pkt_header);
            
            put_skb_to_dbg_pool(skb);

			if ( dma_device_write(g_dma_device_ppe, skb->data, len, skb, chan_no) == len )
			{
				//  remove skb from list
				if ( skb->prev == skb )
				skb->prev = skb->next = ppa_pushthrough_data[i].skb_list = NULL;
				else
				{
					ppa_pushthrough_data[i].skb_list = skb->next;
					ppa_pushthrough_data[i].skb_list->prev = skb->prev;
					skb->prev->next = skb->next;
					skb->prev = skb->next = NULL;
				}
				ppa_pushthrough_data[i].skb_list_len--;
				ppa_pushthrough_data[i].skb_list_size -= skb->len;
			}
			else
			{
				local_irq_restore(sys_flag);
				dev_kfree_skb_any(skb);
				return -1;
			}

			local_irq_restore(sys_flag);

			dev_kfree_skb_any(skb);

			if ( ppa_pushthrough_data[i].callback.start_tx_fn
#if defined(ENABLE_DIRECTPATH_TX_QUEUE_SIZE)
					&& ppa_pushthrough_data[i].skb_list_size <= (ENABLE_DIRECTPATH_TX_QUEUE_SIZE * 3 / 4)
#else
					&& ppa_pushthrough_data[i].skb_list_len <= (ENABLE_DIRECTPATH_TX_QUEUE_PKTS * 3 / 4)
#endif
			)
			ppa_pushthrough_data[i].callback.start_tx_fn(ppa_pushthrough_data[i].netif);
		}
	}

	return 0;
}
#endif

static INLINE void register_netdev_event_handler(void) {
	g_netdev_event_handler_nb.notifier_call = netdev_event_handler;
	register_netdevice_notifier(&g_netdev_event_handler_nb);
}

static INLINE void unregister_netdev_event_handler(void) {
	unregister_netdevice_notifier(&g_netdev_event_handler_nb);
}

#ifdef AVM_DOESNT_USE_IFX_ETHSW_API
static int netdev_event_handler(struct notifier_block *nb, unsigned long event, void *netdev)
{
	struct net_device *netif = (struct net_device *)netdev;
	IFX_ETHSW_HANDLE handle;
	union ifx_sw_param x;

	if ( netif->type != ARPHRD_ETHER )
	return NOTIFY_DONE;

	if ( (netif->flags & IFF_POINTOPOINT) ) //  ppp interface
	return NOTIFY_DONE;

	switch ( event )
	{
		case NETDEV_UP:
		handle = ifx_ethsw_kopen("/dev/switch_api/0");
		memset(&x.MAC_tableAdd, 0x00, sizeof(x.MAC_tableAdd));
		x.MAC_tableAdd.nFId = 0;
		x.MAC_tableAdd.nPortId = 6; //  CPU port
		//if ( netif->name[0] == 'p' && netif->name[1] == 't' && netif->name[2] == 'm' )  //  ptm
		//    x.MAC_tableAdd.nPortId = 11;
		//else if ( netif->name[0] == 'n' && netif->name[1] == 'a' && netif->name[2] == 's' ) //  atm
		//    x.MAC_tableAdd.nPortId = 11;
		//else if ( netif->name[0] == 'e' && netif->name[1] == 't' && netif->name[2] == 'h' && netif->name[3] == '1' )    //  eth1
		//    x.MAC_tableAdd.nPortId = ;
		//else
		//{
		//    int i;
		//
		//    for ( i = 0; i < NUM_ENTITY(ppa_pushthrough_data); i++ )
		//    {
		//        if ( netif == ppa_pushthrough_data[i].netif )
		//            x.MAC_tableAdd.nPortId = 7 + i;
		//    }
		//}
		x.MAC_tableAdd.nAgeTimer = 3;
		x.MAC_tableAdd.bStaticEntry = 1;
		x.MAC_tableAdd.nMAC[0] = netif->dev_addr[0];
		x.MAC_tableAdd.nMAC[1] = netif->dev_addr[1];
		x.MAC_tableAdd.nMAC[2] = netif->dev_addr[2];
		x.MAC_tableAdd.nMAC[3] = netif->dev_addr[3];
		x.MAC_tableAdd.nMAC[4] = netif->dev_addr[4];
		x.MAC_tableAdd.nMAC[5] = netif->dev_addr[5];

		printk(KERN_ERR "[%s]%d\n", __func__, __LINE__);
		ifx_ethsw_kioctl(handle, IFX_ETHSW_MAC_TABLE_ENTRY_ADD, (unsigned int)&x.MAC_tableAdd);
		ifx_ethsw_kclose(handle);
		return NOTIFY_OK;
		case NETDEV_DOWN:
		handle = ifx_ethsw_kopen("/dev/switch_api/0");
		memset(&x.MAC_tableRemove, 0x00, sizeof(x.MAC_tableRemove));
		x.MAC_tableRemove.nFId = 0;
		x.MAC_tableRemove.nMAC[0] = netif->dev_addr[0];
		x.MAC_tableRemove.nMAC[1] = netif->dev_addr[1];
		x.MAC_tableRemove.nMAC[2] = netif->dev_addr[2];
		x.MAC_tableRemove.nMAC[3] = netif->dev_addr[3];
		x.MAC_tableRemove.nMAC[4] = netif->dev_addr[4];
		x.MAC_tableRemove.nMAC[5] = netif->dev_addr[5];
		printk(KERN_ERR "[%s]%d\n", __func__, __LINE__);
		ifx_ethsw_kioctl(handle, IFX_ETHSW_MAC_TABLE_ENTRY_REMOVE, (unsigned int)&x.MAC_tableRemove);
		ifx_ethsw_kclose(handle);
		return NOTIFY_OK;
	}

	return NOTIFY_DONE;
}
#else
static int netdev_event_handler(struct notifier_block *nb __attribute__((unused)), unsigned long event __attribute__((unused)),
		void *netdev __attribute__((unused))) {
	return NOTIFY_DONE;
}
#endif

/*
 * ####################################
 *           Global Function
 * ####################################
 */

int32_t ppa_datapath_generic_hook(PPA_GENERIC_DATAPATH_HOOK_CMD cmd, void *buffer, uint32_t flag __attribute__((unused)))
{
    dbg("ppa_datapath_generic_hook cmd 0x%x\n", cmd );
    switch ( cmd )
    {
    case PPA_GENERIC_DATAPATH_TSET:
        return IFX_SUCCESS;

    case PPA_GENERIC_DATAPATH_ADDR_TO_FPI_ADDR:
        {
            PPA_FPI_ADDR *addr = (PPA_FPI_ADDR *) buffer;
            addr->addr_fpi = sb_addr_to_fpi_addr_convert(addr->addr_orig);
            return IFX_SUCCESS;
        }

    default:
        dbg("ppa_datapath_generic_hook not support cmd 0x%x\n", cmd );
        return IFX_FAILURE;
    }

    return IFX_FAILURE;

}

int atm_showtime_enter(struct port_cell_info *port_cell, void *xdata_addr) {
	unsigned int i, j;

	ASSERT(port_cell != NULL, "port_cell is NULL");
	ASSERT(xdata_addr != NULL, "xdata_addr is NULL");

	for (j = 0; j < ATM_PORT_NUMBER && j < port_cell->port_num; j++)
		if (port_cell->tx_link_rate[j] > 0)
			break;
	for (i = 0; i < ATM_PORT_NUMBER && i < port_cell->port_num; i++)
		g_atm_priv_data.port[i].tx_max_cell_rate =
				port_cell->tx_link_rate[i] > 0 ?
						port_cell->tx_link_rate[i] : port_cell->tx_link_rate[j];

	qsb_global_set();

	for (i = 0; i < g_atm_priv_data.max_connections; i++)
		if ((g_atm_priv_data.connection_table & (1 << i))
				&& g_atm_priv_data.connection[i].vcc != NULL)
			set_qsb(g_atm_priv_data.connection[i].vcc,
					&g_atm_priv_data.connection[i].vcc->qos, i);

	//  TODO: ReTX set xdata_addr
	g_xdata_addr = xdata_addr;

	g_showtime = 1;

	dbg(
			"enter showtime, cell rate: 0 - %d, 1 - %d, xdata addr: 0x%08x",
			g_atm_priv_data.port[0].tx_max_cell_rate, g_atm_priv_data.port[1].tx_max_cell_rate, (unsigned int)g_xdata_addr);

	return IFX_SUCCESS;
}
// EXPORT_SYMBOL(atm_showtime_enter);

static int atm_showtime_exit(void) {
	if (!g_showtime)
		return IFX_ERROR;

	g_showtime = 0;

	//  TODO: ReTX clean state
	g_xdata_addr = NULL;

	dbg("leave showtime");

	return IFX_SUCCESS;
}
// EXPORT_SYMBOL(atm_showtime_exit);

int get_dslwan_qid_with_vcc(struct atm_vcc *vcc, unsigned int prio) {
    int res = 0;
    int conn = find_vcc(vcc);

    if (conn >= 0 && conn < ATM_QUEUE_NUMBER){
       res |= g_atm_priv_data.connection[conn].prio_queue_map[prio > 7 ? 7 : prio];
       res |= (conn << 8);
    }

    return res;
}
   

int get_netif_qid_with_pkt(struct sk_buff *skb, void *arg, int is_atm_vcc) {
	if (is_atm_vcc) {
		int conn = find_vcc((struct atm_vcc *) arg);

		if (conn < 0)
			return -1;

		return (conn << 8)
				| g_atm_priv_data.connection[conn].prio_queue_map[
						skb->priority > 7 ? 7 : skb->priority];
	} else {
		int port = get_eth_port((struct net_device *) arg);
		if (port >= 0)
			return g_eth_prio_queue_map[port][
					skb->priority > 7 ? 7 : skb->priority];
	}

	return 0;
}

int ppe_directpath_send(uint32_t off, struct sk_buff *skb, int32_t len __attribute__((unused)),
		uint32_t flags __attribute__((unused))) {

	uint32_t if_id = PPA_VDEV_ID(off);

//  Careful, no any safety check here.
//  Parameters must be correct.

	dbg_trace_ppa_data("directpath send if:%d", if_id);



#if !defined(ENABLE_DIRECTPATH_TX_QUEUE) || !ENABLE_DIRECTPATH_TX_QUEUE
	if (eth_xmit(skb, if_id /* dummy */, DMA_CHAN_NR_DIRECTPATH_CPU_TO_PPE, if_id /* spid */,
			0 /* lowest priority */)) {
		return -1;
	} else
		return 0;
#else
	if ( !g_directpath_dma_full && !ppa_pushthrough_data[if_id - 3].skb_list )
	{
		if ( eth_xmit(skb, if_id /* dummy */, DMA_CHAN_NR_DIRECTPATH_CPU_TO_PPE, if_id /* spid */, 0 /* lowest priority */) )
		{
			ppa_pushthrough_data[off].tx_pkt_dropped++;
			return -1;
		}
		else
		return 0;
	}
	else
	{
		if (
#if defined(ENABLE_DIRECTPATH_TX_QUEUE_SIZE)
				ppa_pushthrough_data[off].skb_list_size + skb->len + sizeof(struct sw_eg_pkt_header) <= ENABLE_DIRECTPATH_TX_QUEUE_SIZE
#else
				ppa_pushthrough_data[off].skb_list_len < ENABLE_DIRECTPATH_TX_QUEUE_PKTS
#endif
		)
		{
			struct sw_eg_pkt_header pkth = {spid: if_id, //  CPU/app0/app1/app2
				dpid: 2,//  eth0/eth1/app0/app1/app2
				qid: 0,//  0 is lowest priority, 3 is highest priority
				direct: 0}; //  disable direct forward
			unsigned long sys_flag;

			if ( skb_headroom(skb) < sizeof(struct sw_eg_pkt_header) )
			{
				struct sk_buff *new_skb;

				new_skb = alloc_skb_tx(skb->len + sizeof(struct sw_eg_pkt_header));
				if ( !new_skb )
    				goto PPE_DIRECTPATH_SEND_DROP;
				else
				{
					skb_put(new_skb, skb->len + sizeof(struct sw_eg_pkt_header));
					memcpy(new_skb->data + sizeof(struct sw_eg_pkt_header), skb->data, skb->len);
					dev_kfree_skb_any(skb);
					skb = new_skb;
				}
			}
			else
			skb_push(skb, sizeof(struct sw_eg_pkt_header));
			memcpy(skb->data, &pkth, sizeof(struct sw_eg_pkt_header));

			local_irq_save(sys_flag);
			ppa_pushthrough_data[off].skb_list_len++;
			ppa_pushthrough_data[off].skb_list_size += skb->len;
			if ( ppa_pushthrough_data[off].skb_list != NULL )
			{
				ppa_pushthrough_data[off].skb_list->prev->next = skb;
				skb->next = ppa_pushthrough_data[off].skb_list;
				skb->prev = ppa_pushthrough_data[off].skb_list->prev;
				ppa_pushthrough_data[off].skb_list->prev = skb;
			}
			else
			{
				ppa_pushthrough_data[off].skb_list = skb;
				skb->prev = skb->next = skb;
			}
			local_irq_restore(sys_flag);

			ppa_pushthrough_data[off].tx_pkt_queued++;

			return 0;
		}
		else if ( ppa_pushthrough_data[off].callback.stop_tx_fn )
		ppa_pushthrough_data[off].callback.stop_tx_fn(ppa_pushthrough_data[off].netif);

		PPE_DIRECTPATH_SEND_DROP:
		dev_kfree_skb_any(skb);
		ppa_pushthrough_data[off].tx_pkt_dropped++;
		return -1;
	}
#endif
}

int ppe_directpath_rx_stop(uint32_t if_id __attribute__ ((unused)), uint32_t flags __attribute__ ((unused))) {
	return 0;
}

int ppe_directpath_rx_start(uint32_t if_id __attribute__ ((unused)), uint32_t flags __attribute__ ((unused))) {
	return 0;
}

int ppe_clk_change(unsigned int arg __attribute__((unused)), unsigned int flags __attribute__((unused))) {
	unsigned int i;

	TX_QOS_CFG->time_tick = cgu_get_pp32_clock() / 62500;

	qsb_global_set();

	for (i = 0; i < g_atm_priv_data.max_connections; i++)
		if ((g_atm_priv_data.connection_table & (1 << i))
				&& g_atm_priv_data.connection[i].vcc != NULL)
			set_qsb(g_atm_priv_data.connection[i].vcc,
					&g_atm_priv_data.connection[i].vcc->qos, i);

	return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avmnet_swi_ar10_a5_register_xmit (void) {
	struct net_device *dev;
	unsigned int i;
    printk(KERN_ERR "[%s] setup new xmit function \n", __FUNCTION__ );
	for ( i = 0; i < num_registered_eth_dev ; i++) {
		struct net_device_ops * ops;
		dev = g_eth_net_dev[i];
		ops = dev->netdev_ops;
		ops->ndo_start_xmit = &ifx_ppa_eth_hard_start_xmit_a5;
		ops->ndo_tx_timeout = &ifx_ppa_eth_tx_timeout_a5;
		ops->ndo_do_ioctl = &ifx_ppa_eth_ioctl_a5;
	}
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int a5_in_avm_ata_mode(void) {
	// printk(KERN_ERR "[%s]: %s\n", __FUNCTION__, (g_eth_wan_mode == 3 )?"ATA":"NO ATA");
	return (g_eth_wan_mode == 3);
}
/*------------------------------------------------------------------------------------------*\
 * Modul Init: Init/Cleanup API
\*------------------------------------------------------------------------------------------*/
static int ppe_a5_init(void) {
	int ret;
	int i;
	char buf[512];
	struct port_cell_info port_cell = { 0 };
	int showtime = 0;
	void *xdata_addr = NULL;

	if (g_datapath_mod_count){
		printk("A5 (MII0/1 + ATM) driver already loaded\n");
		return -EBUSY;
	}

	g_datapath_mod_count++;

	push_through_devs_supported_by_datapath = 3;

    g_stop_datapath = 1;    // while doing init

    mb();

    printk("Loading A5 (MII0/1 + ATM) driver ...... \n");

	if ( avm_ata_dev )
	    printk("avm_ata_dev: %s \n" , avm_ata_dev);

#if IFX_PPA_DP_DBG_PARAM_ENABLE
    if( ifx_ppa_drv_dp_dbg_param_enable == 1 )
    {
        ethwan = ifx_ppa_drv_dp_dbg_param_ethwan;
        wanitf = ifx_ppa_drv_dp_dbg_param_wanitf;
        ipv6_acc_en= ifx_ppa_drv_dp_dbg_param_ipv6_acc_en;
        wanqos_en= ifx_ppa_drv_dp_dbg_param_wanqos_en;
    }
#endif  //IFX_PPA_DP_DBG_PARAM_ENABLE


    // tell driver to suspend packet flow on all devices
    if(avmnet_hw_config_entry->config->suspend){
        int wait_cycles = 0;
        while( avmnet_hw_config_entry->config->suspend(avmnet_hw_config_entry->config, NULL)){
            wait_cycles++;
            schedule();
        }
        if (wait_cycles)
            printk(KERN_ERR "[%s] had to wait %d cycles to get switch_lock\n", __func__, wait_cycles);
    }else{
        printk(KERN_ERR "[%s] Config error, no suspend function found!\n", __func__);
        BUG();
    }

    // stop NAPI, empty skb backlog queues
    for_each_possible_cpu(i){
        struct softnet_data *queue;
        struct sk_buff *skb, *tmp;

        queue = &per_cpu(g_7port_eth_queue, i);

        napi_disable(&queue->backlog);
        queue->backlog.poll = NULL;

        skb_queue_walk_safe(&(queue->input_pkt_queue), skb, tmp){
            __skb_unlink(skb, &(queue->input_pkt_queue));
            dev_kfree_skb_any(skb);
        }
    }

    stop_7port_dma_gracefully();
	dma_device_unregister(g_dma_device_ppe);
	dma_device_release(g_dma_device_ppe);
	g_dma_device_ppe = NULL;

	init_local_variables();

	init_hw();

	init_communication_data_structures(g_fwcode);

	ret = alloc_dma();
	if (ret)
		goto ALLOC_DMA_FAIL;

	ppa_hook_mpoa_setup = mpoa_setup;
	vcc_map_skb_prio_qos_queue = vcc_map_skb_prio_qos_queue_datapath;
	vcc_set_nr_qos_queues = vcc_set_nr_qos_queues_datapath;

	for (i = 0; i < ATM_PORT_NUMBER; i++) {
		g_atm_priv_data.port[i].dev = atm_dev_register("ifx_atm", &g_ppe_atm_ops, -1, 0UL);
		if (!g_atm_priv_data.port[i].dev) {
			ret = -EIO;
			goto ATM_DEV_REGISTER_FAIL;
		} else {
			g_atm_priv_data.port[i].dev->ci_range.vpi_bits = 8;
			g_atm_priv_data.port[i].dev->ci_range.vci_bits = 16;
			g_atm_priv_data.port[i].dev->link_rate = g_atm_priv_data.port[i].tx_max_cell_rate;
			g_atm_priv_data.port[i].dev->dev_data = (void*) i;
		}
    }
    
    ret = request_irq(IFX_PPE_MBOX_INT0, mailbox0_irq_handler, IRQF_DISABLED, "a5_mailbox0_isr", NULL);
    if ( ret )
    {
        if ( ret == -EBUSY )
            err("IRQ may be occupied by other PPE driver, please reconfig to disable it.\n");
        goto REQUEST_IGU0_IRQ_FAIL;
    }



	ret = request_irq(IFX_PPE_MBOX_INT2, mailbox1_irq_handler, IRQF_DISABLED, "a5_mailbox1_isr", NULL);
	if (ret) {
		if (ret == -EBUSY)
			err( "IRQ may be occupied by other PPE driver, please reconfig to disable it.\n");
		goto REQUEST_IRQ_FAIL;
	}

	/*
	 *  init variable for directpath
	 */

	//don't clear pushthtrough data in Datapath-Init in AVM Setup -> we use common pushthrough_data for all datapath modules
	// memset(ppa_pushthrough_data, 0, sizeof(ppa_pushthrough_data));


	dump_init();

    ret = pp32_start(g_fwcode);

	if (ret)
		goto PP32_START_FAIL;

    *MBOX_IGU0_IER = g_eth_wan_mode ? DMA_TX_CH0_SIG | DMA_TX_CH1_SIG : DMA_TX_CH0_SIG;
	*MBOX_IGU1_IER = g_mailbox_signal_mask;

	qsb_global_set();
	if (g_eth_wan_mode == 0)
		validate_oam_htu_entry();

    /* neue (a5 spezifische) xmit-Funktionen registrieren */
	avmnet_swi_ar10_a5_register_xmit ( );

    /* disable learning, flush mac table*/
	avmnet_swi_7port_disable_learning( avmnet_hw_config_entry->config );


    dbg_ppe_init("show_dma");
    show_fdma_sdma();

	/* (re-)aktivieren der DMA-Ports */
    start_cpu_port();
    // enable_fdma_sdma(); is now done in start_cpu_port

	dbg_ppe_init("show_dma");
    show_fdma_sdma();

    mb();

    if ( g_eth_wan_mode == 0 /* DSL WAN */ ) {

    	avm_wan_mode = AVM_WAN_MODE_ATM;
        avmnet_swi_7port_set_ethwan_dev_by_name( NULL );

        //  turn on DMA TX channels (0, 1) for fast path
        g_dma_device_ppe->tx_chan[0]->open(g_dma_device_ppe->tx_chan[0]); // ppa -> switch( port 6 )
        g_dma_device_ppe->rx_chan[2]->open(g_dma_device_ppe->rx_chan[2]); // ppa -> switch( port 6 )

        //TODO: Test tx_1 ch hier schon oeffnen, da offensichtlich das erste paket verloren geht?!
        g_dma_device_ppe->tx_chan[1]->open(g_dma_device_ppe->tx_chan[1]);
        //TODO: Test Ende

        if (!IS_ERR(&ifx_mei_atm_showtime_check) && &ifx_mei_atm_showtime_check)
            ifx_mei_atm_showtime_check(&showtime, &port_cell, &xdata_addr);
        else
            printk("Weak-Symbol nicht gefunden: ifx_mei_atm_showtime_check\n");
        if (showtime)
            atm_showtime_enter(&port_cell, xdata_addr);

        //  init timer for Data Led
        setup_timer(&g_dsl_led_polling_timer, dsl_led_polling, 0);
#if defined(ENABLE_LED_FRAMEWORK) && ENABLE_LED_FRAMEWORK && defined(CONFIG_IFX_LED)
        ifx_led_trigger_register("dsl_data", &g_data_led_trigger);
#endif
    } else {
        printk(KERN_ERR "[%s] warning: runnung A5 not in DSL mode!\n", __FUNCTION__);
    }

    if ( g_eth_wan_mode == 3 /* ETH WAN */ ) {
        printk(KERN_ERR "[%s] Configuring ETH_WAN (AVM ATA)!\n", __FUNCTION__);
    	avm_wan_mode = AVM_WAN_MODE_ETH;
    }


	register_netdev_event_handler();

	/*  create proc file    */
	//ifx_ppa_drv_proc_init(g_ipv6_acc_en);
	proc_file_create();

    if ( g_eth_wan_mode == 0 /* DSL WAN */ ) {
        if (!IS_ERR(&ifx_mei_atm_showtime_enter) && &ifx_mei_atm_showtime_enter)
            ifx_mei_atm_showtime_enter = atm_showtime_enter;
        else
            printk("Weak-Symbol nicht gefunden: ifx_mei_atm_showtime_enter\n");

        if (!IS_ERR(&ifx_mei_atm_showtime_exit) && &ifx_mei_atm_showtime_exit)
            ifx_mei_atm_showtime_exit = atm_showtime_exit;
        else
            printk("Weak-Symbol nicht gefunden: ifx_mei_atm_showtime_exit\n");

    } else {
        printk(KERN_ERR "[%s] warning: runnung A5 not in DSL mode!\n", __FUNCTION__);
    }

    // replace common NAPI poll function with our one
    for_each_possible_cpu(i) {
        struct softnet_data *queue;

        queue = &per_cpu(g_7port_eth_queue, i);

        skb_queue_head_init(&queue->input_pkt_queue);
        queue->completion_queue = NULL;
        INIT_LIST_HEAD(&queue->poll_list);

        queue->backlog.poll = process_backlog_a5;
        napi_enable(&queue->backlog);
    }

#if 0 //DMA lauft immer
    /* check if some interfaces have been opened by common driver before */
    for (i = 0; i < NUM_ETH_INF; i++)
    	if (g_eth_net_dev[i] && netif_running(g_eth_net_dev[i])){
    		printk(KERN_ERR "[%s] dev=%d is already up and running -> reenable dma\n", __FUNCTION__, i);
    		turn_on_dma_rx(1);
    		break;
    	}
#else
    turn_on_dma_rx(1);
#endif

	if ( ( g_eth_wan_mode == 3 ) && avm_ata_dev ) /* ETH WAN */
        avmnet_swi_7port_set_ethwan_dev_by_name( avm_ata_dev );

    //Fix warning message ---start
    ifx_ppa_drv_get_dslwan_qid_with_vcc_hook = get_dslwan_qid_with_vcc;
    ifx_ppa_drv_get_netif_qid_with_pkt_hook = get_netif_qid_with_pkt;
    ifx_ppa_drv_ppe_clk_change_hook = ppe_clk_change;

	in_avm_ata_mode_hook = a5_in_avm_ata_mode;

	ppa_pushthrough_hook = ppe_directpath_send;

#ifdef CONFIG_IFX_PPA_API_DIRECTPATH
    ifx_ppa_drv_g_ppe_directpath_data = ppa_pushthrough_data;
    ifx_ppa_drv_directpath_send_hook = ppe_directpath_send;
    ifx_ppa_drv_directpath_rx_start_hook = ppe_directpath_rx_start;
    ifx_ppa_drv_directpath_rx_stop_hook = ppe_directpath_rx_stop;
#endif //CONFIG_IFX_PPA_API_DIRECTPATH
    ifx_ppa_drv_datapath_generic_hook = ppa_datapath_generic_hook;

#ifdef AVM_DOESNT_USE_IFX_ETHSW_API
    ifx_ppa_drv_datapath_mac_entry_setting = mac_entry_setting;
#else
    ifx_ppa_drv_datapath_mac_entry_setting = NULL;
#endif
    //Fix warning message ---end

    mb();
    g_stop_datapath = 0;    // restart after init

    // tell driver to resume packet flow on all devices
    if(avmnet_hw_config_entry->config->resume){
        while( avmnet_hw_config_entry->config->resume(avmnet_hw_config_entry->config, NULL) )
            schedule();
    }else{
        printk(KERN_ERR "[%s] Config error, no resume function found!\n", __func__);
        BUG();
    }

	printk("Succeeded!\n");
	print_driver_ver(buf, sizeof(buf), "PPE datapath driver info", VER_FAMILY,
			VER_DRTYPE, VER_INTERFACE, VER_ACCMODE, VER_MAJOR, VER_MID,
			VER_MINOR);
	printk(buf);
	print_fw_ver(buf, sizeof(buf));
	printk(buf);

	return 0;

	PP32_START_FAIL: 
        free_irq(IFX_PPE_MBOX_INT0, NULL);
	REQUEST_IRQ_FAIL: 
        free_irq(IFX_PPE_MBOX_INT2, NULL);

    REQUEST_IGU0_IRQ_FAIL:
	ATM_DEV_REGISTER_FAIL: 
        while (i--)
		    atm_dev_deregister(g_atm_priv_data.port[i].dev);
        ppa_hook_mpoa_setup = NULL;
    	vcc_map_skb_prio_qos_queue = NULL;
	    vcc_set_nr_qos_queues = NULL;

        stop_7port_dma_gracefully();
	    free_dma();
	ALLOC_DMA_FAIL: 
        clear_local_variables();

        g_datapath_mod_count--;
    	avm_wan_mode = AVM_WAN_MODE_NOT_CONFIGURED;

        return ret;
}

static void __exit ppe_a5_exit(void) {
	int i;

	g_stop_datapath = 1;
	mb();

    // tell driver to suspend packet flow on all devices
    if(avmnet_hw_config_entry->config->suspend){
        while( avmnet_hw_config_entry->config->suspend(avmnet_hw_config_entry->config, NULL))
            schedule();
    }else{
        printk(KERN_ERR "[%s] Config error, no suspend function found!\n", __func__);
        BUG();
    }
    // stop NAPI, empty skb backlog queues
    for_each_possible_cpu(i){
        struct softnet_data *queue;
        struct sk_buff *skb, *tmp;

        queue = &per_cpu(g_7port_eth_queue, i);

        napi_disable(&queue->backlog);
        queue->backlog.poll = NULL;

        skb_queue_walk_safe(&(queue->input_pkt_queue), skb, tmp){
            __skb_unlink(skb, &(queue->input_pkt_queue));
            dev_kfree_skb_any(skb);
        }
    }

    //Fix warning message ---start
    ifx_ppa_drv_get_dslwan_qid_with_vcc_hook = NULL;
    ifx_ppa_drv_get_netif_qid_with_pkt_hook = NULL;
    ifx_ppa_drv_ppe_clk_change_hook = NULL;

	in_avm_ata_mode_hook = NULL;

	ppa_pushthrough_hook = NULL;

#ifdef CONFIG_IFX_PPA_API_DIRECTPATH
    ifx_ppa_drv_g_ppe_directpath_data = NULL;
    ifx_ppa_drv_directpath_send_hook = NULL;
    ifx_ppa_drv_directpath_rx_start_hook = NULL;
    ifx_ppa_drv_directpath_rx_stop_hook = NULL;
#endif //CONFIG_IFX_PPA_API_DIRECTPATH
    ifx_ppa_drv_datapath_generic_hook = NULL;
    ifx_ppa_drv_datapath_mac_entry_setting = NULL;
    //Fix warning message ---end
    //
    if ( g_eth_wan_mode == 0 /* DSL WAN */ ) {
        if (!IS_ERR(&ifx_mei_atm_showtime_enter) && &ifx_mei_atm_showtime_enter)
            ifx_mei_atm_showtime_enter = NULL;
        else
            printk("Weak-Symbol nicht gefunden: ifx_mei_atm_showtime_enter\n");

        if (!IS_ERR(&ifx_mei_atm_showtime_exit) && &ifx_mei_atm_showtime_exit)
            ifx_mei_atm_showtime_exit = NULL;
        else
            printk("Weak-Symbol nicht gefunden: ifx_mei_atm_showtime_exit\n");
    }

	proc_file_delete();

	unregister_netdev_event_handler();

#if defined(ENABLE_LED_FRAMEWORK) && ENABLE_LED_FRAMEWORK && defined(CONFIG_IFX_LED)
	ifx_led_trigger_deregister(g_data_led_trigger);
	g_data_led_trigger = NULL;
#endif

	stop_7port_dma_gracefully();

    if ( g_eth_wan_mode == 0 )
	    invalidate_oam_htu_entry();

	pp32_stop();

	free_irq(IFX_PPE_MBOX_INT0 , NULL);
    free_irq(IFX_PPE_MBOX_INT2 , NULL);


#if 0
	for (i = 0; i < num_registered_eth_dev; i++) {
		if (g_eth_net_dev[i]) {
			free_netdev(g_eth_net_dev[i]);
			unregister_netdev(g_eth_net_dev[i]);
			g_eth_net_dev[i] = NULL;
		}
	}
#endif
//    unregister_netdev(&g_eth_net_dev[0]);

	for (i = 0; i < ATM_PORT_NUMBER; i++)
		atm_dev_deregister(g_atm_priv_data.port[i].dev);

	ppa_hook_mpoa_setup = NULL;
	vcc_map_skb_prio_qos_queue = NULL;
	vcc_set_nr_qos_queues = NULL;


	free_dma();

	clear_share_buffer();

	clear_local_variables();

	reinit_7port_common_eth();

    /* disable drop xmit packets*/
    g_stop_datapath = 0;

    // tell driver to resume packet flow on all devices
    if(avmnet_hw_config_entry->config->resume){
        avmnet_hw_config_entry->config->resume(avmnet_hw_config_entry->config, NULL);
    }else{
        printk(KERN_ERR "[%s] Config error, no resume function found!\n", __func__);
        BUG();
    }

	g_datapath_mod_count--;
   	avm_wan_mode = AVM_WAN_MODE_NOT_CONFIGURED;
}

#if defined(CONFIG_IFX_PPA_DATAPATH)

#if !defined(AVM_SETUP_MAC)
static int __init eth0addr_setup(char *line)
{
	ethaddr_setup(0, line);

	return 0;
}

static int __init eth1addr_setup(char *line)
{
	ethaddr_setup(1, line);

	return 0;
}
#endif

static int __init wan_mode_setup(char *line)
{
	if ( strcmp(line, "1") == 0 )
	ethwan = 1;
	else if ( strcmp(line, "2") == 0 )
	ethwan = 2;

	return 0;
}

static int __init wanitf_setup(char *line)
{
	wanitf = simple_strtoul(line, NULL, 0);

	if ( wanitf > 0xFF )
	    wanitf = ~0;

	return 0;
}
static int __init ipv6_acc_en_setup(char *line)
{
	if ( strcmp(line, "0") == 0 )
	ipv6_acc_en = 0;
	else
	ipv6_acc_en = 1;

	return 0;
}

static int __init wanqos_en_setup(char *line)
{
    int tmp_wanqos_en = simple_strtoul(line, NULL, 0);

    if ( isdigit(*line) && tmp_wanqos_en >= 0 && tmp_wanqos_en <= 8 )
        wanqos_en = tmp_wanqos_en;

    return 0;
}

#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_SIZE)
static int __init directpath_tx_queue_size_setup(char *line)
{
	unsigned long i;

	i = simple_strtoul(line, NULL, 0);
	directpath_tx_queue_size = (int)i;

	return 0;
}
#endif

#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_PKTS)
static int __init directpath_tx_queue_pkts_setup(char *line)
{
	unsigned long i;

	i = simple_strtoul(line, NULL, 0);
	directpath_tx_queue_pkts = (int)i;

	return 0;
}
#endif
#endif

module_init(ppe_a5_init);
module_exit(ppe_a5_exit);
#if defined(CONFIG_IFX_PPA_DATAPATH)
__setup("ethaddr=", eth0addr_setup);
__setup("eth1addr=", eth1addr_setup);
__setup("ethwan=", wan_mode_setup);
__setup("wanitf=", wanitf_setup);
__setup("ipv6_acc_en=", ipv6_acc_en_setup);
__setup("wanqos_en=", wanqos_en_setup);
#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_SIZE)
__setup("directpath_tx_queue_size=", directpath_tx_queue_size_setup);
#endif
#if defined(CONFIG_IFX_PPA_DIRECTPATH_TX_QUEUE_PKTS)
__setup("directpath_tx_queue_pkts=", directpath_tx_queue_pkts_setup);
#endif
#endif
MODULE_LICENSE("GPL");
